/**
 * Copyright (c) 2005-2012 EBM WebSourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.admin;

import java.util.Map;
import java.util.Properties;
import java.util.logging.Logger;

import org.objectweb.fractal.fraclet.annotations.Component;
import org.objectweb.fractal.fraclet.annotations.Interface;
import org.objectweb.fractal.fraclet.annotations.Lifecycle;
import org.objectweb.fractal.fraclet.annotations.Requires;
import org.objectweb.fractal.fraclet.types.Contingency;
import org.objectweb.fractal.fraclet.types.Step;
import org.ow2.petals.basisapi.exception.PetalsException;
import org.ow2.petals.clientserverapi.configuration.ContainerConfiguration;
import org.ow2.petals.microkernel.api.admin.PetalsAdminInterface;
import org.ow2.petals.microkernel.api.communication.topology.TopologyService;
import org.ow2.petals.microkernel.api.configuration.ConfigurationService;
import org.ow2.petals.microkernel.api.jbi.management.DeploymentService;
import org.ow2.petals.microkernel.api.jbi.management.InstallationService;

import com.ebmwebsourcing.easycommons.log.LoggingUtil;

/**
 * This class is used for Petals global Administration.
 * @author Adrien Louis - EBM WebSourcing
 * @author Roland Naudin - EBM WebSourcing
 */
@Component(provides = @Interface(name = "service", signature = PetalsAdminInterface.class))
public class PetalsAdminServiceImpl implements PetalsAdminInterface {

    /**
     * Logger wrapper
     */
    protected final LoggingUtil log = new LoggingUtil(Logger.getLogger(Constants.FRACTAL_COMPONENT_LOGGER_NAME));

    /**
     * Deployment service fractal component
     */
    @Requires(name = "deployment")
    protected DeploymentService deployment;

    /**
     * Installation service fractal component
     */
    @Requires(name = "installation")
    protected InstallationService installation;

    /**
     * Topology service fractal component
     */
    @Requires(name = "topology", contingency = Contingency.OPTIONAL)
    protected TopologyService topologyService;

    /**
     * The Configuration service fractal component
     */
    @Requires(name = "configuration")
    private ConfigurationService configurationService;

    /**
     * The container configuration
     */
    private ContainerConfiguration containerConfiguration;

    /**
     * The Petals Stop thread instance, which stops the PEtALS container
     */
    private Thread petalsStopThread;

    @Override
    public Properties retrieveServerProperties() {
        final Properties properties = new Properties();
        final Map<String, String> propertiesMap = this.configurationService.getAllServerProperties();
        for (final Map.Entry<String, String> property : propertiesMap.entrySet()) {
            final String value = property.getValue();
            if (value != null) {
                properties.put(property.getKey(), value);
            } else {

                properties.put(property.getKey(), "");
            }
        }
        return properties;
    }

    @Override
    public void shutdownContainer() throws PetalsException {
        this.log.start();

        this.log.info("PEtALS is shutting down...");

        this.log.info("  1/3 Undeploy Service Assemblies...");

        try {
            this.deployment.shutdown();
        } catch (Exception e) {
            this.log.error("Can not shut down the Deployment service.", e);
        }

        this.log.info("  2/3 Uninstall Shared Libraries and Components...");

        try {
            this.installation.shutdown();
        } catch (Exception e) {
            this.log.error("Can not shut down the Installation service", e);
        }

        this.log.info("  3/3 Stopping the server...");

        // launch the stop of the Petals server in a different thread
        this.petalsStopThread.start();

        this.log.end();
    }

    @Override
    public void stopContainer() throws PetalsException {
        this.log.start();

        this.log.info("PEtALS is stopping...");
        // launch the stop of the Petals server in a different thread
        this.petalsStopThread.start();

        this.log.end();
    }

    @Override
    public void setPetalsStopThread(final Thread petalsStopThread) {
        this.petalsStopThread = petalsStopThread;
    }

    @Lifecycle(step = Step.START)
    public void start() {
        this.log.start();

        this.containerConfiguration = this.configurationService.getContainerConfiguration();

        this.log.end();
    }

    @Lifecycle(step = Step.STOP)
    public void stop() {
        this.log.call();
    }

    @Override
    public boolean ping() {
        return this.topologyService.isContainerReachable(this.containerConfiguration.getName());
    }
}
