/**
 * Copyright (c) 2009-2012 EBM WebSourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.communication.topology;

import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.Timer;
import java.util.TimerTask;
import java.util.logging.Logger;

import org.objectweb.fractal.fraclet.annotations.Component;
import org.objectweb.fractal.fraclet.annotations.Interface;
import org.objectweb.fractal.fraclet.annotations.Lifecycle;
import org.objectweb.fractal.fraclet.annotations.Requires;
import org.objectweb.fractal.fraclet.types.Step;
import org.ow2.petals.basisapi.exception.PetalsException;
import org.ow2.petals.clientserverapi.configuration.ContainerConfiguration;
import org.ow2.petals.clientserverapi.configuration.ContainerConfiguration.ContainerState;
import org.ow2.petals.clientserverapi.topology.exception.AttachContainerException;
import org.ow2.petals.clientserverapi.topology.exception.DetachContainerException;
import org.ow2.petals.clientserverapi.topology.exception.DomainIsNotDynamicException;
import org.ow2.petals.clientserverapi.topology.exception.InconsistencyDomainException;
import org.ow2.petals.clientserverapi.topology.exception.InconsistencyException;
import org.ow2.petals.clientserverapi.topology.exception.InconsistentContainerNameException;
import org.ow2.petals.clientserverapi.topology.exception.InconsistentDomainNameException;
import org.ow2.petals.clientserverapi.topology.exception.InconsistentDomainRegistryImplException;
import org.ow2.petals.clientserverapi.topology.exception.InvalidPassphraseException;
import org.ow2.petals.clientserverapi.topology.exception.NoContainerConfigurationProvidedException;
import org.ow2.petals.clientserverapi.topology.exception.NoContainerNameProvidedException;
import org.ow2.petals.clientserverapi.topology.exception.NoDomainConfigurationProvidedException;
import org.ow2.petals.clientserverapi.topology.exception.NoSecurityPassPhraseProvidedException;
import org.ow2.petals.clientserverapi.topology.exception.TopologyServiceException;
import org.ow2.petals.jmx.api.api.JMXClient;
import org.ow2.petals.jmx.api.api.PetalsJmxApiFactory;
import org.ow2.petals.jmx.api.api.TopologyServiceClient;
import org.ow2.petals.jmx.api.api.exception.ConnectionErrorException;
import org.ow2.petals.jmx.api.api.exception.DuplicatedServiceException;
import org.ow2.petals.jmx.api.api.exception.MissingServiceException;
import org.ow2.petals.microkernel.api.communication.topology.TopologyService;
import org.ow2.petals.microkernel.api.communication.topology.exception.TopologyException;
import org.ow2.petals.microkernel.api.configuration.ConfigurationService;
import org.ow2.petals.microkernel.api.configuration.DomainConfiguration;
import org.ow2.petals.microkernel.api.configuration.DomainConfiguration.DomainMode;
import org.ow2.petals.microkernel.api.implementation.MutableImplementation;
import org.ow2.petals.microkernel.api.implementation.exception.MutableImplException;
import org.ow2.petals.microkernel.api.system.sharedarea.SharedAreaService;
import org.ow2.petals.microkernel.api.system.sharedarea.SharedAreaService.SavedEndpoint;
import org.ow2.petals.microkernel.api.util.ConfigurationServiceUtils;
import org.ow2.petals.microkernel.communication.RemoteCheckerClient;
import org.ow2.petals.microkernel.server.PetalsCompositeController;
import org.ow2.petals.microkernel.server.exception.PetalsCompositeCtrlException;
import org.ow2.petals.topology.generated.Topology;
import org.w3c.dom.Document;

import com.ebmwebsourcing.easycommons.log.LoggingUtil;
import com.ebmwebsourcing.easycommons.uuid.SimpleUUIDGenerator;

/**
 * The default implementation of the topology service.
 *
 * TODO for now, it does not handle well problems during attachment: if an error happens, the best thing to do is to
 * stop the container and restart it...
 * 
 * For it to work we would need: to remember endpoints that were removed in a previous detachment and taking care of
 * well differentiating detached/attached state w.r.t. to that.
 *
 * @author Christophe Hamerling - EBM WebSourcing
 * @author Christophe Deneux
 * @author Victor Noel
 */
@Component(provides = @Interface(name = "service", signature = TopologyService.class))
public class TopologyServiceImpl implements TopologyService {

    /**
     * Pattern defining a description to put in the domain definition about the new domain of a container after
     * detachment.
     */
    private static final String DOMAIN_DESCR_DETACHMENT_PATTERN = "Domain created during container detachment from domain '%s'";

    /**
     * Inner class to ping the remote container.
     * @author Roland Naudin - EBM WebSourcing
     */
    private class TopologyPinger extends TimerTask {

        @Override
        public void run() {
            final String myName = TopologyServiceImpl.this.configurationService.getContainerConfiguration().getName();
            final Set<ContainerConfiguration> containers = TopologyServiceImpl.this.configurationService
                    .getContainersConfiguration();
            for (ContainerConfiguration cc : containers) {
                if (TopologyServiceImpl.this.stopped) {
                    break;
                }
                final String containerName = cc.getName();
                if (!containerName.equals(myName)) {
                    if (TopologyServiceImpl.this.remoteChecker.ping(containerName)) {
                        cc.setState(ContainerState.REACHABLE);
                        TopologyServiceImpl.this.log.debug("Container '" + containerName + "' is reachable");
                    } else {
                        cc.setState(ContainerState.UNREACHABLE);
                        TopologyServiceImpl.this.log.debug("Container '" + containerName + "' is unreachable");
                    }
                }
            }
        }

    }

    /**
     * Logger wrapper.
     */
    private final LoggingUtil log = new LoggingUtil(Logger.getLogger(Constants.FRACTAL_COMPONENT_LOGGER_NAME));
    
    /**
     * The Configuration service fractal component
     */
    @Requires(name = "configuration")
    private ConfigurationService configurationService;

    /**
     * The service 'SharedArea'
     */
    @Requires(name = "sharedArea")
    private SharedAreaService sharedArea;

    /**
     * The service 'PetalsCompositeController'
     */
    @Requires(name = "petalsCompositeController")
    private PetalsCompositeController petalsCompositeController;

    /**
     * Flag to know if the fractal component is stopping
     */
    private volatile boolean stopped;

    /**
     * The Ping service<br>
     * TODO: It must be a ping via the transporter instead of JMX
     */
    @Requires(name = "remote")
    private RemoteCheckerClient remoteChecker;

    /**
     * The timer to ping the remote containers
     */
    private Timer topologyTimer;

    private boolean isCurrentDomainDynamic() {
        return this.configurationService.getDomainConfiguration().getMode().equals(DomainMode.DYNAMIC);
    }

    @Override
    public boolean isContainerReachable(String containerName) {
        return ContainerState.REACHABLE
                .equals(this.configurationService.getContainerConfiguration(containerName).getState());
    }

    @Lifecycle(step = Step.START)
    public void start() {
        this.log.start();

        this.stopped = false;
        this.topologyTimer = new Timer("Topology Pinger");
        if (this.remoteChecker != null) {
            final ContainerConfiguration containerConfig = this.configurationService.getContainerConfiguration();
            this.topologyTimer.schedule(new TopologyPinger(), containerConfig.getTopologyPingerStartDelay(),
                    containerConfig.getTopologyPingerPeriodDelay());
        } else {
            this.log.warning("No remote checker is bound. The topology pinger is not scheduled !!");
        }

        this.log.end();
    }

    @Lifecycle(step = Step.STOP)
    public void stop() {
        this.log.start();

        this.stopped = true;
        this.topologyTimer.cancel();

        this.log.end();
    }

    @Override
    public Set<Map<String, String>> retrieveTopology(final String securityPassPhrase,
            final boolean isSensibleInformationExpected) throws TopologyServiceException {
        this.log.start();
        
        // Check security
        final boolean isSecurityOk = this.configurationService.isSecurityTopologyPassphraseOk(securityPassPhrase);
        if (isSensibleInformationExpected && !isSecurityOk) {
            throw new InvalidPassphraseException();
        }
        if (this.log.isDebugEnabled()) {
            if (isSecurityOk) {
                this.log.debug("Security validated. Sensible information will be returned.");
            } else {
                this.log.debug("Security not validated. Sensible information will NOT be returned.");
            }
        }

        final Set<Map<String, String>> topologyMap = new HashSet<>();
        topologyMap.add(this.convertDomainConfiguration(this.configurationService.getDomainConfiguration(),
                securityPassPhrase));
        for (final ContainerConfiguration cc : this.configurationService.getContainersConfiguration()) {
            topologyMap.add(this.convertContainerConfiguration(cc, isSecurityOk));
        }
        this.log.end();
        return topologyMap;
    }

    /**
     * Convert a domain configuration into a key/value map form.
     * 
     * @param domainConfiguration
     *            The domain configuration to convert.
     * @return The domain configuration as key/value map form.
     * @param securityPassPhrase
     *            Passphrase needed to return sensible information. If <code>null</code>, empty or not matching the one
     *            of the container on which the connection is established, no sensible information will be returned.
     * @throws TopologyException
     *             An error occurs getting configuration properties
     */
    private Map<String, String> convertDomainConfiguration(final DomainConfiguration domainConfiguration,
            final String securityPassPhrase) throws TopologyException {
        this.log.start();

        try {
            final Map<String, String> domainMap = domainConfiguration.toMap();
            domainMap.put(TopologyItem.TOPOLOGY_TYPE, TopologyItem.DOMAIN);

            try {
                final Properties registryProperties = this.petalsCompositeController
                        .getSharedAreaCltConfiguration(securityPassPhrase);
                for (final String propertyName : registryProperties.stringPropertyNames()) {
                    domainMap.put(propertyName, registryProperties.getProperty(propertyName));
                }
            } catch (final PetalsCompositeCtrlException e) {
                throw new TopologyException(e);
            }

            return domainMap;
        } finally {
            this.log.end();
        }
    }

    /**
     * Convert a container configuration into a key/value map form.
     * 
     * @param containerConfiguration
     *            The container configuration to convert.
     * @param isSecurityOk
     *            if <code>true</code>, sensible information are returned. Otherwise, sensible information are skipped.
     * @return The container configuration as key/value map form.
     */
    private Map<String, String> convertContainerConfiguration(final ContainerConfiguration containerConfiguration,
            final boolean isSecurityOk) {
        this.log.start();

        try {
            final Map<String, String> containerMap = containerConfiguration.toMap(isSecurityOk);
            containerMap.put(TopologyItem.TOPOLOGY_TYPE, TopologyItem.CONTAINER);

            return containerMap;
        } finally {
            this.log.end();
        }
    }

    @Override
    public void detachContainer() throws DomainIsNotDynamicException, DetachContainerException {

        final ClassLoader oldClassloader = Thread.currentThread().getContextClassLoader();
        Thread.currentThread().setContextClassLoader(this.getClass().getClassLoader());

        final ContainerConfiguration currentContainerConfig = this.configurationService.getContainerConfiguration();
        final DomainConfiguration currentDomainConfig = this.configurationService.getDomainConfiguration();

        try {
            this.checkIfDetachmentIsPossible();

            // A stand-alone topology is persisted in the working area to be able to restart the
            // container in its own topology. The topology is persisted before to be stored in the
            // shared area, in this way, if an error occurs writing in the registry, a restart is
            // sufficient to restore state and to not join the initial topology.
            // The stand-alone topology is created re-using some configuration information:
            // - the container name to avoid to change the content of the server.properties
            // - the container hostname that has not to be changed
            // - the JMX configuration
            // - the port of the fast remote transporter
            // - the configuration of the web-service service
            // - the shared area client configuration is re-used, and as the domain name has been changed, the
            // shared areas (the shared area leaved, and the new shared area created) are distinct.
            final DomainConfiguration newDomainCfg = new DomainConfiguration();
            newDomainCfg.setName(new SimpleUUIDGenerator().getNewID());
            newDomainCfg.setDescription(String.format(DOMAIN_DESCR_DETACHMENT_PATTERN, currentDomainConfig.getName()));
            newDomainCfg.setMode(DomainMode.DYNAMIC);
            newDomainCfg.setSharedAreaImplementation(currentDomainConfig.getSharedAreaImplementation());
            newDomainCfg.setExtraParameters(currentDomainConfig.getExtraParameters());

            // TODO find a way to factorise this code with the code of attachContainerTo
            // -> the idea would be for the joining to be identical in both case...

            // this will block endpoint resolution and thus new exchange being sent
            try (final AutoCloseable ed = this.petalsCompositeController.pauseEndpointDirectory()) {

                // we will leave the current topology (which is dynamic)
                final Collection<SavedEndpoint> endpoints = this.sharedArea.leaveTopology();

                // Stop the messaging sub-system (i.e. wait for router to process exchanges left
                // and then stop the transporters)
                // Stop the shared area, before reconfiguring the container and restart it using the shared area
                // of a stand-alone topology
                try (final AutoCloseable ms = this.petalsCompositeController.pauseMessagingSubsystem();
                        final AutoCloseable sa = this.petalsCompositeController.pauseSharedAreaComponent()) {
                    // Note: when restarting the shared area, this information will be used
                    // to attach to the topology.
                    this.configurationService.reconfigureDomain(newDomainCfg);

                    // we also need to create a topology and update it in the ConfigurationService
                    // so that loadTopology can work after
                    final Topology standaloneTopology = new Topology();
                    ConfigurationServiceUtils.initializeTopology(standaloneTopology, newDomainCfg,
                            Collections.singletonList(currentContainerConfig));

                    // this validate and update the topology (used by load below)
                    // It will also update the list of known domains!
                    this.configurationService.updateTopology(standaloneTopology);
                }
                // Starting the messaging subsystem will restart transporters

                // and now we can explicitly use the topology!
                this.sharedArea.loadTopology(endpoints);

            } catch (final PetalsException e) {
                // these are PetalsException (and other) thrown by other functions we rely on
                this.log.warning("An error occurs detaching the container from its current domain", e);
                throw new DetachContainerException(currentDomainConfig.getName(), e);
            } catch (final Exception e) {
                this.log.warning("Error while re-starting component", e);
            }
            // starting the endpoint directory will resume new exchanges from us!

            this.log.info("Current container detached from its inital domain");
        } finally {
            Thread.currentThread().setContextClassLoader(oldClassloader);
        }
    }

    /**
     * <p>
     * Only one attachment of the current container can be done in the same time. That's why this method is
     * synchronized. We use a distributed lock to synchronize the attachments of several containers to the same target
     * topology.
     * </p>
     */
    @Override
    public synchronized void attachContainerTo(final ContainerConfiguration targetContainer, final String securityPassPhrase)
            throws NoContainerConfigurationProvidedException, NoContainerNameProvidedException,
            NoDomainConfigurationProvidedException, DomainIsNotDynamicException, InconsistencyException,
            NoSecurityPassPhraseProvidedException, AttachContainerException {

        final ClassLoader oldClassloader = Thread.currentThread().getContextClassLoader();
        Thread.currentThread().setContextClassLoader(this.getClass().getClassLoader());
        try {
            if (targetContainer == null) {
                throw new NoContainerConfigurationProvidedException();
            }

            if (securityPassPhrase == null || securityPassPhrase.isEmpty()) {
                throw new NoSecurityPassPhraseProvidedException();
            }

            final String currentContainerName = this.configurationService.getContainerConfiguration().getName();
            final String targetContainerName = targetContainer.getName();
            if (targetContainerName == null) {
                throw new NoContainerNameProvidedException();
            }

            final String domainName = targetContainer.getDomainName();
            if (domainName == null) {
                throw new NoDomainConfigurationProvidedException();
            }

            this.log.info(
                    String.format("Attaching the current container '%s' to the domain '%s' using the container '%s'",
                            currentContainerName, domainName, targetContainerName));

            // this contains everything needed to connect to another topology including shared area specificities
            final DomainConfiguration targetDomainCfg = checkIfAttachmentIsPossible(targetContainer,
                    securityPassPhrase);

            // this will block endpoint resolution and thus new exchange being sent
            try (final AutoCloseable ed = this.petalsCompositeController.pauseEndpointDirectory()) {

                // we leave the current topology (which is dynamic)
                // this will prevent other containers to send us new exchange
                final Collection<SavedEndpoint> endpoints = this.sharedArea.leaveTopology();

                // Stop the messaging sub-system (i.e. wait for router to process exchanges left
                // and then stop the transporters)
                // Stop the shared area, before reconfiguring the container and restart it using the shared area
                // of the target topology.
                try (final AutoCloseable ms = this.petalsCompositeController.pauseMessagingSubsystem();
                        final AutoCloseable sa = this.petalsCompositeController.pauseSharedAreaComponent()) {
                    // Note: when restarting the shared area, this information will be used
                    // to join the target topology (which is dynamic).
                    this.configurationService.reconfigureDomain(targetDomainCfg);
                }
                // Starting the messaging subsystem will restart transporters

                // and now we can explicitly join the topology.
                // new exchange will arrive from other containers
                this.sharedArea.joinTopology(endpoints);

                // We notify the Petals composite controller that the configuration has changed to adjust
                // states of Fractal components (for example, the Embedded Registry must be stopped if the
                // container is not running in a stand-alone topology).
                this.petalsCompositeController.onConfigurationUpdated();
            } catch (final PetalsException e) {
                // these are PetalsException (and other) thrown by other functions we rely on
                this.log.warning("An error occured while attaching the container to another domain", e);
                throw new AttachContainerException(domainName, currentContainerName, targetContainerName,
                        "An error occurs attaching the container to another domain", e);
            } catch (final Exception e) {
                this.log.warning("Error while re-starting component", e);
            }

            // starting the endpoint directory will resume new exchanges from us!
        } finally {
            Thread.currentThread().setContextClassLoader(oldClassloader);
        }
    }

    /**
     * <p>
     * Check that the container attachment can be realized. During checks, the target domain configuration is retrieved.
     * Checks applied are:
     * <ul>
     * <li>the domain of the current container is 'dynamic'</li>
     * <li>the given container exists in the target topology</li>
     * <li>the target domain mode is 'dynamic'</li>
     * <li>the registry implementation of the target domain must be the same than the one of the current domain</li>
     * </ul>
     * </p>
     * 
     * @param petalsJmxClient
     *            A JMX client to the given target container
     * @param targetContainer
     *            The given container member of the target domain. Not <code>null</code>, and all required fields have
     *            been previously checked
     * @param securityPassPhrase
     *            The security pass-phrase of the given container member of the target domain
     * @return The target domain configuration filled during checks
     * @throws PetalsException
     */
    private DomainConfiguration checkIfAttachmentIsPossible(final ContainerConfiguration targetContainer,
            final String securityPassPhrase)
            throws DomainIsNotDynamicException, InconsistencyException, AttachContainerException {

        // check if the domain of the current container is dynamic
        if (!this.isCurrentDomainDynamic()) {
            throw new DomainIsNotDynamicException();
        }

        final DomainConfiguration targetDomainConf = new DomainConfiguration();

        final String currentContainerName = this.configurationService.getContainerConfiguration().getName();
        final String targetContainerName = targetContainer.getName();
        final String targetDomainName = targetContainer.getDomainName();

        final Set<Map<String, String>> targetTopologyMaps;
        final String sharedAreaImplName;
        try {
            // Create a JMX connection to the container of the domain in which the current container must be
            // attached. The service 'JMXService' cannot be used because the container of the domain to join is not a
            // member of the initial topology of the current container.
            final JMXClient petalsJmxClient = PetalsJmxApiFactory.getInstance().createJMXClient(
                    targetContainer.getHost(),
                    targetContainer.getJmxRMIConnectorPort() == 0 ? JMXClient.DEFAULT_PORT
                            : targetContainer.getJmxRMIConnectorPort(),
                    targetContainer.getUser(), targetContainer.getPassword());
            try {
                final TopologyServiceClient topologyServiceClient = petalsJmxClient.getTopologyServiceClient();
                targetTopologyMaps = topologyServiceClient.retrieveTopology(securityPassPhrase, true);
                sharedAreaImplName = topologyServiceClient.retrieveSharedAreaCltImplementation();
            } catch (final PetalsException e) {
                throw new AttachContainerException(targetDomainName, currentContainerName, targetContainerName,
                        "Can't retrieve topology informations from container used to move", e);
            } finally {
                try {
                    petalsJmxClient.disconnect();
                } catch (final ConnectionErrorException e) {
                    this.log.warning("Error while disconnecting from target container", e);
                }
            }
        } catch (final ConnectionErrorException | DuplicatedServiceException | MissingServiceException e) {
            this.log.warning("An error occurs connecting to the container used to move using JMX", e);
            throw new AttachContainerException(targetContainer.getDomainName(),
                    this.configurationService.getContainerConfiguration().getName(), targetContainer.getName(),
                    "An error occurs connecting to the container used to move using JMX", e);
        }

        if (this.log.isDebugEnabled()) {
            this.log.debug("Target topology maps: " + targetTopologyMaps);
        }

        boolean containerFound = false;
        boolean domainFound = false;
        final Properties domainProperties = new Properties();
        for (final Map<String, String> targetTopologyMap : targetTopologyMaps) {
            final String topologyType = targetTopologyMap.get(TopologyItem.TOPOLOGY_TYPE);
            if (topologyType.equals(TopologyItem.CONTAINER)) {
                if (targetContainerName.equals(targetTopologyMap.get(ContainerPropertyName.CONF_CONTAINER_NAME))) {
                    containerFound = true;
                }
            } else if (topologyType.equals(TopologyItem.DOMAIN)) {
                domainProperties.putAll(targetTopologyMap);
                targetDomainConf.setName(targetTopologyMap.get(DomainPropertyName.CONF_DOMAIN_NAME));
                targetDomainConf.setDescription(targetTopologyMap.get(DomainPropertyName.CONF_DOMAIN_DESCRIPTION));
                if (targetDomainName.equals(targetTopologyMap.get(DomainPropertyName.CONF_DOMAIN_NAME))) {
                    domainFound = true;
                    final String mode = targetTopologyMap.get(DomainPropertyName.CONF_DOMAIN_MODE);
                    if (DomainPropertyValues.MODE_STATIC.equals(mode)) {
                        targetDomainConf.setMode(DomainMode.STATIC);
                    } else {
                        targetDomainConf.setMode(DomainMode.DYNAMIC);
                    }
                    final String targetRegistryImpl = targetTopologyMap
                            .get(DomainPropertyName.CONF_SHARED_AREA_IMPLEMENTATION);
                    targetDomainConf.setSharedAreaImplementation(
                            targetRegistryImpl != null && targetRegistryImpl.isEmpty() ? null : targetRegistryImpl);
                }
            }
        }
        if (!containerFound) {
            throw new InconsistentContainerNameException(targetContainerName);
        }
        if (!targetDomainName.equals(targetDomainConf.getName())) {
            throw new InconsistentDomainNameException(targetDomainConf.getName(), targetDomainName);
        }
        if (!domainFound) {
            // This exception should never happen because a container of the target topology is a member of the expected
            // domain. So if this error occurs, an inconsistency exists on the remote container about domain that
            // should be due to a bug in configuration loading of the remote container.
            throw new InconsistencyDomainException("Unable to retrieve the target domain in the target topology");
        }
        if (!targetDomainConf.getMode().equals(DomainMode.DYNAMIC)) {
            throw new DomainIsNotDynamicException(targetDomainConf.getName());
        }

        // Check the registry implementation consistency
        final DomainConfiguration currentDomainConfiguration = this.configurationService.getDomainConfiguration();
        final String currentRegistryImpl = currentDomainConfiguration.getSharedAreaImplementation();
        final String targetRegistryImpl = targetDomainConf.getSharedAreaImplementation();
        if ((currentRegistryImpl == null && targetRegistryImpl != null)
                || (currentRegistryImpl != null && targetRegistryImpl == null)
                || (currentRegistryImpl != null && targetRegistryImpl != null && !currentRegistryImpl
                        .equals(targetRegistryImpl))) {
            throw new InconsistentDomainRegistryImplException(currentRegistryImpl, targetRegistryImpl);
        }

        final Document topologyPartCfg;
        try {
            // Check that the shared area client implementation is in the classloader
            // Note: as DomainConfiguration.getSharedAreaImplementation() returns null for default implementation loaded
            // from classloader, we must use the JMX API to get the shared area client implementation really used by the
            // domain to join
            final Class<?> sharedAreaImpl = this.getClass().getClassLoader().loadClass(sharedAreaImplName);
            final MutableImplementation sharedAreaImplInstance = (MutableImplementation) sharedAreaImpl.newInstance();
            topologyPartCfg = sharedAreaImplInstance.getTopologyPartConfiguration(domainProperties);
        } catch (final ClassNotFoundException e) {
            throw new AttachContainerException(targetDomainName, currentContainerName, targetContainerName,
                    "Shared area client implementation '" + sharedAreaImplName + "' not found in classloader", e);
        } catch (final InstantiationException | IllegalAccessException | ExceptionInInitializerError
                | SecurityException e) {
            throw new AttachContainerException(targetDomainName, currentContainerName, targetContainerName,
                    "Can't instantiate shared area client implementation '" + sharedAreaImplName + "'", e);
        } catch (final MutableImplException e) {
            throw new AttachContainerException(targetDomainName, currentContainerName, targetContainerName,
                    "Can't retrieve the shared area client implemantation specificy configuration", e);
        }

        // Set the shared area client implementation configuration
        final List<Object> extraParameters = new ArrayList<>();
        if (topologyPartCfg != null) {
            extraParameters.add(topologyPartCfg.getDocumentElement());
        }
        targetDomainConf.setExtraParameters(extraParameters);

        // normally, the configuration service must have already checked that the host and the listen interface are
        // compatible so now we just need to make sure that other containers will be able to contact us.
        final String containerHost = this.configurationService.getContainerConfiguration().getHost();
        boolean myHostIsLoopback = false;
        try {
            myHostIsLoopback = InetAddress.getByName(containerHost).isLoopbackAddress();
        } catch (final UnknownHostException e) {
            this.log.warning("The current container's hostname is not resolvable (" + containerHost
                    + "), other containers may not be able to contact it!", e);
        }
        boolean targetHostIsLoopback = false;
        try {
            targetHostIsLoopback = InetAddress.getByName(targetContainer.getHost()).isLoopbackAddress();
        } catch (final UnknownHostException e) {
            // normally this shouldn't happen since we already created the jmx client...
            this.log.warning("The target container's hostname (" + targetContainer.getHost()
                    + ") is not resolvable, we may have problems to attach to the domain through it.", e);
        }

        if (myHostIsLoopback && !targetHostIsLoopback) {
            this.log.warning("The current container is contactable through the loopback address (" + containerHost
                    + ") but the target container used for attachment seems to be on an external network address ("
                    + targetContainer.getHost() + "), other containes may have problems to contact us!");
        }

        return targetDomainConf;
    }

    /**
     * <p>
     * Check that the container detachment can be realized. Checks applied are:
     * <ul>
     * <li>the domain of the current container is 'dynamic'.</li>
     * </ul>
     * </p>
     * 
     * @throws DomainIsNotDynamicException
     *             The current domain is not <b>dynamic</b>.
     */
    private void checkIfDetachmentIsPossible() throws DomainIsNotDynamicException {
        // check if the domain of the current container is dynamic
        if (!this.isCurrentDomainDynamic()) {
            throw new DomainIsNotDynamicException();
        }
    }

    @Override
    public String retrieveSharedAreaCltImplementation() throws PetalsException {
        return this.petalsCompositeController.getSharedAreaCltImplementation();
    }
}
