/**
 * Copyright (c) 2005-2012 EBM WebSourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.container;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.logging.Logger;

import javax.jbi.JBIException;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.Interface;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.fractal.api.control.ContentController;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalContentException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.api.factory.InstantiationException;
import org.objectweb.fractal.fraclet.annotations.Lifecycle;
import org.objectweb.fractal.fraclet.extensions.Controller;
import org.objectweb.fractal.fraclet.types.Step;
import org.objectweb.fractal.util.Fractal;
import org.ow2.petals.basisapi.exception.PetalsException;
import org.ow2.petals.jbi.descriptor.original.generated.Jbi.SharedLibrary;
import org.ow2.petals.microkernel.api.container.ComponentLifeCycle;
import org.ow2.petals.microkernel.api.container.ContainerService;
import org.ow2.petals.microkernel.api.container.Installer;
import org.ow2.petals.microkernel.api.container.ServiceAssemblyLifeCycle;
import org.ow2.petals.microkernel.api.container.SharedLibraryBean;
import org.ow2.petals.microkernel.api.container.SharedLibraryLifeCycle;
import org.ow2.petals.microkernel.api.server.Binding;
import org.ow2.petals.microkernel.api.server.FractalHelper;
import org.ow2.petals.microkernel.container.exception.ContainerCtrlException;
import org.ow2.petals.microkernel.container.lifecycle.ComponentLifeCycleImpl;
import org.ow2.petals.microkernel.container.lifecycle.InstallerImpl;
import org.ow2.petals.microkernel.container.lifecycle.ServiceAssemblyLifeCycleImpl;
import org.ow2.petals.microkernel.container.lifecycle.SharedLibraryLifeCycleImpl;

import com.ebmwebsourcing.easycommons.log.LoggingUtil;

/**
 * The implementation of {@link ContainerController} that manages Fractal components associated to the JBI artifacts.
 * 
 * TODO move all the constants XXX_ITF to the respective component's interfaces
 * 
 * @author Nicolas SALATGE - EBM WebSourcing
 * @author Christophe DENEUX - Linagora
 */
@org.objectweb.fractal.fraclet.annotations.Component(provides = @org.objectweb.fractal.fraclet.annotations.Interface(name = "service", signature = ContainerController.class))
public class ContainerControllerImpl implements ContainerController {

    /**
     * Fractal name of the container service.
     */
    private static final String FRACTAL_CONTAINER_SERVICE_NAME = "ContainerServiceImpl";

    /**
     * Name of the client interface 'configuration' of the composite {@link FractalHelper#CONTAINER_COMPOSITE}
     */
    private static final String CONTAINER_CONFIGURATION_ITF = "configuration";

    /**
     * Name of the client interface 'configuration' of the Fractal component associated to the JBI component installer (
     * {@link ContainerController#PREFIX_INSTALLER_NAME})
     */
    private static final String INSTALLER_CONFIGURATION_ITF = "configuration";

    /**
     * Name of the client interface 'admin' of the composite {@link FractalHelper#CONTAINER_COMPOSITE}
     */
    private static final String CONTAINER_ADMIN_ITF = "admin";

    /**
     * Name of the client interface 'admin' of the Fractal component associated to the JBI component installer (
     * {@link ContainerController#PREFIX_INSTALLER_NAME})
     */
    private static final String INSTALLER_ADMIN_ITF = "admin";

    /**
     * Name of the client interface 'container' of the Fractal component associated to the JBI component life cycle (
     * {@link ContainerController#PREFIX_COMPONENT_LIFE_CYCLE_NAME})
     */
    private static final String COMPLC_CONTAINER_ITF = "container";

    /**
     * Name of the client interface 'classloader' of the composite {@link FractalHelper#CONTAINER_COMPOSITE}
     */
    private static final String CONTAINER_CLASSLOADER_ITF = "classloader";

    /**
     * Name of the client interface 'classloader' of the Fractal component associated to the JBI component installer (
     * {@link ContainerController#PREFIX_INSTALLER_NAME})
     */
    private static final String INSTALLER_CLASSLOADER_ITF = "classloader";

    /**
     * Name of the client interface 'classloader' of the Fractal component associated to the SA life cycle (
     * {@link ContainerController#PREFIX_SHARED_LIBRARY_LIFE_CYCLE_NAME})
     */
    private static final String SLLC_CLASSLOADER_ITF = "classloader";

    /**
     * Name of the client interface 'router' of the composite {@link FractalHelper#CONTAINER_COMPOSITE}
     */
    private static final String CONTAINER_ROUTER_ITF = "router";

    /**
     * Name of the client interface 'router' of the Fractal component associated to the JBI component installer (
     * {@link ContainerController#PREFIX_INSTALLER_NAME})
     */
    private static final String INSTALLER_ROUTER_ITF = "router";

    /**
     * Name of the client interface 'repository' of the composite {@link FractalHelper#CONTAINER_COMPOSITE}
     */
    private static final String CONTAINER_REPOSITORY_ITF = "repository";

    /**
     * Name of the client interface 'repository' of the Fractal component associated to the JBI component installer (
     * {@link ContainerController#PREFIX_INSTALLER_NAME})
     */
    private static final String INSTALLER_REPOSITORY_ITF = "repository";

    /**
     * Name of the client interface 'endpointDirectory' of the composite {@link FractalHelper#CONTAINER_COMPOSITE}
     */
    private static final String CONTAINER_ENDPOINT_ITF = "endpointDirectory";

    /**
     * Name of the client interface 'endpointDirectory' of the Fractal component associated to the JBI component
     * installer ( {@link ContainerController#PREFIX_INSTALLER_NAME})
     */
    private static final String INSTALLER_ENDPOINT_DIRECTORY_ITF = "endpointDirectory";

    /**
     * Name of the client interface 'jndi' of the composite {@link FractalHelper#CONTAINER_COMPOSITE}
     */
    private static final String CONTAINER_JNDI_ITF = "jndi";

    /**
     * Name of the client interface 'jndi' of the Fractal component associated to the JBI component installer (
     * {@link ContainerController#PREFIX_INSTALLER_NAME})
     */
    private static final String INSTALLER_JNDI_ITF = "jndi";

    /**
     * Name of the client interface 'systemstate' of the composite {@link FractalHelper#CONTAINER_COMPOSITE}
     */
    private static final String CONTAINER_SYSTEMSTATE_ITF = "systemstate";

    /**
     * Name of the client interface 'systemstate' of the Fractal component associated to the JBI component installer (
     * {@link ContainerController#PREFIX_INSTALLER_NAME})
     */
    private static final String INSTALLER_SYSTEMSTATE_ITF = "systemstate";

    /**
     * Name of the client interface 'systemstate' of the Fractal component associated to the JBI component life cycle (
     * {@link ContainerController#PREFIX_COMPONENT_LIFE_CYCLE_NAME})
     */
    private static final String COMPLC_SYSTEMSTATE_ITF = "systemstate";

    /**
     * Name of the client interface 'installer' of the Fractal component associated to the JBI component life cycle (
     * {@link ContainerController#PREFIX_COMPONENT_LIFE_CYCLE_NAME})
     */
    private static final String COMPLC_INSTALLER_ITF = "installer";

    /**
     * Name of the client interface 'systemstate' of the Fractal component associated to the SA life cycle (
     * {@link ContainerController#PREFIX_SERVICE_ASSEMBLY_LIFE_CYCLE_NAME})
     */
    private static final String SALC_SYSTEMSTATE_ITF = "systemstate";

    /**
     * Name of the client interface 'containerService' of the Fractal component associated to the JBI component
     * installer ( {@link ContainerController#PREFIX_INSTALLER_NAME})
     */
    private static final String INSTALLER_CONTAINER_ITF = "container";

    /**
     * the log.
     */
    private final LoggingUtil log = new LoggingUtil(Logger.getLogger(FRACTAL_COMPONENT_LOGGER_NAME));

    /**
     * The component.
     */
    @Controller
    private Component comp;

    /**
     * The parent composite {@link FractalHelper#CONTAINER_COMPOSITE}
     */
    private Component containerComposite;

    /**
     * The content controller of the composite {@link FractalHelper#CONTAINER_COMPOSITE}
     */
    private ContentController containerCompositeCC;

    /**
     * The Fractal component 'Container Service'
     */
    private Component containerServiceComponent;

    /**
     * The binding controller of the Fractal component 'Container Service'
     */
    private BindingController containerServiceComponentBC;

    @Override
    public Installer createInstaller(
            final org.ow2.petals.jbi.descriptor.original.generated.Component componentDescription)
            throws ContainerCtrlException {

        try {
            // Creation of the installer
            final Component installerComponent = FractalHelper.createPrimitiveComponent(InstallerImpl.class);

            // Creation of the list of bindings of the component with others components
            final List<Binding> listOfBindings = new ArrayList<Binding>();
            // TODO is that really needed to do such synchronization?!
            synchronized (this.containerCompositeCC) {
                listOfBindings.add(new Binding(INSTALLER_CONFIGURATION_ITF, (Interface) this.containerCompositeCC
                        .getFcInternalInterface(CONTAINER_CONFIGURATION_ITF)));
                listOfBindings.add(new Binding(INSTALLER_ADMIN_ITF, (Interface) this.containerCompositeCC
                        .getFcInternalInterface(CONTAINER_ADMIN_ITF)));
                listOfBindings.add(new Binding(INSTALLER_CLASSLOADER_ITF, (Interface) this.containerCompositeCC
                        .getFcInternalInterface(CONTAINER_CLASSLOADER_ITF)));
                listOfBindings.add(new Binding(INSTALLER_ROUTER_ITF, (Interface) this.containerCompositeCC
                        .getFcInternalInterface(CONTAINER_ROUTER_ITF)));
                listOfBindings.add(new Binding(INSTALLER_REPOSITORY_ITF, (Interface) this.containerCompositeCC
                        .getFcInternalInterface(CONTAINER_REPOSITORY_ITF)));
                listOfBindings.add(new Binding(INSTALLER_ENDPOINT_DIRECTORY_ITF, (Interface) this.containerCompositeCC
                        .getFcInternalInterface(CONTAINER_ENDPOINT_ITF)));
                listOfBindings.add(new Binding(INSTALLER_JNDI_ITF, (Interface) this.containerCompositeCC
                        .getFcInternalInterface(CONTAINER_JNDI_ITF)));
                listOfBindings.add(new Binding(INSTALLER_SYSTEMSTATE_ITF, (Interface) this.containerCompositeCC
                        .getFcInternalInterface(CONTAINER_SYSTEMSTATE_ITF)));
                listOfBindings.add(new Binding(INSTALLER_CONTAINER_ITF, (Interface) this.containerServiceComponent
                        .getFcInterface(ContainerService.FRACTAL_SRV_ITF_NAME)));
            }

            // Add the new installer in the composite 'Container' ...
            final String name = componentDescription.getIdentification().getName();

            FractalHelper.addComponent(installerComponent, this.containerComposite, listOfBindings,
                    PREFIX_INSTALLER_NAME + name);

            final Installer installerItf = (Installer) installerComponent
                    .getFcInterface(Installer.FRACTAL_SRV_ITF_NAME);

            // ... and bind it to the container service
            this.containerServiceComponentBC.bindFc(ContainerService.INSTALLERS_ITF_PREFIX + name, installerItf);

            FractalHelper.startComponent(installerComponent);

            return installerItf;

        } catch (final InstantiationException e) {
            throw new ContainerCtrlException(e);
        } catch (final NoSuchInterfaceException e) {
            throw new ContainerCtrlException(e);
        } catch (final IllegalContentException e) {
            throw new ContainerCtrlException(e);
        } catch (final IllegalLifeCycleException e) {
            throw new ContainerCtrlException(e);
        } catch (final IllegalBindingException e) {
            throw new ContainerCtrlException(e);
        }
    }

    @Override
    public ComponentLifeCycle createComponentLifeCycle(
            final org.ow2.petals.jbi.descriptor.original.generated.Component componentDescription)
            throws ContainerCtrlException {

        try {
            // Creation of the component life cycle
            final Component componentLifeCycle = FractalHelper.createPrimitiveComponent(ComponentLifeCycleImpl.class);

            final String name = componentDescription.getIdentification().getName();

            final Component installer = FractalHelper.getRecursiveComponentByName(this.containerCompositeCC,
                    PREFIX_INSTALLER_NAME + name);

            // Creation of the list of bindings of the component with others components
            final List<Binding> listOfBindings = new ArrayList<Binding>();
            // TODO is that really needed to do such synchronization?!
            synchronized (this.containerCompositeCC) {
                listOfBindings.add(new Binding(COMPLC_CONTAINER_ITF, (Interface) (Interface) this.containerServiceComponent
                        .getFcInterface(ContainerService.FRACTAL_SRV_ITF_NAME)));
                listOfBindings.add(new Binding(COMPLC_SYSTEMSTATE_ITF, (Interface) this.containerCompositeCC
                        .getFcInternalInterface(CONTAINER_SYSTEMSTATE_ITF)));
                listOfBindings.add(new Binding(COMPLC_INSTALLER_ITF,
                        (Interface) installer.getFcInterface(Installer.FRACTAL_SRV_ITF_NAME)));
            }

            // Add the new life cycle in the composite 'Container' ...
            FractalHelper.addComponent(componentLifeCycle, this.containerComposite, listOfBindings,
                    PREFIX_COMPONENT_LIFE_CYCLE_NAME + name);

            final ComponentLifeCycle clcItf = (ComponentLifeCycle) componentLifeCycle
                    .getFcInterface(ComponentLifeCycle.FRACTAL_SRV_ITF_NAME);

            // ... and bind it to the container service
            this.containerServiceComponentBC.bindFc(ContainerService.COMPONENT_LIFECYCLES_ITF_PREFIX + name, clcItf);

            FractalHelper.startComponent(componentLifeCycle);
            
            return clcItf;

        } catch (final InstantiationException e) {
            throw new ContainerCtrlException(e);
        } catch (final NoSuchInterfaceException e) {
            throw new ContainerCtrlException(e);
        } catch (final IllegalContentException e) {
            throw new ContainerCtrlException(e);
        } catch (final IllegalLifeCycleException e) {
            throw new ContainerCtrlException(e);
        } catch (final IllegalBindingException e) {
            throw new ContainerCtrlException(e);
        }
    }

    @Override
    public void removeComponentLifeCycle(final String name) throws ContainerCtrlException {
        removeComponent(name, PREFIX_COMPONENT_LIFE_CYCLE_NAME, ContainerService.COMPONENT_LIFECYCLES_ITF_PREFIX);
    }

    private void removeComponent(final String name, final String prefixComponentName, final String prefixBindingName)
            throws ContainerCtrlException {
        try {

            // Get the component
            final Component fractalComponent = FractalHelper.getRecursiveComponentByName(this.containerCompositeCC,
                    prefixComponentName + name);

            // Stop the Fractal component
            FractalHelper.stopComponent(fractalComponent);

            // Remove of all bind
            unbindAllFc(fractalComponent);

            this.containerServiceComponentBC.unbindFc(prefixBindingName + name);

            // Remove the component
            this.containerCompositeCC.removeFcSubComponent(fractalComponent);

        } catch (final NoSuchInterfaceException e) {
            throw new ContainerCtrlException(e);
        } catch (final IllegalContentException e) {
            throw new ContainerCtrlException(e);
        } catch (final IllegalLifeCycleException e) {
            throw new ContainerCtrlException(e);
        } catch (final IllegalBindingException e) {
            throw new ContainerCtrlException(e);
        }
    }

    @Override
    public void removeInstaller(final String name) throws ContainerCtrlException {
        removeComponent(name, PREFIX_INSTALLER_NAME, ContainerService.INSTALLERS_ITF_PREFIX);
    }

    @Override
    public ServiceAssemblyLifeCycle createServiceAssemblyLifeCycle(
            final org.ow2.petals.jbi.descriptor.original.generated.ServiceAssembly sa) throws ContainerCtrlException {

        try {

            // Creation of the service assembly life cycle
            final Component serviceAssemblyLifeCycle = FractalHelper
                    .createPrimitiveComponent(ServiceAssemblyLifeCycleImpl.class);

            // Creation of the list of bindings of the component with others components
            final List<Binding> listOfBindings = new ArrayList<Binding>();
            // TODO is that really needed to do such synchronization?!
            synchronized (this.containerCompositeCC) {
                listOfBindings.add(new Binding(SALC_SYSTEMSTATE_ITF,
                        (Interface) this.containerCompositeCC.getFcInternalInterface(CONTAINER_SYSTEMSTATE_ITF)));
            }

            final String name = sa.getIdentification().getName();

            // Add the new life cycle in the framework
            FractalHelper.addComponent(serviceAssemblyLifeCycle, this.containerComposite, listOfBindings,
                    PREFIX_SERVICE_ASSEMBLY_LIFE_CYCLE_NAME + name);

            final ServiceAssemblyLifeCycle saLcItf = (ServiceAssemblyLifeCycle) serviceAssemblyLifeCycle
                    .getFcInterface(ServiceAssemblyLifeCycle.FRACTAL_SRV_ITF_NAME);

            // ... and bind it to the container service
            this.containerServiceComponentBC.bindFc(ContainerService.SA_LIFECYCLES_ITF_PREFIX + name, saLcItf);

            FractalHelper.startComponent(serviceAssemblyLifeCycle);

            return saLcItf;

        } catch (final InstantiationException e) {
            throw new ContainerCtrlException(e);
        } catch (final NoSuchInterfaceException e) {
            throw new ContainerCtrlException(e);
        } catch (final IllegalContentException e) {
            throw new ContainerCtrlException(e);
        } catch (final IllegalLifeCycleException e) {
            throw new ContainerCtrlException(e);
        } catch (final IllegalBindingException e) {
            throw new ContainerCtrlException(e);
        }
    }

    @Override
    public void removeServiceAssemblyLifeCycle(final String name) throws ContainerCtrlException {
        removeComponent(name, PREFIX_SERVICE_ASSEMBLY_LIFE_CYCLE_NAME, ContainerService.SA_LIFECYCLES_ITF_PREFIX);
    }

    @Override
    public SharedLibraryLifeCycle createSharedLibraryLifeCycle(final SharedLibrary sl) throws ContainerCtrlException {

        try {
            // Creation of the shared library life cycle
            final Component sharedLibraryLifeCycle = FractalHelper
                    .createPrimitiveComponent(SharedLibraryLifeCycleImpl.class);

            // Creation of the list of bindings of the component with others components
            final List<Binding> listOfBindings = new ArrayList<Binding>();
            // TODO is that really needed to do such synchronization?!
            synchronized (this.containerCompositeCC) {
                listOfBindings.add(new Binding(SLLC_CLASSLOADER_ITF, (Interface) this.containerCompositeCC
                        .getFcInternalInterface(CONTAINER_CLASSLOADER_ITF)));
            }

            // Add the new shared library life cycle in the framework
            final String name = SharedLibraryBean.getSharedLibraryName(sl.getIdentification().getName(),
                    sl.getVersion());

            FractalHelper.addComponent(sharedLibraryLifeCycle, this.containerComposite, listOfBindings,
                    PREFIX_SHARED_LIBRARY_LIFE_CYCLE_NAME + name);

            final SharedLibraryLifeCycle slLcItf = (SharedLibraryLifeCycle) sharedLibraryLifeCycle
                    .getFcInterface(SharedLibraryLifeCycle.FRACTAL_SRV_ITF_NAME);

            // ... and bind it to the container service
            this.containerServiceComponentBC.bindFc(ContainerService.SL_LIFECYCLES_ITF_PREFIX + name, slLcItf);

            FractalHelper.startComponent(sharedLibraryLifeCycle);

            return slLcItf;

        } catch (final InstantiationException e) {
            throw new ContainerCtrlException(e);
        } catch (final NoSuchInterfaceException e) {
            throw new ContainerCtrlException(e);
        } catch (final IllegalContentException e) {
            throw new ContainerCtrlException(e);
        } catch (final IllegalLifeCycleException e) {
            throw new ContainerCtrlException(e);
        } catch (final IllegalBindingException e) {
            throw new ContainerCtrlException(e);
        }
    }

    @Override
    public void removeSharedLibraryLifeCycle(final String slName, final String slVersion) throws ContainerCtrlException {
        final String sharedLibraryName = SharedLibraryBean.getSharedLibraryName(slName, slVersion);
        removeComponent(sharedLibraryName, PREFIX_SHARED_LIBRARY_LIFE_CYCLE_NAME,
                ContainerService.SL_LIFECYCLES_ITF_PREFIX);
    }

    /**
     * Initialization of the petals component.
     * 
     * @throws PetalsException
     *             : impossible to start the component
     * @throws JBIException
     *             : JBI error
     * @throws NoSuchInterfaceException
     *             : Fractal Error
     * @throws ADLException
     *             : Fractal error
     * 
     */
    @Lifecycle(step = Step.START)
    public void start() throws PetalsException, NoSuchInterfaceException {
        this.log.start();

        try {
            this.containerComposite = FractalHelper.getParentComponent(this.comp);
            this.containerCompositeCC = Fractal.getContentController(this.containerComposite);
            this.containerServiceComponent = FractalHelper.getComponentByName(this.containerCompositeCC,
                    FRACTAL_CONTAINER_SERVICE_NAME);
            this.containerServiceComponentBC = Fractal.getBindingController(this.containerServiceComponent);

        } catch (final PetalsException e) {
            this.log.error("Failed to start Container service", e);
            throw e;
        } catch (final NoSuchInterfaceException e) {
            this.log.error("Failed to start Container service", e);
            throw e;
        } finally {
            this.log.end();
        }
    }

    /**
     * Stop of the petals component.
     * 
     */
    @Lifecycle(step = Step.STOP)
    public void stop() {
        this.log.call();
    }

    @Override
    public void stopAllJbiArtefacts() throws ContainerCtrlException {

        try {

            final List<Component> sas = FractalHelper.getComponentListByPrefix(containerCompositeCC,
                    ContainerController.PREFIX_SERVICE_ASSEMBLY_LIFE_CYCLE_NAME);
            Collections.reverse(sas);
            for (final Component component : sas) {
                FractalHelper.stopComponent(component);
            }
            final List<Component> components = FractalHelper.getComponentListByPrefix(this.containerCompositeCC,
                    ContainerController.PREFIX_COMPONENT_LIFE_CYCLE_NAME);
            Collections.reverse(components);
            for (final Component component : components) {
                FractalHelper.stopComponent(component);
            }

            final List<Component> installers = FractalHelper.getComponentListByPrefix(containerCompositeCC,
                    ContainerController.PREFIX_INSTALLER_NAME);
            Collections.reverse(installers);
            for (final Component component : installers) {
                FractalHelper.stopComponent(component);
            }

            final List<Component> sls = FractalHelper.getComponentListByPrefix(containerCompositeCC,
                    ContainerController.PREFIX_SHARED_LIBRARY_LIFE_CYCLE_NAME);
            Collections.reverse(sls);
            for (final Component component : sls) {
                FractalHelper.stopComponent(component);
            }

        } catch (final NoSuchInterfaceException e) {
            throw new ContainerCtrlException(e);
        } catch (final IllegalLifeCycleException e) {
            throw new ContainerCtrlException(e);
        }
    }

    @Override
    public void startAllJbiArtefacts() throws ContainerCtrlException {

        try {

            final List<Component> sls = FractalHelper.getComponentListByPrefix(containerCompositeCC,
                    ContainerController.PREFIX_SHARED_LIBRARY_LIFE_CYCLE_NAME);
            for (final Component component : sls) {
                FractalHelper.startComponent(component);
            }

            final List<Component> installers = FractalHelper.getComponentListByPrefix(containerCompositeCC,
                    ContainerController.PREFIX_INSTALLER_NAME);
            for (final Component component : installers) {
                FractalHelper.startComponent(component);
            }

            final List<Component> components = FractalHelper.getComponentListByPrefix(this.containerCompositeCC,
                    ContainerController.PREFIX_COMPONENT_LIFE_CYCLE_NAME);
            for (final Component component : components) {
                FractalHelper.startComponent(component);
            }

            final List<Component> sas = FractalHelper.getComponentListByPrefix(containerCompositeCC,
                    ContainerController.PREFIX_SERVICE_ASSEMBLY_LIFE_CYCLE_NAME);
            for (final Component component : sas) {
                FractalHelper.startComponent(component);
            }

        } catch (final NoSuchInterfaceException e) {
            throw new ContainerCtrlException(e);
        } catch (final IllegalLifeCycleException e) {
            throw new ContainerCtrlException(e);
        }
    }

    private void unbindAllFc(Component component) throws ContainerCtrlException {
        try {
            final BindingController lifeCycleBindingController = Fractal.getBindingController(component);
            for (String fc : lifeCycleBindingController.listFc()) {
                lifeCycleBindingController.unbindFc(fc);
            }
        } catch (final NoSuchInterfaceException e) {
            throw new ContainerCtrlException(e);
        } catch (final IllegalLifeCycleException e) {
            throw new ContainerCtrlException(e);
        } catch (final IllegalBindingException e) {
            throw new ContainerCtrlException(e);
        }
    }
}
