/**
 * Copyright (c) 2005-2012 EBM WebSourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.container;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.concurrent.ConcurrentHashMap;
import java.util.logging.Logger;

import javax.jbi.JBIException;
import javax.jbi.management.ComponentLifeCycleMBean;
import javax.jbi.management.InstallerMBean;
import javax.management.InstanceAlreadyExistsException;
import javax.management.InstanceNotFoundException;
import javax.management.MBeanException;
import javax.management.MBeanRegistrationException;
import javax.management.MBeanServer;
import javax.management.NotCompliantMBeanException;
import javax.management.ObjectName;
import javax.management.RuntimeOperationsException;
import javax.management.modelmbean.InvalidTargetObjectTypeException;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.fraclet.annotations.Lifecycle;
import org.objectweb.fractal.fraclet.annotations.Requires;
import org.objectweb.fractal.fraclet.extensions.Controller;
import org.objectweb.fractal.fraclet.types.Cardinality;
import org.objectweb.fractal.fraclet.types.Contingency;
import org.objectweb.fractal.fraclet.types.Step;
import org.ow2.petals.basisapi.exception.PetalsException;
import org.ow2.petals.jbi.descriptor.original.generated.ComponentType;
import org.ow2.petals.jbi.descriptor.original.generated.Jbi.SharedLibrary;
import org.ow2.petals.microkernel.api.communication.JMXService;
import org.ow2.petals.microkernel.api.container.ComponentLifeCycle;
import org.ow2.petals.microkernel.api.container.ContainerService;
import org.ow2.petals.microkernel.api.container.Installer;
import org.ow2.petals.microkernel.api.container.ServiceAssemblyLifeCycle;
import org.ow2.petals.microkernel.api.container.ServiceUnitLifeCycle;
import org.ow2.petals.microkernel.api.container.SharedLibraryBean;
import org.ow2.petals.microkernel.api.container.SharedLibraryLifeCycle;
import org.ow2.petals.microkernel.api.container.exception.ComponentLifecycleNotFoundException;
import org.ow2.petals.microkernel.api.container.exception.ContainerServiceException;
import org.ow2.petals.microkernel.api.container.exception.InstallerNotFoundException;
import org.ow2.petals.microkernel.api.container.exception.SharedLibraryLifecycleNotFoundException;
import org.ow2.petals.microkernel.api.server.FractalHelper;
import org.ow2.petals.microkernel.container.exception.ContainerCtrlException;
import org.ow2.petals.microkernel.jbi.management.MBeanNameBuilder;
import org.ow2.petals.microkernel.server.MBeanHelper;

import com.ebmwebsourcing.easycommons.log.LoggingUtil;
import com.google.common.base.Predicate;
import com.google.common.collect.Maps;

/**
 * This {@link ContainerServiceImpl} implementation allow to create and handle JBI containers.
 * @author Nicolas SALATGE - EBM WebSourcing
 */
@org.objectweb.fractal.fraclet.annotations.Component(provides = @org.objectweb.fractal.fraclet.annotations.Interface(name = ContainerService.FRACTAL_SRV_ITF_NAME, signature = ContainerService.class))
public class ContainerServiceImpl implements ContainerService {

    /**
     * the log.
     */
    private final LoggingUtil log = new LoggingUtil(Logger.getLogger(FRACTAL_COMPONENT_LOGGER_NAME));

    /**
     * The container controller
     */
    @Requires(name = "containerController")
    private ContainerController containerController;

    /**
     * The JMX service
     */
    @Requires(name = "jmx")
    private JMXService jmxService;

    /**
     * The builder of MBean names
     */
    @Requires(name = "mbean-name-builder")
    private MBeanNameBuilder mbeanNameBuilder;

    /**
     * The MBean server
     */
    private MBeanServer mbeanServer;

    /**
     * The list of installModule Fractal components
     */
    @Requires(name = ContainerService.INSTALLERS_ITF_PREFIX, cardinality = Cardinality.COLLECTION, contingency = Contingency.OPTIONAL)
    private final Map<String, Installer> installers = new ConcurrentHashMap<String, Installer>();

    /**
     * The list of component life cycle Fractal components
     */
    @Requires(name = ContainerService.COMPONENT_LIFECYCLES_ITF_PREFIX, cardinality = Cardinality.COLLECTION, contingency = Contingency.OPTIONAL)
    private final Map<String, ComponentLifeCycle> componentLifecycles = new ConcurrentHashMap<String, ComponentLifeCycle>();

    /**
     * The service assembly life cycles.
     */
    @Requires(name = ContainerService.SA_LIFECYCLES_ITF_PREFIX, cardinality = Cardinality.COLLECTION, contingency = Contingency.OPTIONAL)
    private final Map<String, ServiceAssemblyLifeCycle> serviceAssemblyLifeCycles = new ConcurrentHashMap<String, ServiceAssemblyLifeCycle>();

    /**
     * The shared library life cycles.
     */
    @Requires(name = ContainerService.SL_LIFECYCLES_ITF_PREFIX, cardinality = Cardinality.COLLECTION, contingency = Contingency.OPTIONAL)
    private final Map<String, SharedLibraryLifeCycle> sharedLibraryLifeCycles = new ConcurrentHashMap<String, SharedLibraryLifeCycle>();

    /**
     * The component.
     */
    @Controller
    private Component comp;

    /**
     * The parent component.
     */
    private Component parentcontainer;

    @Override
    public Component getParentContainer() {
        return this.parentcontainer;
    }

    @Override
    public Installer createInstaller(
            final org.ow2.petals.jbi.descriptor.original.generated.Component componentDescription)
            throws ContainerServiceException {

        try {
            return this.containerController.createInstaller(componentDescription);
        } catch (final ContainerCtrlException e) {
            throw new ContainerServiceException(e);
        }
    }

    @Override
    public ObjectName registerInstallerMBean(final String name) throws InstallerNotFoundException,
            ContainerServiceException {

        try {
            final Installer installer = this.getInstallerByName(name);
            if (installer != null) {
                final ObjectName installerMbeanName = this.mbeanNameBuilder.createInstallerMBeanName(name);
                this.mbeanServer.registerMBean(MBeanHelper.createCustomMBean(installer, InstallerMBean.class),
                        installerMbeanName);
                return installerMbeanName;
            } else {
                throw new InstallerNotFoundException(name);
            }
        } catch (final RuntimeOperationsException e) {
            throw new ContainerServiceException(e);
        } catch (final InvalidTargetObjectTypeException e) {
            throw new ContainerServiceException(e);
        } catch (final InstanceAlreadyExistsException e) {
            throw new ContainerServiceException(e);
        } catch (final MBeanRegistrationException e) {
            throw new ContainerServiceException(e);
        } catch (final NotCompliantMBeanException e) {
            throw new ContainerServiceException(e);
        } catch (final InstanceNotFoundException e) {
            throw new ContainerServiceException(e);
        } catch (final MBeanException e) {
            throw new ContainerServiceException(e);
        }
    }

    @Override
    public void unregisterInstallerMBean(final String name) throws InstallerNotFoundException,
            ContainerServiceException {

        final ObjectName objectName = this.mbeanNameBuilder.createInstallerMBeanName(name);
        try {
            this.mbeanServer.unregisterMBean(objectName);
        } catch (final InstanceNotFoundException e) {
            throw new InstallerNotFoundException(name, e);
        } catch (final MBeanRegistrationException e) {
            throw new ContainerServiceException("Failed to unregistered the binding Component '" + name + "'", e);
        }
    }

    @Override
    public ComponentLifeCycle createComponentLifeCycle(
            final org.ow2.petals.jbi.descriptor.original.generated.Component componentDescription)
            throws ContainerServiceException {

        try {
            final ComponentLifeCycle clc = this.containerController.createComponentLifeCycle(componentDescription);
            return clc;
        } catch (final ContainerCtrlException e) {
            throw new ContainerServiceException(e);
        }
    }

    @Override
    public ObjectName registerComponentLifeCycleMBean(final String name) throws ComponentLifecycleNotFoundException,
            ContainerServiceException {

        try {
            final ComponentLifeCycle componentLifecycle = this.getComponentLifecycleByName(name);
            if (componentLifecycle != null) {
                final ObjectName componentLifecycleMbeanName;
                if (checkComponentType(name, ComponentType.BINDING_COMPONENT)) {
                    componentLifecycleMbeanName = this.mbeanNameBuilder.createBindingComponentMBeanName(name);
                } else if (checkComponentType(name, ComponentType.SERVICE_ENGINE)) {
                    componentLifecycleMbeanName = this.mbeanNameBuilder.createEngineComponentMBeanName(name);
                } else {
                    throw new ContainerServiceException("Type of the lifecycle not known or null");
                }
                this.mbeanServer.registerMBean(
                        MBeanHelper.createCustomMBean(componentLifecycle, ComponentLifeCycleMBean.class),
                        componentLifecycleMbeanName);
                return componentLifecycleMbeanName;
            } else {
                throw new ComponentLifecycleNotFoundException(name);
            }
        } catch (final RuntimeOperationsException e) {
            throw new ContainerServiceException(e);
        } catch (final InvalidTargetObjectTypeException e) {
            throw new ContainerServiceException(e);
        } catch (final InstanceAlreadyExistsException e) {
            throw new ContainerServiceException(e);
        } catch (final MBeanRegistrationException e) {
            throw new ContainerServiceException(e);
        } catch (final NotCompliantMBeanException e) {
            throw new ContainerServiceException(e);
        } catch (final InstanceNotFoundException e) {
            throw new ContainerServiceException(e);
        } catch (final MBeanException e) {
            throw new ContainerServiceException(e);
        }
    }

    @Override
    public void unregisterComponentLifeCycleMBean(final String name) throws ContainerServiceException {
        if (this.componentLifecycles.containsKey(name)) {
            final ObjectName objectName;
            if (checkComponentType(name, ComponentType.SERVICE_ENGINE)) {
                objectName = this.mbeanNameBuilder.createEngineComponentMBeanName(name);
            } else if (checkComponentType(name, ComponentType.BINDING_COMPONENT)) {
                objectName = this.mbeanNameBuilder.createBindingComponentMBeanName(name);
            } else {
                throw new ContainerServiceException("Type of the lifecycle not known");
            }
            try {
                this.mbeanServer.unregisterMBean(objectName);
            } catch (final InstanceNotFoundException e) {
                throw new ComponentLifecycleNotFoundException(name, e);
            } catch (final MBeanRegistrationException e) {
                throw new ContainerServiceException("Failed to unregistered the binding Component '" + name + "'", e);
            }
        } else {
            throw new ComponentLifecycleNotFoundException(name);
        }
    }

    @Override
    public void removeComponentLifeCycle(final String name) throws PetalsException {
        try {
            this.containerController.removeComponentLifeCycle(name);
        } catch (final ContainerCtrlException e) {
            throw new PetalsException(e);
        }
    }

    @Override
    public void removeInstaller(final String name) throws PetalsException {
        try {
            this.containerController.removeInstaller(name);
        } catch (final ContainerCtrlException e) {
            throw new PetalsException(e);
        }
    }

    @Override
    public ServiceAssemblyLifeCycle createServiceAssemblyLifeCycle(
            final org.ow2.petals.jbi.descriptor.original.generated.ServiceAssembly sa) throws ContainerServiceException {

        try {
            return this.containerController.createServiceAssemblyLifeCycle(sa);
        } catch (final ContainerCtrlException e) {
            throw new ContainerServiceException(e);
        }
    }

    @Override
    public void removeServiceAssemblyLifeCycle(final String saName) throws PetalsException {
        try {
            this.containerController.removeServiceAssemblyLifeCycle(saName);
        } catch (final ContainerCtrlException e) {
            throw new PetalsException(e);
        }
    }

    @Override
    public SharedLibraryLifeCycle createSharedLibraryLifeCycle(final SharedLibrary sl) throws ContainerServiceException {
        try {
            return this.containerController.createSharedLibraryLifeCycle(sl);
        } catch (final ContainerCtrlException e) {
            throw new ContainerServiceException(e);
        }
    }

    @Override
    public void removeSharedLibraryLifeCycle(final String slName, final String slVersion) throws PetalsException {
        try {
            final SharedLibraryLifeCycle sharedLibraryLifecycle = this.getSharedLibraryByName(slName, slVersion);
            if (sharedLibraryLifecycle != null) {
                sharedLibraryLifecycle.unLoadSharedLibrary();
                this.containerController.removeSharedLibraryLifeCycle(slName, slVersion);
            } else {
                throw new SharedLibraryLifecycleNotFoundException(slName, slVersion);
            }
        } catch (final ContainerCtrlException e) {
            throw new PetalsException(e);
        }
    }

    /**
     * Initialization of the petals component.
     * 
     * @throws PetalsException
     *             : impossible to start the component
     * @throws JBIException
     *             : JBI error
     * @throws NoSuchInterfaceException
     *             : Fractal Error
     * @throws ADLException
     *             : Fractal error
     * 
     */
    @Lifecycle(step = Step.START)
    public void start() throws PetalsException, NoSuchInterfaceException {
        this.log.start();

        try {
            this.parentcontainer = FractalHelper.getParentComponent(this.comp);
            this.mbeanServer = this.jmxService.getLocalJMXServer();
        } catch (PetalsException e) {
            this.log.error("Failed to start Container service", e);
            throw e;
        } catch (NoSuchInterfaceException e) {
            this.log.error("Failed to start Container service", e);
            throw e;
        } finally {
            this.log.end();
        }
    }

    /**
     * Stop of the petals component.
     * 
     */
    @Lifecycle(step = Step.STOP)
    public void stop() {
        this.log.call();
    }

    @Override
    public Installer getInstallerByName(final String name) {
        return this.installers.get(name);
    }

    @Override
    public Map<String, Installer> getInstallers() {
        return Collections.unmodifiableMap(this.installers);
    }

    @Override
    public ComponentLifeCycle getComponentLifecycleByName(final String name) {
        return this.componentLifecycles.get(name);
    }

    @Override
    public Map<String, ComponentLifeCycle> getComponentLifecycles() {
        return Collections.unmodifiableMap(this.componentLifecycles);
    }

    @Override
    public Collection<String> getBindingComponentsNames() {
        return Collections.unmodifiableCollection(filterByType(ComponentType.BINDING_COMPONENT,
                this.componentLifecycles).keySet());
    }

    @Override
    public Collection<String> getServiceEnginesNames() {
        return Collections.unmodifiableCollection(filterByType(ComponentType.SERVICE_ENGINE, this.componentLifecycles)
                .keySet());
    }

    @Override
    public boolean isBindingComponent(final String componentName) {
        return this.checkComponentType(componentName, ComponentType.BINDING_COMPONENT);
    }

    @Override
    public boolean isServiceEngine(final String componentName) {
        return this.checkComponentType(componentName, ComponentType.SERVICE_ENGINE);
    }

    /**
     * Check that a JBI component is a component of the given type.
     * 
     * @param componentName
     *            The JBI component identifier
     * @param expectedType
     *            The expected type of the component
     * @return <code>true</code> if the JBI component has the give type, <code>false</code> otherwise.
     */
    private boolean checkComponentType(final String componentName, final ComponentType expectedType) {
        final Installer installer = this.getInstallerByName(componentName);
        if (installer == null) {
            return false;
        } else {
            return expectedType.equals(installer.getJBIDescriptor().getComponent().getType());
        }
    }

    @Override
    public ServiceAssemblyLifeCycle getServiceAssemblyByName(final String saName) {
        return this.serviceAssemblyLifeCycles.get(saName);
    }

    @Override
    public Map<String, ServiceAssemblyLifeCycle> getServiceAssemblies() {
        return Collections.unmodifiableMap(this.serviceAssemblyLifeCycles);
    }

    @Override
    public SharedLibraryLifeCycle getSharedLibraryByName(final String slName, final String slVersion) {
        return this.sharedLibraryLifeCycles.get(SharedLibraryBean.getSharedLibraryName(slName, slVersion));
    }

    @Override
    public Map<SharedLibraryBean, SharedLibraryLifeCycle> getSharedLibraries() {
        final Map<SharedLibraryBean, SharedLibraryLifeCycle> result = new HashMap<SharedLibraryBean, SharedLibraryLifeCycle>(
                this.sharedLibraryLifeCycles.size());
        for (final SharedLibraryLifeCycle sllc : this.sharedLibraryLifeCycles.values()) {
            result.put(sllc.getSharedLibraryBean(), sllc);
        }
        return Collections.unmodifiableMap(result);
    }

    @Override
    public Collection<ServiceUnitLifeCycle> getServiceUnitsLifeCyclesByComponent(String component) {
        final List<ServiceUnitLifeCycle> suLCs = new ArrayList<ServiceUnitLifeCycle>();

        for (final ServiceAssemblyLifeCycle saLC : this.getServiceAssemblies().values()) {
            for (ServiceUnitLifeCycle suLC : saLC.getServiceUnitLifeCycles()) {
                if (suLC.getTargetComponentName().equals(component)) {
                    suLCs.add(suLC);
                }
            }
        }

        return suLCs;
    }

    private Map<String, ComponentLifeCycle> filterByType(final ComponentType expectedType,
            final Map<String, ComponentLifeCycle> map) {
        return Maps.filterEntries(map, new Predicate<Entry<String, ComponentLifeCycle>>() {
            @Override
            public boolean apply(Entry<String, ComponentLifeCycle> input) {
                return ContainerServiceImpl.this.checkComponentType(input.getKey(),expectedType);
            }
        });
    }
}
