/**
 * Copyright (c) 2005-2012 EBM WebSourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.container.lifecycle;

import java.util.concurrent.Callable;

import javax.jbi.JBIException;
import javax.jbi.component.ServiceUnitManager;

import org.ow2.petals.jbi.descriptor.original.generated.ServiceUnit;
import org.ow2.petals.microkernel.api.container.Installer;
import org.ow2.petals.microkernel.api.container.IsolatingThread;
import org.ow2.petals.microkernel.api.container.ServiceUnitLifeCycle;

/**
 * This class manage the life cycle of deployed service units.
 * 
 * Careful, this is not a Fractal component, but it behaves more or less like the other lifecycle classes which are
 * fractal components. Because of its double relation to both SAs and Components, it was hard to model that with the
 * Fractal component model, so we kept it as a Java class for now.
 * 
 * Also, the isolation of the execution of the actions done by this class is taken care of by the
 * {@link IsolatingThread} coming from the {@link Installer}.
 * 
 * @author Olivier Fabre - EBM WebSourcing
 * @author Frederic Gardes - EBM WebSourcing
 * @author Victor Noel
 */
public class ServiceUnitLifeCycleImpl implements ServiceUnitLifeCycle {

    private enum ServiceUnitState {
        SHUTDOWN, STOPPED, STARTED;
    }

    private final ServiceUnit serviceUnit;

    private final String serviceUnitRootPath;

    private final ServiceUnitManager suManager;

    private final String name;

    private final IsolatingThread separateThread;

    private final ClassLoader classLoader;

    private ServiceUnitState state;

    public ServiceUnitLifeCycleImpl(final ServiceUnit serviceUnit, final String serviceUnitRootPath,
            final ServiceUnitManager suManager, final IsolatingThread separateThread) {
        this.suManager = suManager;
        this.serviceUnit = serviceUnit;
        this.serviceUnitRootPath = serviceUnitRootPath;
        this.separateThread = separateThread;
        this.name = serviceUnit.getIdentification().getName();
        this.classLoader = suManager.getClass().getClassLoader();

        state = ServiceUnitState.SHUTDOWN;
    }

    @Override
    public String getTargetComponentName() {
        return serviceUnit.getTarget().getComponentName();
    }

    @Override
    public String getServiceUnitRootPath() {
        return serviceUnitRootPath;
    }

    @Override
    public String getSuName() {
        return name;
    }

    @Override
    public void init() throws JBIException {
        this.separateThread.execute(new Callable<Void>() {
            @Override
            public Void call() throws Exception {
                if (state == ServiceUnitState.SHUTDOWN) {
                    suManager.init(name, serviceUnitRootPath);
                    state = ServiceUnitState.STOPPED;
                }
                return null;
            }
        }, this.classLoader);
    }

    @Override
    public void shutDown() throws JBIException {
        this.separateThread.execute(new Callable<Void>() {
            @Override
            public Void call() throws Exception {
                if (state == ServiceUnitState.STOPPED) {
                    suManager.shutDown(name);
                    state = ServiceUnitState.SHUTDOWN;
                }
                return null;
            }
        }, this.classLoader);
    }

    @Override
    public void start() throws JBIException {
        this.separateThread.execute(new Callable<Void>() {
            @Override
            public Void call() throws Exception {
                if (state == ServiceUnitState.STOPPED) {
                    suManager.start(name);
                    state = ServiceUnitState.STARTED;
                }
                return null;
            }
        }, this.classLoader);
    }

    @Override
    public void stop() throws JBIException {
        this.separateThread.execute(new Callable<Void>() {
            @Override
            public Void call() throws Exception {
                if (state == ServiceUnitState.STARTED) {
                    suManager.stop(name);
                    state = ServiceUnitState.STOPPED;
                }
                return null;
            }
        }, this.classLoader);
    }
}
