/**
 * Copyright (c) 2015-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.container.lifecycle.util;

import java.lang.Thread.UncaughtExceptionHandler;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.ThreadFactory;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

import javax.jbi.JBIException;

import org.ow2.petals.commons.log.PetalsExecutionContext;
import org.ow2.petals.microkernel.api.container.IsolatingThread;
import org.ow2.petals.microkernel.container.lifecycle.ComponentLifeCycleImpl;
import org.ow2.petals.microkernel.container.lifecycle.InstallerImpl;
import org.ow2.petals.microkernel.container.lifecycle.ServiceAssemblyLifeCycleImpl;

import com.ebmwebsourcing.easycommons.lang.UncheckedException;
import com.ebmwebsourcing.easycommons.log.LoggingUtil;

/**
 * 
 * This is used to isolate the execution of the some of the actions in the {@link InstallerImpl},
 * {@link ComponentLifeCycleImpl} or {@link ServiceAssemblyLifeCycleImpl} using the given classloader and in a
 * restricted time.
 * 
 * The main reason of existence if this mechanism is really the restricted time... all the rest could have happened in
 * the calling thread of {@link #execute(Callable)}.
 * 
 * {@link #execute(Callable)} ensures the order of execution, that only one action at a time is executed (both thanks to
 * the {@link ExecutorService}'s queue and its only one thread) and that the actual action does not take more than the
 * restricted timeout to execute (not including the waiting time in the queue).
 * 
 * @author Victor Noel
 *
 */
public class SeparateIsolatingThread implements IsolatingThread {

    private ExecutorService executor;

    private final long timeout;

    private final String name;

    private final LoggingUtil log;

    /**
     * 
     * @param name
     *            of the thread
     * @param timeout
     *            timeout before actions are cancelled
     * @param log
     */
    public SeparateIsolatingThread(final String name, final long timeout, final LoggingUtil log) {
        assert name != null;
        assert log != null;
        assert timeout >= 0;

        this.name = name;
        this.log = log;
        this.timeout = timeout;

        initialize();
    }

    /**
     * Initialize executor (call it only after {@link #shutdown()} has been executed!)
     */
    public void initialize() {
        assert this.executor == null;
        final ThreadFactory dtf = Executors.defaultThreadFactory();
        // settings from Executors.newSingleThreadExecutor
        this.executor = new ThreadPoolExecutor(1, 1, 0L, TimeUnit.MILLISECONDS, new LinkedBlockingQueue<Runnable>(),
                new ThreadFactory() {
                    @Override
                    public Thread newThread(Runnable r) {
                        final Thread t = dtf.newThread(r);
                        t.setName(name);
                        t.setUncaughtExceptionHandler(new UncaughtExceptionHandler() {
                            @Override
                            public void uncaughtException(Thread t, Throwable e) {
                                log.error("Uncaught exception in isolating thread " + t.getName(), e);
                            }
                        });
                        return t;
                    }
                });
    }

    /**
     * Shutdown the executor and interrupt active tasks
     */
    public void shutdown() {
        // when this method is called, normally there can't be new task executed anymore (because the caller is a
        // fractal component being stopped, which can't happen if other call are made to the component at the same time)
        this.executor.shutdownNow();
        this.executor = null;
    }

    @Override
    public <T> T execute(final Callable<T> what, final ClassLoader ccl) throws JBIException {

        Future<T> f = null;
        try {
            // we use an array as a boolean holder (i.e. the object doesn't change but it's content can)
            final Boolean[] sync = new Boolean[] { false };
            f = executor.submit(new Callable<T>() {
                @Override
                public T call() throws Exception {
                    synchronized (sync) {
                        // let's use a simple state to allow the waiter to recognise our notify as such
                        sync[0] = true;
                        sync.notify();
                    }
                    // Set the context classloader (we do it here in case a component
                    // implementation has been playing with it).
                    Thread.currentThread().setContextClassLoader(ccl);
                    // Let's be sure there is nothing in the ExecutionContext from a previous thread (shouldn't be the
                    // case but who knows...)
                    PetalsExecutionContext.clear();
                    return what.call();
                }
            });

            // we wait because we want the timeout on the get to be considered
            // for the execution itself and not the time waiting in the queue.
            synchronized (sync) {
                // let's use a loop to prevent spurious wake-up not caused by the above notify
                while (!sync[0]) {
                    sync.wait();
                }
            }

            return f.get(timeout, TimeUnit.MILLISECONDS);
        } catch (final InterruptedException e) {
            // this can happen when we are shutdown (even though there is no reason this actually happen because we are
            // called by a fractal component and their stop method is not called before all the executing methods are
            // finished)
            throw new JBIException("Interruption when executing management task for " + name
                    + ". The lifecycle must have been stopped.");
        } catch (final ExecutionException e) {
            if (e.getCause() instanceof JBIException) {
                throw (JBIException) e.getCause();
            } else {
                throw new UncheckedException("This should never happen", e);
            }
        } catch (final TimeoutException e) {
            throw new JBIException(
                    "Timeout exceeded when executing management task for " + name + ". Assuming the task is blocked.");
        } finally {
            // let's cancel it in case there was a timeout (and if there wasn't, this has no effect)
            if (f != null) {
                // we must interrupt, if not, this is useless as the task is obviously already being executed, let's
                // just hope that the component code actually take interruptions into account...
                f.cancel(true);
            }
        }
    }
}
