/**
 * Copyright (c) 2005-2012 EBM WebSourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.jbi.management.admin;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;
import java.util.Map;
import java.util.SortedSet;
import java.util.TreeSet;
import java.util.logging.Logger;

import javax.jbi.JBIException;
import javax.jbi.management.MBeanNames;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.management.MBeanServer;
import javax.management.MalformedObjectNameException;
import javax.management.ObjectName;

import org.objectweb.fractal.fraclet.annotations.Component;
import org.objectweb.fractal.fraclet.annotations.Interface;
import org.objectweb.fractal.fraclet.annotations.Lifecycle;
import org.objectweb.fractal.fraclet.annotations.Requires;
import org.objectweb.fractal.fraclet.types.Step;
import org.ow2.petals.jmx.api.api.AdminServiceClient;
import org.ow2.petals.jmx.api.api.DeploymentServiceClient;
import org.ow2.petals.jmx.api.api.InstallationServiceClient;
import org.ow2.petals.microkernel.api.communication.JMXService;
import org.ow2.petals.microkernel.api.configuration.ConfigurationService;
import org.ow2.petals.microkernel.api.container.ComponentLifeCycle;
import org.ow2.petals.microkernel.api.container.ContainerService;
import org.ow2.petals.microkernel.api.container.Installer;
import org.ow2.petals.microkernel.api.extension.PetalsExtensionController;
import org.ow2.petals.microkernel.api.jbi.management.AdminService;
import org.ow2.petals.microkernel.api.system.sharedarea.SharedAreaService;
import org.ow2.petals.microkernel.extension.PostExtensionsManager;
import org.ow2.petals.microkernel.extension.PreExtensionsManager;
import org.ow2.petals.microkernel.jbi.management.MBeanNameBuilder;
import org.ow2.petals.microkernel.server.MBeanHelper;
import org.ow2.petals.microkernel.util.ParameterCheckHelper;

import com.ebmwebsourcing.easycommons.log.LoggingUtil;

/**
 * TODO AdminServiceImpl : check the exceptions catch.
 * 
 * @author Adrien LOUIS - EBM WebSourcing
 * @author wjoseph - EBM WebSourcing
 */
@Component(provides = @Interface(name = "service", signature = AdminService.class))
public class AdminServiceImpl implements AdminService {

    /**
     * Logger wrapper
     */
    protected final LoggingUtil log = new LoggingUtil(Logger.getLogger(AdminService.COMPONENT_LOGGER_NAME));

    /**
     * The JMX Fractal service
     */
    @Requires(name = "jmx")
    private JMXService jmxService;

    @Requires(name = "mbean-name-builder")
    private MBeanNameBuilder mBeanNames;

    @Requires(name = "containerService")
    private ContainerService containerService;

    @Requires(name = "configuration")
    private ConfigurationService configurationService;

    @Requires(name = "preExtensionsMgr")
    private PreExtensionsManager preExtensionsMgr;

    @Requires(name = "postExtensionsMgr")
    private PostExtensionsManager postExtensionsMgr;

    @Requires(name = "sharedArea")
    private SharedAreaService sharedArea;

    private MBeanServer mbeanServer;

    @Override
    public ObjectName[] getBindingComponents() {
        this.log.start();
        try {
            final List<ObjectName> result = new ArrayList<ObjectName>();
            for (final String name : this.containerService.getBindingComponentsNames()) {
                result.add(this.mBeanNames.createBindingComponentMBeanName(name));
            }
            return result.toArray(new ObjectName[0]);
        } finally {
            this.log.end();
        }
    }

    @Override
    public ObjectName[] getEngineComponents() {
        this.log.start();
        try {
            final List<ObjectName> result = new ArrayList<ObjectName>();
            for (final String name : this.containerService.getServiceEnginesNames()) {
                result.add(this.mBeanNames.createEngineComponentMBeanName(name));
            }
            return result.toArray(new ObjectName[0]);
        } finally {
            this.log.end();
        }
    }

    @Override
    public ObjectName getComponentByName(final String name) {
        this.log.call();

        ParameterCheckHelper.isNullOrEmptyParameterWithLog(name, "name cannot be null", this.log);

        final ComponentLifeCycle componentLifecycle = this.containerService.getComponentLifecycleByName(name);
        if (componentLifecycle != null) {
            if (this.containerService.isBindingComponent(name)) {
                return this.mBeanNames.createBindingComponentMBeanName(name);
            } else if (this.containerService.isServiceEngine(name)) {
                return this.mBeanNames.createEngineComponentMBeanName(name);
            } else {
                return null;
            }
        } else {
            return null;
        }
    }

    @Override
    public String getSystemInfo() {

        final StringBuilder componentVersions = new StringBuilder();
        componentVersions.append("Petals ESB µKernel").append(' ')
                .append(AdminServiceImpl.class.getPackage().getImplementationVersion())
                .append(System.getProperty("line.separator"));

        // Mutable implementations
        componentVersions.append(this.sharedArea.getImplementationName()).append(' ')
                .append(this.sharedArea.getImplementationVersion()).append(System.getProperty("line.separator"));

        // Extensions:
        // - 1st: pre-extensions ordered by name
        // - 2nd: post-extensions ordered by name
        AdminServiceImpl.addComponentVersions(this.preExtensionsMgr.getExtensionControllers(), componentVersions);
        AdminServiceImpl.addComponentVersions(this.postExtensionsMgr.getExtensionControllers(), componentVersions);

        // System
        componentVersions.append(System.getProperty("java.runtime.name")).append(' ')
                .append(System.getProperty("java.runtime.version")).append(' ')
                .append(System.getProperty("java.vm.vendor")).append(System.getProperty("line.separator"));
        componentVersions.append(System.getProperty("os.name")).append(' ').append(System.getProperty("os.version"))
                .append(' ').append(System.getProperty("os.arch")).append(System.getProperty("line.separator"));

        return componentVersions.toString();
    }

    private static void addComponentVersions(
            final Map<PetalsExtensionController, org.objectweb.fractal.api.Component> extensionControllers,
            final StringBuilder componentVersions) {

        final SortedSet<PetalsExtensionController> orderedExtensions = new TreeSet<PetalsExtensionController>(
                new Comparator<PetalsExtensionController>() {

                    @Override
                    public int compare(final PetalsExtensionController extCtrl1,
                            final PetalsExtensionController extCtrl2) {
                        return extCtrl1.getExtensionName().compareTo(extCtrl2.getExtensionName());
                    }
                });
        orderedExtensions.addAll(extensionControllers.keySet());

        for (final PetalsExtensionController extensionController : orderedExtensions) {
            componentVersions.append(extensionController.getExtensionName()).append(' ')
                    .append(extensionController.getExtensionVersion()).append(System.getProperty("line.separator"));
        }
    }

    @Override
    public ObjectName getSystemService(String serviceName) {
        this.log.call();

        ParameterCheckHelper.isNullOrEmptyParameterWithLog(serviceName,
                "serviceName cannot be null", this.log);

        ObjectName result = null;
        ObjectName[] systemServices = this.getSystemServices();

        for (ObjectName objectName : systemServices) {
            if (objectName.getCanonicalName().equals(serviceName)) {
                result = objectName;
                break;
            }
        }
        return result;
    }

    @Override
    public ObjectName[] getSystemServices() {
        this.log.call();

        ObjectName[] names = new ObjectName[3];

        try {
            names[0] = MBeanHelper.retrieveServiceMBean(
                    AdminServiceClient.ADMIN_SERVICE_MBEAN_NAME, this.mbeanServer);
            names[1] = MBeanHelper.retrieveServiceMBean(
                    DeploymentServiceClient.DEPLOYMENT_SERVICE_MBEAN_NAME,
                    this.mbeanServer);
            names[2] = MBeanHelper.retrieveServiceMBean(
                    InstallationServiceClient.INSTALLATION_SERVICE_MBEAN_NAME,
                    this.mbeanServer);
        } catch (MalformedObjectNameException e) {
            this.log.warning("A system service name can't be retrieved !", e);
        }

        return names;
    }

    @Override
    public boolean isBinding(String componentName) {
        ParameterCheckHelper.isNullOrEmptyParameterWithLog(componentName,
                "componentName cannot be null", this.log);

        return this.containerService.isBindingComponent(componentName);
    }

    @Override
    public boolean isEngine(String componentName) {
        ParameterCheckHelper.isNullOrEmptyParameterWithLog(componentName,
                "componentName cannot be null", this.log);

        return this.containerService.isServiceEngine(componentName);
    }

    @Override
    public boolean isExchangeWithConsumerOkayForComponent(final String providerComponent,
            final ServiceEndpoint providerEndpoint, final MessageExchange exchange) throws JBIException {
        this.log.start();
        try {
            if (configurationService.getContainerConfiguration().isExchangeValidation()) {
                final Installer installer = this.containerService.getInstallerByName(providerComponent);
                if (installer == null) {
                    throw new JBIException("Cannot find a component '" + providerComponent + "' in the container");
                }

                return installer.getComponent().isExchangeWithConsumerOkay(providerEndpoint, exchange);
            } else {
                return true;
            }
        } finally {
            this.log.end();
        }
    }

    @Override
    public boolean isExchangeWithProviderOkayForComponent(final String consumerComponent,
            final ServiceEndpoint providerEndpoint, final MessageExchange exchange) throws JBIException {
        this.log.start();
        try {
            if (configurationService.getContainerConfiguration().isExchangeValidation()) {
                final Installer installer = this.containerService.getInstallerByName(consumerComponent);
                if (installer == null) {
                    throw new JBIException("Cannot find a component '" + consumerComponent + "' in the container");
                }

                return installer.getComponent().isExchangeWithProviderOkay(providerEndpoint, exchange);
            } else {
                return true;
            }
        } finally {
            this.log.end();
        }
    }

    @Override
    public MBeanNames getMBeanNames() {
        return this.mBeanNames;
    }

    @Override
    public MBeanServer getMBeanServer() {
        return this.mbeanServer;
    }

    // ----------------------------------------------------------------------
    // Fractal LifecycleController implementation
    // ----------------------------------------------------------------------

    @Lifecycle(step = Step.START)
    public void start() {
        this.log.start();

        this.mbeanServer = this.jmxService.getLocalJMXServer();

        this.log.end();
    }

    @Lifecycle(step = Step.STOP)
    public void stop() {
        this.log.call();

    }

}
