/**
 * Copyright (c) 2006-2012 EBM WebSourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.jbi.management.recovery;

import java.io.File;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.concurrent.Callable;
import java.util.concurrent.CompletionService;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorCompletionService;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.logging.Logger;

import org.objectweb.fractal.fraclet.annotations.Interface;
import org.objectweb.fractal.fraclet.annotations.Lifecycle;
import org.objectweb.fractal.fraclet.annotations.Requires;
import org.objectweb.fractal.fraclet.types.Step;
import org.ow2.petals.clientserverapi.configuration.ContainerConfiguration;
import org.ow2.petals.microkernel.api.configuration.ConfigurationService;
import org.ow2.petals.microkernel.api.container.ContainerService;
import org.ow2.petals.microkernel.api.jbi.management.ManagementException;
import org.ow2.petals.microkernel.api.jbi.messaging.RouterService;
import org.ow2.petals.microkernel.api.jbi.messaging.endpoint.EndpointDirectoryService;
import org.ow2.petals.microkernel.api.system.SystemStateService;
import org.ow2.petals.microkernel.api.system.repository.RepositoryService;
import org.ow2.petals.microkernel.extension.InstallationExtensionsManager;
import org.ow2.petals.microkernel.jbi.management.recovery.task.RecoverComponentTask;
import org.ow2.petals.microkernel.jbi.management.recovery.task.RecoverSaTask;
import org.ow2.petals.microkernel.jbi.management.recovery.task.RecoverSlTask;
import org.ow2.petals.microkernel.jbi.management.recovery.task.RecoveryResult;
import org.ow2.petals.microkernel.jbi.management.task.deployment.DeploymentTaskFactory;
import org.ow2.petals.microkernel.jbi.management.task.installation.InstallationTaskFactory;
import org.ow2.petals.systemstate.generated.Component;
import org.ow2.petals.systemstate.generated.ServiceAssembly;
import org.ow2.petals.systemstate.generated.SharedLibrary;

import com.ebmwebsourcing.easycommons.log.LoggingUtil;

/**
 * Allows to restore system state after shutdown or crash.
 * @author Olivier Fabre - EBM WebSourcing
 */
@org.objectweb.fractal.fraclet.annotations.Component(provides = @Interface(name = "service", signature = SystemRecoveryService.class))
public class SystemRecoveryServiceImpl implements SystemRecoveryService {

    /**
     * logger wrapper
     */
    private final LoggingUtil log = new LoggingUtil(Logger.getLogger(SystemRecoveryService.COMPONENT_LOGGER_NAME));;

    private DeploymentTaskFactory deploymentTaskFactory;

    private InstallationTaskFactory installationTaskFactory;

    /**
     * Configuration Service
     */
    @Requires(name = "configuration")
    private ConfigurationService configurationService;

    /**
     * The endpoint directory service is used to manage all the endpoint interfaces over the topology
     */
    @Requires(name = "endpointDirectory")
    private EndpointDirectoryService endpointDirectory;

    /**
     * JBI messaging : Router Service.
     */
    @Requires(name = "router")
    private RouterService router;

    /**
     * Platform :: Repository Service.
     */
    @Requires(name = "repository")
    private RepositoryService repositoryService;

    /**
     * Platform :: SystemState Service.
     */
    @Requires(name = "systemstate")
    private SystemStateService systemStateService;

    /**
     * Container :: Container Service.
     */
    @Requires(name = "container")
    private ContainerService containerService;

    @Requires(name = "installationExtensionManager")
    private InstallationExtensionsManager installationExtensionManager;

    /**
     * Thread pool to recover JBI artifacts.
     */
    private ThreadPoolExecutor threadPoolExecutor;

    /**
     * The excecutor completion service bound to the Thread Pool Executor.
     */
    private CompletionService<RecoveryResult> ecs;

    @Override
    public void recoverAllEntities() throws ManagementException {
        this.log.start();

        // TODO remove all local endpoints because we don't remove the endpoints from the registry (on shutdown, and of
        // course neither on crash), so before restarting everything, we should clean the registry!

        this.recoverAllSharedLibrary();
        this.recoverAllComponent();
        this.recoverAllServiceAssembly();

        this.log.end();
    }

    /**
     * Recover all components in the state in which they was at the last container stop.
     * 
     * @throws ManagementException
     */
    private void recoverAllComponent() throws ManagementException {
        this.log.start();

        // Retrieve all component directories in the components petals
        // repository. This list will be used to delete components that have
        // been unzipped in the Petals repository but no more persisted in the
        // system recovery file (system-state.xml). These unknown component
        // directories will be put in the Petals "lost+found" directory.
        final List<File> unknownDirs = new ArrayList<File>(Arrays.asList(this.repositoryService
                .getComponentsDirectory().listFiles()));

        // Retrieve all components to recover
        final List<Component> componentStateHolders = this.systemStateService
                .getComponentStateHolders();

        int counter = 0;
        // Recover state for each component
        for (final Component componentStateHolder : componentStateHolders) {

            this.log.debug("Recover Component '" + componentStateHolder.getName() + "'");
            final Callable<RecoveryResult> recoverComponentTask = new RecoverComponentTask(unknownDirs,
                    componentStateHolder, this.repositoryService, this.systemStateService, this.containerService,
                    this.installationTaskFactory, this.log);
            ecs.submit(recoverComponentTask);
            counter++;
        }

        for (int i = 0; i < counter; i++) {
            try {
                RecoveryResult result = ecs.take().get();
                this.log.debug("Component '" + result.getJbiArtifactName() + "' recovered");
            } catch (InterruptedException e) {
                this.log.error("Component recovering has been interrupted");
            } catch (ExecutionException e) {
                this.log.error(e.getMessage(), e);
            }
        }

        // Move all unknown components from Petals repository to "lost+found"
        // directory
        this.moveUnknownDirsToLostPlusFoundDir(unknownDirs);

        this.log.end();

    }

    /**
     * Recover all service assemblies in the state in which they was at the last container stop.
     * 
     * @throws ManagementException
     */
    private void recoverAllServiceAssembly() throws ManagementException {
        this.log.start();

        // Retrieve all SA directories in the SA petals repository. This list
        // will be used to delete SA that have been unzipped in the Petals
        // repository but no more persisted in the system recovery file
        // (system-state.xml). These unknown SA directories will be put in the
        // Petals "lost+found" directory.
        final List<File> unknownDirs = new ArrayList<File>(Arrays.asList(this.repositoryService
                .getServiceAssembliesDirectory().listFiles()));

        // Retrieve all service assemblies to recover
        final List<ServiceAssembly> saStateHolders = this.systemStateService
                .getServiceAssemblyStateHolders();

        // Recover state for each service assembly
        for (final ServiceAssembly saStateHolder : saStateHolders) {

            this.log.debug("Recover Service Assembly '" + saStateHolder.getName() + "'");
            final Callable<RecoveryResult> recoverSaTask = new RecoverSaTask(unknownDirs, saStateHolder,
                    this.repositoryService, this.systemStateService, this.containerService, this.deploymentTaskFactory,
                    this.log);
            ecs.submit(recoverSaTask);
        }

        for (int i = 0; i < saStateHolders.size(); i++) {
            try {
                RecoveryResult result = ecs.take().get();
                this.log.debug("Service Assembly '" + result.getJbiArtifactName() + "' recovered");
            } catch (InterruptedException e) {
                this.log.error("Service Assembly recovering has been interrupted");
            } catch (ExecutionException e) {
                this.log.error(e.getMessage(), e);
            }
        }

        // Move all unknown SA from Petals repository to
        // "lost+found" directory
        this.moveUnknownDirsToLostPlusFoundDir(unknownDirs);

        this.log.end();
    }

    /**
     * Recover all shared libraries in the state in which they was at the last container stop.
     * 
     * @throws ManagementException
     */
    private void recoverAllSharedLibrary() throws ManagementException {
        this.log.start();

        // Retrieve all sl directories in the sl petals repository. This list
        // will be used to delete sl that have been unzipped in the Petals
        // repository but no more persisted in the system recory file
        // (system-state.xml). These unknown sl directories will be put in the
        // Petals "lost+found" directory.
        final List<File> unknownDirs = new ArrayList<File>(Arrays.asList(this.repositoryService
                .getSharedLibrariesDirectory().listFiles()));

        // Retrieve all shared library to recover
        final List<SharedLibrary> sharedLibraryStateHolders = this.systemStateService
                .getSharedLibraryStateHolders();

        // Recover state for each shared library
        for (final SharedLibrary sharedLibraryState : sharedLibraryStateHolders) {

            this.log.debug("Recover Shared Library '" + sharedLibraryState.getName() + "'");
            final Callable<RecoveryResult> recoverSLTask = new RecoverSlTask(unknownDirs, sharedLibraryState,
                    this.repositoryService, this.systemStateService, this.containerService,
                    this.installationTaskFactory, this.log);
            ecs.submit(recoverSLTask);
        }

        for (int i = 0; i < sharedLibraryStateHolders.size(); i++) {
            try {
                RecoveryResult result = ecs.take().get();
                this.log.debug("Shared Library '" + result.getJbiArtifactName() + "' recovered");
            } catch (InterruptedException e) {
                this.log.error("Shared Library recovering has been interrupted");
            } catch (ExecutionException e) {
                this.log.error(e.getMessage(), e);
            }
        }

        // Now, move all unknown SL from Petals repository to
        // "lost+found" directory
        this.moveUnknownDirsToLostPlusFoundDir(unknownDirs);

        this.log.end();
    }

    /**
     * Move all unknown components from Petals repository to "lost+found"
     * directory
     * 
     * @param unknownDirs
     *            the {@link List} of unknown dirs that have to be moved.
     */
    private void moveUnknownDirsToLostPlusFoundDir(List<File> unknownDirs) {
        if (unknownDirs != null) {
            for (final File unknownDir : unknownDirs) {
                if (!unknownDir.isFile()) {
                    if (!unknownDir.exists()) {
                        this.log.warning("The unknown dir (" + unknownDir.getAbsolutePath()
                                + ") doesn't exist.");
                    }
                    final boolean moveDir = unknownDir.renameTo(new File(this.repositoryService
                            .getLostPlusFoundDirectory(), unknownDir.getName()));
                    if (!moveDir) {
                        this.log.warning("Failed to move unknown directory ("
                                + unknownDir.getAbsolutePath() + ") to lost+found directory");
                    }
                }
            }
        }
    }

    /**
     * Start FRACTAL component.
     */
    @Lifecycle(step = Step.START)
    public void start() {
        this.log.start();

        // Initialize task factory
        ContainerConfiguration containerConfiguration = this.configurationService
                .getContainerConfiguration();
        this.installationTaskFactory = new InstallationTaskFactory(this.systemStateService, this.repositoryService,
                this.log, this.router, containerConfiguration, this.containerService,
                this.installationExtensionManager.getExtensions());
        this.deploymentTaskFactory = new DeploymentTaskFactory(this.endpointDirectory, null, this.systemStateService,
                this.repositoryService, this.log, this.containerService, containerConfiguration,
                this.installationExtensionManager.getExtensions());

        /*
         * Initialize the recovering workers. As we use an unbounded queue, the size of the tread pool is driven by the
         * core pool size. The max pool size has no effect.
         */
        this.threadPoolExecutor = new ThreadPoolExecutor(
                containerConfiguration.getRecoveryCorePoolSize(),
                containerConfiguration.getRecoveryCorePoolSize(),
                containerConfiguration.getRecoveryKeepAliveTime(), TimeUnit.SECONDS,
                new LinkedBlockingQueue<Runnable>(), new SystemRecoveryThreadFactory(this.log));

        this.ecs = new ExecutorCompletionService<RecoveryResult>(threadPoolExecutor);

        this.log.end();
    }

    /**
     * Stop the FRACTAL component
     */
    @Lifecycle(step = Step.STOP)
    public void stop() {
        this.log.call();

        this.threadPoolExecutor.shutdownNow();
    }

}
