/**
 * Copyright (c) 2006-2012 EBM WebSourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.jbi.management.recovery.task;

import java.io.File;
import java.util.List;

import javax.jbi.JBIException;
import javax.jbi.management.LifeCycleMBean;

import org.ow2.petals.jbi.descriptor.original.JBIDescriptorBuilder;
import org.ow2.petals.jbi.descriptor.original.generated.Jbi;
import org.ow2.petals.microkernel.api.container.ComponentLifeCycle;
import org.ow2.petals.microkernel.api.container.ContainerService;
import org.ow2.petals.microkernel.api.container.Installer;
import org.ow2.petals.microkernel.api.jbi.management.Context;
import org.ow2.petals.microkernel.api.jbi.management.ManagementException;
import org.ow2.petals.microkernel.api.jbi.management.TaskProcessor;
import org.ow2.petals.microkernel.api.system.SystemStateService;
import org.ow2.petals.microkernel.api.system.repository.RepositoryService;
import org.ow2.petals.microkernel.jbi.management.task.installation.InstallationTaskFactory;
import org.ow2.petals.systemstate.generated.Component;

import com.ebmwebsourcing.easycommons.log.LoggingUtil;

/**
 * The task processing the recovery of a component
 * 
 * @author Olivier Fabre - EBM WebSourcing
 * @author Christophe DENEUX - LIlnagora
 * 
 */
public class RecoverComponentTask extends RecoverAbstractInstallationTask {

    private final Component componentStateHolder;

    private final InstallationTaskFactory installationTaskFactory;

    public RecoverComponentTask(final List<File> unknownDirs, final Component componentStateHolder,
            final RepositoryService repositoryService, final SystemStateService systemStateService,
            final ContainerService containerService, final InstallationTaskFactory installationTaskFactory,
            final LoggingUtil log) {
        super(unknownDirs, repositoryService, systemStateService, containerService, log);
        this.componentStateHolder = componentStateHolder;
        this.installationTaskFactory = installationTaskFactory;
    }

    @Override
    public RecoveryResult call() throws ManagementException {

        try {
            final File installFile = this.repositoryService
                    .getComponentInstallDirectory(this.componentStateHolder.getName());
            final Jbi descriptor = JBIDescriptorBuilder.getInstance().buildJavaJBIDescriptorFromFolder(installFile);

            this.recoverComponent(installFile, descriptor, this.componentStateHolder.getInstallState(),
                    this.componentStateHolder.getLifecycleState(), this.componentStateHolder.getName());

            // Delete component from the unknownDirs list because this
            // component is known now...
            if (this.unknownDirs != null) {
                final File knownComp = installFile.getParentFile();
                if (knownComp.exists()) {
                    this.unknownDirs.remove(knownComp);
                }
            }

            return new RecoveryResult(this.componentStateHolder.getName());
        } catch (final Exception e) {
            final StringBuilder msgBuffer = new StringBuilder("A component consistency is corrupted: ");
            msgBuffer.append(this.componentStateHolder.getName());
            // try to set the states of the component to UNKNOWN
            try {
                this.systemStateService.updateComponentLifeCycleState(this.componentStateHolder.getName(),
                        LifeCycleMBean.UNKNOWN);
                this.systemStateService.updateComponentInstallationState(this.componentStateHolder.getName(),
                        LifeCycleMBean.UNKNOWN);
                msgBuffer.append("\nIts state changes to UNKNOWN");
            } catch (final Exception e1) {
                // do nothing
            }
            msgBuffer.append("\nPlease check the repository consistency.");
            throw new ManagementException(msgBuffer.toString(), e);
        }
    }

    /**
     * Recover the given component. Recovering a component consist in redoing installation process of a component (see
     * <code>InstallationServiceMBean</code>), calling the "install" method on the created Installer and putting the
     * component in its previous state by calling "start" or/and "stop" on the created component lifecycle.
     * 
     * @param installationRoot
     * @param installState
     * @param lifecycleState
     * @throws ManagementException
     */
    private void recoverComponent(final File installFile, final Jbi descriptor, final String installState,
            final String lifecycleState, final String compName) throws ManagementException {
        this.log.start();

        // Create installation execution context
        Context installationContext = null;

        installationContext = this.createRecoveryContext(descriptor, installFile);
        
        // Recover component installer
        this.recoverComponent(installationContext);

        final String componentName = installationContext.getDescriptor().getComponent().getIdentification().getName();
        this.log.info("Installer of the component '" + componentName + "' loaded");

        try {
            if (Installer.INSTALLED.equals(installState)) {
                // Re-process install
                final Installer componentInstaller = this.containerService.getInstallerByName(componentName);
                componentInstaller.install();

                final ComponentLifeCycle componentLifecycle = this.containerService
                        .getComponentLifecycleByName(componentName);
                if (LifeCycleMBean.STOPPED.equals(lifecycleState)) {
                    // Re-process start followed by stop
                    componentLifecycle.start();
                    componentLifecycle.stop();
                } else if (LifeCycleMBean.STARTED.equals(lifecycleState)) {
                    // Re-process start
                    componentLifecycle.start();
                } else if (LifeCycleMBean.UNKNOWN.equals(lifecycleState)) {
                    this.log.warning("The component '" + compName + "' was in state " + LifeCycleMBean.UNKNOWN
                            + ", it is recovered in state " + LifeCycleMBean.SHUTDOWN);
                    try {
                        this.systemStateService.updateComponentLifeCycleState(compName, LifeCycleMBean.SHUTDOWN);
                    } catch (final Exception e) {
                        this.log.error("The component '" + compName + "' failed to be recovered at state "
                                + LifeCycleMBean.SHUTDOWN, e);
                    }
                }
            }
        } catch (final JBIException e) {
            throw new ManagementException("Failed to recover Component '" + compName + "'", e);
        } finally {
            this.log.end();
        }
    }

    /**
     * Load the installer for a previously successfully installed component. Used only by the Recovery Service.
     * 
     * @param installationContext
     *            the installation context for the recovery process : handles descriptor, archive URI and installRoot
     *            information
     * @throws ManagementException
     *             if the component installer can't be recovered
     */
    private void recoverComponent(final Context installationContext) throws ManagementException {
        final Jbi descriptor = installationContext.getDescriptor();

        // Create installation processor and fill with tasks
        final TaskProcessor processor = this.createComponentRecoveryProcessor(installationContext);

        // Launch the recover processor
        final Throwable processResult = processor.process();
        if (processResult != null) {
            throw new ManagementException("Failed to recover Component '"
                    + descriptor.getComponent().getIdentification().getName() + "'", processResult);
        }
    }

    /**
     * Create the installation processor and add all required task for the complete recovery of a component
     * 
     * @param installationContext
     *            the installation context
     * @return the filled installation processor
     */
    private TaskProcessor createComponentRecoveryProcessor(final Context installationContext) {
        final TaskProcessor processor = new TaskProcessor(installationContext, this.log);
        processor.addTask(this.installationTaskFactory.getRegisterComponentInSLTask());
        processor.addTask(this.installationTaskFactory.getCreateAndRegisterInstallerMBeanTask());
        return processor;
    }
}
