/**
 * Copyright (c) 2007-2012 EBM WebSourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.jbi.management.task;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URL;
import java.net.URLConnection;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.concurrent.atomic.AtomicLong;

import org.apache.commons.io.FileUtils;
import org.apache.commons.io.IOUtils;
import org.ow2.petals.basisapi.exception.PetalsException;
import org.ow2.petals.clientserverapi.configuration.ContainerConfiguration;
import org.ow2.petals.microkernel.api.jbi.management.Context;

import com.ebmwebsourcing.easycommons.log.LoggingUtil;

/**
 * This task download entity (component, shared lib, service assembly) archive into the work directory, only if the URL
 * protocol is http: or ftp: <br/>
 * Inputs : archive {@link URI} <br/>
 * Outputs : downloaded archive {@link File}<br/>
 * The task has got a non-robust approach : it stops at the first subtask failed<br />
 * and have the undo capability.<br />
 * 
 * @author Olivier Fabre - EBM WebSourcing
 * @author Frederic Gardes - EBM WebSourcing
 */
public class DownloadTask extends AbstractFileManipulationTask {

    /**
     * Unique download identifier, it is used in temporary path name to prevent
     * concurrent downloading.
     */
    private final static AtomicLong DOWNLOAD_ID = new AtomicLong(0L);

    /**
     * DateFormat used for repository folders
     */
    private final static SimpleDateFormat FOLDER_DATE_FORMAT = new SimpleDateFormat(
            "MM-dd-yyyy-hhmmss");

    public DownloadTask(final LoggingUtil log, final ContainerConfiguration containerConfiguration) {
        super(log, containerConfiguration);
    }

    private static String getCurrentDateAsString() {
        synchronized (FOLDER_DATE_FORMAT) {
            return DownloadTask.FOLDER_DATE_FORMAT.format(new Date(System.currentTimeMillis()));
        }
    }

    @Override
    public void execute(Context context) throws Exception {
        URL archiveUrl = context.getArchiveUrl();
        File workArchiveFile = this.downloadEntityPackage(archiveUrl);
        context.setArchiveFile(workArchiveFile);
    }

    @Override
    public void undo(Context context) throws Exception {
        File downloadFile = context.getArchiveFile();
        if (downloadFile != null) {
            try {
                FileUtils.forceDelete(downloadFile);
            } catch (IOException e) {
                this.log.error("Failed to clean downloaded artefact " + downloadFile);
            }
        }
    }

    /**
     * Download the JBI artifact to the Petals work directory.
     * 
     * @param archiveUrl
     * @return
     * @throws <PetalsException>
     */
    private File downloadEntityPackage(URL archiveUrl) throws PetalsException {

        File file = null;
        try {
            final URLConnection urlConnection = archiveUrl.openConnection();

            // We check if the remote resource exists.
            verifyThatResourceExist(archiveUrl, urlConnection);

            // Retrieve the filename to be record in work repository
            String fileName = retrieveFileName(archiveUrl);

            // Records the zip file and returns its new URL
            file = recordTheZipFileInWorkDirectory(archiveUrl, urlConnection, fileName);

        } catch (final IOException e) {
            throw new PetalsException("Can't read distant file: " + archiveUrl);
        }
        return file;
    }

    private File recordTheZipFileInWorkDirectory(URL archiveUrl, URLConnection urlConnection,
            String fileName) throws FileNotFoundException, IOException, MalformedURLException {
        FileOutputStream fileOutputStream = null;
        InputStream inputStream = null;
        File file;

        try {
            inputStream = urlConnection.getInputStream();
            inputStream = archiveUrl.openStream();

            file = new File(this.getWorkDirectory(), fileName + "-downloaded-"
                    + DownloadTask.getCurrentDateAsString() + "-"
                    + DownloadTask.DOWNLOAD_ID.getAndIncrement() + ".zip");
            fileOutputStream = new FileOutputStream(file);
            IOUtils.copy(inputStream, fileOutputStream);
        } finally {
            try {
                if (inputStream != null) {
                    inputStream.close();
                }
                if (fileOutputStream != null) {
                    fileOutputStream.close();
                }
            } catch (IOException e) {
                this.log.warning(
                        "Failed to clean the downloaded archive '" + archiveUrl.toExternalForm()
                                + "'", e);
            }
        }

        return file;
    }

    private String retrieveFileName(URL archiveUrl) {
        final String path = archiveUrl.getPath();
        String fileName = path.substring(path.lastIndexOf('/') + 1, path.lastIndexOf('.'));
        if (!fileName.contains(".")) {
            // try to get the file name with the query
            final String query = archiveUrl.getQuery();
            if (query != null) {
                fileName = query.substring(query.lastIndexOf('=') + 1);
            }
        }
        return fileName;
    }

    private void verifyThatResourceExist(URL archiveUrl, final URLConnection urlConnection)
            throws IOException, PetalsException {
        if ("http".equalsIgnoreCase(archiveUrl.getProtocol())) {
            final HttpURLConnection httpUrlConnection = (HttpURLConnection) urlConnection;
            final int responseCode = httpUrlConnection.getResponseCode();
             this.log.debug("HTTP Response code: " + responseCode);
            if (responseCode != HttpURLConnection.HTTP_OK) {
                throw new PetalsException("Unable to download the remote resource " + archiveUrl
                        + " [HTTP " + responseCode + "]");
            } else {
                // We check that the resource is a binary file
                final String contentType = httpUrlConnection.getContentType();
                if (contentType != null) {
                    if (!contentType.startsWith(MimeTypeConstants.APPLICATION_X_ZIP)
                            && !contentType
                                    .startsWith(MimeTypeConstants.APPLICATION_X_ZIP_COMPRESSED)
                            && !contentType.startsWith(MimeTypeConstants.APPLICATION_ZIP)
                            && !contentType
                                    .startsWith(MimeTypeConstants.APPLICATION_ZIP_COMPRESSED)) {
                        throw new PetalsException("The remote resource [" + archiveUrl
                                + "] is not a valid JBI archive: " + contentType);
                    }
                } else {
                    throw new PetalsException("The remote resource [" + archiveUrl.toExternalForm()
                            + "] is not a valid JBI archive. It's content-type is undefined.");
                }
            }
        }
    }

}
