/**
 * Copyright (c) 2006-2012 EBM WebSourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.jbi.management.task.deployment.deploy;

import java.io.File;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.ow2.petals.basisapi.exception.PetalsException;
import org.ow2.petals.clientserverapi.configuration.ContainerConfiguration;
import org.ow2.petals.jbi.descriptor.original.generated.Jbi;
import org.ow2.petals.jbi.descriptor.original.generated.ServiceUnit;
import org.ow2.petals.microkernel.api.jbi.management.Context;
import org.ow2.petals.microkernel.api.jbi.management.ManagementException;
import org.ow2.petals.microkernel.api.system.repository.RepositoryService;
import org.ow2.petals.microkernel.jbi.management.task.AbstractLoggableTask;

import com.ebmwebsourcing.easycommons.log.LoggingUtil;

/**
 * This task extracts service unit archives into the repository.
 * <p>
 * Inputs : {@link Jbi}, parent service assembly install root {@link File} <br/>
 * Outputs : SU install roots {@link Map} of {@link File} <br/>
 * The task has a non-robust approach: it stops at the first subtask failed and has no undo capability.
 * </p>
 * 
 * @author Olivier Fabre - EBM WebSourcing
 * @author Frederic Gardes - EBM WebSourcing
 */
public class ExtractSUsTask extends AbstractLoggableTask {

    /**
     * Platform Component :: Repository Service
     */
    protected RepositoryService repositoryService;

    public ExtractSUsTask(final LoggingUtil log, final RepositoryService repositoryService,
            final ContainerConfiguration containerConfiguration) {
        super(log, containerConfiguration);
        this.repositoryService = repositoryService;
    }

    @Override
    public void execute(Context context) throws Exception {
        Map<String, File> suInstallFiles = this.expandSUs(context.getInstallFile(), context.getDescriptor());
        context.setSuInstallFiles(suInstallFiles);
    }

    /**
     * Explode service unit packages into the repository
     * 
     * @param saInstallRoot
     *            the root {@link URI} of the service assembly
     * @param descriptor
     *            the service assembly descriptor
     * @return the map containing all su install roots
     * @throws ManagementException
     *             if the repository service doesn't manage to expend SU archive
     * @throws PetalsException
     * @throws MalformedURLException
     * @throws URISyntaxException
     */
    protected Map<String, File> expandSUs(File saInstallFile, Jbi descriptor) throws ManagementException {
        Map<String, File> suInstallFiles = new HashMap<String, File>();
        List<ServiceUnit> sus = descriptor.getServiceAssembly().getServiceUnit();
        for (ServiceUnit unit : sus) {
            String suZipName = unit.getTarget().getArtifactsZip().trim();
            File suArchiveFile = new File(saInstallFile, suZipName);
            File suInstallFile = this.expandSUIntoSA(unit.getIdentification().getName(), suArchiveFile,
                    descriptor.getServiceAssembly().getIdentification().getName());
            suInstallFiles.put(unit.getIdentification().getName(), suInstallFile);
        }
        return suInstallFiles;
    }

    /**
     * Expand a service unit package into its service assembly installRoot.
     * 
     * @param suId
     *            the unique identifier of the su
     * @param suZipLocation
     *            URI representing the su zip location
     * @param saId
     *            the unique identifier of the sa
     * @return SU installationRoot, null if it fails
     * @throws ManagementException
     *             if it fails to expend the su package to the repository
     */
    protected File expandSUIntoSA(String suId, File suZipLocation, String saId) throws ManagementException {
        try {
            return this.repositoryService.explodeSUIntoSAInstallDirectory(suId, suZipLocation, saId);
        } catch (IOException e) {
            throw new ManagementException("Unexpected error expending service JBI installation "
                    + "package into repository", e);
        }
    }

    @Override
    public void undo(Context context) throws Exception {
        // The folder will be removed by the undo of the CreateSARepository task
    }
}
