/**
 * Copyright (c) 2006-2012 EBM WebSourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.jbi.management.task.deployment.deploy;

import java.io.File;
import java.net.URI;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.ow2.petals.jbi.descriptor.original.generated.Jbi;
import org.ow2.petals.jbi.descriptor.original.generated.ServiceUnit;
import org.ow2.petals.microkernel.api.jbi.management.Context;
import org.ow2.petals.microkernel.api.jbi.management.Task;
import org.ow2.petals.microkernel.api.system.repository.RepositoryService;

/**
 * This task retrieves the SA install root and all SUs install roots.
 * <p>
 * Inputs : {@link Jbi} <br/>
 * Outputs : SA install root {@link File}, SU install roots {@link Map} of {@link File}<br/>
 * The task has a non-robust approach : it stops at the first subtask failed and has no undo capability.
 * </p>
 * 
 * @author Olivier Fabre - EBM WebSourcing
 * @author Frederic Gardes - EBM WebSourcing
 */
public class RetrieveSAAndSUInstallRootsTask implements Task {

    /**
     * Platform Component :: Repository Service
     */
    protected RepositoryService repositoryService;

    public RetrieveSAAndSUInstallRootsTask(RepositoryService repositoryService) {
        super();
        this.repositoryService = repositoryService;
    }

    @Override
    public void execute(Context context) throws Exception {
        Jbi descriptor = context.getDescriptor();
        File saInstallFile = this.repositoryService
                .getServiceAssemblyInstallDirectory(descriptor.getServiceAssembly().getIdentification().getName());
        Map<String, File> suInstallFiles = this.retrieveSUInstallFiles(saInstallFile, descriptor);
        context.setInstallFile(saInstallFile);
        context.setSuInstallFiles(suInstallFiles);
    }

    /**
     * Retrieve all SU installation root {@link URI}
     * 
     * @param descriptor
     *            the service assembly descriptor
     * @param installationRoot
     *            the SA installation root {@link URI}
     * @return the service unit install roots {@link Map}
     */
    protected Map<String, File> retrieveSUInstallFiles(File saInstallFile, Jbi descriptor) {
        Map<String, File> suInstallFiles = new HashMap<String, File>();
        List<ServiceUnit> sus = descriptor.getServiceAssembly().getServiceUnit();
        for (ServiceUnit unit : sus) {
            String suName = unit.getIdentification().getName();
            suInstallFiles.put(suName, new File(saInstallFile, suName));
        }
        return suInstallFiles;
    }

    @Override
    public void undo(Context context) throws Exception {
        // nothing to do
    }

}
