/**
 * Copyright (c) 2006-2012 EBM WebSourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.jbi.management.task.deployment.undeploy;

import javax.xml.namespace.QName;

import org.ow2.petals.clientserverapi.configuration.ContainerConfiguration;
import org.ow2.petals.clientserverapi.jbi.messaging.endpoint.exception.EndpointDirectoryException;
import org.ow2.petals.jbi.descriptor.original.generated.Connection;
import org.ow2.petals.jbi.descriptor.original.generated.ServiceAssembly;
import org.ow2.petals.microkernel.api.container.ServiceAssemblyLifeCycle;
import org.ow2.petals.microkernel.api.jbi.management.Context;
import org.ow2.petals.microkernel.api.jbi.messaging.endpoint.EndpointDirectoryService;
import org.ow2.petals.microkernel.jbi.management.task.AbstractLoggableTask;

import com.ebmwebsourcing.easycommons.log.LoggingUtil;

/**
 * This task remove all connections from the registry.
 * <p>
 * Inputs : {@link ServiceAssemblyLifeCycle} <br/>
 * The task has a non-robust approach: it tries to do all the subtasks and
 * doesn't have the undo capability.
 * </p>
 * 
 * @author Olivier Fabre - EBM WebSourcing
 * @author Frederic Gardes - EBM WebSourcing
 */
public class RemoveAllConnectionsTask extends AbstractLoggableTask {

    protected final EndpointDirectoryService endpointDirectory;

    public RemoveAllConnectionsTask(final LoggingUtil log, final EndpointDirectoryService endpointDirectory,
            final ContainerConfiguration containerConfiguration) {
        super(log, containerConfiguration);
        this.endpointDirectory = endpointDirectory;
    }

    @Override
    public void execute(final Context context) throws Exception {
        if (context.getSaLifeCycle() != null) {
            this.removeConnections(context.getSaLifeCycle());
        }
    }

    /**
     * Remove all connections registered during the SA deployment process
     * 
     * @param saLifeCycle
     *            the service assembly life cycle
     * @throws Exception
     *             if problem occurs during the deletion of a specific
     *             connection
     */
    protected void removeConnections(final ServiceAssemblyLifeCycle saLifeCycle) {
        final ServiceAssembly saDesc = saLifeCycle.getServiceAssembly();
        if ((saDesc != null) && (saDesc.getConnections() != null)) {
            for (final Connection connection : saDesc.getConnections().getConnection()) {
                if ((connection != null) && (connection.getConsumer() != null)) {
                    final QName consInterface = connection.getConsumer().getInterfaceName();
                    final QName consService = connection.getConsumer().getServiceName();
                    final String consEndpoint = connection.getConsumer().getEndpointName();
                    final QName provService = connection.getProvider().getServiceName();
                    final String provEndpoint = connection.getProvider().getEndpointName();
                    if ((provEndpoint != null) && (provService != null)) {
                        if (consInterface != null) {
                            try {
                                this.endpointDirectory.deregisterConnection(consInterface,
                                        provService, provEndpoint);
                            } catch (final EndpointDirectoryException e) {
                                this.log
                                        .error(
                                                "Unable to unregister a connection with a resolved interface",
                                                e);
                            }
                        } else if ((consService != null) && (consEndpoint != null)) {
                            try {
                                this.endpointDirectory.deregisterConnection(consService,
                                        consEndpoint, provService, provEndpoint);
                            } catch (final EndpointDirectoryException e) {
                                this.log
                                        .error(
                                                "Unable to unregister a connection with an unresolved interface",
                                                e);
                            }
                        }
                    }
                }
            }
        }
    }

    @Override
    public void undo(final Context context) throws Exception {
        // TODO : recreate all the connections
    }
}
