/**
 * Copyright (c) 2006-2012 EBM WebSourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.jbi.management.task.deployment.undeploy;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import javax.jbi.JBIException;
import javax.jbi.component.ServiceUnitManager;
import javax.jbi.management.DeploymentException;

import org.ow2.petals.clientserverapi.configuration.ContainerConfiguration;
import org.ow2.petals.microkernel.api.container.ContainerService;
import org.ow2.petals.microkernel.api.container.Installer;
import org.ow2.petals.microkernel.api.container.ServiceAssemblyLifeCycle;
import org.ow2.petals.microkernel.api.container.ServiceUnitLifeCycle;
import org.ow2.petals.microkernel.api.jbi.management.Context;
import org.ow2.petals.microkernel.api.jbi.management.XMLResult;
import org.ow2.petals.microkernel.jbi.management.task.AbstractLoggableTask;

import com.ebmwebsourcing.easycommons.log.LoggingUtil;

/**
 * This task executes the undeployment process of a service unit, for each
 * Service Unit included into the service assembly.
 * 
 * <p>
 * Inputs : {@link ServiceAssemblyLifeCycle}, {@link XMLResult}.<br />
 * The task has got a mixed approach: it tries to do all the subtasks<br />
 * and doesn't have the undo capability. But if the service assembly isn't over
 * shutdown or unknown state, to implement the JBI specification, it stops and
 * throws an exception to launch the undo process.
 * </p>
 * 
 * @author Olivier Fabre - EBM WebSourcing
 * @author Frederic Gardes - EBM WebSourcing
 */
public class UndeployAllSUsTask extends AbstractLoggableTask {

    /**
     * JBI Container :: Container service
     */
    protected ContainerService containerService;

    public UndeployAllSUsTask(final LoggingUtil log, final ContainerService containerService,
            final ContainerConfiguration containerConfiguration) {
        super(log, containerConfiguration);
        this.containerService = containerService;
    }

    @Override
    public void execute(Context context) throws Exception {
        if ((context.getSaLifeCycle() != null) && (context.getXmlResult() != null)) {
            this.undeploySUS(context.getSaLifeCycle(), context.getXmlResult());
        }
    }

    /**
     * Undeploy all SUs by calling the "undeploySU" method of the corresponding
     * SUDeploymentManager.
     * 
     * @param saLifeCycle
     *            the parent {@link ServiceAssemblyLifeCycle}
     * @param xmlResult
     *            xml task result
     * @throws Exception
     *             if an error occurs during the undeployment of a service unit
     */
    protected void undeploySUS(final ServiceAssemblyLifeCycle saLifeCycle, XMLResult xmlResult)
            throws JBIException {
        // Undeploy the SUs only if the SA is in SHUTDOWN or UNKNOW state
        if (saLifeCycle.isShutdownState()) {
            // JBI spec specifies that service units must be undeployed in the
            // reverse order than the order of the initial declared list
            List<ServiceUnitLifeCycle> suLifeCycles = new ArrayList<ServiceUnitLifeCycle>(
                    saLifeCycle.getServiceUnitLifeCycles());
            Collections.reverse(suLifeCycles);
            StringBuilder componentOutput = new StringBuilder();
            for (ServiceUnitLifeCycle suLifeCycle : suLifeCycles) {
                String xmlReturn = this.undeploySU(suLifeCycle);
                // unregister the SU from the SA
                saLifeCycle.unregisterSU(suLifeCycle);
                componentOutput.append(xmlReturn);
            }
            xmlResult.addComponentTaskResult(componentOutput);
        } else {
            throw new JBIException(
                    "Service Assembly can't be undeployed in this state: " + saLifeCycle.getCurrentState());
        }
    }

    /**
     * Process the undeployment of the given service unit.
     * 
     * @param suLifeCycle
     *            the service unit life cycle
     * @throws Exception
     *             if an error occurs during the undeployment process : can't
     *             retrieve the {@link Installer}, can't retrieve the component
     *             {@link ServiceUnitManager} (SUM) or an exception is thrown
     *             during the call of the undeploy method of the SUM
     */
    protected String undeploySU(ServiceUnitLifeCycle suLifeCycle) {
        String xmlReturn = null;

        // Retrieve the component life cycle
        Installer installer = this.containerService.getInstallerByName(suLifeCycle
                .getTargetComponentName());
        if (installer == null) {
            this.log.error("The component '" + suLifeCycle.getTargetComponentName()
                    + "' is not deployed. The service unit cannot be undeployed from it.");
        } else {
            // Retrieve the service unit manager of the component
            ServiceUnitManager sum = installer.getComponent().getServiceUnitManager();
            if (sum == null) {
                this.log.error("The service unit manager is null for the component '"
                        + suLifeCycle.getTargetComponentName()
                        + "'. The service units can not be undeployed from it.");
            } else {
                // Set the context class loader to the component's one just
                // for the method invocation
                try {
                    Thread.currentThread().setContextClassLoader(sum.getClass().getClassLoader());
                    xmlReturn = sum.undeploy(suLifeCycle.getSuName(), suLifeCycle
                            .getServiceUnitRootPath());
                } catch (DeploymentException e) {
                    this.log.error("The service unit manager of the component '"
                            + suLifeCycle.getTargetComponentName()
                            + "' cannot  undeploy a service unit.");
                } finally {
                    Thread.currentThread().setContextClassLoader(this.getClass().getClassLoader());
                }
            }
        }

        return xmlReturn;
    }

    @Override
    public void undo(Context context) throws Exception {
        // TODO : redeploy all the service units
    }
}
