/**
 * Copyright (c) 2013-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.jbi.messaging.endpoint;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Logger;

import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;
import javax.xml.transform.TransformerException;

import org.objectweb.fractal.fraclet.annotations.Component;
import org.objectweb.fractal.fraclet.annotations.Interface;
import org.objectweb.fractal.fraclet.annotations.Requires;
import org.ow2.petals.clientserverapi.jbi.messaging.endpoint.EndpointDirectoryServiceConstants;
import org.ow2.petals.clientserverapi.jbi.messaging.endpoint.exception.EndpointDirectoryException;
import org.ow2.petals.jbi.servicedesc.endpoint.Location;
import org.ow2.petals.jbi.servicedesc.endpoint.PetalsServiceEndpoint;
import org.ow2.petals.jbi.servicedesc.endpoint.PetalsServiceEndpoint.EndpointType;
import org.ow2.petals.microkernel.api.container.ContainerService;
import org.ow2.petals.microkernel.api.container.Installer;
import org.ow2.petals.microkernel.api.jbi.messaging.endpoint.EndpointDirectoryService;
import org.ow2.petals.microkernel.api.jbi.messaging.endpoint.exception.EndpointAlreadyExistsException;
import org.ow2.petals.microkernel.api.jbi.messaging.endpoint.exception.EndpointDoesNotExistsException;
import org.ow2.petals.microkernel.api.system.sharedarea.SharedAreaService;
import org.ow2.petals.microkernel.api.system.sharedarea.exception.SharedAreaException;
import org.w3c.dom.Document;
import org.w3c.dom.DocumentFragment;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

import com.ebmwebsourcing.easycommons.log.LoggingUtil;
import com.ebmwebsourcing.easycommons.xml.XMLHelper;

/**
 * The registry implementation based on the Petals Registry Overlay.
 * 
 * @author Christophe DENEUX - Linagora
 */

@Component(provides = @Interface(name = EndpointDirectoryService.FRACTAL_SRV_ITF_NAME, signature = EndpointDirectoryService.class))
public class EndpointDirectoryServiceImpl implements EndpointDirectoryService {

    /**
     * The shared area service used as distributed database
     */
    @Requires(name = EndpointDirectoryService.SHARED_AREA_FRACTAL_ITF_NAME)
    protected SharedAreaService sharedAreaService;

    @Requires(name = EndpointDirectoryService.CONTAINER_FRACTAL_ITF_NAME)
    protected ContainerService containerService;

    /**
     * Logging utilities
     */
    private final LoggingUtil log = new LoggingUtil(Logger.getLogger(COMPONENT_LOGGER_NAME));

    @Override
    public List<Map<String, Object>> getAllInternalEndpoints() {
        this.log.start();

        try {
            final Collection<PetalsServiceEndpoint> internalEndpoints = this.getInternalEndpoints();
            return EndpointDirectoryServiceImpl.convertEndpoints(internalEndpoints);
        } catch (final EndpointDirectoryException e) {
            this.log.warning("An error ocurs getting internal endpoints", e);
            return Collections.emptyList();
        } finally {
            this.log.end();
        }
    }

    @Override
    public List<Map<String, Object>> getAllExternalEndpoints() {
        this.log.start();

        try {
            final Collection<PetalsServiceEndpoint> externalEndpoints = this.getExternalEndpoints();
            return EndpointDirectoryServiceImpl.convertEndpoints(externalEndpoints);
        } catch (final EndpointDirectoryException e) {
            this.log.warning("An error ocurs getting external endpoints", e);
            return Collections.emptyList();
        } finally {
            this.log.end();
        }
    }

    /**
     * Convert a {@link PetalsServiceEndpoint} in its JMX representation. See
     * {@link EndpointDirectoryService#getAllEndpoints()}, {@link EndpointDirectoryService#getAllInternalEndpoints()} or
     * {@link EndpointDirectoryService#getAllExternalEndpoints()}
     * 
     * @param serviceEndpoints
     *            A list of service endpoints to convert
     * @return The service endpoints converted. Not <code>null</code>. Can be empty.
     */
    private static List<Map<String, Object>> convertEndpoints(
            final Collection<PetalsServiceEndpoint> serviceEndpoints) {

        final List<Map<String, Object>> convertedEndpoints = new ArrayList<Map<String, Object>>(
                serviceEndpoints.size());
        for (final PetalsServiceEndpoint serviceEndpoint : serviceEndpoints) {
            final Map<String, Object> endpointMap = new HashMap<String, Object>();
            endpointMap.put(EndpointDirectoryServiceConstants.Endpoint.COMPONENT_NAME, serviceEndpoint
                    .getLocation().getComponentName());
            endpointMap.put(EndpointDirectoryServiceConstants.Endpoint.CONTAINER_NAME, serviceEndpoint
                    .getLocation().getContainerName());
            endpointMap.put(EndpointDirectoryServiceConstants.Endpoint.ENDPOINT_NAME,
                    serviceEndpoint.getEndpointName());
            endpointMap.put(EndpointDirectoryServiceConstants.Endpoint.SERVICE_NAME, serviceEndpoint
                    .getServiceName().toString());

            final EndpointType endpointType = serviceEndpoint.getType();
            if (endpointType.equals(EndpointType.INTERNAL)) {
                endpointMap.put(EndpointDirectoryServiceConstants.Endpoint.TYPE,
                        EndpointDirectoryServiceConstants.EndpointType.INTERNAL);
            } else if (endpointType.equals(EndpointType.EXTERNAL)) {
                endpointMap.put(EndpointDirectoryServiceConstants.Endpoint.TYPE,
                        EndpointDirectoryServiceConstants.EndpointType.EXTERNAL);
            }

            final QName[] interfaces = serviceEndpoint.getInterfaces();
            final List<String> convertedInterfaceNames = new ArrayList<String>();
            if (interfaces.length > 0) {
                for (final QName interfaceName : interfaces) {
                    convertedInterfaceNames.add(interfaceName.toString());
                }
            } else {
                convertedInterfaceNames.add(EndpointDirectoryServiceConstants.Interface.UNDEFINED);
            }
            endpointMap.put(EndpointDirectoryServiceConstants.Endpoint.INTERFACE_NAMES,
                    convertedInterfaceNames.toArray(new String[convertedInterfaceNames.size()]));

            convertedEndpoints.add(endpointMap);
        }

        return convertedEndpoints;
    }

    @Override
    public List<Map<String, Object>> getAllEndpoints() {

        this.log.start();

        try {
            final List<Map<String, Object>> allEndpointsMap = this.getAllExternalEndpoints();
            allEndpointsMap.addAll(this.getAllInternalEndpoints());

            return allEndpointsMap;
        } finally {
            this.log.end();
        }
    }

    @Override
    public String getDescription(final String serviceName, final String endpointName)
            throws EndpointDirectoryException {

        if (this.log.isDebugEnabled()) {
            this.log.start("with params serviceName = " + serviceName + ", endpointName = " + endpointName);
        }

        try {
            final Document description = this.sharedAreaService.getDescription(QName.valueOf(serviceName),
                    endpointName);
            return XMLHelper.createStringFromDOMDocument(description);
        } catch (final SharedAreaException | TransformerException | RuntimeException e) {
            throw new EndpointDirectoryException(e);
        }
    }

    @Override
    public Document getDescription(final ServiceEndpoint endpoint) throws EndpointDirectoryException {
        try {
            if (this.sharedAreaService.isRegisteredEndpoint(endpoint.getServiceName(), endpoint.getEndpointName())) {
                return this.sharedAreaService.getDescription(endpoint);
            } else {
                throw new EndpointDoesNotExistsException(endpoint.getEndpointName(), endpoint.getServiceName());
            }
        } catch (final SharedAreaException e) {
            throw new EndpointDirectoryException(e);
        }
    }

    @Override
    public Collection<PetalsServiceEndpoint> getInternalEndpoints() throws EndpointDirectoryException {
        try {
            return this.sharedAreaService.getEndpoints(EndpointType.INTERNAL);
        } catch (final SharedAreaException e) {
            throw new EndpointDirectoryException(e);
        }
    }

    @Override
    public Collection<PetalsServiceEndpoint> getExternalEndpoints() throws EndpointDirectoryException {
        try {
            return this.sharedAreaService.getEndpoints(EndpointType.EXTERNAL);
        } catch (final SharedAreaException e) {
            throw new EndpointDirectoryException(e);
        }
    }

    @Override
    public Collection<PetalsServiceEndpoint> getEndpoints() throws EndpointDirectoryException {
        try {
            return this.sharedAreaService.getEndpoints(null);
        } catch (final SharedAreaException e) {
            throw new EndpointDirectoryException(e);
        }
    }

    @Override
    public PetalsServiceEndpoint activateEndpoint(final QName serviceName, final String endpointName,
            final QName[] interfaces, final Document description, final Location location,
            final boolean canOverwrite) throws EndpointDirectoryException {

        try {
            final PetalsServiceEndpoint endpoint = this.sharedAreaService.registerInternalEndpoint(endpointName,
                    serviceName, interfaces, location, description, canOverwrite);
            if (endpoint == null) {
                throw new EndpointAlreadyExistsException(endpointName, serviceName);
            }
            return endpoint;
        } catch (final SharedAreaException e) {
            throw new EndpointDirectoryException(e);
        }

    }

    @Override
    public void deactivateEndpoint(final String endpointName, final QName serviceName)
            throws EndpointDirectoryException {

        assert endpointName != null;
        assert serviceName != null;

        if ((endpointName == null) || (serviceName == null)) {
            throw new EndpointDirectoryException("Can not deactivate null endpoint");
        }

        try {
            final PetalsServiceEndpoint edpt = this.sharedAreaService.deregisterEndpoint(serviceName, endpointName);
            if (edpt == null) {
                throw new EndpointDoesNotExistsException(endpointName, serviceName);
            }
        } catch (final SharedAreaException e) {
            throw new EndpointDirectoryException(e);
        }

    }

    @Override
    public void registerExternalEndpoint(final ServiceEndpoint externalSvcEp, final Location location,
            final Document description, final boolean canOverride) throws EndpointDirectoryException {

        if (this.log.isDebugEnabled()) {
            this.log.start("with params externalSvcEp = " + externalSvcEp);
        }

        try {
            final PetalsServiceEndpoint endpoint = this.sharedAreaService.registerExternalEndpoint(externalSvcEp,
                    location, description, canOverride);
            if (endpoint != null) {
                throw new EndpointAlreadyExistsException(externalSvcEp.getEndpointName(),
                        externalSvcEp.getServiceName());
            }
        } catch (final SharedAreaException e) {
            throw new EndpointDirectoryException(e);
        } finally {
            this.log.end();
        }

    }

    @Override
    public void deregisterExternalEndpoint(final ServiceEndpoint externalSvcEp) throws EndpointDirectoryException {
        this.deactivateEndpoint(externalSvcEp.getEndpointName(), externalSvcEp.getServiceName());
    }

    @Override
    public void registerConnection(final QName consInterface, final QName provService, final String provEndpoint)
            throws EndpointDirectoryException {
        throw new UnsupportedOperationException();
    }

    @Override
    public void registerConnection(final QName consService, final String consEndpoint, final QName provService,
            final String provEndpoint) throws EndpointDirectoryException {
        throw new UnsupportedOperationException();
    }

    @Override
    public void deregisterConnection(final QName consInterface, final QName provService, final String provEndpoint)
            throws EndpointDirectoryException {
        throw new UnsupportedOperationException();
    }

    @Override
    public void deregisterConnection(final QName consService, final String consEndpoint, final QName provService,
            final String provEndpoint) throws EndpointDirectoryException {
        throw new UnsupportedOperationException();
    }

    @Override
    public Collection<PetalsServiceEndpoint> getInternalEndpointsForInterface(final QName interfaceName)
            throws EndpointDirectoryException {

        if (this.log.isDebugEnabled()) {
            this.log.start("with params interfaceName = " + interfaceName);
        }

        try {
            return this.sharedAreaService.getEndpointsForInterface(interfaceName, EndpointType.INTERNAL);
        } catch (final SharedAreaException e) {
            throw new EndpointDirectoryException(e);
        } finally {
            this.log.end();
        }

    }

    @Override
    public Collection<PetalsServiceEndpoint> getInternalEndpointsForService(final QName serviceName)
            throws EndpointDirectoryException {

        if (this.log.isDebugEnabled()) {
            this.log.start("with params serviceName = " + serviceName);
        }

        try {
            return this.sharedAreaService.getEndpointsForService(serviceName, EndpointType.INTERNAL);
        } catch (final SharedAreaException e) {
            throw new EndpointDirectoryException(e);
        } finally {
            this.log.end();
        }

    }

    @Override
    public Collection<PetalsServiceEndpoint> getExternalEndpointsForInterface(final QName interfaceName)
            throws EndpointDirectoryException {

        if (this.log.isDebugEnabled()) {
            this.log.start("with params interfaceName = " + interfaceName);
        }

        try {
            return this.sharedAreaService.getEndpointsForInterface(interfaceName, EndpointType.EXTERNAL);
        } catch (final SharedAreaException e) {
            throw new EndpointDirectoryException(e);
        } finally {
            this.log.end();
        }
    }

    @Override
    public Collection<PetalsServiceEndpoint> getExternalEndpointsForService(final QName serviceName)
            throws EndpointDirectoryException {

        if (this.log.isDebugEnabled()) {
            this.log.start("with params serviceName = " + serviceName);
        }

        try {
            return this.sharedAreaService.getEndpointsForService(serviceName, EndpointType.EXTERNAL);
        } catch (final SharedAreaException e) {
            throw new EndpointDirectoryException(e);
        } finally {
            this.log.end();
        }
    }

    @Override
    public PetalsServiceEndpoint getEndpoint(final QName service, final String endpointName)
            throws EndpointDirectoryException {
        assert service != null;
        assert endpointName != null;

        if (this.log.isDebugEnabled()) {
            this.log.start("with params serviceName = " + service.toString() + ", endpointName = " + endpointName);
        }

        try {
            return this.sharedAreaService.getEndpoint(service, endpointName);
        } catch (final SharedAreaException e) {
            throw new EndpointDirectoryException(e);
        } finally {
            this.log.end();
        }
    }

    @Override
    public PetalsServiceEndpoint getInternalEndpoint(final QName service, final String endpointName)
            throws EndpointDirectoryException {
        this.log.call();

        final PetalsServiceEndpoint endpoint = getEndpoint(service, endpointName);
        if (endpoint != null && EndpointType.INTERNAL.equals(endpoint.getType())) {
            return endpoint;
        } else {
            return null;
        }
    }

    @Override
    public PetalsServiceEndpoint getExternalEndpoint(final QName service, final String endpointName)
            throws EndpointDirectoryException {
        this.log.call();

        final PetalsServiceEndpoint endpoint = getEndpoint(service, endpointName);
        if (endpoint != null && EndpointType.EXTERNAL.equals(endpoint.getType())) {
            return endpoint;
        } else {
            return null;
        }
    }

    @Override
    public ServiceEndpoint resolveEndpointReference(final DocumentFragment epr) throws EndpointDirectoryException {
        final Node firstChild = epr.getFirstChild();
        if (firstChild instanceof Element && PetalsServiceEndpoint.EPR_NAME.equals(firstChild.getLocalName())
                && PetalsServiceEndpoint.JBI_NAMESPACE.equals(firstChild.getNamespaceURI())) {
            final Element e = (Element) firstChild;
            final String endpointName = e.getAttribute(PetalsServiceEndpoint.ENDPOINT_NAME);
            if (endpointName.trim().isEmpty()) {
                throw new EndpointDirectoryException("Missing endpoint name attribute in EPR");
            }
            // this is meant to be a xsd:QName ([prefix:]localName)
            final String service = e.getAttribute(PetalsServiceEndpoint.SERVICE_NAME);
            final String[] split = service.split(":");
            final QName serviceName;
            if (split.length == 0) {
                throw new EndpointDirectoryException("Missing service name attribute in EPR");
            } else if (split.length == 1) {
                serviceName = new QName(split[0]);
            } else if (split.length == 2) {
                serviceName = new QName(e.lookupNamespaceURI(split[0]), split[1]);
            } else {
                throw new EndpointDirectoryException("Invalid service name attribute in EPR: '" + service + "'");
            }
            try {
                final PetalsServiceEndpoint endpoint = this.sharedAreaService.getEndpoint(serviceName, endpointName);
                if (endpoint != null) {
                    return endpoint;
                }
            } catch (SharedAreaException e1) {
                throw new EndpointDirectoryException(e1);
            }
        }

        // Here we are most certainly resolving an EPR for an external endpoint...
        // -> a BC (usually) will give us the corresponding internal endpoint

        // TODO PETALSESBCONT-378 it should be using all the nodes available in the bus!
        // For example using something similar to the ExchangeCheckerClient? or using hazelcast?
        for (final Installer installer : this.containerService.getInstallers().values()) {
            final ServiceEndpoint endpoint = installer.getComponent().resolveEndpointReference(epr);
            if (endpoint != null) {
                // we pack it so that we have the location
                return new DynamicEndpointImpl(endpoint.getEndpointName(), endpoint.getServiceName(),
                        endpoint.getInterfaces(), installer.getLocation());
            }
        }

        return null;
    }
}
