/**
 * Copyright (c) 2005-2012 EBM WebSourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.jbi.messaging.exchange;

import java.net.URI;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.concurrent.TimeUnit;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.component.ComponentContext;
import javax.jbi.messaging.DeliveryChannel;
import javax.jbi.messaging.ExchangeStatus;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessageExchange.Role;
import javax.jbi.messaging.MessageExchangeFactory;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;
import javax.xml.transform.Source;

import org.ow2.easywsdl.wsdl.api.abstractItf.AbsItfOperation.MEPPatternConstants;
import org.ow2.petals.clientserverapi.configuration.ContainerConfiguration;
import org.ow2.petals.jbi.messaging.PetalsDeliveryChannel;
import org.ow2.petals.jbi.messaging.exchange.AbstractMessageExchangeFactory;
import org.ow2.petals.jbi.messaging.exchange.MessageExchangeWrapper;
import org.ow2.petals.jbi.messaging.exchange.PetalsMessageExchange;
import org.ow2.petals.jbi.messaging.exchange.impl.DefaultMessageExchangeFactory;
import org.ow2.petals.jbi.messaging.exchange.types.InOnlyImpl;
import org.ow2.petals.jbi.messaging.exchange.types.InOptionalOutImpl;
import org.ow2.petals.jbi.messaging.exchange.types.InOutImpl;
import org.ow2.petals.jbi.messaging.exchange.types.RobustInOnlyImpl;
import org.ow2.petals.jbi.servicedesc.endpoint.Location;
import org.ow2.petals.microkernel.api.jbi.component.PetalsComponentContext;
import org.ow2.petals.microkernel.api.jbi.messaging.exception.RoutingException;
import org.ow2.petals.microkernel.jbi.messaging.routing.module.EndpointResolverModule;

import com.ebmwebsourcing.easycommons.log.LoggingUtil;
import com.ebmwebsourcing.easycommons.xml.SourceHelper;

/**
 * 
 * TODO resolve jbi logical endpoint (it is meant to be based on the SU... see also {@link EndpointResolverModule} where
 * there is some related code...)
 * 
 * @author Adrien LOUIS - EBM WebSourcing
 * @author Olivier Fabre - EBM WebSourcing
 * @author Roland Naudin - EBM WebSourcing
 * @author Victor Noel
 */
public class DeliveryChannelImpl implements PetalsDeliveryChannel {

    /**
     * Used to recognise the role that did the sendSync in case of crossing sendSync
     * 
     * TODO all of this is quite complex, maybe we should store this kind of information directly in the
     * {@link MessageExchange} implementation in order to prevent components to touch the properties as well as more
     * simply store the sendsync state...
     */
    private static final String PROPERTY_SENDSYNC_AS_CONSUMER = "org.ow2.petals.microkernel.jbi.messaging.exchange.DeliveryChannelImpl.SendSyncAsConsumer";

    /**
     * Used to recognise the role that did the sendSync in case of crossing sendSync
     */
    private static final String PROPERTY_SENDSYNC_AS_PROVIDER = "org.ow2.petals.microkernel.jbi.messaging.exchange.DeliveryChannelImpl.SendSyncAsProvider";

    private final PetalsComponentContext componentContext;

    private final Location location;

    private final LoggingUtil log;

    /**
     * Map of exchanges that are blocked during a synchronous send
     * 
     * The key is the message exchange id, and the value is the local (in case of distributed implementation) wrapper
     */
    private final ConcurrentMap<String, StoredSendSyncExchange> pendingSyncExchanges = new ConcurrentHashMap<>();

    private final BlockingQueue<PetalsMessageExchange> exchangesQueue;

    /**
     * Offering timeout used by component queues
     */
    private final long transporterQueueOfferingTimeout;

    private volatile boolean opened = true;

    public DeliveryChannelImpl(final Location location, final PetalsComponentContext context, final Logger logger) {

        this.location = location;
        this.log = new LoggingUtil(logger);

        final ContainerConfiguration containerConfiguration = context.getContainerConfiguration();

        final int transporterQueueMaxSize = containerConfiguration.getTransporterQueueMaxSize();
        if (this.log.isDebugEnabled()) {
            this.log.debug("Using a queue bounded to " + transporterQueueMaxSize);
        }

        this.transporterQueueOfferingTimeout = containerConfiguration.getTransporterQueueOfferingTimeout();
        if (this.log.isDebugEnabled()) {
            this.log.debug(
                    "The offering timeout of component queues is set to " + transporterQueueOfferingTimeout + "'.");
        }

        this.componentContext = context;

        this.exchangesQueue = new ArrayBlockingQueue<>(transporterQueueMaxSize);
    }

    @Override
    public ComponentContext getContext() {
        return this.componentContext;
    }

    @Override
    public Location getLocation() {
        return this.location;
    }

    @Override
    public Logger getLogger() {
        return this.log.getLogger();
    }

    private void checkDeliveryChannelIsOpened() throws MessagingException {
        if (!this.opened) {
            throw new MessagingException("DeliveryChannel is closed");
        }
    }

    @Override
    public MessageExchangeWrapper accept() throws MessagingException {
        this.log.call();

        return accept(0);
    }

    @Override
    public MessageExchangeWrapper accept(final long timeoutMS) throws MessagingException {
        this.log.call(timeoutMS);

        if (timeoutMS < 0) {
            throw new IllegalArgumentException("Timeout must be positive");
        }

        this.checkDeliveryChannelIsOpened();

        PetalsMessageExchange exchange = null;
        try {
            exchange = poll(timeoutMS);
        } catch (final InterruptedException e) {
            // The thread receiving message was interrupted: most certainly it happened because the JBI component was
            // stopped

            // Even if one of the causes is an InterruptedException,
            // we must throw the MessagingException according to the specification
            this.log.debug(
                    "accept() was interrupted, converting to a MessagingException as defined in the JBI specification.");
            throw new MessagingException("accept() was interrupted", e);
        }

        this.log.end();

        // the idea is that we are giving the exchange to its owner, so the observer role is the role of the
        // exchange, hence we can simply call the constructor of MessageExchangeWrapper.
        return exchange != null ? createWrapper(exchange) : null;
    }

    private PetalsMessageExchange poll(final long timeoutMS) throws InterruptedException {
        this.log.start();

        this.log.debug("Trying to receive an exchange with timeout " + timeoutMS + "ms");

        final PetalsMessageExchange exchange;
        if (timeoutMS > 0) {
            // blocking call with timeout
            exchange = this.exchangesQueue.poll(timeoutMS, TimeUnit.MILLISECONDS);
        } else {
            // blocking call without timeout
            exchange = this.exchangesQueue.take();
        }

        if (exchange != null) {
            this.log.end("Exchange Id " + exchange.getExchangeId());
        } else {
            this.log.end("No exchange");
        }

        return exchange;
    }

    public void offerExchange(final PetalsMessageExchange exchange) throws RoutingException {

        if (!this.opened) {
            throw new RoutingException("Delivery channel is closed, can't deliver message to component.");
        }

        final String exchangeId = exchange.getExchangeId();

        // is that a message that someone is synchronously waiting for?
        // note: we must check the property and not the map, because the entry can be removed from the map
        // when a send times out.
        final String sendSyncRolePropertyName;
        // See sendSyncInNMR method for explanation about prefix
        final String prefix;
        if (exchange.isConsumer()) {
            prefix = "C";
            sendSyncRolePropertyName = PROPERTY_SENDSYNC_AS_CONSUMER;
        } else {
            prefix = "P";
            sendSyncRolePropertyName = PROPERTY_SENDSYNC_AS_PROVIDER;
        }
        // the router is responsible of switching the roles of the exchange, so it must have the same role as the one it
        // had when I sent the message
        final boolean answerToSendSync = exchange.getProperty(PetalsDeliveryChannel.PROPERTY_SENDSYNC) != null
                && exchange.getProperty(sendSyncRolePropertyName) != null;

        // there is 3 possible cases:
        // 1) it is a sendsync
        // 1a) and it has timed out
        // 1b) and it hasn't
        // 2) it is not a send sync
        if (answerToSendSync) {
            // case 1) we must notify the waiting thread if it didn't timeout
            if (this.log.isDebugEnabled()) {
                this.log.debug(exchangeId + " is an answer to a send sync");
            }

            // we reset the properties
            exchange.setProperty(sendSyncRolePropertyName, null);
            final boolean stillSendSync;
            if (sendSyncRolePropertyName.equals(PROPERTY_SENDSYNC_AS_CONSUMER)) {
                stillSendSync = exchange.getProperty(PROPERTY_SENDSYNC_AS_PROVIDER) != null;
            } else {
                stillSendSync = exchange.getProperty(PROPERTY_SENDSYNC_AS_CONSUMER) != null;
            }
            // except if we know it was a sendSync that sent us this exchange!
            if (!stillSendSync) {
                exchange.setProperty(PetalsDeliveryChannel.PROPERTY_SENDSYNC, null);
            }

            // we get the exchange from sendSync if it's there
            // Note: if it has timed out, it is not in the map anymore...
            final StoredSendSyncExchange stored = this.pendingSyncExchanges.remove(prefix + exchangeId);

            // let's put the new one instead
            final boolean timedout;
            if (stored == null) {
                // case 1a) it timed out
                timedout = true;
            } else {
                // case 1b) someone is waiting for it
                synchronized (stored) {
                    // it is still possible that we timed out just after removing the exchange
                    if (stored.isTimeout()) {
                        timedout = true;
                    } else {
                        // let's release the wait() of the sendSync method
                        stored.setAnswer(exchange);
                        timedout = false;
                    }
                }
            }

            // let's verify again, after removing the exchange, if the channel was closed
            if (!this.opened) {
                throw new RoutingException("Delivery channel is closed, can't deliver message to component.");
            }

            if (timedout) {
                if (this.log.isInfoEnabled()) {
                    // let's log that, it shouldn't happen often and an admin would like to know if this happen
                    this.log.info(exchangeId + " arrived too late: the send sync already timed out.");
                }
                // Normally we should just forget about this answer (according to JBI...) but let's notify the sender if
                // possible, it will ease diagnostics
                throw new RoutingException("Exchange was ignored: it had already timed out.");
            } else {
                if (this.log.isDebugEnabled()) {
                    this.log.debug(exchangeId + " has not timed out: we notified the thread waiting for this answer");
                }
            }
        } else {
            // case 2) we must put in in the delivery channel of the correct component
            if (this.log.isDebugEnabled()) {
                this.log.debug(exchangeId + " is just a normal exchange (not a send sync answer)");
            }

            try {
                if (!this.exchangesQueue.offer(exchange, this.transporterQueueOfferingTimeout, TimeUnit.MILLISECONDS)) {
                    throw new RoutingException("Couldn't add exchange to the Component queue, not enough space.");
                }
            } catch (final InterruptedException e) {
                // in local mode, this means the sender itself is interrupted
                // in remote mode, this means the NIO agent has been interrupted
                // TODO is that the correct action to throw an exception then?
                throw new RoutingException(
                        "Interrupted while putting the exchange in the queue: it was not delivered!");
            }
        }
    }

    public boolean hasExchangesInQueue() {
        return !this.exchangesQueue.isEmpty();
    }

    /**
     * Copied from {@link AbstractMessageExchangeFactory}.
     * 
     * TODO we should be able to reuse the factory directly...
     */
    private MessageExchangeWrapper createWrapper(final PetalsMessageExchange exchange) {
        final URI pattern = exchange.getPattern();
        if (MEPPatternConstants.IN_ONLY.equals(pattern)) {
            return new InOnlyImpl(exchange);
        } else if (MEPPatternConstants.ROBUST_IN_ONLY.equals(pattern)) {
            return new RobustInOnlyImpl(exchange);
        } else if (MEPPatternConstants.IN_OUT.equals(pattern)) {
            return new InOutImpl(exchange);
        } else if (MEPPatternConstants.IN_OPTIONAL_OUT.equals(pattern)) {
            return new InOptionalOutImpl(exchange);
        } else {
            if (pattern != null) {
                this.log.warning("Creating an exchange with unknown pattern: " + pattern);
            }
            return new MessageExchangeWrapper(exchange);
        }
    }

    /**
     * This method wake up all threads that are waiting for a message. This method deregister all endpoints for this
     * component.
     */
    @Override
    public synchronized void close() throws MessagingException {
        this.log.start();

        this.checkDeliveryChannelIsOpened();

        // let's mark it as closed first so that:
        // 1) we don't get new exchanges in the queue
        // 2) we don't get new pending sync exchange
        this.opened = false;

        // we should first deactivate ALL the activated endpoints of the component
        this.componentContext.deactivateAllEndpoints();

        // then we can interrupt current waiting threads in send sync
        final Iterator<StoredSendSyncExchange> storeds = pendingSyncExchanges.values().iterator();
        while (storeds.hasNext()) {
            storeds.next().interrupt();
            storeds.remove();
        }

        // and then, let's reinject the messages in the queue
        final List<PetalsMessageExchange> leftovers = new ArrayList<>(this.exchangesQueue.size());
        this.exchangesQueue.drainTo(leftovers);
        
        for (final PetalsMessageExchange exchange : leftovers) {
            reinjectIntoNMR(exchange);
        }

        this.log.end();
    }

    @Override
    public MessageExchangeFactory createExchangeFactory() {
        this.log.call();
        return DefaultMessageExchangeFactory.create(this);
    }

    @Override
    public MessageExchangeFactory createExchangeFactory(final QName interfaceName) {
        this.log.call(interfaceName);
        return DefaultMessageExchangeFactory.createForInterface(this, interfaceName);
    }

    @Override
    public MessageExchangeFactory createExchangeFactory(final ServiceEndpoint endpoint) {
        this.log.call(endpoint);
        return DefaultMessageExchangeFactory.createForEndpoint(this, endpoint);
    }

    @Override
    public MessageExchangeFactory createExchangeFactoryForService(final QName serviceName) {
        this.log.call(serviceName);
        return DefaultMessageExchangeFactory.createForService(this, serviceName);
    }

    @Override
    public boolean isOpened() {
        return this.opened;
    }

    @Override
    public void send(final MessageExchange exchange) throws MessagingException {
        this.log.start("Exchange Id:" + exchange.getExchangeId());

        // Note: ownership is never returned on it (except in case of errer), there will be ownership on an answer, if
        // there is one, but it will be a different object (see accept)
        final MessageExchangeWrapper wrapper = (MessageExchangeWrapper) exchange;
        final PetalsMessageExchange mex = wrapper.removeOwnership();
        try {
            sendInNMR(mex);
        } catch (final MessagingException e) {
            wrapper.returnOwnership(mex);
            throw e;
        }
    }

    @Override
    public boolean sendSync(final MessageExchange exchange) throws MessagingException {
        return this.sendSync(exchange, 0);
    }

    /**
     * Note: the timeout actually starts when the Router take the lead (i.e., once
     * {@link #sendInNMR(PetalsMessageExchange)} has been called).
     */
    @Override
    public boolean sendSync(final MessageExchange exchange, final long timeoutMS)
            throws MessagingException {
        this.log.call("Exchange Id:" + exchange.getExchangeId() + " - Timeout:" + timeoutMS);

        if (timeoutMS < 0) {
            throw new IllegalArgumentException("Timeout must be superior or equal to 0.");
        }
        
        // the exchange can only be a MessageExchangeWrapper
        final MessageExchangeWrapper exchangeDecorator = (MessageExchangeWrapper) exchange;
        final PetalsMessageExchange messageExchange = exchangeDecorator.removeOwnership();

        if (!messageExchange.isActive()) {
            throw new MessagingException("Sending an acknowledgment in synchronous mode is forbidden");
        }

        PetalsMessageExchange answer = null;
        try {
            answer = sendSyncInNMR(messageExchange, timeoutMS);
            // if we timed out
            if (answer == null) {
                // in case the exchange was terminated because of the end of an MEP in the router
                messageExchange.setTermination(null);
                // let's put it in error (as defined by spec)
                messageExchange.setStatus(ExchangeStatus.ERROR);
                // so that we are sure nobody will try to send it again, we terminate the exchange
                messageExchange.setTermination("Timeout during sendSync().");
            }
        } catch (final InterruptedException e) {
            // Even if one of the causes is an InterruptedException,
            // we must throw the MessagingException according to the specification
            this.log.debug("sendSync() was interrupted: setting the message in ERROR and converting"
                    + " to a MessagingException as defined in the JBI specification.");
            // in case the exchange was terminated because of the end of an MEP in the router
            messageExchange.setTermination(null);
            // let's put it in error (as defined by spec)
            exchange.setStatus(ExchangeStatus.ERROR);
            // so that we are sure nobody will try to send it again, we terminate the exchange
            messageExchange.setTermination("Interruption during sendSync().");
            // let's return the ownership since it wasn't sent
            exchangeDecorator.returnOwnership(messageExchange);
            throw new MessagingException(e);
        }

        // if it is a timeout, the ownership is not returned!
        if (answer != null) {
            exchangeDecorator.returnOwnership(answer);
        }

        this.log.end();

        return answer != null;
    }

    /**
     * This terminates the exchange, sets needed properties, logs it and passes it to the router.
     * 
     * It is meant to be called only from the public {@link DeliveryChannel} channel methods,
     * {@link #reinjectIntoNMR(PetalsMessageExchange)} takes care of terminating things by itself for example
     */
    private void sendInNMR(final PetalsMessageExchange exchange) throws MessagingException {

        this.checkDeliveryChannelIsOpened();

        final String termination = exchange.getTermination();
        if (termination != null) {
            throw new MessagingException(
                    "MessageExchange is terminated, it can not be sent any more (reason: " + termination + ")");
        }

        this.logMessage(exchange);

        // check if the exchange is terminated and mark it as such
        // The idea is that an exchange is terminated once it has been sent by the last actor of the exchange
        if (exchange.isDone()) {
            exchange.setTermination("MEP ended with Done.");
        } else if (exchange.isError()) {
            exchange.setTermination("MEP ended with Error.");
        }
        
        try {
            this.componentContext.getRouter().sendInNMR(exchange);
        } catch (final RoutingException e) {
            this.log.debug("Error while sending: setting the message in ERROR");
            // in case the exchange was terminated because of the end of an MEP in the router
            exchange.setTermination(null);
            // we do not set the error itself since we are throwing the exception
            exchange.setStatus(ExchangeStatus.ERROR);
            exchange.setTermination("Routing error during send.");
            throw new MessagingException(e);
        }
    }

    /**
     * This takes care of waiting for the answer only.
     */
    private PetalsMessageExchange sendSyncInNMR(PetalsMessageExchange exchange, final long timeout)
            throws MessagingException, InterruptedException {
        this.log.start();

        // two phases:
        // A) sending the exchange
        // B) waiting for an answer (or timing out)
        // each phase is linked to a role the exchange has, there is two cases:
        // i) sending consumer exchange, receiving provider exchange
        // ii) sending provider exchange, receiving consumer exchange
        // this information helps us reading the logs because of the interleaving of threads happening when we are
        // waiting for an answer and this answer is sent

        final String exchangeId = exchange.getExchangeId();

        final String logHint = exchangeId + "/" + exchange.getRole();

        // phase A)
        if (this.log.isDebugEnabled()) {
            this.log.debug("Sending synchronously exchange " + logHint);
        }

        // we store it (before sending it, because we need to synchronize on it)
        final StoredSendSyncExchange stored = new StoredSendSyncExchange(exchange);

        // This prefix is needed because two sendSync can happen on the same exchange from the same component with two
        // different role and it will create a confusion between the two!
        final String prefix;
        final String sendSyncRolePropertyName;
        if (exchange.isConsumer()) {
            prefix = "C";
            sendSyncRolePropertyName = PROPERTY_SENDSYNC_AS_CONSUMER;
        } else {
            prefix = "P";
            sendSyncRolePropertyName = PROPERTY_SENDSYNC_AS_PROVIDER;
        }

        final PetalsMessageExchange answer;
        synchronized (stored) {
            try {
                // this must be put in the map after we started to synchronize on it so that offerExchange do not start
                // to do things on it before us (this shouldn't happen because we haven't sent it yet but well,
                // it's not wrong...)
                final StoredSendSyncExchange removed = this.pendingSyncExchanges.put(prefix + exchangeId, stored);
                this.log.assertOrLog(removed == null,
                        "A different exchange with the same ID was already in processing");

                // and this one is useful to know who did the sendSync
                exchange.setProperty(sendSyncRolePropertyName, true);

                // set the property as defined in the JBI specifications 5.5.2.1.3
                // we also use it to store the deadline, because it is used by some of the other modules
                // It will be unset when the answer will be received in offerExchange()
                if (timeout > 0) {
                    // Note: actually, the timeout is meant to be used only for the wait, but anyway, we set here a not
                    // so precise value to be used by the other modules (not used currently)
                    exchange.setProperty(PetalsDeliveryChannel.PROPERTY_SENDSYNC, System.currentTimeMillis() + timeout);
                } else {
                    exchange.setProperty(PetalsDeliveryChannel.PROPERTY_SENDSYNC, Long.MAX_VALUE);
                }

                // we delegate to the router
                sendInNMR(exchange);

                // phase B) we wait for the answer

                // let's check again in case the send took some time
                checkDeliveryChannelIsOpened();

                // Note: we must keep the lock until the wait to be sure
                // that any notifyAll() does not happen before we start waiting
                // and after the handling of timeout to avoid intermediary situation where we think we timed out and the
                // exchange arrived just after...

                // So let's wait for an answer
                answer = stored.waitForAnswerOrTimeout(timeout);

                if (answer == null) {
                    this.log.warning("Failed to send synchronously the exchange '" + exchangeId + "': timeout occured");

                    // let's remove it from the map now that it has timed out
                    final StoredSendSyncExchange removed2 = this.pendingSyncExchanges.remove(prefix + exchangeId);
                    this.log.assertOrLog(removed2 != null,
                            "sendSync timed out but the stored exchange is not in the map");
                } else {
                    if (this.log.isDebugEnabled()) {
                        this.log.debug(logHint + " was waiting, we got an answer");
                    }
                }
            } finally {
                // who knows what could have happened, for example an exception from sendInNMR or an interruption: we
                // need to remove that so that there is no leftovers in case of problems
                this.pendingSyncExchanges.remove(prefix + exchangeId);
            }
        }

        this.log.end();

        return answer;
    }

    private void reinjectIntoNMR(final PetalsMessageExchange exchange) {
        final String logPrefix = "Can't reinject left-overs exchange '" + exchange.getExchangeId() + "' on shutdown: ";
        try {
            if (exchange.getTermination() != null) {
                // this can happen if the sender has timed out for example
                this.log.debug(logPrefix + "already terminated (reason: " + exchange.getTermination() + ").");
            } else if (exchange.isNewExchange()) {
                this.log.debug(
                        "Reinjecting a new exchange during component shutdown: '" + exchange.getExchangeId() + "'.");
                // we simply reset the elected provider endpoint
                exchange.setElectedProviderEndpoint(null);

                // in this case it will be sent to another provider
                // so we first need to inverse the roles
                final Role originalRole = exchange.getRole();
                if (exchange.isConsumer()) {
                    exchange.setRole(Role.PROVIDER);
                } else {
                    exchange.setRole(Role.CONSUMER);
                }

                try {
                    this.componentContext.getRouter().sendInNMR(exchange);
                } catch (final RoutingException e) {
                    this.log.warning(logPrefix + "error while sending. Sending back the error to the consumer", e);
                    try {
                        // first let's, again, switch back the role
                        exchange.setRole(originalRole);
                        // we are a new exchange, so still active
                        exchange.setError(e);
                        exchange.setTermination("Routing error during reinjection.");
                        // in this case it will be sent back to the consumer!
                        this.componentContext.getRouter().sendInNMR(exchange);
                    } catch (final RoutingException e2) {
                        this.log.warning("Can't send back '" + exchange.getExchangeId()
                                + "' to the consumer: error while sending.", e2);
                    }
                }
            } else {
                // we can't do anything about that...
                final String logPrefix2 = logPrefix + "not a new exchange. ";
                if (exchange.isActive()) {
                    this.log.warning(logPrefix2 + "Sending back the error to the consumer");
                    try {
                        // we are a new exchange, so still active
                        exchange.setError(
                                new MessagingException("The receiver has been shutdown before the end of the MEP."));
                        exchange.setTermination("Can't reinject a non-new exchange on component shutdown.");
                        // in this case it will be sent back to the consumer!
                        this.componentContext.getRouter().sendInNMR(exchange);
                    } catch (final RoutingException e2) {
                        this.log.warning("Can't send back '" + exchange.getExchangeId()
                                + "' to the consumer: error while sending.", e2);
                    }
                } else {
                    this.log.warning(logPrefix2 + "MEP is ended, nothing can be done.");
                }
            }
        } catch (final IllegalStateException e) {
            // this shouldn't happen but well, let's log it...
            this.log.error(logPrefix + "not the owner (this shouldn't happen, maybe a bug...).");
        }
    }

    /**
     * Log the current message, it can be an Out, Fault or In message.
     */
    private void logMessage(final PetalsMessageExchange exchange) {
        final Logger logger = this.log.getLogger();
        try {
            if (logger.isLoggable(Level.FINEST)) {
                final String exchangeId = exchange.getExchangeId();
                if (exchange.isActive()) {
                    NormalizedMessage normalizedMessage = exchange.getMessage(PetalsMessageExchange.OUT_MSG);
                    String messageType = PetalsMessageExchange.OUT_MSG;
                    if (normalizedMessage == null) {
                        normalizedMessage = exchange.getFault();
                        messageType = "Fault";
                    }
                    if (normalizedMessage == null) {
                        normalizedMessage = exchange.getMessage(PetalsMessageExchange.IN_MSG);
                        messageType = PetalsMessageExchange.IN_MSG;
                    }

                    logger.fine("Message " + messageType + " with Id '" + exchangeId + "' to send");
                    logger.fine("Operation : " + exchange.getOperation());
                    logger.fine("MEP : " + exchange.getPattern());
                    logger.fine("Consumer location: " + exchange.getConsumerLocation());

                    for (final Object propertyName : exchange.getPropertyNames()) {
                        logger.fine("Exchange property '" + propertyName + "' : "
                                + exchange.getProperty((String) propertyName));
                    }

                    for (final Object propertyName : normalizedMessage.getPropertyNames()) {
                        logger.fine("Message property '" + propertyName + "' : "
                                + normalizedMessage.getProperty((String) propertyName));
                    }

                    Source content = normalizedMessage.getContent();
                    if (content != null) {
                        logger.fine("Message Content : " + SourceHelper.toString(content));
                    } else {
                        logger.fine("Message Content is null");
                    }

                    for (final Object attachmentName : normalizedMessage.getAttachmentNames()) {
                        logger.fine("Got a message attachment named '" + attachmentName + "'");
                    }
                } else if (exchange.isDone()) {
                    logger.fine("Message acknowledgement DONE with Id '" + exchangeId + "' to send");
                    for (final Object propertyName : exchange.getPropertyNames()) {
                        logger.fine("Exchange property '" + propertyName + "' : "
                                + exchange.getProperty((String) propertyName));
                    }
                } else {
                    logger.fine("Message acknowledgement ERROR with Id '" + exchangeId + "' to send");
                    logger.fine("Message Error : " + exchange.getError().getMessage());
                }
            }
        } catch (final Exception e) {
            logger.log(Level.FINE, "Failed to log the content of the message to send", e);
        }
    }

    /**
     * Careful, all the calls to the methods of an object of this class must be done inside synchronized block on the
     * object!
     */
    private class StoredSendSyncExchange {

        private final Thread waitingThread;

        private PetalsMessageExchange exchange;

        private boolean received = false;

        private boolean timeout = false;

        public StoredSendSyncExchange(final PetalsMessageExchange exchange) {
            this.exchange = exchange;
            this.waitingThread = Thread.currentThread();
        }

        public void interrupt() {
            this.waitingThread.interrupt();
        }

        public boolean isTimeout() {
            return this.timeout;
        }

        /**
         * Note: it's not "really" needed to specify synchronized on this method (because the call to it is), but let's
         * be on the safe side for future changes to this class
         */
        public synchronized void setAnswer(final PetalsMessageExchange exchange) {
            this.received = true;
            this.exchange = exchange;

            // let's wake-up the wait() waiting for this one
            // we could have used notify, but it's the same since there is only one thread waiting
            this.notifyAll();
        }

        /**
         * Note: it's not "really" needed to specify synchronized on this method (because the call to it is), but let's
         * be on the safe side for future changes to this class
         * 
         * @return <code>null</code> if it timed out!
         * 
         */
        public synchronized PetalsMessageExchange waitForAnswerOrTimeout(final long timeout)
                throws InterruptedException {
            log.start();

            log.assertOrLog(this.waitingThread == Thread.currentThread(),
                    "waitForAnswerOrTimeout() is meant to be called only by the thread that created the stored send sync exchange.");

            // the timeout is meant to be used only for waiting for an answer
            // so we compute it really here
            final Long deadline;
            if (timeout > 0) {
                deadline = System.currentTimeMillis() + timeout;
            } else {
                deadline = null;
            }

            // wait for a response or a timeout reached: interruption will just throw an exception
            if (log.isDebugEnabled()) {
                log.debug("Let's wait for an answer, " + (deadline == null ? "no timeout"
                        : "timeout left is " + (deadline - System.currentTimeMillis())) + "...");
            }
            // the loop protects from spurious wakeup and is exited when we got an answer or timed out
            while (true) {
                // we can start to wait directly, because of the synchronize, there is no way
                // that the notify happened before we called wait!
                try {
                    if (deadline != null) {
                        final long waitTime = deadline - System.currentTimeMillis();
                        // if timeout is <= 0, then we just don't want to wait because we already timed out
                        // (wait(0) means no timeout!)
                        if (waitTime > 0) {
                            this.wait(waitTime);
                        }
                    } else {
                        this.wait();
                    }
                } catch (final InterruptedException e) {
                    log.debug("The wait was interrupted.");
                    throw e;
                }

                log.debug("The wait was released so let's check the condition...");

                // we check if the answer was set in the wrapper
                if (this.received) {
                    if (log.isDebugEnabled()) {
                        log.debug("... the condition was met, we can stop waiting!");
                    }
                    log.end();
                    return this.exchange;
                } else if (deadline != null) {
                    log.debug("... the condition wasn't met, let's check timeout...");
                    // we check for timeout ONLY if we haven't got the answer since what matters is to get the answer
                    if (deadline < System.currentTimeMillis()) {
                        log.debug("... and yes, we timed out!");
                        // let's remember it for offerExchange
                        this.timeout = true;
                        log.end();
                        return null;
                    }
                }

                log.debug("... the condition wasn't met and we didn't timeout: "
                        + "it must have been the famous mysterious spurious wake-up!");
            }
        }
    }
}
