/**
 * Copyright (c) 2008-2012 EBM WebSourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.jbi.messaging.routing.module;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.StringTokenizer;
import java.util.logging.Logger;

import javax.jbi.messaging.MessageExchange;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;

import org.objectweb.fractal.fraclet.annotations.Interface;
import org.objectweb.fractal.fraclet.annotations.Lifecycle;
import org.objectweb.fractal.fraclet.annotations.Requires;
import org.objectweb.fractal.fraclet.types.Contingency;
import org.objectweb.fractal.fraclet.types.Step;
import org.ow2.petals.clientserverapi.configuration.ContainerConfiguration;
import org.ow2.petals.clientserverapi.jbi.messaging.endpoint.exception.EndpointDirectoryException;
import org.ow2.petals.jbi.messaging.exchange.PetalsMessageExchange;
import org.ow2.petals.jbi.servicedesc.endpoint.Location;
import org.ow2.petals.jbi.servicedesc.endpoint.PetalsServiceEndpoint;
import org.ow2.petals.microkernel.api.communication.topology.TopologyService;
import org.ow2.petals.microkernel.api.configuration.ConfigurationService;
import org.ow2.petals.microkernel.api.jbi.messaging.endpoint.EndpointDirectoryService;
import org.ow2.petals.microkernel.api.jbi.messaging.exception.RoutingException;
import org.ow2.petals.microkernel.jbi.messaging.control.ExchangeCheckerClient;
import org.ow2.petals.microkernel.jbi.messaging.routing.module.endpoint.EndpointOrderer;
import org.ow2.petals.microkernel.transport.util.TransportSendContext;

import com.ebmwebsourcing.easycommons.log.LoggingUtil;

/**
 * Router module to resolve address.
 * @author Christophe HAMERLING - EBM WebSourcing
 */
@org.objectweb.fractal.fraclet.annotations.Component(provides = @Interface(name = "service", signature = SenderModule.class))
public class EndpointResolverModule implements SenderModule {

    /**
     * the highest strategy.
     */
    public static final String HIGHEST = "highest";

    /**
     * Number of parameters defined in strategy.
     */
    public static final int NUMBER_STRATEGY_PARAMETERS = 4;

    /**
     * Constant used to specify the strategy to use.
     */
    public static final String PROPERTY_STRATEGY_PROTOCOLS = "org.ow2.petals.routing.strategy";

    /**
     * the load balancing (random strategy).
     */
    public static final String RANDOM = "random";

    /**
     * the strategy separator.
     */
    public static final String STRATEGY_SEPARATOR = ",";

    /**
     * the default configured strategy parameters.
     */
    private List<Object> defaultStrategyParameters;

    /**
     * The Topology service fractal component
     */
    @Requires(name = "configuration")
    private ConfigurationService configurationService;

    /**
     * the endpoint directory service.
     */
    @Requires(name = "endpointDirectory")
    private EndpointDirectoryService endpointDirectory;

    /**
     * FIXME : Use it instaed of the registry!
     */
    @Requires(name = "checker")
    private ExchangeCheckerClient exchangeChecker;

    /**
     * Logger wrapper.
     */
    private final LoggingUtil log = new LoggingUtil(
            Logger.getLogger(Constants.FRACTAL_COMPONENT_LOGGER_NAME_ENDPOINT_RESOLVER));

    /**
     * The Topology service fractal component
     */
    @Requires(name = "topology", contingency = Contingency.OPTIONAL)
    private TopologyService topologyService;

    /**
     * The elected endpoint orderer
     */
    private EndpointOrderer endpointOrderer;

    /**
     * Analyze a strategy.
     * 
     * @param strategy
     * @return
     * @throws RoutingException
     */
    private final static List<Object> analyseStrategy(final String strategy)
            throws RoutingException {
        final List<Object> result = new ArrayList<Object>();
        final List<String> parameters = new ArrayList<String>();

        final StringTokenizer st = new StringTokenizer(strategy, STRATEGY_SEPARATOR);
        while (st.hasMoreTokens()) {
            parameters.add(st.nextToken().trim());
        }

        if (parameters.size() != NUMBER_STRATEGY_PARAMETERS) {
            throw new RoutingException("4 parameters are required for a routing strategy '"
                    + strategy + "'");
        }

        final String type = parameters.get(0).toLowerCase();
        final Float localWeighting;
        final Float remoteActiveWeighting;
        final Float remoteInactiveWeighting;
        try {
            localWeighting = Float.valueOf(parameters.get(1));
            remoteActiveWeighting = Float.valueOf(parameters.get(2));
            remoteInactiveWeighting = Float.valueOf(parameters.get(3));
        } catch (final NumberFormatException e) {
            throw new RoutingException("Impossible to convert a "
                    + "parameter to a float for the routing strategy '" + strategy + "'");
        }

        if ((!type.equals(HIGHEST)) && (!type.equals(RANDOM))) {
            throw new RoutingException("the " + type + " routing strategy is unknown");
        }
        if ((localWeighting < 0) || (remoteActiveWeighting < 0) || (remoteInactiveWeighting < 0)) {
            throw new RoutingException("Negative parameter is forbidden in routing strategy");
        }
        if (localWeighting == 0 && remoteActiveWeighting == 0 && remoteInactiveWeighting == 0) {
            throw new RoutingException("Weightings '0,0,0' is forbidden");
        }

        result.add(type);
        result.add(localWeighting);
        result.add(remoteActiveWeighting);
        result.add(remoteInactiveWeighting);

        return result;
    }

    /**
     * Get the strategy and analyze it. If the strategy is defined as message
     * property, it is used. Otherwise the default configured strategy is used.
     * 
     * @param exchange
     *            the message exchange in which the strategy to use can be
     *            provided. Must be non-null
     * @throws RoutingException
     *             impossible to create the endpoint chooser strategy
     * @return the list of parameters for the strategy
     */
    private List<Object> getStrategy(final MessageExchange exchange)
            throws RoutingException {

        final String strategy = (String) exchange.getProperty(PROPERTY_STRATEGY_PROTOCOLS);
        final List<Object> result;
        if (strategy == null) {
            result = defaultStrategyParameters;
        } else {
            result = analyseStrategy(strategy);

        }
        return result;
    }

    @Override
    public void electEndpoints(final List<TransportSendContext> electedDestinations,
            final PetalsMessageExchange exchange) throws RoutingException {
        this.log.call();

        if (exchange.isConsumer()) {
            for (final PetalsServiceEndpoint endpoint : this.resolveEndpoints(exchange)) {
                electedDestinations.add(new TransportSendContext(endpoint));
            }
        } else {
            electedDestinations.add(new TransportSendContext(exchange.getConsumerLocation()));
        }
    }

    /**
     * Start the Fractal component
     * 
     * @throws RoutingException
     */
    @Lifecycle(step = Step.START)
    public void start() throws RoutingException {
        this.log.start();

        final ContainerConfiguration containerConfiguration = this.configurationService
                .getContainerConfiguration();

        try {
            this.defaultStrategyParameters = analyseStrategy(containerConfiguration.getRouterStrategy());
        } catch (final RoutingException e) {
            this.log.error("Failed to start Endpoint Resolver module", e);
            throw e;
        }

        this.endpointOrderer = new EndpointOrderer(this.log, containerConfiguration,
                this.topologyService);

        this.log.end();
    }

    /**
     * Stop the Fractal component
     */
    @Lifecycle(step = Step.STOP)
    public void stop() {
        this.log.call();
    }

    /**
     * Return only the enabled endPoints among the potentials endpoints.
     * 
     * @param type
     *            the type of research (by service or interface)
     * @param consumer
     *            the consumer
     * @param givenName
     *            the given service or interface name
     * @param exchange
     *            the exchange
     * @return the list of valide endpoints
     * @throws RegistryException
     */
    private void removeEndpointsNotOK(final Collection<PetalsServiceEndpoint> potentialEndpoints,
            final PetalsMessageExchange exchange) throws EndpointDirectoryException {
        this.log.call();

        final Iterator<PetalsServiceEndpoint> iterator = potentialEndpoints.iterator();
        while (iterator.hasNext()) {
            final PetalsServiceEndpoint providerEP = iterator.next();
            if (!this.isExchangeOK(providerEP, exchange)) {
                iterator.remove();
            }
        }
    }

    /**
     * Get the real endpoint matching the given endpoint.
     * 
     * @param endpoint
     *            the endpoint in the message exchange. May be null
     * @return the targeted endpoint. May be null
     * @throws RoutingException
     *             impossible to get the endpoint
     */
    private PetalsServiceEndpoint getEndpointForEndpoint(final ServiceEndpoint edpt)
            throws RoutingException {
        this.log.call();

        if (edpt instanceof PetalsServiceEndpoint) {
            PetalsServiceEndpoint endpoint = (PetalsServiceEndpoint) edpt;
            switch (endpoint.getType()) {
                case EXTERNAL:
                    throw new RoutingException("Can't route a message to an external endpoint");
                case INTERNAL:
                    return endpoint;
            }
        }
        // if it's not a PetalsServiceEndpoint, it's a fake one, either used for service connections resolution or
        // typical resolution
        // TODO shouldn't we consider the use of an endpoint as a way to do a typical resolution an error?
        // TODO support service connections, see EndpointDirectoryService and PETALSESBCONT-391
        // Note: service connections are for a given SA, maybe this is not the right place to do this?

        try {
            return this.endpointDirectory.getInternalEndpoint(edpt.getServiceName(), edpt.getEndpointName());
        } catch (final EndpointDirectoryException e) {
            throw new RoutingException(e);
        }
    }

    /**
     * Get the endpoints matching the given interface.
     * 
     * @param interfaceName
     *            the given interface Name in the message exchange. May be null
     * @param exchange
     *            the exchange.
     * @param component
     *            the component.
     * @return the targeted endpoint following the specific strategy used. May
     *         be null
     * @throws RoutingException
     *             impossible to get the endpoint
     */
    private Collection<PetalsServiceEndpoint> getEndpointsForInterface(final QName interfaceName,
            final PetalsMessageExchange exchange) throws RoutingException {
        this.log.call();

        final Collection<PetalsServiceEndpoint> retrievedEndpoints;
        try {
            retrievedEndpoints = this.endpointDirectory.getInternalEndpointsForInterface(interfaceName);
            removeEndpointsNotOK(retrievedEndpoints, exchange);
        } catch (final EndpointDirectoryException e) {
            throw new RoutingException(e);
        }

        if (retrievedEndpoints.isEmpty()) {
            throw new RoutingException("No endpoint(s) matching the target interface '" + interfaceName
                    + "' for Message Exchange with id '" + exchange.getExchangeId() + "'");
        } else if (retrievedEndpoints.size() == 1) {
            return retrievedEndpoints;
        } else {
            return this.endpointOrderer.orderEndpoints(retrievedEndpoints, getStrategy(exchange));
        }
    }

    /**
     * Get the endpoints matching the given service.
     * 
     * @param serviceName
     *            the given service Name in the message exchange.
     * @param exchange
     *            the exchange.
     * @param component
     *            the component.
     * @return the targeted endpoint following the specific strategy used. May
     *         be null
     * @throws RoutingException
     *             impossible to get the endpoint
     */
    private Collection<PetalsServiceEndpoint> getEndpointsForService(final QName serviceName,
            final PetalsMessageExchange exchange) throws RoutingException {
        this.log.call();

        final Collection<PetalsServiceEndpoint> retrievedEndpoints;
        try {
            retrievedEndpoints = this.endpointDirectory.getInternalEndpointsForService(serviceName);
            removeEndpointsNotOK(retrievedEndpoints, exchange);
        } catch (final EndpointDirectoryException e) {
            throw new RoutingException(e);
        }

        if (retrievedEndpoints.isEmpty()) {
            throw new RoutingException("No endpoint(s) matching the target service '" + serviceName
                    + "' for Message Exchange with id '" + exchange.getExchangeId() + "'");
        } else if (retrievedEndpoints.size() == 1) {
            return retrievedEndpoints;
        } else {
            return this.endpointOrderer.orderEndpoints(retrievedEndpoints, getStrategy(exchange));
        }
    }

    /**
     * Control that the exchange is possible. Verify if the consumer and
     * provider are agree to realize the exchange.
     * 
     * @param consumer
     *            the consumer
     * @param providerEP
     *            the provider
     * @param exchange
     *            the exchange
     * @return true if ok
     */
    private boolean isExchangeOK(final PetalsServiceEndpoint providerEP, final PetalsMessageExchange exchange) {
        this.log.call();

        boolean result = true;
        if (this.configurationService.getContainerConfiguration().isExchangeValidation()) {
            // this method is only called if this is a CONSUMER exchange
            this.log.assertOrLog(exchange.isConsumer(), "Expecting consumer exchange");
            final Location consumer = exchange.getConsumerLocation();
            final Location provider = providerEP.getLocation();
            if (!this.exchangeChecker.isExchangeWithProviderOkayForComponent(consumer, providerEP, exchange)
                    || !this.exchangeChecker.isExchangeWithConsumerOkayForComponent(provider, providerEP, exchange)) {
                result = false;
            }
        }

        return result;
    }

    /**
     * Resolve the address if necessary. Resolution consist in finding an endpoint satisfying <code>interface</code> or
     * <code>service</code>.
     * 
     * @param exchange
     *            the exchange
     * @throws RoutingException
     *             no destination can be found, or problem accessing the registry
     */
    private Collection<PetalsServiceEndpoint> resolveEndpoints(final PetalsMessageExchange exchange)
            throws RoutingException {
        this.log.start();

        // get message exchange values
        final ServiceEndpoint givenEndpoint = exchange.getEndpoint();
        final QName serviceName = exchange.getService();
        final QName interfaceName = exchange.getInterfaceName();

        // Construct a list of potential matching endpoints.
        // If the provider endpoint has been set in the messageExchange,
        // this list contains only one element, which is this endpoint.

        Collection<PetalsServiceEndpoint> electedEndpoints = null;

        if (exchange.getElectedProviderEndpoint() != null) {
            // Case 0: the endpoint was previously elected, let's simply keep it
            electedEndpoints = Collections.singletonList(exchange.getElectedProviderEndpoint());
        } else if (givenEndpoint != null) {
            // Case 1 : The endpoint is explicit, let's resolve it if needed
            final PetalsServiceEndpoint targetEndpoint = this.getEndpointForEndpoint(givenEndpoint);
            if (targetEndpoint != null) {
                electedEndpoints = Collections.singletonList(targetEndpoint);
            }
        } else if (serviceName != null) {
            // Case 2 : The endpoint is implicit, let's resolve it based on the service
            electedEndpoints = this.getEndpointsForService(serviceName, exchange);
            if (interfaceName != null) {
                // control that the interface name matches
                final Iterator<PetalsServiceEndpoint> endpointsIterator = electedEndpoints.iterator();
                while (endpointsIterator.hasNext()) {
                    final PetalsServiceEndpoint endpoint = endpointsIterator.next();
                    final QName[] interfaces = endpoint.getInterfaces();
                    boolean found = false;
                    for (final QName name : interfaces) {
                        if (interfaceName.equals(name)) {
                            found = true;
                            break;
                        }
                    }
                    // if there was no interfaces, let's keep it then (see PETALSESBCONT-398)
                    if (!found && interfaces.length > 0) {
                        endpointsIterator.remove();
                    }
                }
            }
        } else if (interfaceName != null) {
            // Case 3 : The endpoint is implicit, let's resolve it based on the interface
            electedEndpoints = this.getEndpointsForInterface(interfaceName, exchange);
        }

        if (electedEndpoints == null || electedEndpoints.isEmpty()) {
            throw new RoutingException(
                    "Failed to find an endpoint for the Message Exchange with id '"
                            + exchange.getExchangeId() + "'");
        }

        if (this.log.isDebugEnabled()) {
            this.log.debug("Elected endpoints:");
            for (final PetalsServiceEndpoint electedEndpoint : electedEndpoints) {
                this.log.debug("\t - " + electedEndpoint.toString());
            }
        }

        this.log.end();
        return electedEndpoints;
    }

}
