/**
 * Copyright (c) 2013-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */

package org.ow2.petals.microkernel.server;

import java.util.Collections;
import java.util.Iterator;
import java.util.Properties;
import java.util.ServiceLoader;
import java.util.logging.Logger;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.Interface;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.fractal.api.control.ContentController;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalContentException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.api.control.LifeCycleController;
import org.objectweb.fractal.api.control.NameController;
import org.objectweb.fractal.api.factory.InstantiationException;
import org.objectweb.fractal.fraclet.annotations.Lifecycle;
import org.objectweb.fractal.fraclet.annotations.Requires;
import org.objectweb.fractal.fraclet.extensions.Controller;
import org.objectweb.fractal.fraclet.types.Step;
import org.objectweb.fractal.util.Fractal;
import org.ow2.petals.microkernel.admin.PetalsAdminServiceImpl;
import org.ow2.petals.microkernel.api.admin.PetalsAdminInterface;
import org.ow2.petals.microkernel.api.configuration.ConfigurationService;
import org.ow2.petals.microkernel.api.implementation.exception.MutableImplException;
import org.ow2.petals.microkernel.api.jbi.messaging.RouterService;
import org.ow2.petals.microkernel.api.server.FractalHelper;
import org.ow2.petals.microkernel.api.system.sharedarea.SharedAreaService;
import org.ow2.petals.microkernel.api.system.sharedarea.exception.SharedAreaException;
import org.ow2.petals.microkernel.container.ContainerController;
import org.ow2.petals.microkernel.container.exception.ContainerCtrlException;
import org.ow2.petals.microkernel.extension.PreExtensionsManager;
import org.ow2.petals.microkernel.server.exception.BindingInterfaceException;
import org.ow2.petals.microkernel.server.exception.CreateComponentException;
import org.ow2.petals.microkernel.server.exception.FractalComponentNotFoundException;
import org.ow2.petals.microkernel.server.exception.FractalCompositeNotFoundException;
import org.ow2.petals.microkernel.server.exception.InvalidCompBasedArchiException;
import org.ow2.petals.microkernel.server.exception.NoSharedAreaImplException;
import org.ow2.petals.microkernel.server.exception.PetalsCompositeCtrlException;
import org.ow2.petals.microkernel.server.exception.TooManySharedAreaImplException;

import com.ebmwebsourcing.easycommons.log.LoggingUtil;

/**
 * The controller implementation of the Fractal composite 'Petals'
 * 
 * @author Christophe DENEUX - Linagora
 * 
 */
@org.objectweb.fractal.fraclet.annotations.Component(provides = @org.objectweb.fractal.fraclet.annotations.Interface(name = "service", signature = PetalsCompositeController.class))
public class PetalsCompositeControllerImpl implements PetalsCompositeController {

    /**
     * The configuration service
     */
    @Requires(name = "configuration")
    private ConfigurationService configurationService;

    /**
     * The pre-extension manager service
     */
    @Requires(name = "preExtensionsMgr")
    private PreExtensionsManager preExtensionsMgr;

    @Controller
    private Component comp;

    /**
     * The composite 'Petals'
     */
    private Component petalsComposite;

    /**
     * The content controller of the composite 'Petals'
     */
    private ContentController petalsCompositeCC;

    /**
     * The shared area client implementation
     */
    private Class<?> sharedAreaImplDef = null;

    private final LoggingUtil log = new LoggingUtil(Logger.getLogger(PetalsCompositeController.COMPONENT_LOGGER_NAME));

    /**
     * <p>
     * Starts the composite controller.
     * </p>
     * <p>
     * When starting, the composite controller will create the following Fractal component, bind interfaces and start
     * it:
     * <ol>
     * <li>the composite 'Communication',</li>
     * <li>the composite 'System',</li>
     * <li>the composite 'ExtensionManager',</li>
     * <li>the composite 'JBI-Management',</li>
     * <li>the component 'org.ow2.petals.microkernel.admin.PetalsAdminServiceImpl',</li>
     * <li>the composite 'JBI-Messaging',</li>
     * <li>the composite 'Transporter',</li>
     * <li>the composite 'Container'.</li>
     * </ol>
     * </p>
     * 
     * @throws PetalsCompositeCtrlException
     *             An error occurs during the instantiation of a composite or a component.
     * @throws InvalidCompBasedArchiException
     *             The component based architecture is invalid.
     */
    @Lifecycle(step = Step.START)
    public void start() throws PetalsCompositeCtrlException, InvalidCompBasedArchiException {
        this.log.start();

        try {
            try {
                this.petalsComposite = FractalHelper.getParentComponent(this.comp);
                this.petalsCompositeCC = Fractal.getContentController(petalsComposite);

                // Creates composites
                this.log.debug("Creating composites");

                this.log.debug("Creates, binds and starts the composite " + FractalHelper.SYSTEM_COMPOSITE);
                final Component systemComposite = this.createAndAddCompositeSystem();
                final Component configuration = FractalHelper.getComponentByName(this.petalsCompositeCC,
                        FractalHelper.CONFIGURATION_COMPONENT);
                final Interface configurationServiceItf = (Interface) configuration
                        .getFcInterface(ConfigurationService.FRACTAL_SRV_ITF_NAME);
                this.bindCompositeSystemItf(systemComposite, configurationServiceItf);
                this.log.debug("Starts composite " + FractalHelper.SYSTEM_COMPOSITE);
                FractalHelper.startComponent(systemComposite);
                this.log.debug("Composite '" + FractalHelper.SYSTEM_COMPOSITE + "' created, bound and started.");

                this.log.debug("Creates, binds and starts the composite " + FractalHelper.COMMUNICATION_COMPOSITE);
                final Component communicationComposite = this.createAndAddComponent(
                        FractalHelper.COMMUNICATION_COMPOSITE, FractalHelper.COMMUNICATION_COMPOSITE,
                        this.petalsComposite);
                final Interface sharedAreaServiceItf = (Interface) systemComposite
                        .getFcInterface(PetalsCompositeController.SYSTEM_SHARED_AREA_FRACTAL_SRV_ITF_NAME);
                final Component petalsCompositeController = FractalHelper.getComponentByName(this.petalsCompositeCC,
                        FractalHelper.PETALS_COMPOSITE_CTRL_COMPONENT);
                final Interface petalsCompositeCtrlItf = (Interface) petalsCompositeController
                        .getFcInterface(PetalsCompositeController.FRACTAL_SRV_ITF_NAME);
                this.bindCompositeCommunicationItf(communicationComposite, configurationServiceItf,
                        sharedAreaServiceItf, petalsCompositeCtrlItf);
                this.log.debug("Starts composite " + FractalHelper.COMMUNICATION_COMPOSITE);
                FractalHelper.startComponent(communicationComposite);
                this.log.debug("Composite '" + FractalHelper.COMMUNICATION_COMPOSITE + "' created, bound and started.");

                // Before starting other components, we update the topology with the dynamic topology store in the
                // registry
                final SharedAreaService sharedAreadService = (SharedAreaService) sharedAreaServiceItf;

                try {
                    sharedAreadService.loadTopology(Collections.<SharedAreaService.SavedEndpoint> emptyList());
                } catch (final SharedAreaException e) {
                    throw new PetalsCompositeCtrlException(e);
                }

                this.log.debug("Creates remaining composites/components");
                final Component extensionsManagerComposite = this.createAndAddComponent(
                        FractalHelper.POST_EXTENSIONSMANAGER_COMPOSITE, FractalHelper.POST_EXTENSIONSMANAGER_COMPOSITE,
                        this.petalsComposite);
                final Component jbiManagementComposite = this.createAndAddComponent(
                        FractalHelper.JBI_MANAGEMENT_COMPOSITE, FractalHelper.JBI_MANAGEMENT_COMPOSITE,
                        this.petalsComposite);
                final Component petalsAdminServiceComponent = this.createAndAddComponent(PetalsAdminServiceImpl.class,
                        FractalHelper.PETALSADMIN_COMPONENT, this.petalsComposite);
                final Component jbiMessagingComposite = this.createAndAddComponent(
                        FractalHelper.JBI_MESSAGING_COMPOSITE, FractalHelper.JBI_MESSAGING_COMPOSITE,
                        this.petalsComposite);
                final Component transporterComposite = this.createAndAddComponent(FractalHelper.TRANSPORTERS_COMPOSITE,
                        FractalHelper.TRANSPORTERS_COMPOSITE, this.petalsComposite);
                this.log.debug("Composites created");

                // Bind interfaces
                this.log.debug("Binds interfaces of remaining composites/components");
                final Interface topologyServiceItf = (Interface) communicationComposite
                        .getFcInterface(PetalsCompositeController.COMMUNICATION_TOPOLOGY_FRACTAL_SRV_ITF_NAME);
                final Interface jmxServiceItf = (Interface) communicationComposite
                        .getFcInterface(PetalsCompositeController.COMMUNICATION_JMX_FRACTAL_SRV_ITF_NAME);
                final Interface jndiServiceItf = (Interface) communicationComposite
                        .getFcInterface(PetalsCompositeController.COMMUNICATION_JNDI_FRACTAL_SRV_ITF_NAME);
                final Interface systemStateServiceItf = (Interface) systemComposite
                        .getFcInterface(PetalsCompositeController.SYSTEM_SYSTEMSTATE_FRACTAL_SRV_ITF_NAME);
                final Interface classloaderServiceItf = (Interface) systemComposite
                        .getFcInterface(PetalsCompositeController.SYSTEM_CLASSLOADER_FRACTAL_SRV_ITF_NAME);
                final Interface repositoryServiceItf = (Interface) systemComposite
                        .getFcInterface(PetalsCompositeController.SYSTEM_REPOSITORY_FRACTAL_SRV_ITF_NAME);
                final Interface deploymentServiceItf = (Interface) jbiManagementComposite
                        .getFcInterface(PetalsCompositeController.JBIMANAGEMENT_DEPLOYMENT_FRACTAL_SRV_ITF_NAME);
                final Interface installationServiceItf = (Interface) jbiManagementComposite
                        .getFcInterface(PetalsCompositeController.JBIMANAGEMENT_INSTALLATION_FRACTAL_SRV_ITF_NAME);
                final Interface componentInstallationServiceItf = (Interface) jbiManagementComposite
                        .getFcInterface(PetalsCompositeController.JBIMANAGEMENT_COMPONENTINSTALLATION_FRACTAL_SRV_ITF_NAME);
                final Interface atomicDeploymentServiceItf = (Interface) jbiManagementComposite
                        .getFcInterface(PetalsCompositeController.JBIMANAGEMENT_ATOMICDEPLOYMENT_FRACTAL_SRV_ITF_NAME);
                final Interface adminServiceItf = (Interface) jbiManagementComposite
                        .getFcInterface(PetalsCompositeController.JBIMANAGEMENT_ADMIN_FRACTAL_SRV_ITF_NAME);
                final Interface containerServiceItf = (Interface) jbiManagementComposite
                        .getFcInterface(PetalsCompositeController.JBIMANAGEMENT_CONTAINER_FRACTAL_SRV_ITF_NAME);
                final Interface routerServiceItf = (Interface) jbiMessagingComposite
                        .getFcInterface(PetalsCompositeController.JBIMESSAGING_ROUTER_FRACTAL_SRV_ITF_NAME);
                final Interface transportListenerServiceItf = (Interface) jbiMessagingComposite
                        .getFcInterface(PetalsCompositeController.JBIMESSAGING_TRANSPORTLISTENER_FRACTAL_SRV_ITF_NAME);
                final Component preExtensionMgrComposite = FractalHelper.getComponentByName(this.petalsCompositeCC,
                        FractalHelper.PRE_EXTENSIONSMANAGER_COMPOSITE);
                final Interface preExtensionsManagerItf = (Interface) preExtensionMgrComposite
                        .getFcInterface(PetalsCompositeController.PREEXTENSIONSMANAGER_FRACTAL_SRV_ITF_NAME);
                final Interface postExtensionsManagerItf = (Interface) extensionsManagerComposite
                        .getFcInterface(PetalsCompositeController.POSTEXTENSIONSMANAGER_EXTENSIONSMANAGER_FRACTAL_SRV_ITF_NAME);
                final Interface installationExtensionsManagerItf = (Interface) extensionsManagerComposite
                        .getFcInterface(PetalsCompositeController.POSTEXTENSIONSMANAGER_INSTALLATIONEXTENSIONSMANAGER_FRACTAL_SRV_ITF_NAME);
                final Interface petalsAdminServiceItf = (Interface) petalsAdminServiceComponent
                        .getFcInterface(PetalsAdminInterface.FRACTAL_SRV_ITF_NAME);
                final Interface transporterLocalServiceItf = (Interface) transporterComposite
                        .getFcInterface(PetalsCompositeController.TRANSPORTER_TRANSPORTERLOCAL_FRACTAL_SRV_ITF_NAME);
                final Interface transporterTcpServiceItf = (Interface) transporterComposite
                        .getFcInterface(PetalsCompositeController.TRANSPORTER_TRANSPORTERTCP_FRACTAL_SRV_ITF_NAME);
                final Interface endpointDirectoryServiceItf = (Interface) jbiMessagingComposite
                        .getFcInterface(PetalsCompositeController.JBIMESSAGING_ENDPOINT_DIRECTORY_FRACTAL_SRV_ITF_NAME);

                this.bindCompositePostExtensionsManagerItf(extensionsManagerComposite, configurationServiceItf,
                        topologyServiceItf, jmxServiceItf, systemStateServiceItf, endpointDirectoryServiceItf,
                        deploymentServiceItf, installationServiceItf, componentInstallationServiceItf,
                        atomicDeploymentServiceItf, adminServiceItf, containerServiceItf, petalsAdminServiceItf);
                this.bindCompositeJbiManagementItf(jbiManagementComposite, configurationServiceItf, jmxServiceItf,
                        systemStateServiceItf, classloaderServiceItf, repositoryServiceItf, sharedAreaServiceItf,
                        routerServiceItf, installationExtensionsManagerItf, jndiServiceItf, preExtensionsManagerItf,
                        postExtensionsManagerItf, endpointDirectoryServiceItf);
                this.bindComponentPetalsAdminService(petalsAdminServiceComponent, configurationServiceItf,
                        topologyServiceItf, deploymentServiceItf, installationServiceItf);
                this.bindCompositeJbiMessagingItf(jbiMessagingComposite, configurationServiceItf, topologyServiceItf,
                        jndiServiceItf, jmxServiceItf, transporterLocalServiceItf,
                        transporterTcpServiceItf, sharedAreaServiceItf, adminServiceItf, containerServiceItf);
                this.bindCompositeTransporterItf(transporterComposite, configurationServiceItf,
                        transportListenerServiceItf, routerServiceItf);
                this.log.debug("Interfaces of composites bound");

                // Starts composites
                this.log.debug("Starts remaining composites/component");

                this.log.debug("Starts composite " + FractalHelper.POST_EXTENSIONSMANAGER_COMPOSITE);
                FractalHelper.startComponent(extensionsManagerComposite);

                this.log.debug("Starts composite " + FractalHelper.JBI_MANAGEMENT_COMPOSITE);
                FractalHelper.startComponent(jbiManagementComposite);

                this.log.debug("Starts composite " + FractalHelper.PETALSADMIN_COMPONENT);
                FractalHelper.startComponent(petalsAdminServiceComponent);

                this.startTransporters();
                this.startJbiMessaging();

                this.log.debug("Composites started");

            } catch (final NoSuchInterfaceException e) {
                throw new InvalidCompBasedArchiException(e);
            } catch (final IllegalLifeCycleException e) {
                throw new InvalidCompBasedArchiException(e);
            } finally {
                this.log.end();
            }
        } catch (final InvalidCompBasedArchiException e) {
            this.log.error(e.getMessage(), e);
            throw e;
        } catch (final PetalsCompositeCtrlException e) {
            this.log.error(e.getMessage(), e);
            throw e;
        } catch (final Exception e) {
            this.log.error("An unexpected error occurs.", e);
            throw new PetalsCompositeCtrlException(e);
        }
    }

    private final Component createAndAddComponent(final String componentDefinition, final String componentName,
            final Component parentComposite) throws CreateComponentException {
        // Create the Fractal composite from its name
        final Component composite;
        try {
            composite = FractalHelper.createCompositeComponent(componentDefinition);
            addComponentToPetals(composite, componentName, parentComposite);
        } catch (final InstantiationException e) {
            throw new CreateComponentException(componentName, e);
        }
        return composite;

    }

    private final Component createAndAddComponent(final Class<?> componentClass, final String componentName,
            final Component parentComposite) throws CreateComponentException {
        // Create the Fractal component from its class
        final Component component;
        try {
            component = FractalHelper.createPrimitiveComponent(componentClass);
            addComponentToPetals(component, componentName, parentComposite);
        } catch (final InstantiationException e) {
            throw new CreateComponentException(componentName, e);
        }
        return component;
    }

    /**
     * Create a composite or component into the composite {@value FractalHelper#PETALS_COMPOSITE}, from its name.
     * 
     * @param componentDefinition
     *            Definition of the composite or component to create
     * @param componentName
     *            Name of the composite or component in the Fractal environment.
     * @param parentComposite
     *            The composite in which the composite or component will be created.
     * @return The composite or component created
     * @throws CreateComponentException
     *             An error occurs during the instantiation of the composite
     */
    private final void addComponentToPetals(final Component component, final String componentName,
            final Component parentComposite) throws CreateComponentException {
        this.log.start(componentName);
        try {
            // Add the composite (or component) in the parent component
            FractalHelper.addComponent(component, parentComposite, null, componentName);
        } catch (final NoSuchInterfaceException e) {
            throw new CreateComponentException(componentName, e);
        } catch (final IllegalContentException e) {
            throw new CreateComponentException(componentName, e);
        } catch (final IllegalLifeCycleException e) {
            throw new CreateComponentException(componentName, e);
        } catch (final IllegalBindingException e) {
            throw new CreateComponentException(componentName, e);
        } finally {
            this.log.end(componentName);
        }
    }

    /**
     * <p>
     * Create the composite {@value FractalHelper#SYSTEM_COMPOSITE} into the composite
     * {@value FractalHelper#PETALS_COMPOSITE}.
     * </p>
     * <p>
     * The composite creation is done with following ordered steps:
     * <ol>
     * <li>load the core of the composite 'System' described by a Fractal ADL file,</li>
     * <li>the dummy registry implementation is removed. This dummy registration implementation is used to simplify the
     * creation of the Fractal architecture by loading an ADL file instead coding all,</li>
     * <li>the implementation of the sub-component 'Registry' is retrieve from:
     * <ol>
     * <li>the configured property,</li>
     * <li>the classloader,</li>
     * <li>and, loaded in the composite 'System'</li>
     * <li>the server interface {@value PetalsCompositeController#SYSTEM_SHARED_AREA_FRACTAL_SRV_ITF_NAME} of the
     * composite is created,</li>
     * <li>all interfaces are bound.</li>
     * </ol>
     * </p>
     * 
     * @param petalsComposite
     *            The composite {@value FractalHelper#PETALS_COMPOSITE} in which the composite or component will be
     *            created.
     * @return The composite or component created
     * @throws CreateComponentException
     *             An error occurs during the instantiation of the composite
     * @throws NoSharedAreaImplException
     *             No registry implementation found
     * @throws TooManySharedAreaImplException
     *             Too many registry implementations found
     * @throws BindingInterfaceException
     *             An error occurs during the binding of the interfaces
     */
    private final Component createAndAddCompositeSystem() throws CreateComponentException, NoSharedAreaImplException,
            TooManySharedAreaImplException, BindingInterfaceException {
        this.log.start();

        // Create the base composite 'System'
        final Component systemComposite = this.createAndAddComponent(FractalHelper.SYSTEM_COMPOSITE,
                FractalHelper.SYSTEM_COMPOSITE, this.petalsComposite);

        // Remove the dummy implementation of the registry. Next it will be
        // replaced by the expected one.
        final BindingController systemCompositeBC;
        try {
            systemCompositeBC = Fractal.getBindingController(systemComposite);
            systemCompositeBC.unbindFc(PetalsCompositeController.SYSTEM_SHARED_AREA_FRACTAL_INT_SRV_ITF_NAME);

            final ContentController systemCompositeCC = Fractal.getContentController(systemComposite);
            for (final Component systemCompositeSubComponent : systemCompositeCC.getFcSubComponents()) {
                final String systemCompositeSubComponentName = Fractal.getNameController(systemCompositeSubComponent)
                        .getFcName();
                if (systemCompositeSubComponentName.equals(FractalHelper.SHARED_AREA_COMPONENT)) {
                    systemCompositeCC.removeFcSubComponent(systemCompositeSubComponent);
                    break;
                }
            }
        } catch (final NoSuchInterfaceException e) {
            throw new CreateComponentException(FractalHelper.SYSTEM_COMPOSITE, e);
        } catch (final IllegalBindingException e) {
            throw new CreateComponentException(FractalHelper.SYSTEM_COMPOSITE, e);
        } catch (final IllegalLifeCycleException e) {
            throw new CreateComponentException(FractalHelper.SYSTEM_COMPOSITE, e);
        } catch (final IllegalContentException e) {
            throw new CreateComponentException(FractalHelper.SYSTEM_COMPOSITE, e);
        }

        // Retrieve the shared area implementation from configuration and looking up the classloader
        final String sharedAreaImplProperty = this.configurationService.getDomainConfiguration()
                .getSharedAreaImplementation();
        if (sharedAreaImplProperty != null) {
            this.log.info("Shared area implementation property found: " + sharedAreaImplProperty);
            try {
                this.sharedAreaImplDef = getClass().getClassLoader().loadClass(sharedAreaImplProperty);
            } catch (final ClassNotFoundException e) {
                // TODO error instead... ?
                this.log.warning(sharedAreaImplProperty + " is not a valid class name");
            }
        }

        if (this.sharedAreaImplDef == null) {
            final ServiceLoader<SharedAreaService> serviceLoader = ServiceLoader.load(SharedAreaService.class);
            final Iterator<SharedAreaService> servicesIterator = serviceLoader.iterator();
            if (servicesIterator.hasNext()) {
                this.sharedAreaImplDef = servicesIterator.next().getClass();
                if (servicesIterator.hasNext()) {
                    this.log.error("Too many shared area implementations found.");
                    throw new TooManySharedAreaImplException();
                }
            } else {
                this.log.error("No shared area implementation found.");
                throw new NoSharedAreaImplException();
            }
        }

        this.log.info("Shared area implementation to used: " + this.sharedAreaImplDef.getName());

        // Instantiates the shared area implementation as a Fractal component
        try {
            final Component sharedAreaImpl = this.createAndAddComponent(this.sharedAreaImplDef,
                    FractalHelper.SHARED_AREA_COMPONENT, systemComposite);

            this.log.debug("Shared area: Fractal component created.");

            // Next we bind client interfaces of the registry to the right client interfaces of the composite 'System'
            final ContentController systemCompositeCC = Fractal.getContentController(systemComposite);

            final BindingController sharedAreaImplBC = Fractal.getBindingController(sharedAreaImpl);

            // We bind the client interface 'configuration' of the component 'SharedAreaServiceImpl' with the client
            // interface 'configuration' of the composite 'System'
            sharedAreaImplBC.bindFc(SharedAreaService.CONFIGURATION_FRACTAL_ITF_NAME, systemCompositeCC
                    .getFcInternalInterface(PetalsCompositeController.SYSTEM_CONFIGURATION_FRACTAL_INT_ITF_NAME));

            // We bind the server interface 'service' of the component 'SharedAreaServiceImpl' with the server interface
            // 'registry' of the composite 'System'
            final Interface sharedAreaImplServiceItf = (Interface) sharedAreaImpl
                    .getFcInterface(SharedAreaService.FRACTAL_SRV_ITF_NAME);
            systemCompositeBC.bindFc(PetalsCompositeController.SYSTEM_SHARED_AREA_FRACTAL_INT_SRV_ITF_NAME,
                    sharedAreaImplServiceItf);

            this.log.debug("Shared area: interfaces bound.");

            this.log.info("Shared area loaded");

        } catch (final NoSuchInterfaceException e) {
            this.log.error(
                    "Error binding interfaces of the shared area implementation '" + this.sharedAreaImplDef.getName()
                            + "'", e);
            throw new BindingInterfaceException(this.sharedAreaImplDef.getName(), e);
        } catch (final IllegalLifeCycleException e) {
            this.log.error(
                    "Error binding interfaces of the shared area implementation '" + this.sharedAreaImplDef.getName()
                            + "'", e);
            throw new BindingInterfaceException(this.sharedAreaImplDef.getName(), e);
        } catch (final IllegalBindingException e) {
            this.log.error(
                    "Error binding interfaces of the shared area implementation '" + this.sharedAreaImplDef.getName()
                            + "'", e);
            throw new BindingInterfaceException(this.sharedAreaImplDef.getName(), e);
        }

        return systemComposite;

    }

    /**
     * Bind interfaces of the composite 'Communication'
     * 
     * @param communicationComposite
     *            The composite 'Communication'
     * @param configurationServiceItf
     *            The server interface {@value ConfigurationService#FRACTAL_SRV_ITF_NAME} offered by the component
     *            {@value FractalHelper#CONFIGURATION_COMPONENT}.
     * @param registryServiceItf
     *            The server interface {@value EndpointRegistry#FRACTAL_SRV_ITF_NAME} offered by the component
     *            {@value FractalHelper#REGISTRY_COMPONENT}.
     * @param petalsCompositeCtrlItf
     *            The server interface {@value PetalsCompositeController#FRACTAL_SRV_ITF_NAME} offered by the component
     *            {@value FractalHelper#PETALS_COMPOSITE_CTRL_COMPONENT}.
     * @throws BindingInterfaceException
     *             An error occurs during interface binding of the composite
     */
    private final void bindCompositeCommunicationItf(final Component communicationComposite,
            final Interface configurationServiceItf, final Interface registryServiceItf,
            final Interface petalsCompositeCtrlItf) throws BindingInterfaceException {
        this.log.start(FractalHelper.COMMUNICATION_COMPOSITE);
        try {
            // We bind the client interface 'configuration' of the composite 'Communication' with the server interface
            // 'service' of the component 'ConfigurationServiceImpl'
            final BindingController communicationCompositeBC = Fractal.getBindingController(communicationComposite);
            communicationCompositeBC.bindFc(PetalsCompositeController.COMMUNICATION_CONFIGURATION_FRACTAL_ITF_NAME,
                    configurationServiceItf);
            communicationCompositeBC.bindFc(PetalsCompositeController.COMMUNICATION_SHARED_AREA_FRACTAL_ITF_NAME,
                    registryServiceItf);
            communicationCompositeBC.bindFc(
                    PetalsCompositeController.COMMUNICATION_PETALS_COMPOSITE_CTRL_FRACTAL_ITF_NAME,
                    petalsCompositeCtrlItf);

        } catch (final NoSuchInterfaceException e) {
            throw new BindingInterfaceException(FractalHelper.COMMUNICATION_COMPOSITE, e);
        } catch (final IllegalLifeCycleException e) {
            throw new BindingInterfaceException(FractalHelper.COMMUNICATION_COMPOSITE, e);
        } catch (final IllegalBindingException e) {
            throw new BindingInterfaceException(FractalHelper.COMMUNICATION_COMPOSITE, e);
        } finally {
            this.log.end(FractalHelper.COMMUNICATION_COMPOSITE);
        }
    }

    /**
     * Bind interfaces of the composite {@value FractalHelper#SYSTEM_COMPOSITE}
     * 
     * @param systemComposite
     *            The composite 'System'
     * @param configurationServiceItf
     *            The server interface {@value ConfigurationService#FRACTAL_SRV_ITF_NAME} offered by the component
     *            {@value FractalHelper#CONFIGURATION_COMPONENT}.
     * @throws BindingInterfaceException
     *             An error occurs during interface binding of the composite
     */
    private final void bindCompositeSystemItf(final Component systemComposite, final Interface configurationServiceItf)
            throws BindingInterfaceException {
        this.log.start(FractalHelper.SYSTEM_COMPOSITE);
        try {
            // We bind the client interface 'configuration' of the composite 'System' with the server interface
            // 'service' of the component 'ConfigurationServiceImpl'
            final BindingController systemCompositeBC = Fractal.getBindingController(systemComposite);
            systemCompositeBC.bindFc(PetalsCompositeController.SYSTEM_CONFIGURATION_FRACTAL_ITF_NAME,
                    configurationServiceItf);

        } catch (final NoSuchInterfaceException e) {
            throw new BindingInterfaceException(FractalHelper.SYSTEM_COMPOSITE, e);
        } catch (final IllegalLifeCycleException e) {
            throw new BindingInterfaceException(FractalHelper.SYSTEM_COMPOSITE, e);
        } catch (final IllegalBindingException e) {
            throw new BindingInterfaceException(FractalHelper.SYSTEM_COMPOSITE, e);
        } finally {
            this.log.end(FractalHelper.SYSTEM_COMPOSITE);
        }
    }

    /**
     * Bind interfaces of the composite {@value FractalHelper#EXTENSIONSMANAGER_COMPOSITE}
     * 
     * @param extensionsManagerComposite
     *            The composite 'ExtensionsManager'
     * @param configurationServiceItf
     *            The server interface {@value ConfigurationService#FRACTAL_SRV_ITF_NAME} offered by the component
     *            {@value FractalHelper#CONFIGURATION_COMPONENT}.
     * @param topologyServiceItf
     *            The server interface {@value PetalsCompositeController#COMMUNICATION_JMX_FRACTAL_SRV_ITF_NAME} offered
     *            by the the composite {@value FractalHelper#COMMUNICATION_COMPOSITE}.
     * @param jmxServiceItf
     *            The server interface {@value PetalsCompositeController#COMMUNICATION_JMX_FRACTAL_SRV_ITF_NAME} offered
     *            by the the composite {@value FractalHelper#COMMUNICATION_COMPOSITE}.
     * @param systemStateServiceItf
     *            The server interface {@value PetalsCompositeController#SYSTEM_SYSTEMSTATE_FRACTAL_SRV_ITF_NAME}
     *            offered by the the composite {@value FractalHelper#SYSTEM_COMPOSITE}.
     * @param endpointDirectoryServiceItf
     *            The server interface
     *            {@value PetalsCompositeController#JBIMESSAGING_ENDPOINT_DIRECTORY_FRACTAL_ITF_NAME} offered by the the
     *            composite {@value FractalHelper#JBI_MESSAGING_COMPOSITE}.
     * @param deploymentServiceItf
     *            The server interface {@value PetalsCompositeController#JBIMANAGEMENT_DEPLOYMENT_FRACTAL_SRV_ITF_NAME}
     *            offered by the the composite {@value FractalHelper#JBI_MANAGEMENT_COMPOSITE}.
     * @param installationServiceItf
     *            The server interface
     *            {@value PetalsCompositeController#JBIMANAGEMENT_INSTALLATION_FRACTAL_SRV_ITF_NAME} offered by the the
     *            composite {@value FractalHelper#JBI_MANAGEMENT_COMPOSITE}.
     * @param componentInstallationServiceItf
     *            The server interface
     *            {@value PetalsCompositeController#JBIMANAGEMENT_COMPONENTINSTALLATION_FRACTAL_SRV_ITF_NAME} offered by
     *            the the composite {@value FractalHelper#JBI_MANAGEMENT_COMPOSITE}.
     * @param atomicDeploymentServiceItf
     *            The server interface
     *            {@value PetalsCompositeController#JBIMANAGEMENT_ATOMICDEPLOYMENT_FRACTAL_SRV_ITF_NAME} offered by the
     *            the composite {@value FractalHelper#JBI_MANAGEMENT_COMPOSITE}.
     * @param adminServiceItf
     *            The server interface {@value PetalsCompositeController#JBIMANAGEMENT_ADMIN_FRACTAL_SRV_ITF_NAME}
     *            offered by the the composite {@value FractalHelper#JBI_MANAGEMENT_COMPOSITE}.
     * @param containerServiceItf
     *            The server interface {@value PetalsCompositeController#JBIMANAGEMENT_ADMIN_FRACTAL_SRV_ITF_NAME}
     *            offered by the the composite {@value FractalHelper#JBI_MANAGEMENT_COMPOSITE}.
     * @param petalsAdminServiceItf
     *            The server interface {@value PetalsAdminInterface#FRACTAL_SRV_ITF_NAME} offered by the the component
     *            {@value FractalHelper#PETALSADMIN_COMPONENT} .
     * @throws BindingInterfaceException
     *             An error occurs during interface binding of the composite
     */
    private final void bindCompositePostExtensionsManagerItf(final Component extensionsManagerComposite,
            final Interface configurationServiceItf, final Interface topologyServiceItf, final Interface jmxServiceItf,
            final Interface systemStateServiceItf, final Interface endpointDirectoryServiceItf,
            final Interface deploymentServiceItf, final Interface installationServiceItf,
            final Interface componentInstallationServiceItf, final Interface atomicDeploymentServiceItf,
            final Interface adminServiceItf, final Interface containerServiceItf, final Interface petalsAdminServiceItf)
            throws BindingInterfaceException {
        this.log.start(FractalHelper.POST_EXTENSIONSMANAGER_COMPOSITE);
        try {
            // We bind the client interface 'configuration' of the composite 'PostExtensionsManager' with the server
            // interface 'service' of the component 'ConfigurationServiceImpl'
            final BindingController extensionsManagerCompositeBC = Fractal
                    .getBindingController(extensionsManagerComposite);
            extensionsManagerCompositeBC.bindFc(
                    PetalsCompositeController.POSTEXTENSIONSMANAGER_CONFIGURATION_FRACTAL_ITF_NAME,
                    configurationServiceItf);

            // We bind the client interface 'jmx' of the composite 'PostExtensionsManager' with the server interface
            // 'jmx' of the composite 'Communication'
            extensionsManagerCompositeBC.bindFc(PetalsCompositeController.POSTEXTENSIONSMANAGER_JMX_FRACTAL_ITF_NAME,
                    jmxServiceItf);

            // We bind the client interface 'topology' of the composite 'PostExtensionsManager' with the server
            // interface 'topology' of the composite 'Communication'
            extensionsManagerCompositeBC.bindFc(
                    PetalsCompositeController.POSTEXTENSIONSMANAGER_TOPOLOGY_FRACTAL_ITF_NAME, topologyServiceItf);

            // We bind the client interface 'systemstate' of the composite 'PostExtensionsManager' with the server
            // interface 'systemstate' of the composite 'System'
            extensionsManagerCompositeBC
                    .bindFc(PetalsCompositeController.POSTEXTENSIONSMANAGER_SYSTEMSTATE_FRACTAL_ITF_NAME,
                            systemStateServiceItf);

            // We bind the client interface 'endpointDirectory' of the composite 'PostExtensionsManager' with the server
            // interface 'endpointDirectory' of the composite 'JBI-Messaging'
            extensionsManagerCompositeBC.bindFc(
                    PetalsCompositeController.POSTEXTENSIONSMANAGER_ENDPOINT_DIRECTORY_FRACTAL_ITF_NAME,
                    endpointDirectoryServiceItf);

            // We bind the client interface 'deployment' of the composite 'PostExtensionsManager' with the server
            // interface 'deploymentService' of the composite 'JBI-Management'
            extensionsManagerCompositeBC.bindFc(
                    PetalsCompositeController.POSTEXTENSIONSMANAGER_DEPLOYMENT_FRACTAL_ITF_NAME, deploymentServiceItf);

            // We bind the client interface 'installation' of the composite 'PostExtensionsManager' with the server
            // interface 'installationService' of the composite 'JBI-Management'
            extensionsManagerCompositeBC.bindFc(
                    PetalsCompositeController.POSTEXTENSIONSMANAGER_INSTALLATION_FRACTAL_ITF_NAME,
                    installationServiceItf);

            // We bind the client interface 'componentInstallation' of the composite 'PostExtensionsManager' with the
            // server interface 'componentInstallationService' of the composite 'JBI-Management'
            extensionsManagerCompositeBC.bindFc(
                    PetalsCompositeController.POSTEXTENSIONSMANAGER_COMPONENTINSTALLATION_FRACTAL_ITF_NAME,
                    componentInstallationServiceItf);

            // We bind the client interface 'atomicDeployment' of the composite 'PostExtensionsManager' with the server
            // interface 'atomicDeploymentService' of the composite 'JBI-Management'
            extensionsManagerCompositeBC.bindFc(
                    PetalsCompositeController.POSTEXTENSIONSMANAGER_ATOMICDEPLOYMENT_FRACTAL_ITF_NAME,
                    atomicDeploymentServiceItf);

            // We bind the client interface 'admin' of the composite 'PostExtensionsManager' with the server interface
            // 'adminService' of the composite 'JBI-Management'
            extensionsManagerCompositeBC.bindFc(PetalsCompositeController.POSTEXTENSIONSMANAGER_ADMIN_FRACTAL_ITF_NAME,
                    adminServiceItf);

            // We bind the client interface 'containerService' of the composite 'PostExtensionsManager' with the server
            // interface 'containerService' of the composite 'JBI-Management'
            extensionsManagerCompositeBC.bindFc(
                    PetalsCompositeController.POSTEXTENSIONSMANAGER_CONTAINERSERVICE_FRACTAL_ITF_NAME,
                    containerServiceItf);

            // We bind the client interface 'petalsAdmin' of the composite 'PostExtensionsManager' with the server
            // interface 'service' of the component 'PetalsAdmin'
            extensionsManagerCompositeBC
                    .bindFc(PetalsCompositeController.POSTEXTENSIONSMANAGER_PETALSADMIN_FRACTAL_ITF_NAME,
                            petalsAdminServiceItf);

        } catch (final NoSuchInterfaceException e) {
            throw new BindingInterfaceException(FractalHelper.POST_EXTENSIONSMANAGER_COMPOSITE, e);
        } catch (final IllegalLifeCycleException e) {
            throw new BindingInterfaceException(FractalHelper.POST_EXTENSIONSMANAGER_COMPOSITE, e);
        } catch (final IllegalBindingException e) {
            throw new BindingInterfaceException(FractalHelper.POST_EXTENSIONSMANAGER_COMPOSITE, e);
        } finally {
            this.log.end(FractalHelper.POST_EXTENSIONSMANAGER_COMPOSITE);
        }
    }

    /**
     * Bind interfaces of the composite {@value FractalHelper#JBI_MANAGEMENT_COMPOSITE}
     * 
     * @param jbiManagementComposite
     *            The composite {@value FractalHelper#JBI_MANAGEMENT_COMPOSITE}
     * @param configurationServiceItf
     *            The server interface {@value ConfigurationService#FRACTAL_SRV_ITF_NAME} offered by the component
     *            {@value FractalHelper#CONFIGURATION_COMPONENT}.
     * @param jmxServiceItf
     *            The server interface {@value PetalsCompositeController#COMMUNICATION_JMX_FRACTAL_SRV_ITF_NAME} offered
     *            by the the composite {@value FractalHelper#COMMUNICATION_COMPOSITE}.
     * @param systemStateServiceItf
     *            The server interface {@value PetalsCompositeController#SYSTEM_SYSTEMSTATE_FRACTAL_SRV_ITF_NAME}
     *            offered by the the composite {@value FractalHelper#SYSTEM_COMPOSITE}.
     * @param classloaderServiceItf
     *            The server interface {@value PetalsCompositeController#SYSTEM_CLASSLOADER_FRACTAL_SRV_ITF_NAME}
     *            offered by the the composite {@value FractalHelper#SYSTEM_COMPOSITE}.
     * @param repositoryServiceItf
     *            The server interface {@value PetalsCompositeController#SYSTEM_REPOSITORY_FRACTAL_SRV_ITF_NAME} offered
     *            by the the composite {@value FractalHelper#SYSTEM_COMPOSITE}.
     * @param sharedAreaServiceItf
     *            The server interface {@value PetalsCompositeController#SYSTEM_SHARED_AREA_FRACTAL_SRV_ITF_NAME}
     *            offered by the the composite {@value FractalHelper#SYSTEM_COMPOSITE}.
     * @param routerServiceItf
     *            The server interface {@value PetalsCompositeController#JBIMANAGEMENT_ROUTER_FRACTAL_ITF_NAME} offered
     *            by the the composite {@value FractalHelper#JBI_MESSAGING_COMPOSITE}.
     * @param installationExtensionsManagerItf
     *            The server interface
     *            {@value PetalsCompositeController#JBIMANAGEMENT_INSTALLATIONEXTENSIONSMANAGER_FRACTAL_ITF_NAME}
     *            offered by the the composite {@value FractalHelper#POST_EXTENSIONSMANAGER_COMPOSITE}.
     * @param jndiServiceItf
     *            The server interface {@value PetalsCompositeController#COMMUNICATION_JNDI_FRACTAL_SRV_ITF_NAME}
     *            offered by the the composite {@value FractalHelper#COMMUNICATION_COMPOSITE}.
     * @param endpointDirectoryServiceItf
     *            The server interface
     *            {@value PetalsCompositeController#JBIMESSAGING_ENDPOINT_DIRECTORY_FRACTAL_ITF_NAME} offered by the the
     *            composite {@value FractalHelper#JBI_MESSAGING_COMPOSITE}.
     * @param preExtensionManagerItf
     *            The server interface
     *            {@value PetalsCompositeController#JBIMANAGEMENT_PREEXTENSIONSMANAGER_FRACTAL_ITF_NAME} offered by the
     *            the composite {@value FractalHelper#PRE_EXTENSIONSMANAGER_COMPOSITE}.
     * @param postExtensionManagerItf
     *            The server interface
     *            {@value PetalsCompositeController#JBIMANAGEMENT_POSTEXTENSIONSMANAGER_FRACTAL_ITF_NAME} offered by the
     *            the composite {@value FractalHelper#POST_EXTENSIONSMANAGER_COMPOSITE}.
     * @throws BindingInterfaceException
     *             An error occurs during interface binding of the composite
     */
    private final void bindCompositeJbiManagementItf(final Component jbiManagementComposite,
            final Interface configurationServiceItf, final Interface jmxServiceItf,
            final Interface systemStateServiceItf, final Interface classloaderServiceItf,
            final Interface repositoryServiceItf, final Interface sharedAreaServiceItf,
            final Interface routerServiceItf, final Interface installationExtensionsManagerItf,
            final Interface jndiServiceItf, final Interface preExtensionsManagerItf,
            final Interface postExtensionsManagerItf, final Interface endpointDirectoryServiceItf)
            throws BindingInterfaceException {
        this.log.start(FractalHelper.JBI_MANAGEMENT_COMPOSITE);
        try {
            // We bind the client interface 'configuration' of the composite 'JBI-Management' with the server interface
            // 'service' of the component 'ConfigurationServiceImpl'
            final BindingController jbiManagementCompositeBC = Fractal.getBindingController(jbiManagementComposite);
            jbiManagementCompositeBC.bindFc(PetalsCompositeController.JBIMANAGEMENT_CONFIGURATION_FRACTAL_ITF_NAME,
                    configurationServiceItf);

            // We bind the client interface 'jmx' of the composite 'JBI-Management' with the server interface 'jmx' of
            // the composite 'Communication'
            jbiManagementCompositeBC
                    .bindFc(PetalsCompositeController.JBIMANAGEMENT_JMX_FRACTAL_ITF_NAME, jmxServiceItf);

            // We bind the client interface 'systemstate' of the composite 'JBI-Management' with the server interface
            // 'systemstate' of the composite 'System'
            jbiManagementCompositeBC.bindFc(PetalsCompositeController.JBIMANAGEMENT_SYSTEMSTATE_FRACTAL_ITF_NAME,
                    systemStateServiceItf);

            // We bind the client interface 'classloader' of the composite 'JBI-Management' with the server interface
            // 'classloader' of the composite 'System'
            jbiManagementCompositeBC.bindFc(PetalsCompositeController.JBIMANAGEMENT_CLASSLOADER_FRACTAL_ITF_NAME,
                    classloaderServiceItf);

            // We bind the client interface 'repository' of the composite 'JBI-Management' with the server interface
            // 'repository' of the composite 'System'
            jbiManagementCompositeBC.bindFc(PetalsCompositeController.JBIMANAGEMENT_REPOSITORY_FRACTAL_ITF_NAME,
                    repositoryServiceItf);

            // We bind the client interface 'endpointDirectory' of the composite 'JBI-Management' with the server
            // interface 'endpointDirectory' of the composite 'JBI-Messaging'
            jbiManagementCompositeBC.bindFc(
                    PetalsCompositeController.JBIMANAGEMENT_ENDPOINT_DIRECTORY_FRACTAL_ITF_NAME,
                    endpointDirectoryServiceItf);

            // We bind the client interface 'sharedArea' of the composite 'JBI-Management' with the server
            // interface 'sharedArea' of the composite 'System'
            jbiManagementCompositeBC.bindFc(PetalsCompositeController.JBIMANAGEMENT_SHARED_AREA_FRACTAL_ITF_NAME,
                    sharedAreaServiceItf);

            // We bind the client interface 'router' of the composite 'JBI-Management' with the server interface
            // 'router' of the composite 'JBI-Messaging'
            jbiManagementCompositeBC.bindFc(PetalsCompositeController.JBIMANAGEMENT_ROUTER_FRACTAL_ITF_NAME,
                    routerServiceItf);

            // We bind the client interface 'installationExtensionManager' of the composite 'JBI-Management' with the
            // server interface 'installationExtensionManager' of the composite 'ExtensionsManager'
            jbiManagementCompositeBC.bindFc(
                    PetalsCompositeController.JBIMANAGEMENT_INSTALLATIONEXTENSIONSMANAGER_FRACTAL_ITF_NAME,
                    installationExtensionsManagerItf);

            // We bind the client interface 'jndi' of the composite 'JBI-Management' with the server interface 'jndi' of
            // the composite 'Communication'
            jbiManagementCompositeBC.bindFc(PetalsCompositeController.JBIMANAGEMENT_JNDI_FRACTAL_ITF_NAME,
                    jndiServiceItf);

            // We bind the client interface 'preExtensionManager' of the composite 'JBI-Management' with the server
            // interface 'preExtensionManager' of the component 'PreExtensionsManager'
            jbiManagementCompositeBC.bindFc(
                    PetalsCompositeController.JBIMANAGEMENT_PREEXTENSIONSMANAGER_FRACTAL_ITF_NAME,
                    preExtensionsManagerItf);

            // We bind the client interface 'postExtensionManager' of the composite 'JBI-Management' with the server
            // interface 'extensionsManager' of the composite 'PostExtensionsManager'
            jbiManagementCompositeBC.bindFc(
                    PetalsCompositeController.JBIMANAGEMENT_POSTEXTENSIONSMANAGER_FRACTAL_ITF_NAME,
                    postExtensionsManagerItf);

        } catch (final NoSuchInterfaceException e) {
            throw new BindingInterfaceException(FractalHelper.JBI_MANAGEMENT_COMPOSITE, e);
        } catch (final IllegalLifeCycleException e) {
            throw new BindingInterfaceException(FractalHelper.JBI_MANAGEMENT_COMPOSITE, e);
        } catch (final IllegalBindingException e) {
            throw new BindingInterfaceException(FractalHelper.JBI_MANAGEMENT_COMPOSITE, e);
        } finally {
            this.log.end(FractalHelper.JBI_MANAGEMENT_COMPOSITE);
        }
    }

    /**
     * Bind interfaces of the composite {@value FractalHelper#PETALSADMIN_COMPONENT}
     * 
     * @param petalsAdminServiceComponent
     *            The composite {@value FractalHelper#JBI_MANAGEMENT_COMPOSITE}
     * @param configurationServiceItf
     *            The server interface {@value ConfigurationService#FRACTAL_SRV_ITF_NAME} offered by the component
     *            {@value FractalHelper#CONFIGURATION_COMPONENT}.
     * @param topologyServiceItf
     *            The server interface {@value PetalsCompositeController#COMMUNICATION_TOPOLOGY_FRACTAL_SRV_ITF_NAME}
     *            offered by the the composite {@value FractalHelper#SYSTEM_COMPOSITE}.
     * @param deploymentServiceItf
     *            The server interface {@value PetalsCompositeController#JBIMANAGEMENT_DEPLOYMENT_FRACTAL_SRV_ITF_NAME}
     *            offered by the the composite {@value FractalHelper#JBI_MANAGEMENT_COMPOSITE}.
     * @param installationServiceItf
     *            The server interface
     *            {@value PetalsCompositeController#JBIMANAGEMENT_INSTALLATION_FRACTAL_SRV_ITF_NAME} offered by the the
     *            composite {@value FractalHelper#JBI_MANAGEMENT_COMPOSITE}.
     * @throws BindingInterfaceException
     *             An error occurs during interface binding of the composite
     */
    private final void bindComponentPetalsAdminService(final Component petalsAdminServiceComponent,
            final Interface configurationServiceItf, final Interface topologyServiceItf,
            final Interface deploymentServiceItf, final Interface installationServiceItf)
            throws BindingInterfaceException {
        this.log.start(FractalHelper.PETALSADMIN_COMPONENT);
        try {
            // We bind the client interface 'configuration' of the component 'PetalsAdminServiceImpl' with the server
            // interface 'service' of the component 'ConfigurationServiceImpl'
            final BindingController petalsAdminServiceComponentBC = Fractal
                    .getBindingController(petalsAdminServiceComponent);
            petalsAdminServiceComponentBC.bindFc(PetalsCompositeController.PETALSADMIN_CONFIGURATION_FRACTAL_ITF_NAME,
                    configurationServiceItf);

            // We bind the client interface 'topology' of the composite 'PetalsAdminServiceImpl' with the server
            // interface 'topology' of the composite 'System'
            petalsAdminServiceComponentBC.bindFc(PetalsCompositeController.PETALSADMIN_TOPOLOGY_FRACTAL_ITF_NAME,
                    topologyServiceItf);

            // We bind the client interface 'deployment' of the composite 'PetalsAdminServiceImpl' with the server
            // interface 'deploymentService' of the composite 'JBI-Management'
            petalsAdminServiceComponentBC.bindFc(PetalsCompositeController.PETALSADMIN_DEPLOYMENT_FRACTAL_ITF_NAME,
                    deploymentServiceItf);

            // We bind the client interface 'installation' of the composite 'PetalsAdminServiceImpl' with the server
            // interface 'installationService' of the composite 'JBI-Management'
            petalsAdminServiceComponentBC.bindFc(PetalsCompositeController.PETALSADMIN_INSTALLATION_FRACTAL_ITF_NAME,
                    installationServiceItf);

        } catch (final NoSuchInterfaceException e) {
            throw new BindingInterfaceException(FractalHelper.PETALSADMIN_COMPONENT, e);
        } catch (final IllegalLifeCycleException e) {
            throw new BindingInterfaceException(FractalHelper.PETALSADMIN_COMPONENT, e);
        } catch (final IllegalBindingException e) {
            throw new BindingInterfaceException(FractalHelper.PETALSADMIN_COMPONENT, e);
        } finally {
            this.log.end(FractalHelper.PETALSADMIN_COMPONENT);
        }
    }

    /**
     * Bind interfaces of the composite {@value FractalHelper#JBI_MESSAGING_COMPOSITE}
     * 
     * @param jbiManagementComposite
     *            The composite {@value FractalHelper#JBI_MANAGEMENT_COMPOSITE}
     * @param configurationServiceItf
     *            The server interface {@value ConfigurationService#FRACTAL_SRV_ITF_NAME} offered by the component
     *            {@value FractalHelper#CONFIGURATION_COMPONENT}.
     * @param topologyServiceItf
     *            The server interface {@value PetalsCompositeController#COMMUNICATION_TOPOLOGY_FRACTAL_SRV_ITF_NAME}
     *            offered by the the composite {@value FractalHelper#COMMUNICATION_COMPOSITE}.
     * @param jndiServiceItf
     *            The server interface {@value PetalsCompositeController#COMMUNICATION_JNDI_FRACTAL_SRV_ITF_NAME}
     *            offered by the the composite {@value FractalHelper#COMMUNICATION_COMPOSITE}.
     * @param jmxServiceItf
     *            The server interface {@value PetalsCompositeController#COMMUNICATION_JMX_FRACTAL_SRV_ITF_NAME} offered
     *            by the the composite {@value FractalHelper#COMMUNICATION_COMPOSITE}.
     * @param transporterLocalServiceItf
     *            The server interface {@value PetalsCompositeController#TRANSPORTER_LOCAL_FRACTAL_SRV_ITF_NAME} offered
     *            by the the composite {@value FractalHelper#TRANSPORTERS_COMPOSITE}.
     * @param transporterTcpServiceItf
     *            The server interface {@value PetalsCompositeController#TRANSPORTER_TCP_FRACTAL_SRV_ITF_NAME} offered
     *            by the the composite {@value FractalHelper#TRANSPORTERS_COMPOSITE}.
     * @param registryServiceItf
     *            The server interface {@value PetalsCompositeController#SYSTEM_SHARED_AREA_FRACTAL_SRV_ITF_NAME}
     *            offered by the the composite {@value FractalHelper#SYSTEM_COMPOSITE}.
     * @throws BindingInterfaceException
     *             An error occurs during interface binding of the composite
     */
    private final void bindCompositeJbiMessagingItf(final Component jbiManagementComposite,
            final Interface configurationServiceItf, final Interface topologyServiceItf,
            final Interface jndiServiceItf, final Interface jmxServiceItf,
            final Interface transporterLocalServiceItf, final Interface transporterTcpServiceItf,
            final Interface registryServiceItf, final Interface adminServiceItf, final Interface containerServiceItf)
                    throws BindingInterfaceException {
        this.log.start(FractalHelper.JBI_MESSAGING_COMPOSITE);
        try {
            // We bind the client interface 'configuration' of the composite 'JBI-Messaging' with the server interface
            // 'service' of the component 'ConfigurationServiceImpl'
            final BindingController jbiMessagingCompositeBC = Fractal.getBindingController(jbiManagementComposite);
            jbiMessagingCompositeBC.bindFc(PetalsCompositeController.JBIMESSAGING_CONFIGURATION_FRACTAL_ITF_NAME,
                    configurationServiceItf);

            // We bind the client interface 'topology' of the composite 'JBI-Messaging' with the server interface
            // 'topology' of the composite 'System'
            jbiMessagingCompositeBC.bindFc(PetalsCompositeController.JBIMESSAGING_TOPOLOGY_FRACTAL_ITF_NAME,
                    topologyServiceItf);

            // We bind the client interface 'jndi' of the composite 'JBI-Messaging' with the server interface 'jndi' of
            // the composite 'Communication'
            jbiMessagingCompositeBC
                    .bindFc(PetalsCompositeController.JBIMESSAGING_JNDI_FRACTAL_ITF_NAME, jndiServiceItf);

            // We bind the client interface 'jmx' of the composite 'JBI-Messaging' with the server interface 'jmx' of
            // the composite 'Communication'
            jbiMessagingCompositeBC.bindFc(PetalsCompositeController.JBIMESSAGING_JMX_FRACTAL_ITF_NAME, jmxServiceItf);

            // We bind the client interface 'transporter-local' of the composite 'JBI-Messaging' with the server
            // interface 'transporter-local' of the composite 'Transporter'
            jbiMessagingCompositeBC.bindFc(PetalsCompositeController.JBIMESSAGING_TRANSPORTERLOCAL_FRACTAL_ITF_NAME,
                    transporterLocalServiceItf);

            // We bind the client interface 'transporter-tcp' of the composite 'JBI-Messaging' with the server interface
            // 'transporter-tcp' of the composite 'Transporter'
            jbiMessagingCompositeBC.bindFc(PetalsCompositeController.JBIMESSAGING_TRANSPORTERTCP_FRACTAL_ITF_NAME,
                    transporterTcpServiceItf);

            // We bind the client interface 'sharedArea' of the composite 'JBI-Messaging' with the server interface
            // 'sharedArea' of the composite 'System'
            jbiMessagingCompositeBC.bindFc(PetalsCompositeController.JBIMESSAGING_SHARED_AREA_FRACTAL_ITF_NAME,
                    registryServiceItf);

            // We bind the client interface 'container' of the composite 'JBI-Messaging' with the server interface
            // 'containerService' of the composite 'System'
            jbiMessagingCompositeBC.bindFc(PetalsCompositeController.JBIMESSAGING_CONTAINER_FRACTAL_ITF_NAME,
                    containerServiceItf);

            // We bind the client interface 'admin' of the composite 'JBI-Messaging' with the server interface
            // 'admin' of the composite 'JBI-Management'
            jbiMessagingCompositeBC.bindFc(PetalsCompositeController.JBIMESSAGING_ADMIN_FRACTAL_ITF_NAME,
                    adminServiceItf);

        } catch (final NoSuchInterfaceException e) {
            throw new BindingInterfaceException(FractalHelper.JBI_MESSAGING_COMPOSITE, e);
        } catch (final IllegalLifeCycleException e) {
            throw new BindingInterfaceException(FractalHelper.JBI_MESSAGING_COMPOSITE, e);
        } catch (final IllegalBindingException e) {
            throw new BindingInterfaceException(FractalHelper.JBI_MESSAGING_COMPOSITE, e);
        } finally {
            this.log.end(FractalHelper.JBI_MESSAGING_COMPOSITE);
        }
    }

    /**
     * Bind interfaces of the composite {@value FractalHelper#TRANSPORTERS_COMPOSITE}
     * 
     * @param transporterComposite
     *            The composite 'System'
     * @param configurationServiceItf
     *            The server interface {@value ConfigurationService#FRACTAL_SRV_ITF_NAME} offered by the component
     *            {@value FractalHelper#CONFIGURATION_COMPONENT}.
     * @param transportListenerServiceItf
     *            The server interface
     *            {@value PetalsCompositeController#JBIMESSAGING_TRANSPORTERLISTENER_FRACTAL_SRV_ITF_NAME} offered by
     *            the the composite {@value FractalHelper#JBI_MESSAGING_COMPOSITE}.
     * @throws BindingInterfaceException
     *             An error occurs during interface binding of the composite
     */
    private final void bindCompositeTransporterItf(final Component transporterComposite,
            final Interface configurationServiceItf, final Interface transportListenerServiceItf,
            final Interface routerServiceItf) throws BindingInterfaceException {
        this.log.start(FractalHelper.TRANSPORTERS_COMPOSITE);
        try {
            // We bind the client interface 'configuration' of the composite 'Transporter' with the server interface
            // 'service' of the component 'ConfigurationServiceImpl'
            final BindingController systemCompositeBC = Fractal.getBindingController(transporterComposite);
            systemCompositeBC.bindFc(PetalsCompositeController.TRANSPORTER_CONFIGURATION_FRACTAL_ITF_NAME,
                    configurationServiceItf);

            // We bind the client interface 'transportlistener' of the composite 'Transporter' with the server interface
            // 'transportlistener' of the composite 'JBI-Messaging'
            systemCompositeBC.bindFc(PetalsCompositeController.TRANSPORTER_TRANSPORTLISTENER_FRACTAL_ITF_NAME,
                    transportListenerServiceItf);

            // We bind the client interface 'router' of the composite 'Transporter' with the server interface
            // 'router' of the composite 'JBI-Messaging'
            systemCompositeBC.bindFc(PetalsCompositeController.TRANSPORTER_ROUTER_FRACTAL_ITF_NAME, routerServiceItf);

        } catch (final NoSuchInterfaceException e) {
            throw new BindingInterfaceException(FractalHelper.TRANSPORTERS_COMPOSITE, e);
        } catch (final IllegalLifeCycleException e) {
            throw new BindingInterfaceException(FractalHelper.TRANSPORTERS_COMPOSITE, e);
        } catch (final IllegalBindingException e) {
            throw new BindingInterfaceException(FractalHelper.TRANSPORTERS_COMPOSITE, e);
        } finally {
            this.log.end(FractalHelper.TRANSPORTERS_COMPOSITE);
        }
    }

    /**
     * Stops the registry manager.
     */
    @Lifecycle(step = Step.STOP)
    public void stop() {
        this.log.call();
    }

    @Override
    public void stopJbiArtefacts() throws PetalsCompositeCtrlException {
        this.log.start();
        try {
            final Component containerCtrlComponent = FractalHelper.getRecursiveComponentByName(this.petalsCompositeCC,
                    FractalHelper.CONTAINER_CONTROLLER_COMPONENT);
            if (containerCtrlComponent != null) {

                final ContainerController containerController = (ContainerController) containerCtrlComponent
                        .getFcInterface(ContainerController.FRACTAL_SRV_ITF_NAME);

                containerController.stopAllJbiArtefacts();

            } else {
                throw new FractalComponentNotFoundException(FractalHelper.CONTAINER_CONTROLLER_COMPONENT);
            }
        } catch (final NoSuchInterfaceException e) {
            throw new PetalsCompositeCtrlException(e);
        } catch (final ContainerCtrlException e) {
            throw new PetalsCompositeCtrlException(e);
        } finally {
            this.log.end();
        }
    }

    @Override
    public void stopJbiMessaging() throws PetalsCompositeCtrlException {
        this.stopComponent(FractalHelper.JBI_MESSAGING_COMPOSITE);
    }

    @Override
    public void startJbiMessaging() throws PetalsCompositeCtrlException {
        this.startComponent(FractalHelper.JBI_MESSAGING_COMPOSITE);
    }

    @Override
    public AutoCloseable pauseEndpointDirectory() throws PetalsCompositeCtrlException {
        return this.pauseComponent(FractalHelper.ENDPOINT_DIRECTORY_COMPONENT);
    }

    private void stopTransportersIndividually() throws PetalsCompositeCtrlException {

        // Stop the local transporter
        this.stopComponent(FractalHelper.LOCAL_TRANSPORTER_COMPONENT);

        // Stop the remote transporter
        this.stopComponent(FractalHelper.TCP_TRANSPORTER_COMPONENT);
    }

    private void startTransportersIndividually() throws PetalsCompositeCtrlException {

        // Stop the remote transporter
        this.startComponent(FractalHelper.TCP_TRANSPORTER_COMPONENT);

        // Start the local transporter
        this.startComponent(FractalHelper.LOCAL_TRANSPORTER_COMPONENT);
    }

    @Override
    public void startTransporters() throws PetalsCompositeCtrlException {
        this.startComponent(FractalHelper.TRANSPORTERS_COMPOSITE);
    }

    private final void stopComponent(final String componentName) throws PetalsCompositeCtrlException {
        this.log.start();
        try {
            this.log.debug(String.format("Stop fractal component '%s' ...", componentName));
            final Component composite = FractalHelper
                    .getRecursiveComponentByName(this.petalsCompositeCC, componentName);
            if (composite != null) {
                if (this.log.isDebugEnabled()) {
                    this.logComponentState(composite, "");
                }
                FractalHelper.stopComponent(composite);
            } else {
                throw new FractalCompositeNotFoundException(componentName);
            }
            this.log.debug(String.format("Fractal component '%s' stopped", componentName));

        } catch (final NoSuchInterfaceException | IllegalLifeCycleException e) {
            throw new PetalsCompositeCtrlException(String.format("An error occurs stopping the component '%s'.",
                    componentName), e);
        } finally {
            this.log.end();
        }
    }

    private AutoCloseable pauseComponent(final String componentName) throws PetalsCompositeCtrlException {
        this.stopComponent(componentName);
        return new AutoCloseable() {
            private boolean closed = false;

            @Override
            public void close() throws Exception {
                if (!closed) {
                    startComponent(componentName);
                    closed = true;
                }
            }
        };
    }

    private final void startComponent(final String componentName) throws PetalsCompositeCtrlException {
        this.log.start();
        try {
            this.log.debug(String.format("Start fractal component '%s' ...", componentName));
            final Component composite = FractalHelper
                    .getRecursiveComponentByName(this.petalsCompositeCC, componentName);
            if (composite != null) {
                FractalHelper.startComponent(composite);
            } else {
                throw new FractalCompositeNotFoundException(componentName);
            }
            this.log.debug(String.format("Fractal component '%s' started", componentName));

        } catch (final NoSuchInterfaceException | IllegalLifeCycleException e) {
            throw new PetalsCompositeCtrlException(String.format("An error occurs starting the component '%s'.",
                    componentName), e);
        } finally {
            this.log.end();
        }
    }

    @Override
    public AutoCloseable pauseMessagingSubsystem() throws PetalsCompositeCtrlException {
        this.log.start();
        try {
            this.log.debug("Waiting end of pending message exchanges ...");
            final Component routerComponent = FractalHelper.getRecursiveComponentByName(this.petalsCompositeCC,
                    FractalHelper.ROUTER_COMPONENT);
            if (routerComponent != null) {
                final RouterService router = (RouterService) routerComponent
                        .getFcInterface(RouterService.FRACTAL_SRV_ITF_NAME);

                if (router.waitEndOfPendingExchange(this.configurationService.getContainerConfiguration()
                        .getRouterPauseTrafficDelay())) {
                    this.log.debug("All pending message exchanges are terminated.");
                } else {
                    this.log.warning("Some pending message exchanges are not terminated and can be lost or can be terminated with error.");
                }
            } else {
                throw new FractalComponentNotFoundException(FractalHelper.ROUTER_COMPONENT);
            }

            this.stopTransportersIndividually();

            return new AutoCloseable() {
                private boolean closed;

                @Override
                public void close() throws Exception {
                    if (!closed) {
                        startTransportersIndividually();
                        closed = true;
                    }
                }
            };
        } catch (final NoSuchInterfaceException e) {
            throw new PetalsCompositeCtrlException("An error occurs pausing the messaging sub-system", e);
        } finally {
            this.log.end();
        }
    }

    @Override
    public AutoCloseable pauseSharedAreaComponent() throws PetalsCompositeCtrlException {
        return this.pauseComponent(FractalHelper.SHARED_AREA_COMPONENT);
    }


    @Override
    public String getSharedAreaCltImplementation() {
        return this.sharedAreaImplDef.getName();
    }

    @Override
    public Properties getSharedAreaCltConfiguration(final String providedTopologyPassphrase)
            throws PetalsCompositeCtrlException {
        try {
            final Component sharedAreaComponent = FractalHelper.getRecursiveComponentByName(this.petalsCompositeCC,
                    FractalHelper.SHARED_AREA_COMPONENT);
            if (sharedAreaComponent != null) {
                final SharedAreaService sharedArea = (SharedAreaService) sharedAreaComponent
                        .getFcInterface(SharedAreaService.FRACTAL_SRV_ITF_NAME);
                return sharedArea.getConfiguration(providedTopologyPassphrase);
            } else {
                throw new FractalComponentNotFoundException(FractalHelper.SHARED_AREA_COMPONENT);
            }

        } catch (final NoSuchInterfaceException e) {
            throw new PetalsCompositeCtrlException(
                    "An error occurs getting the configuration of the shared area mutable implementation", e);
        } catch (final MutableImplException e) {
            throw new PetalsCompositeCtrlException(
                    "An error occurs getting the configuration of the shared area mutable implementation", e);
        }
    }

    @Override
    public void stopSystemRecoveryService() throws PetalsCompositeCtrlException {
        this.stopComponent(FractalHelper.SYSTEMRECOVERY_COMPONENT);
    }

    @Override
    public void startSystemRecoveryService() throws PetalsCompositeCtrlException {
        this.startComponent(FractalHelper.SYSTEMRECOVERY_COMPONENT);
    }

    @Override
    public void onConfigurationUpdated() throws PetalsCompositeCtrlException {
        this.preExtensionsMgr.onConfigurationUpdated();
    }

    /**
     * Log the component status (started or stopped) of a Fractal component.
     * 
     * @param component
     * @throws NoSuchInterfaceException
     */
    private void logComponentState(final Component component, final String space) throws NoSuchInterfaceException {

        final LifeCycleController componentLC = Fractal.getLifeCycleController(component);
        final NameController componentNC = Fractal.getNameController(component);
        this.log.debug(space + "- " + componentNC.getFcName() + " is " + componentLC.getFcState());

        ContentController componentCC = null;
        try {
            componentCC = Fractal.getContentController(component);
        } catch (final NoSuchInterfaceException e) {
            // No content controller (ie. no sub components)
            componentCC = null;
        }

        if (componentCC != null) {
            final Component[] componentSubComponents = componentCC.getFcSubComponents();
            for (final Component componentSubComponent : componentSubComponents) {
                this.logComponentState(componentSubComponent, space + "\t");
            }
        }
    }

}
