/**
 * Copyright (c) 2005-2012 Fossil E-Commerce, 2005-2012 EBM WebSourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.system.classloader;

import java.io.IOException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.List;
import java.util.Map.Entry;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.logging.Logger;

import org.objectweb.fractal.fraclet.annotations.Component;
import org.objectweb.fractal.fraclet.annotations.Interface;
import org.objectweb.fractal.fraclet.annotations.Lifecycle;
import org.objectweb.fractal.fraclet.annotations.Requires;
import org.objectweb.fractal.fraclet.types.Step;
import org.ow2.petals.basisapi.exception.PetalsException;
import org.ow2.petals.microkernel.api.configuration.ConfigurationService;

import com.ebmwebsourcing.easycommons.log.LoggingUtil;

/**
 * @since Petals 1.0
 * @author Rafael Marins - Fossil E-Commerce 
 * @author Adrien LOUIS - EBM WebSourcing
 */
@Component(provides = @Interface(name = "service", signature = ClassLoaderService.class))
public class ClassLoaderServiceImpl implements ClassLoaderService {

    /**
     * the common loader name
     */
    public static final String COMMON_LOADER = "common";

    private final ConcurrentMap<String, SharedLibraryClassLoader> sharedLibraryClassLoaders = new ConcurrentHashMap<String, SharedLibraryClassLoader>();

    /**
     * The class loaders table.
     */
    private final ConcurrentMap<String, PetalsClassLoader> classLoaders = new ConcurrentHashMap<String, PetalsClassLoader>();

    /**
     * Parent ClassLoader for all JBI ClassLoaders.
     */
    private ClassLoader parentClassLoader;

    private final LoggingUtil log = new LoggingUtil(Logger.getLogger(Constants.FRACTAL_COMPONENT_LOGGER_NAME));

    /**
     * The Configuration service fractal component
     */
    @Requires(name = "configuration")
    private ConfigurationService configurationService;

    // ----------------------------------------------------------------------
    // ClassLoadingService interface implementation
    // ----------------------------------------------------------------------

    /**
     * @see org.ow2.petals.system.classloader.ClassLoaderService#containsClassLoader(java.lang.String)
     */
    public boolean containsClassLoader(String classLoaderId) {
        this.log.call();

        return this.classLoaders.containsKey(classLoaderId);
    }

    @Override
    public URLClassLoader createComponentClassLoader(String classLoaderId, URL[] baseUrls, List<String> classpathUrls,
            boolean useParentFirst, List<String> sharedLibraryNames) throws PetalsException {
        this.log.start();

        final PetalsClassLoader cl;
        try {
            cl = new ComponentClassLoader(baseUrls, classpathUrls, new SharedLibrariesClassLoader(parentClassLoader,
                    sharedLibraryNames, sharedLibraryClassLoaders), useParentFirst);
        } catch (final IOException e) {
            throw new PetalsException("Error while creating a PEtALS class loader", e);
        }

        this.classLoaders.put(classLoaderId, cl);

        this.log.end();
        return cl;
    }

    @Override
    public URLClassLoader createSharedLibraryClassLoader(String classLoaderId, URL[] baseUrls,
            List<String> classpathUrls, boolean useParentFirst) throws PetalsException {
        this.log.start();

        final SharedLibraryClassLoader cl;
        try {
            cl = new SharedLibraryClassLoader(baseUrls, classpathUrls, this.parentClassLoader, useParentFirst);
        } catch (IOException e) {
            throw new PetalsException("Error while creating a class loader", e);
        }

        this.classLoaders.put(classLoaderId, cl);
        this.sharedLibraryClassLoaders.put(classLoaderId, cl);

        this.log.end();
        return cl;
    }

    @Override
    public void deleteClassLoader(String classLoaderId) {
        this.log.start();

        final PetalsClassLoader classLoader = this.classLoaders.remove(classLoaderId);

        if (classLoader == null) {
            this.log.warning("No classloader with Id " + classLoaderId + " loaded. Can not remove it");
            return;
        }

        if (classLoader instanceof SharedLibraryClassLoader) {
            this.sharedLibraryClassLoaders.remove(classLoaderId);
        }

        try {
            classLoader.close();
        } catch (IOException e) {
            this.log.warning("Couldn't properly close the classloader " + classLoaderId, e);
        }

        this.log.end();
    }

    @Lifecycle(step = Step.START)
    public void start() {
        this.log.start();

        boolean isIsolated = this.configurationService.getContainerConfiguration()
                .isIsolateJBIClassLoaders();

        if (isIsolated) {
            this.parentClassLoader = ClassLoader.getSystemClassLoader();

        } else {
            this.parentClassLoader = ClassLoaderServiceImpl.class.getClassLoader();
        }

        this.log.end();
    }

    @Lifecycle(step = Step.STOP)
    public void stop() {
        this.log.start();

        // All the classloaders, SL and Component, are in this.classLoaders
        for (Entry<String, PetalsClassLoader> e : this.classLoaders.entrySet()) {
            try {
                e.getValue().close();
            } catch (IOException ex) {
                this.log.warning("Couldn't properly close the classloader " + e.getKey(), ex);
            }
        }

        this.log.end();
    }
}
