/**
 * Copyright (c) 2005-2012 EBM WebSourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.system.classloader;

import java.io.File;
import java.io.IOException;
import java.net.URISyntaxException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.Enumeration;
import java.util.List;

import org.ow2.petals.microkernel.system.classloader.factory.URLFactory;
import org.ow2.petals.microkernel.system.classloader.locator.LocatorAbstract;

/**
 * 
 * This class is subclassed by {@link ComponentClassLoader} and {@link SharedLibraryClassLoader}.
 * 
 * It implements the functionality of self-first resolution (JBI spec) where the parent is used second after loading the
 * class from this classloader.
 * 
 * It also provides methods that skip the parent classloader during resolution: they are used by
 * {@link SharedLibrariesClassLoader} to avoid querying the parent classloader (which is shared by all instance of the
 * present class) multiple times.
 * 
 * @since Petals 1.0
 * @author Adrien Louis - EBM WebSourcing
 * @author ddesjardins - EBM WebSourcing
 * @author Rafael Marins - Fossil E-Commerce
 * @author Olivier Fabre - EBM WebSourcing
 * @author Victor Noël - Linagora
 */
public abstract class PetalsClassLoader extends URLClassLoader {

    static {
        ClassLoader.registerAsParallelCapable();
    }

    /**
     * Standard Java package binary name.
     */
    private static final String JAVA_PACKAGE = "java.";

    /**
     * Standard Java extensions package binary name.
     */
    private static final String JAVAX_PACKAGE = "javax.";

    /**
     * Bases URLs (not in classloader but contain urls of the classloader)
     */
    private final URL[] bases;

    /**
     * To use or not the parent first class loading delegation model.
     */
    private final boolean parentFirst;

    /**
     * Create a class loader for petals
     * 
     * @param baseUrls
     *            installation root path in petals
     * @param classpathLocations
     *            jar defined in the classpath part of the jbi descriptor
     * @param parent
     * @throws IOException
     */
    public PetalsClassLoader(final URL[] baseUrls, final List<String> classpathLocations, final ClassLoader parent,
            final boolean parentFirst) throws IOException {
        super(baseUrls, parent);
        this.parentFirst = parentFirst;
        this.bases = baseUrls;
        initialize(classpathLocations);
    }

    /**
     * @return Returns the bases.
     */
    public URL[] getBases() {
        return bases;
    }

    /**
     * @return Returns a String representation of the classpath used by this
     *         classloader
     */
    public String getClasspath() {
        final URL[] urls = getURLs();
        final StringBuilder cp = new StringBuilder();
        for (int i = 0; i < urls.length; i++) {
            final String url = urls[i].getFile();
            // do not add URL with !/ inside (they are urls inside jars)
            if (url.indexOf("!/") == -1) {
                cp.append(File.pathSeparator + url);
            }
        }
        return cp.toString();
    }
    

    // ----------------------------------------------------------------------
    // overriding methods
    // ----------------------------------------------------------------------


    /**
     * Note: this implementation is used by {@link #getResourceAsStream(String)} from super-class so we don't need to
     * reimplement the latter!
     */
    @Override
    public URL getResource(final String resourceName) {
        return getResource(resourceName, true);
    }

    /**
     * The incluseParent parameter is used by {@link SharedLibrariesClassLoader} to avoid including parent when querying
     * many SLs classloader (since they all have the same parent).
     */
    URL getResource(final String resourceName, final boolean includesParent) {
        URL resource = null;
        // parent first
        if (isParentFirst() && includesParent) {
            // the super method is parent first by default
            resource = super.getResource(resourceName);
        } else {
            // self first
            resource = findResource(resourceName);
            if (resource == null && includesParent) {
                resource = this.getParent().getResource(resourceName);
            }
        }
        return resource;
    }

    @Override
    public Enumeration<URL> getResources(final String name) throws IOException {
        return getResources(name, true);
    }

    /**
     * The incluseParent parameter is used by {@link SharedLibrariesClassLoader} to avoid including parent when querying
     * many SLs classloader (since they all have the same parent).
     */
    @SuppressWarnings("restriction")
    Enumeration<URL> getResources(final String name, final boolean includesParent) throws IOException {
        if (isParentFirst() && includesParent) {
            // parent first
            // the super method is parent first by default
            return super.getResources(name);
        } else {
            // self first
            final Enumeration<URL> resources = findResources(name);
            if (includesParent) {
                @SuppressWarnings("rawtypes")
                final Enumeration[] tmp = new Enumeration[2];
                tmp[0] = resources;
                tmp[1] = getParent().getResources(name);
                // copied from super-method
                return new sun.misc.CompoundEnumeration<>(tmp);
            } else {
                return resources;
            }
        }
    }

    /**
     * Note: this implementation is missing the super-class performance counter stuffs...
     */
    @Override
    protected Class<?> loadClass(final String className, final boolean resolve) throws ClassNotFoundException {
        return loadClass(className, resolve, true);
    }

    /**
     * The incluseParent parameter is used by {@link SharedLibrariesClassLoader} to avoid including parent when querying
     * many SLs classloader (since they all have the same parent).
     */
    protected Class<?> loadClass(final String className, final boolean resolve, final boolean includesParent)
            throws ClassNotFoundException {

        // see parent class for explanation on the lock
        synchronized (getClassLoadingLock(className)) {

            // first let's check if the class is already loaded
            Class<?> clazz = findLoadedClass(className);

            if (clazz == null) {
                // parent first
                if (useParentFirst(className) && includesParent) {
                    // we must pass the parameter for resolve if not our method
                    // will be called by the implementation of loadClass(Class)!
                    // Note: parent first is the default implemented behaviour, let's use it!
                    clazz = super.loadClass(className, false);
                } else {
                    // self first
                    try {
                        clazz = findClass(className);
                    } catch (final ClassNotFoundException | SecurityException e) {
                        if (includesParent) {
                            // let's ask the parent (note: the parent is always there in our case)
                            clazz = this.getParent().loadClass(className);
                        } else {
                            // in that case we couldn't find it
                            throw new ClassNotFoundException(className);
                        }
                    }
                }
            }

            // clazz can't be null, if it is there is a bug somewhere
            assert clazz != null;

            if (resolve) {
                resolveClass(clazz);
            }

            return clazz;
        }
    }

    /**
     * For package, there is no meaning in parent or self-first.
     * 
     * But we must override it to make it accessible from {@link SharedLibrariesClassLoader}.
     */
    @Override
    protected Package getPackage(String name) {
        return super.getPackage(name);
    }

    /**
     * Their neither there is no meaning in parent or self-first (normally...)
     * 
     * But we must override it to make it accessible from {@link SharedLibrariesClassLoader}.
     */
    @Override
    protected Package[] getPackages() {
        return super.getPackages();
    }

    // ----------------------------------------------------------------------
    // implementation specific methods
    // ----------------------------------------------------------------------
    
    /**
     * @return Returns the useParentFirst.
     */
    public boolean isParentFirst() {
        return parentFirst;
    }

    /**
     * @return Returns a String representation of the PetalsClassLoader
     */
    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();

        sb.append("classloader : " + getClass().getName() + "\n");
        sb.append("\tmodules bases (not in loader!) : \n");
        for (int i = 0; i < bases.length; i++) {
            sb.append("\t\t -" + bases[i] + "\n");
        }
        sb.append("\trepositories :\n");
        URL[] rep = getURLs();
        for (int i = 0; i < rep.length; i++) {
            sb.append("\t\t -" + rep[i] + "\n");
        }
        sb.append("\tparent : " + getParent() + "\n");

        return sb.toString();
    }

    /**
     * Add specified location into the repository. If location is found in
     * multiple URLs of the bases, new URLs will be added multiple times.
     * 
     * @param location
     *            an entry name (for a jar) or a path name (for a directory)
     * @throws IOException
     *             when constructed URL is malformed
     */
    private final void addInRepository(String location, final LocatorAbstract[] locators,
            final URLFactory[] factories)
            throws IOException {
        // test existence in each base URL
        for (int i = 0; i < bases.length; i++) {
            if (locators[i].hasDirectory(location)) {
                if (location.charAt(location.length() - 1) != '/') {
                    location = location + "/";
                }
                addURL(factories[i].getURL(location));
            } else if (locators[i].hasFile(location)) {
                addURL(factories[i].getURL(location));
            }
        }
    }

    /**
     * Return the package name of a class
     * 
     * @param className
     *            complete name of a class
     * @return package name
     */
    protected String getPackageName(String className) {
        int lastSeparator = className.lastIndexOf('.');
        return className.substring(0, lastSeparator);
    }

    /**
     * Method to know if we choose to use the parent class loader first.
     * 
     * Only to be used for class loading, not resources
     * 
     * @param className
     *            resource name
     * @param usingForClass
     *            resource as a class
     * @return boolean
     */
    protected boolean useParentFirst(final String className) {

        boolean useParentFirst = isParentFirst();

        if (!useParentFirst) {
            // always use parent class loader when loading java packages
            useParentFirst = className.startsWith(JAVA_PACKAGE) || className.startsWith(JAVAX_PACKAGE);
        }

        return useParentFirst;
    }

    /**
     * Initialize the petals class loader
     * 
     * @param classpathLocations
     *            elements of the classpath
     */
    private final void initialize(final List<String> classpathLocations) throws IOException {
        final URLFactory[] factories = new URLFactory[bases.length];
        final LocatorAbstract[] locators = new LocatorAbstract[bases.length];
        // Create factories and locator for each base URL
        for (int i = 0; i < bases.length; i++) {
            factories[i] = URLFactory.getFactory(bases[i]);
            try {
                locators[i] = LocatorAbstract.getLocator(bases[i]);
            } catch (final URISyntaxException e) {
                throw new IOException("Invalid URL: " + bases[i]);
            }
        }
        // Add all classpathLocations into the repository
        for (String location : classpathLocations) {
            addInRepository(location, locators, factories);
        }
    }
}
