/**
 * Copyright (c) 2005-2012 Fossil E-Commerce, 2005-2012 EBM WebSourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.system.classloader;

import java.io.IOException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * This class represents a set of SharedLibrary classLoader.
 * 
 * It is a classloade that delegates to either shared libary classloaders or the parent jbi classloader. When delegating
 * to shared libs, it optimize the querying by asking the SL classloader not to query their parent every time but the
 * first.
 * 
 * (Note that actually, some code from {@link ClassLoader} or {@link URLClassLoader} will anyway call the original
 * methods on the SL classloaders so the optimization will not always be applied. In any case, the behaviour will still
 * be correct!).
 * 
 * @since Petals 1.0
 * @author Adrien LOUIS - EBM WebSourcing
 * @author ddesjardins - EBM WebSourcing
 * @author Victor Noël - Linagora
 */
public class SharedLibrariesClassLoader extends URLClassLoader {

    static {
        // we are parallel capable because parent class is and the SharedLibraryClassLoader are too!
        // moreover, we only do delegation for loadClass, so they take care of handling the lock
        ClassLoader.registerAsParallelCapable();
    }

    private List<String> sharedLibraryNames;

    private Map<String, SharedLibraryClassLoader> sharedLibraryClassLoaders;

    /**
     * We need both the shared library map and the names separated because the SLs can be reloaded at runtime.
     */
    public SharedLibrariesClassLoader(final ClassLoader parent, final List<String> sharedLibraryNames,
            final Map<String, SharedLibraryClassLoader> sharedLibraryClassLoaders) {
        super(new URL[0], parent);
        this.sharedLibraryNames = sharedLibraryNames;
        this.sharedLibraryClassLoaders = sharedLibraryClassLoaders;
    }

    /**
     * Search in all referenced SharedLibrariesClassLoaders for the requested resource.
     * 
     * Note: {@link URLClassLoader#getResourceAsStream(String)} calls this method, so no need to reimplement it in the
     * present class.
     */
    @Override
    public URL getResource(final String resourceName) {
        if (!sharedLibraryNames.isEmpty()) {
            // the first time it will be true and we will also get resources from the parent
            // and then after not (this relies on the fact that they all share the same parent classloader!
            boolean includesParent = true;
            for (final String slName : sharedLibraryNames) {
                final SharedLibraryClassLoader element = sharedLibraryClassLoaders.get(slName);
                if (element != null) {
                    final URL url = element.getResource(resourceName, includesParent);
                    if (url != null) {
                        return url;
                    }
                    includesParent = false;
                }
            }
            return null;
        } else {
            // if there is no SLs, then let's look in the parent classloader directly by calling the super-class
            // note: there is no resources in this classloader, only in the parent
            return super.getResource(resourceName);
        }
    }

    @SuppressWarnings("restriction")
    @Override
    public Enumeration<URL> getResources(final String resourceName)
            throws IOException {
        if (!sharedLibraryNames.isEmpty()) {
            final List<Enumeration<URL>> res = new ArrayList<Enumeration<URL>>(sharedLibraryNames.size());
            // the first time it will be true and we will also get resources from the parent
            // and then after not (this relies on the fact that they all share the same parent classloader!
            boolean includesParent = true;
            for (final String slName : sharedLibraryNames) {
                final SharedLibraryClassLoader slc = sharedLibraryClassLoaders.get(slName);
                if (slc != null) {
                    res.add(slc.getResources(resourceName, includesParent));
                    includesParent = false;
                }
            }
            return new sun.misc.CompoundEnumeration<>(res.toArray(new Enumeration[res.size()]));
        } else {
            // if there is no SLs, then let's look in the parent classloader directly by calling the super-class
            // note: there is no resources in this classloader, only in the parent
            return super.getResources(resourceName);
        }
    }

    /**
     * Search in all referenced SharedLibrariesClassLoaders for the requested
     * class. If the class is not found, search in the parent classLoader.
     * 
     * @see java.lang.ClassLoader#loadClass(java.lang.String, boolean)
     */
    @Override
    protected Class<?> loadClass(final String className, final boolean resolve)
            throws ClassNotFoundException {

        // here no need to use the class-lock (see parent class for explanation) as we delegate to other classloaders
        if (!sharedLibraryNames.isEmpty()) {
            // the first time it will be true and we will also get classes from the parent
            // and then after not (this relies on the fact that they all share the same parent classloader as
            // explained in the JBI specifications
            boolean includesParent = true;
            for (final String slName : sharedLibraryNames) {
                final SharedLibraryClassLoader element = sharedLibraryClassLoaders.get(slName);
                if (element != null) {
                    try {
                        return element.loadClass(className, resolve, includesParent);
                    } catch (final ClassNotFoundException | SecurityException e) {
                        // Nothing to do because we have to search in all SLClassloaders
                        includesParent = false;
                    }
                }
            }
            // if not found (and the list wasn't empty!!) it means it's not either in the parent classloader
            throw new ClassNotFoundException(className);
        } else {
            // if there is no SLs, then let's look in the parent classloader directly by calling the super-class
            // note: there is no classes in this classloader, only in the parent
            return super.loadClass(className, resolve);
        }
    }

    @Override
    protected Package getPackage(String name) {
        Package pkg = super.getPackage(name);
        if (pkg == null) {
            // it would best to cache the result as our parent does by defining the package and storing it, but it's
            // not possible to add it to the field package from the super class, and we can't have our own because only
            // ClassLoader can create instances of Package!
            for (final String slName : sharedLibraryNames) {
                final SharedLibraryClassLoader element = sharedLibraryClassLoaders.get(slName);
                if (element != null) {
                    Package pkg2 = element.getPackage(name);
                    if (pkg2 != null) {
                        return pkg2;
                    }
                }
            }
            return null;
        } else {
            return pkg;
        }
    }

    @Override
    protected Package[] getPackages() {
        // it would best to cache the result as our parent does by defining the package and storing it, but it's
        // not possible to add it to the field package from the super class, and we can't have our own because only
        // ClassLoader can create instances of Package!

        final Map<String, Package> map = new HashMap<String, Package>();

        for (final Package pkg : super.getPackages()) {
            map.put(pkg.getName(), pkg);
        }

        for (final String slName : sharedLibraryNames) {
            final SharedLibraryClassLoader element = sharedLibraryClassLoaders.get(slName);
            if (element != null) {
                for (final Package pkg : element.getPackages()) {
                    if (map.get(pkg) == null) {
                        map.put(pkg.getName(), pkg);
                    }
                }
            }
        }

        return map.values().toArray(new Package[map.size()]);
    }

    /**
     * @return Returns a String representation of the
     *         SharedLibrairiesClassLoader
     */
    @Override
    public String toString() {
        final StringBuilder sb = new StringBuilder();

        sb.append("classloader : " + getClass().getName() + "\n");

        synchronized (sharedLibraryClassLoaders) {
            sb.append("\t " + sharedLibraryClassLoaders.values());
        }

        return sb.toString();
    }
}
