/**
 * Copyright (c) 1999 Bull S.A., 2006-2012 EBM WebSourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.system.classloader.locator;

import java.io.IOException;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.List;

/**
 * A <code>Locator</code> is used to hide System specific when looking for a
 * file.
 * 
 * @author Guillaume Sauthier
 */
public abstract class LocatorAbstract {

    /**
     * Returns true when file was found.
     * 
     * @param path
     *            the path to the file to look up
     * 
     * @return true when file was found, otherwise false.
     */
    public abstract boolean hasFile(String path);

    /**
     * Returns true when directory was found.
     * 
     * @param path
     *            the path to the directory to look up
     * 
     * @return true when directory was found, otherwise false.
     */
    public abstract boolean hasDirectory(String path);

    /**
     * Returns a list of filename stored in path.
     * 
     * @param path
     *            the path to the directory where looking for files
     * 
     * @return a list of filename stored in path.
     */
    public abstract List<String> listContent(String path);

    /**
     * Return a new Locator in function of the URL type. an URL pointing to a
     * jar file will return a <code>JarFileLocator</code> and an URL pointing to
     * a directory file will return a <code>DirFileLocator</code>.
     * 
     * @param url
     *            the base URL
     * 
     * @return a new Locator in function of the URL type.
     * 
     * @throws IOException
     *             when cannot find a specialized locator for the given URL.
     * @throws URISyntaxException
     *             If the URL is invalid
     */
    public static LocatorAbstract getLocator(URL url) throws IOException, URISyntaxException {
        String path = url.getPath();
        if (path.matches(".*\\..ar")) {
            // jar detected if path ends with .?ar (*.jar, *.war, *.ear)
            return new JarFileLocator(url);
        } else if (path.endsWith("/")) {
            // directory detected if url ends with a separator /
            return new DirLocator(url);
        } else {
            String err = "Unsupported URL '" + url + "' support "
                    + "only jar archive and directory";
            throw new IOException(err);
        }
    }
}
