/**
 * Copyright (c) 2009-2012 Capgemini, 2009-2012 EBM WebSourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.transport.platform.nio.client;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectOutputStream;
import java.net.InetSocketAddress;
import java.nio.channels.SelectionKey;
import java.nio.channels.SocketChannel;
import java.util.concurrent.Semaphore;
import java.util.concurrent.TimeUnit;
import java.util.logging.Logger;

import org.ow2.petals.microkernel.transport.platform.nio.selector.NioClient;
import org.ow2.petals.microkernel.transport.platform.nio.selector.NioSelector;
import org.ow2.petals.microkernel.transport.platform.nio.selector.OutputSocketChannelContext;

import com.ebmwebsourcing.easycommons.log.LoggingUtil;

/**
 * A client agent of the NIO transporter.
 * @author Christophe DENEUX - Capgemini Sud
 */
public class NioClientAgent implements NioClient {

    private final LoggingUtil log;

    /**
     * The client socket channel on which network IO are sent.
     */
    private SocketChannel clientSocketChannel;

    /**
     * <p>
     * The socket selector to use to process network IO.
     * </p>
     * <p>
     * Note: The selector agent does not have to be stopped by a client agent,
     * because it is needed by the server agent that is already running.
     * </p>
     */
    private NioSelector selectorAgent;

    /**
     * A semaphore used to report the end of the connection establishment, or
     * the end of a write, or an error during connection or writing. This
     * semaphore is acquired before to invoke asynchronously the action to do
     * (connection or write). It is also acquire just after the action
     * submission to wait the end of action. On the action end, the semaphore is
     * released.
     */
    private Semaphore isDoneSignal = new Semaphore(1);

    /**
     * Flag indicating that this NIO Client can be used to send data. A NIO
     * client is delared unusable if an error occurs or when the TCP connection
     * is closed on the server side.
     */
    private volatile boolean isUsable = true;

    /**
     * Exception detected and returned by the selector through callbacks
     * methods.
     */
    private IOException detectedException = null;

    /**
     * The client socket channel context. Needed to be able to write bytes
     * buffer in its queue.
     */
    private OutputSocketChannelContext clientSocketChannelContext;

    /**
     * String representation of the client socket, used in trace purpose. We
     * need to store it because it is not always available, for example in case
     * of a closed socket.
     */
    private String clientSocketStr;

    /**
     * Stream supporting the object output stream used to serialize the message
     * exchanges.
     */
    private ByteArrayOutputStream bytesStream;

    /**
     * 
     * NIO Client constructor.
     * 
     * @param logger
     * @param selectorAgent
     * @param serializationCacheCleanPeriod
     * @throws IOException
     */
    public NioClientAgent(final Logger logger, final NioSelector selectorAgent) throws IOException {
        this.log = new LoggingUtil(logger);
        this.selectorAgent = selectorAgent;
    }

    /**
     * Realize a connection to the server part. It's a blocking call until the
     * connection is realized.
     * 
     * @param host
     *            IP adress of the NIO server agent
     * @param port
     *            TCP port of the NIO server agent
     * @param connectionTimeout
     *            Connection timeout in millisecond
     * @throws IOException
     */
    public void connect(final String host, final int port, final long connectionTimeout)
            throws IOException {
        this.log.start();

        try {
            final InetSocketAddress clientSocket = new InetSocketAddress(host, port);

            this.clientSocketChannel = SocketChannel.open();
            this.clientSocketChannel.configureBlocking(false);

            // TODO: Set socket options
            // this.clientSocketChannel.socket().setTcpNoDelay(true);
            this.clientSocketChannel.socket().setKeepAlive(true);
            // this.clientSocketChannel.socket().setTrafficClass(0x04 | 0x08 |
            // 0x010);
            // this.clientSocketChannel.socket().setPerformancePreferences(0, 1,
            // 2);
            // TODO: IP buffer must be aligned with TCP buffer to avoid
            // fragmentation

            this.isDoneSignal.acquire();

            if (!this.clientSocketChannel.connect(clientSocket)) {
                // The connection is not established immediately.
                // We register to OP_CONNECT and ...
                this.selectorAgent
                        .register(SelectionKey.OP_CONNECT, this.clientSocketChannel, this);

                // ... we wait the connection establishment trying to acquire
                // the semaphore. When the semaphore is acquired, we must
                // release it to be able to reuse it.
                if (this.isDoneSignal.tryAcquire(connectionTimeout, TimeUnit.MILLISECONDS)) {
                    this.isDoneSignal.release();

                    // We check if an error occurs during connection
                    if (this.detectedException != null) {
                        throw this.detectedException;
                    }
                } else {
                    // Connection timeout
                    this.isDoneSignal.release();
                    this.disconnect();
                    throw new IOException("A connection timeout occurs on socket (" + clientSocket
                            + ").");
                }
            } else {
                // The connection is established immediately. We release the
                // semaphore acquired before connection establishment.
                this.isDoneSignal.release();
            }

            this.clientSocketStr = this.clientSocketChannel.socket().toString();

            this.log.info("Connection established to: " + this.clientSocketStr);

        } catch (final InterruptedException e) {
            // The NIO Client is stopped during the TCP connection establishment
            this.log.warning("Connection interrupted.");
        }

        this.log.end();
    }

    @Override
    public void onIsConnected(final OutputSocketChannelContext oscc) {
        this.clientSocketChannelContext = oscc;
        this.bytesStream = new ByteArrayOutputStream();

        try {
            // OP_READ must be registered to detect disconnection on the
            // remote side.
            this.selectorAgent.register(SelectionKey.OP_READ, this.clientSocketChannel,
                    this.clientSocketChannelContext);
        } catch (Exception e) {
            this.log
                    .warning("An error occurs during OP_READ registration, perhaps disconnection of the remote side won't be detected ("
                            + e.getMessage() + ").");
        }

        this.isDoneSignal.release();
    }

    @Override
    public void onConnectionError(final IOException exception) {
        this.detectedException = exception;
        this.isDoneSignal.release();
    }

    @Override
    public void send(final Object obj, final long timeout) throws IOException, InterruptedException {
        this.log.start();

        try {
            ObjectOutputStream objectOutputStream = new ObjectOutputStream(this.bytesStream);
            objectOutputStream.writeUnshared(obj);
            objectOutputStream.flush();

            final byte[] buffer = this.bytesStream.toByteArray();
            // We prepare the bytes stream for the next send
            this.bytesStream.reset();

            if (this.clientSocketChannelContext.offer(buffer)) {

                this.isDoneSignal.acquire();
                this.detectedException = null;

                // Bytes are available to be sent, we notify the selector.
                // Don't forget OP_READ to be notified when the socket is closed
                // by the server side
                this.selectorAgent.register(SelectionKey.OP_WRITE | SelectionKey.OP_READ,
                        this.clientSocketChannel, this.clientSocketChannelContext);

                // ... we wait the end of the bytes send trying to acquire the
                // semaphore. When the semaphore is acquired, we must
                // release it to be able to reuse it.
                if (this.isDoneSignal.tryAcquire(timeout, TimeUnit.MILLISECONDS)) {
                    this.isDoneSignal.release();

                    // We check if an error occurs during connection
                    if (this.detectedException != null) {
                        throw this.detectedException;
                    }
                } else {
                    // Write timeout
                    this.isDoneSignal.release();
                    this.disconnect();
                    throw new IOException("A write timeout occurs on socket ("
                            + this.clientSocketStr + ").");
                }

                // The OP_WRITE event is unregistered by the "onIsWritten"
                // callback to avoid extra loop in the selector.

                // We check if an error occurs during connection
                if (this.detectedException != null) {
                    throw this.detectedException;
                }

            } else {
                throw new IOException(
                        "Unable to transmit data to the NIO selector (the queue is full).");
            }
        } catch (final IOException e) {
            this.isUsable = false;
            throw e;
        }

        this.log.end();
    }

    @Override
    public void onIsWritten() throws IOException {
        // Data are written, we unregister the OP_WRITE event. We unregister the
        // OP_WRITE event in the callback to avoid extra loop in the selector.
        // Don't forget to keep OP_READ registered to be notified when the
        // socket is closed by the server side.
        this.selectorAgent.register(SelectionKey.OP_READ, this.clientSocketChannel,
                this.clientSocketChannelContext);

        this.isDoneSignal.release();
    }

    @Override
    public void onWriteError(final IOException exception) throws IOException {
        // As the socket channel is closed by the selector agent when an error
        // occurs, it is not needed to deregister the OP_WRITE event.

        this.detectedException = exception;

        this.isDoneSignal.release();
    }

    @Override
    public void onClose(final SocketChannel socketChannel) {
        this.isUsable = false;
        this.log.warning("Connection closed by the server side ("
                + socketChannel.socket().getInetAddress() + ").");
        try {
            this.disconnect();
        } catch (final IOException e) {
            this.log.warning(e.getMessage(), e);
        }
    }

    public void disconnect() throws IOException {
        this.isUsable = false;
        this.clientSocketChannel.close();

        this.log.info("Client disconnected: " + this.clientSocketStr);
    }

    /**
     * <p>
     * This client can be used (i.e. the TCP connection is valid) ?
     * </p>
     * <p>
     * This method is thread-safe.
     * </p>
     * 
     * @return
     */
    public boolean isUsable() {
        return this.isUsable;
    }

}
