/**
 * Copyright (c) 2009-2012 Capgemini, 2009-2012 EBM WebSourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.transport.platform.nio.selector;

import java.io.IOException;
import java.net.SocketException;
import java.nio.channels.SelectionKey;
import java.nio.channels.SocketChannel;
import java.util.Queue;
import java.util.concurrent.LinkedBlockingQueue;

/**
 * The socket channel context of an output socket. It contains following
 * information:
 * <ul>
 * <li>the socket channel,</li>
 * <li>its pending writes.</li>
 * </ul>
 * 
 * @author Christophe DENEUX - Capgemini Sud
 */
public class OutputSocketChannelContext extends AbstractSocketChannelContext {

    /**
     * Remaining externalized message exchanges as bytes array to write on the
     * client socket.
     */
    private Queue<byte[]> writesQueue = new LinkedBlockingQueue<byte[]>();

    /**
     * The NIO client to invoke when bytes are written or a socket is detected
     * as closed.
     */
    private NioClient nioClient;

    /**
     * The current serialized message exchange to write. As it can be written in
     * several part because it size is greater than the direct buffer size, it
     * is needed to keep its reference.
     */
    private byte[] currentBuffer;

    /**
     * Constructor to use on server socket channel
     * 
     * @param socketChannel
     *            The socket channel on which a context must be created.
     * @param nioClient
     *            The NIO Client agent using this output socket context
     * @throws SocketException
     */
    public OutputSocketChannelContext(final SocketChannel socketChannel, final NioClient nioClient)
            throws SocketException {
        // TODO: The size of the direct buffer to use to write should be
        // configured instead of using the SO_SNDBUF size.
        super(socketChannel, socketChannel.socket().getSendBufferSize());
        this.nioClient = nioClient;
    }

    /**
     * Add the data to send to a queue waiting its send.
     * 
     * @param buffer
     *            The data to send.
     * @return <code>true</code> if data have been correctly added to the queue;
     *         <code>false</code> if no more space in the queue.
     */
    public final boolean offer(final byte[] buffer) {

        return this.writesQueue.offer(buffer);

    }

    /**
     * This method is called by the NIO selector to write available data on the
     * socket channel. If needed, data are cut in several packet to be
     * multiplexed into the NIO selector and to use the same direct byte buffer.
     * 
     * @param key
     */
    public final void write(final SelectionKey key) {

        if (!this.isIOPending) {
            // No write is pending,we prepare a new one
            this.currentBuffer = this.writesQueue.poll();

            if (this.currentBuffer != null) {
                // We prepare the number of remaining bytes to write in the buffer
                // The header size is not counted as it is put now
                this.msgExRemainingBytes = this.currentBuffer.length;

                // We add the header (size of the externalized message exchange)
                // Note: It is not needed to clear the buffer because the clear
                // is done by the last write of the previous serialized message
                // exchange send.
                this.directBuffer.putInt(this.currentBuffer.length);

            } else {
                // Nothing to do because no more data to send
                this.msgExRemainingBytes = 0;
            }
        }

        try {

            if (this.msgExRemainingBytes > 0) {
                // In case of remaining bytes existing in the direct buffer, we
                // complete the buffer with remaining bytes of the next
                // serialized message exchange part before to try a re-send.
                // In case of an empty direct buffer, we push the next
                // serialized message exchange part into it.
                // In case of the header has been included in the direct buffer
                // in this loop, we must take into account the header size to
                // compute available space in the direct buffer.

                final int nbBytesToPut = Math.min(this.msgExRemainingBytes,
                        this.directBuffer.capacity() - this.directBuffer.position());
                this.directBuffer.put(this.currentBuffer, this.currentBuffer.length
                        - this.msgExRemainingBytes, nbBytesToPut);

                this.directBuffer.flip();

                // We write that we have to write
                this.socketChannel.write(this.directBuffer);

                // We manage partial write. If all bytes have been written, we
                // clear the buffer to write the next message exchange part on
                // the next NIO selector iteration. Otherwise, we do nothing:
                // remaining bytes of the direct buffer will be sent on the next
                // NIO selector iteration.
                if (this.directBuffer.hasRemaining()) {

                    // Remaining bytes will be resend on the next NIO selector
                    // iteration. We compact the buffer to be able to complete
                    // it with remaining bytes of data.
                    this.directBuffer.compact();

                    // TODO: To prevent infinite loop if the TCP send buffer is
                    // full for a long time because of network problems, it is
                    // needed to detect it and do something. In this case, the
                    // socket will be closed and few message exchanges can be
                    // lost.
                } else {
                    // All bytes have been written, we clear the buffer to write
                    // the next message exchange part.
                    this.directBuffer.clear();
                }

                this.msgExRemainingBytes -= nbBytesToPut;
            }

            if (this.msgExRemainingBytes > 0) {
                // The write is done partially
                this.isIOPending = true;
            } else {
                // The write is completed
                this.isIOPending = false;
                this.nioClient.onIsWritten();
            }
        } catch (final IOException e) {
            // In error occurs during write, we close the connection
            try {
                this.socketChannel.close();
                this.isIOPending = false;
                key.cancel();
                this.nioClient.onWriteError(e);
            } catch (final IOException e1) {
                // NOP
            }
        }
    }

    public final void onClose(final SelectionKey key) {
        this.nioClient.onClose(this.socketChannel);
    }
}
