/**
 * Copyright (c) 2014-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.communication.topology;

import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.util.Collection;
import java.util.Collections;
import java.util.Properties;

import org.easymock.classextension.EasyMock;
import org.junit.Rule;
import org.junit.Test;
import org.ow2.petals.basisapi.exception.PetalsException;
import org.ow2.petals.clientserverapi.configuration.ContainerConfiguration;
import org.ow2.petals.clientserverapi.topology.exception.AttachContainerException;
import org.ow2.petals.clientserverapi.topology.exception.DomainIsNotDynamicException;
import org.ow2.petals.clientserverapi.topology.exception.InconsistentContainerNameException;
import org.ow2.petals.clientserverapi.topology.exception.InconsistentDomainNameException;
import org.ow2.petals.clientserverapi.topology.exception.InconsistentDomainRegistryImplException;
import org.ow2.petals.clientserverapi.topology.exception.InvalidPassphraseException;
import org.ow2.petals.clientserverapi.topology.exception.NoContainerConfigurationProvidedException;
import org.ow2.petals.clientserverapi.topology.exception.NoContainerNameProvidedException;
import org.ow2.petals.clientserverapi.topology.exception.NoDomainConfigurationProvidedException;
import org.ow2.petals.clientserverapi.topology.exception.NoSecurityPassPhraseProvidedException;
import org.ow2.petals.jmx.api.api.TopologyServiceClient;
import org.ow2.petals.jmx.api.api.exception.ConnectionErrorException;
import org.ow2.petals.jmx.api.mock.junit.PetalsJmxApiJunitRule;
import org.ow2.petals.launcher.api.server.conf.ConfigurationProperties;
import org.ow2.petals.microkernel.api.communication.topology.TopologyService;
import org.ow2.petals.microkernel.api.configuration.ConfigurationService;
import org.ow2.petals.microkernel.api.system.sharedarea.SharedAreaService;
import org.ow2.petals.microkernel.api.system.sharedarea.exception.SharedAreaException;
import org.ow2.petals.microkernel.configuration.ConfigurationServiceImpl;
import org.ow2.petals.microkernel.configuration.ConfigurationTestUtils;
import org.ow2.petals.microkernel.configuration.LocalServerPropertiesBuilder;
import org.ow2.petals.microkernel.server.PetalsCompositeController;
import org.ow2.petals.microkernel.system.sharedarea.DummySharedAreaImpl;
import org.ow2.petals.topology.generated.Container;
import org.ow2.petals.topology.generated.Domain;
import org.ow2.petals.topology.generated.DomainMode;
import org.ow2.petals.topology.generated.JmxService;
import org.ow2.petals.topology.generated.Topology;
import org.ow2.petals.topology.generated.TransportService;

import com.ebmwebsourcing.easycommons.lang.reflect.ReflectionHelper;

/**
 * Unit tests of {@link TopologyServiceImpl#attachContainerTo(ContainerConfiguration, String)}
 * 
 * @author Christophe DENEUX - Linagora
 * 
 */
public class TopologyServiceImpl_AttachContainerToTest extends AbstractTopologyServiceImplTest {

    private final static String TARGET_DOMAIN_NAME = "target-domain-name";

    private final static String TARGET_CONTAINER_NAME = "target-container-name";

    /**
     * Port of the JMX server of the target container
     */
    private final static Integer TARGET_JMX_PORT = Integer.valueOf(7070);

    /**
     * Hostname of the target container
     */
    private final static String TARGET_JMX_HOST = "localhost";

    private final static String TARGET_JMX_USER = "petals-user";

    private final static String TARGET_JMX_PWD = "petals-pwd";

    private final static int TARGET_CONTAINER_TCP_PORT = ConfigurationTestUtils.CONTAINER_TCP_PORT_0 + 1;

    private final static String TARGET_PASS_PHRASE = "pass-phrase";

    @Rule
    public final PetalsJmxApiJunitRule jmxApiOnTargetContainer = new PetalsJmxApiJunitRule(TARGET_JMX_HOST,
            TARGET_JMX_PORT, null, null);

    @Rule
    public final LocalServerPropertiesBuilder localServerPropertiesBuilder = new LocalServerPropertiesBuilder();

    @Rule
    public final LocalServerPropertiesBuilder targetServerPropertiesBuilder = new LocalServerPropertiesBuilder();

    /**
     * Instantiate and start the topology service
     * 
     * @param cfgSvc
     * @return
     * @throws InvocationTargetException
     * @throws SharedAreaException
     * @throws InterruptedException
     */
    private final TopologyService createAndStartTopologySvc(final ConfigurationService cfgSvc,
            final Topology targetTopology, final PetalsCompositeController petalsCompositeCtrlMock)
            throws InvocationTargetException, SharedAreaException, InterruptedException {

        final SharedAreaService sharedAreaMock = EasyMock.createMock(SharedAreaService.class);
        Collection<SharedAreaService.SavedEndpoint> endpoints = Collections.<SharedAreaService.SavedEndpoint> emptyList();
        EasyMock.expect(sharedAreaMock.leaveTopology()).andReturn(endpoints);
        sharedAreaMock.joinTopology(endpoints);
        EasyMock.replay(sharedAreaMock);

        final TopologyService topoSvc = new TopologyServiceImpl();
        ReflectionHelper.setPrivateField(topoSvc, "configurationService", cfgSvc);
        ReflectionHelper.setPrivateField(topoSvc, "sharedArea", sharedAreaMock);
        if (petalsCompositeCtrlMock != null) {
            ReflectionHelper.setPrivateField(topoSvc, "petalsCompositeController", petalsCompositeCtrlMock);
        }
        this.startTopologyService(topoSvc);

        return topoSvc;
    }

    /**
     * <p>
     * Check {@link TopologyServiceImpl#attachContainerTo(ContainerConfiguration, String)} against:
     * <ul>
     * <li>no container configuration is provided</li>
     * </ul>
     * </p>
     * <p>
     * Expected results: An error occurs about the missing container configuration
     * </p>
     */
    @Test(expected = NoContainerConfigurationProvidedException.class)
    public void testNoContainerConfiguration() throws Exception {

        // Create a basic sample topology
        final Topology basicCurrentTopology = ConfigurationTestUtils.createBasicSampleTopology();

        // Create and initialize the configuration service
        final Properties currentServerProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);
        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);
        try {
            cfgSvc.loadConfiguration(currentServerProps, basicCurrentTopology);

            final TopologyService topoSvc = createAndStartTopologySvc(cfgSvc, null, null);

            try {
                topoSvc.attachContainerTo(null, null);
            } finally {
                this.stopTopologyService(topoSvc);
            }
        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check {@link TopologyServiceImpl#attachContainerTo(ContainerConfiguration, String)} against:
     * <ul>
     * <li>no container name is provided in the given container configuration</li>
     * </ul>
     * </p>
     * <p>
     * Expected results: An error occurs about the missing container name
     * </p>
     */
    @Test(expected = NoContainerNameProvidedException.class)
    public void testNoContainerName() throws Exception {

        // Create a basic sample topology
        final Topology basicCurrentTopology = ConfigurationTestUtils.createBasicSampleTopology();

        // Create and initialize the configuration service
        final Properties currentServerProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);
        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);
        try {
            cfgSvc.loadConfiguration(currentServerProps, basicCurrentTopology);

            final TopologyService topoSvc = createAndStartTopologySvc(cfgSvc, null, null);

            try {
                final ContainerConfiguration targetContainerConfiguration = new ContainerConfiguration();

                topoSvc.attachContainerTo(targetContainerConfiguration, TARGET_PASS_PHRASE);
            } finally {
                this.stopTopologyService(topoSvc);
            }
        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check {@link TopologyServiceImpl#attachContainerTo(ContainerConfiguration, String)} against:
     * <ul>
     * <li>no security pass-phrase is provided</li>
     * </ul>
     * </p>
     * <p>
     * Expected results: An error occurs about the missing security pass-phrase
     * </p>
     */
    @Test(expected = NoSecurityPassPhraseProvidedException.class)
    public void testNoSecurityPassPhrase() throws Exception {

        // Create a basic sample topology
        final Topology basicCurrentTopology = ConfigurationTestUtils.createBasicSampleTopology();

        // Create and initialize the configuration service
        final Properties currentServerProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);
        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);
        try {
            cfgSvc.loadConfiguration(currentServerProps, basicCurrentTopology);

            final TopologyService topoSvc = createAndStartTopologySvc(cfgSvc, null, null);

            try {
                final ContainerConfiguration targetContainerConfiguration = new ContainerConfiguration();
                targetContainerConfiguration.setName(TARGET_CONTAINER_NAME);

                topoSvc.attachContainerTo(targetContainerConfiguration, null);
            } finally {
                this.stopTopologyService(topoSvc);
            }
        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check {@link TopologyServiceImpl#attachContainerTo(ContainerConfiguration, String)} against:
     * <ul>
     * <li>an invalid pass-phrase is provided</li>
     * </ul>
     * </p>
     * <p>
     * Expected results: An error occurs about the invalid pass-phrase
     * </p>
     */
    @Test(expected = InvalidPassphraseException.class)
    public void testInvalidPassPhrase() throws Exception {

        // Create a basic target topology for the remote container
        final Topology basicTargetTopology = TopologyServiceImpl_AttachContainerToTest.createBasicTargetTopology();
        basicTargetTopology.getDomain().setMode(DomainMode.DYNAMIC);
        final Properties targetServerProps = this.targetServerPropertiesBuilder
                .createMinimalServerProperties(TARGET_CONTAINER_NAME);
        targetServerProps.put(ConfigurationProperties.TOPOLOGY_PASSPHRASE, TARGET_PASS_PHRASE);
        final ConfigurationService targetCfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(targetCfgSvc);
        try {
            targetCfgSvc.loadConfiguration(targetServerProps, basicTargetTopology);
            final PetalsCompositeController targetPetalsCompositeCtrl = EasyMock
                    .createMock(PetalsCompositeController.class);
            EasyMock.expect(targetPetalsCompositeCtrl.getSharedAreaCltConfiguration(TARGET_PASS_PHRASE))
                    .andReturn(new Properties()).anyTimes();
            EasyMock.replay(targetPetalsCompositeCtrl);
            final TopologyService targetTopoSvc = new TopologyServiceImpl();
            ReflectionHelper.setPrivateField(targetTopoSvc, "configurationService", targetCfgSvc);
            ReflectionHelper.setPrivateField(targetTopoSvc, "petalsCompositeController", targetPetalsCompositeCtrl);
            this.startTopologyService(targetTopoSvc);
            try {
                // Create a basic sample topology for the current container
                final Topology basicCurrentTopology = ConfigurationTestUtils.createBasicSampleTopology();
                final Domain currentDomain = basicCurrentTopology.getDomain();
                currentDomain.setMode(DomainMode.DYNAMIC);
                currentDomain.setRegistryImplementation("a.registry.implementation");

                // Create and initialize the configuration service
                final Properties currentServerProps = this.localServerPropertiesBuilder
                        .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);
                final ConfigurationService currentCfgSvc = new ConfigurationServiceImpl();
                ConfigurationTestUtils.startConfigurationService(currentCfgSvc);
                try {
                    currentCfgSvc.loadConfiguration(currentServerProps, basicCurrentTopology);

                    final TopologyService currentTopoSvc = createAndStartTopologySvc(currentCfgSvc, null, null);

                    try {
                        final ContainerConfiguration targetContainerConfiguration = new ContainerConfiguration();
                        targetContainerConfiguration.setName(TARGET_CONTAINER_NAME);
                        targetContainerConfiguration.setDomainName(TARGET_DOMAIN_NAME);
                        targetContainerConfiguration.setHost(TARGET_JMX_HOST);
                        targetContainerConfiguration.setJmxRMIConnectorPort(TARGET_JMX_PORT);

                        final TopologyServiceClient jmxTopologyServiceClientMock = EasyMock
                                .createMock(TopologyServiceClient.class);
                        EasyMock.expect(jmxTopologyServiceClientMock.retrieveTopology("invalid-pass-phrase", true))
                                .andReturn(targetTopoSvc.retrieveTopology("invalid-pass-phrase", true)).anyTimes();
                        EasyMock.expect(jmxTopologyServiceClientMock.retrieveSharedAreaCltImplementation())
                                .andReturn(DummySharedAreaImpl.class.getName()).anyTimes();
                        
                        EasyMock.replay(jmxTopologyServiceClientMock);
                        this.jmxApiOnTargetContainer.setTopologyServiceClient(jmxTopologyServiceClientMock);

                        currentTopoSvc.attachContainerTo(targetContainerConfiguration, "invalid-pass-phrase");
                    } finally {
                        this.stopTopologyService(currentTopoSvc);
                    }
                } finally {
                    ConfigurationTestUtils.stopConfigurationService(currentCfgSvc);
                }
            } finally {
                this.stopTopologyService(targetTopoSvc);
            }
        } finally {
            ConfigurationTestUtils.stopConfigurationService(targetCfgSvc);
        }
    }

    /**
     * <p>
     * Check {@link TopologyServiceImpl#attachContainerTo(ContainerConfiguration, String)} against:
     * <ul>
     * <li>an empty security pass-phrase is provided</li>
     * </ul>
     * </p>
     * <p>
     * Expected results: An error occurs about the missing security pass-phrase
     * </p>
     */
    @Test(expected = NoSecurityPassPhraseProvidedException.class)
    public void testEmptySecurityPassPhrase() throws Exception {

        // Create a basic sample topology
        final Topology basicCurrentTopology = ConfigurationTestUtils.createBasicSampleTopology();

        // Create and initialize the configuration service
        final Properties currentServerProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);
        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);
        try {
            cfgSvc.loadConfiguration(currentServerProps, basicCurrentTopology);

            final TopologyService topoSvc = createAndStartTopologySvc(cfgSvc, null, null);

            try {
                final ContainerConfiguration targetContainerConfiguration = new ContainerConfiguration();
                targetContainerConfiguration.setName(TARGET_CONTAINER_NAME);

                topoSvc.attachContainerTo(targetContainerConfiguration, "");
            } finally {
                this.stopTopologyService(topoSvc);
            }
        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check {@link TopologyServiceImpl#attachContainerTo(ContainerConfiguration, String)} against:
     * <ul>
     * <li>the domain of the given container configuration is not set</li>
     * </ul>
     * </p>
     * <p>
     * Expected results: An error occurs about the missing domain information of the given container
     * </p>
     */
    @Test(expected = NoDomainConfigurationProvidedException.class)
    public void testNoDomain() throws Exception {

        // Create a basic sample topology
        final Topology basicCurrentTopology = ConfigurationTestUtils.createBasicSampleTopology();

        // Create and initialize the configuration service
        final Properties currentServerProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);
        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);
        try {
            cfgSvc.loadConfiguration(currentServerProps, basicCurrentTopology);

            final TopologyService topoSvc = createAndStartTopologySvc(cfgSvc, null, null);

            try {
                final ContainerConfiguration targetContainerConfiguration = new ContainerConfiguration();
                targetContainerConfiguration.setName(TARGET_CONTAINER_NAME);

                topoSvc.attachContainerTo(targetContainerConfiguration, TARGET_PASS_PHRASE);
            } finally {
                this.stopTopologyService(topoSvc);
            }
        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check {@link TopologyServiceImpl#attachContainerTo(ContainerConfiguration, String)} against:
     * <ul>
     * <li>the domain of the current container is static</li>
     * </ul>
     * </p>
     * <p>
     * Expected results: An error occurs because the domain mode is not <b>dynamic</b>
     * </p>
     */
    @Test(expected = DomainIsNotDynamicException.class)
    public void testCurrentDomainNotDynamic() throws Exception {

        // Create a basic sample topology
        final Topology basicCurrentTopology = ConfigurationTestUtils.createBasicSampleTopology();
        basicCurrentTopology.getDomain().setMode(DomainMode.STATIC);

        // Create and initialize the configuration service
        final Properties currentServerProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);
        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);
        try {
            cfgSvc.loadConfiguration(currentServerProps, basicCurrentTopology);

            final TopologyService topoSvc = createAndStartTopologySvc(cfgSvc, null, null);

            try {
                final ContainerConfiguration targetContainerConfiguration = new ContainerConfiguration();
                targetContainerConfiguration.setName(TARGET_CONTAINER_NAME);
                targetContainerConfiguration.setDomainName(TARGET_DOMAIN_NAME);
                targetContainerConfiguration.setHost(TARGET_JMX_HOST);
                targetContainerConfiguration.setJmxRMIConnectorPort(TARGET_JMX_PORT);

                topoSvc.attachContainerTo(targetContainerConfiguration, TARGET_PASS_PHRASE);
            } finally {
                this.stopTopologyService(topoSvc);
            }
        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check {@link TopologyServiceImpl#attachContainerTo(ContainerConfiguration, String)} against:
     * <ul>
     * <li>the target container is not running</li>
     * </ul>
     * </p>
     * <p>
     * Expected results: An error occurs because the container can't be contacted
     * </p>
     */
    @Test(expected = ConnectionErrorException.class)
    public void testTargetContainerNotRunning() throws Throwable {

        // Create a basic sample topology
        final Topology basicCurrentTopology = ConfigurationTestUtils.createBasicSampleTopology();
        basicCurrentTopology.getDomain().setMode(DomainMode.DYNAMIC);

        // Create and initialize the configuration service
        final Properties currentServerProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);
        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);
        try {
            cfgSvc.loadConfiguration(currentServerProps, basicCurrentTopology);

            final TopologyService topoSvc = createAndStartTopologySvc(cfgSvc, null, null);

            try {
                final ContainerConfiguration targetContainerConfiguration = new ContainerConfiguration();
                targetContainerConfiguration.setName(TARGET_CONTAINER_NAME);
                targetContainerConfiguration.setDomainName(TARGET_DOMAIN_NAME);
                targetContainerConfiguration.setHost("host-not-running");
                targetContainerConfiguration.setJmxRMIConnectorPort(TARGET_JMX_PORT);

                topoSvc.attachContainerTo(targetContainerConfiguration, TARGET_PASS_PHRASE);
            } catch (final AttachContainerException e) {
                // It's the expected direct exception. The error connection exception is the cause of this exception
                throw e.getCause();
            } finally {
                this.stopTopologyService(topoSvc);
            }
        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check {@link TopologyServiceImpl#attachContainerTo(ContainerConfiguration, String)} against:
     * <ul>
     * <li>the domain in which the container must be attached is not the expected one</li>
     * </ul>
     * </p>
     * <p>
     * Expected results: An error occurs about the inconsistency about domain name
     * </p>
     */
    @Test(expected = InconsistentDomainNameException.class)
    public void testTargetDomainNotFound() throws Exception {
        
        // Create a basic target topology for the remote container
        final Topology basicTargetTopology = TopologyServiceImpl_AttachContainerToTest.createBasicTargetTopology();
        final Properties targetServerProps = this.targetServerPropertiesBuilder
                .createMinimalServerProperties(TARGET_CONTAINER_NAME);
        targetServerProps.put(ConfigurationProperties.TOPOLOGY_PASSPHRASE, TARGET_PASS_PHRASE);
        final ConfigurationService targetCfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(targetCfgSvc);
        try {
            targetCfgSvc.loadConfiguration(targetServerProps, basicTargetTopology);
            final PetalsCompositeController targetPetalsCompositeCtrl = EasyMock
                    .createMock(PetalsCompositeController.class);
            EasyMock.expect(targetPetalsCompositeCtrl.getSharedAreaCltConfiguration(TARGET_PASS_PHRASE))
                    .andReturn(new Properties()).anyTimes();
            EasyMock.replay(targetPetalsCompositeCtrl);
            final TopologyService targetTopoSvc = new TopologyServiceImpl();
            ReflectionHelper.setPrivateField(targetTopoSvc, "configurationService", targetCfgSvc);
            ReflectionHelper.setPrivateField(targetTopoSvc, "petalsCompositeController", targetPetalsCompositeCtrl);
            this.startTopologyService(targetTopoSvc);
            try {
                // Create a basic sample topology for the current container
                final Topology basicCurrentTopology = ConfigurationTestUtils.createBasicSampleTopology();
                basicCurrentTopology.getDomain().setMode(DomainMode.DYNAMIC);
        
                // Create and initialize the configuration service
                final Properties currentServerProps = this.localServerPropertiesBuilder
                        .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);
                final ConfigurationService currentCfgSvc = new ConfigurationServiceImpl();
                ConfigurationTestUtils.startConfigurationService(currentCfgSvc);
                try {
                    currentCfgSvc.loadConfiguration(currentServerProps, basicCurrentTopology);
        
                    final TopologyService currentTopoSvc = createAndStartTopologySvc(currentCfgSvc, null, null);
        
                    try {
                        final ContainerConfiguration targetContainerConfiguration = new ContainerConfiguration();
                        targetContainerConfiguration.setName(TARGET_CONTAINER_NAME);
                        targetContainerConfiguration.setDomainName("inconsistent-domain-name");
                        targetContainerConfiguration.setHost(TARGET_JMX_HOST);
                        targetContainerConfiguration.setJmxRMIConnectorPort(TARGET_JMX_PORT);
                        
                        final TopologyServiceClient jmxTopologyServiceClientMock = EasyMock
                                .createMock(TopologyServiceClient.class);
                        EasyMock.expect(jmxTopologyServiceClientMock.retrieveTopology(TARGET_PASS_PHRASE, true))
                                .andReturn(targetTopoSvc.retrieveTopology(TARGET_PASS_PHRASE, true)).anyTimes();
                        EasyMock.expect(jmxTopologyServiceClientMock.retrieveSharedAreaCltImplementation())
                                .andReturn(DummySharedAreaImpl.class.getName()).anyTimes();
                        EasyMock.replay(jmxTopologyServiceClientMock);
                        this.jmxApiOnTargetContainer.setTopologyServiceClient(jmxTopologyServiceClientMock);
        
                        currentTopoSvc.attachContainerTo(targetContainerConfiguration, TARGET_PASS_PHRASE);
                    } finally {
                        this.stopTopologyService(currentTopoSvc);
                    }
                } finally {
                    ConfigurationTestUtils.stopConfigurationService(currentCfgSvc);
                }
            } finally {
                this.stopTopologyService(targetTopoSvc);
            }
        } finally {
            ConfigurationTestUtils.stopConfigurationService(targetCfgSvc);
        }
    }

    /**
     * <p>
     * Check {@link TopologyServiceImpl#attachContainerTo(ContainerConfiguration, String)} against:
     * <ul>
     * <li>the given container is not found in the target topology</li>
     * </ul>
     * </p>
     * <p>
     * Expected results: An error occurs about the inconsistency about the container name
     * </p>
     */
    @Test(expected = InconsistentContainerNameException.class)
    public void testTargetContainerNotFound() throws Exception {

        // Create a basic target topology for the remote container
        final Topology basicTargetTopology = TopologyServiceImpl_AttachContainerToTest.createBasicTargetTopology();
        final Properties targetServerProps = this.targetServerPropertiesBuilder
                .createMinimalServerProperties(TARGET_CONTAINER_NAME);
        targetServerProps.put(ConfigurationProperties.TOPOLOGY_PASSPHRASE, TARGET_PASS_PHRASE);
        final ConfigurationService targetCfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(targetCfgSvc);
        try {
            targetCfgSvc.loadConfiguration(targetServerProps, basicTargetTopology);
            final PetalsCompositeController targetPetalsCompositeCtrl = EasyMock
                    .createMock(PetalsCompositeController.class);
            EasyMock.expect(targetPetalsCompositeCtrl.getSharedAreaCltConfiguration(TARGET_PASS_PHRASE))
                    .andReturn(new Properties()).anyTimes();
            EasyMock.replay(targetPetalsCompositeCtrl);
            final TopologyService targetTopoSvc = new TopologyServiceImpl();
            ReflectionHelper.setPrivateField(targetTopoSvc, "configurationService", targetCfgSvc);
            ReflectionHelper.setPrivateField(targetTopoSvc, "petalsCompositeController", targetPetalsCompositeCtrl);
            this.startTopologyService(targetTopoSvc);
            try {
                // Create a basic sample topology for the current container
                final Topology basicCurrentTopology = ConfigurationTestUtils.createBasicSampleTopology();
                basicCurrentTopology.getDomain().setMode(DomainMode.DYNAMIC);

                // Create and initialize the configuration service
                final Properties currentServerProps = this.localServerPropertiesBuilder
                        .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);
                final ConfigurationService currentCfgSvc = new ConfigurationServiceImpl();
                ConfigurationTestUtils.startConfigurationService(currentCfgSvc);
                try {
                    currentCfgSvc.loadConfiguration(currentServerProps, basicCurrentTopology);

                    final TopologyService currentTopoSvc = createAndStartTopologySvc(currentCfgSvc, null, null);

                    try {
                        final ContainerConfiguration targetContainerConfiguration = new ContainerConfiguration();
                        targetContainerConfiguration.setName("target-container-name-not-exist");
                        targetContainerConfiguration.setDomainName(TARGET_DOMAIN_NAME);
                        targetContainerConfiguration.setHost(TARGET_JMX_HOST);
                        targetContainerConfiguration.setJmxRMIConnectorPort(TARGET_JMX_PORT);

                        final TopologyServiceClient jmxTopologyServiceClientMock = EasyMock
                                .createMock(TopologyServiceClient.class);
                        EasyMock.expect(jmxTopologyServiceClientMock.retrieveTopology(TARGET_PASS_PHRASE, true))
                                .andReturn(targetTopoSvc.retrieveTopology(TARGET_PASS_PHRASE, true)).anyTimes();
                        EasyMock.expect(jmxTopologyServiceClientMock.retrieveSharedAreaCltImplementation())
                                .andReturn(DummySharedAreaImpl.class.getName()).anyTimes();
                        EasyMock.replay(jmxTopologyServiceClientMock);
                        this.jmxApiOnTargetContainer.setTopologyServiceClient(jmxTopologyServiceClientMock);

                        currentTopoSvc.attachContainerTo(targetContainerConfiguration, TARGET_PASS_PHRASE);
                    } finally {
                        this.stopTopologyService(currentTopoSvc);
                    }
                } finally {
                    ConfigurationTestUtils.stopConfigurationService(currentCfgSvc);
                }
            } finally {
                this.stopTopologyService(targetTopoSvc);
            }
        } finally {
            ConfigurationTestUtils.stopConfigurationService(targetCfgSvc);
        }
    }

    /**
     * <p>
     * Check {@link TopologyServiceImpl#attachContainerTo(ContainerConfiguration, String)} against:
     * <ul>
     * <li>the domain of the remote container is static</li>
     * </ul>
     * </p>
     * <p>
     * Expected results: An error occurs because the remote domain mode is not <b>dynamic</b>
     * </p>
     */
    @Test(expected = DomainIsNotDynamicException.class)
    public void testTargetDomainNotDynamic() throws PetalsException, InvocationTargetException, IOException, InterruptedException {

        // Create a basic target topology for the remote container
        final Topology basicTargetTopology = TopologyServiceImpl_AttachContainerToTest.createBasicTargetTopology();
        basicTargetTopology.getDomain().setMode(DomainMode.STATIC);
        final Properties targetServerProps = this.targetServerPropertiesBuilder
                .createMinimalServerProperties(TARGET_CONTAINER_NAME);
        targetServerProps.put(ConfigurationProperties.TOPOLOGY_PASSPHRASE, TARGET_PASS_PHRASE);
        final ConfigurationService targetCfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(targetCfgSvc);
        try {
            targetCfgSvc.loadConfiguration(targetServerProps, basicTargetTopology);
            final PetalsCompositeController targetPetalsCompositeCtrl = EasyMock
                    .createMock(PetalsCompositeController.class);
            EasyMock.expect(targetPetalsCompositeCtrl.getSharedAreaCltConfiguration(TARGET_PASS_PHRASE))
                    .andReturn(new Properties()).anyTimes();
            EasyMock.replay(targetPetalsCompositeCtrl);
            final TopologyService targetTopoSvc = new TopologyServiceImpl();
            ReflectionHelper.setPrivateField(targetTopoSvc, "configurationService", targetCfgSvc);
            ReflectionHelper.setPrivateField(targetTopoSvc, "petalsCompositeController", targetPetalsCompositeCtrl);
            this.startTopologyService(targetTopoSvc);
            try {
                // Create a basic sample topology for the current container
                final Topology basicCurrentTopology = ConfigurationTestUtils.createBasicSampleTopology();
                basicCurrentTopology.getDomain().setMode(DomainMode.DYNAMIC);

                // Create and initialize the configuration service
                final Properties currentServerProps = this.localServerPropertiesBuilder
                        .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);
                final ConfigurationService currentCfgSvc = new ConfigurationServiceImpl();
                ConfigurationTestUtils.startConfigurationService(currentCfgSvc);
                try {
                    currentCfgSvc.loadConfiguration(currentServerProps, basicCurrentTopology);

                    final TopologyService currentTopoSvc = createAndStartTopologySvc(currentCfgSvc, null, null);

                    try {
                        final ContainerConfiguration targetContainerConfiguration = new ContainerConfiguration();
                        targetContainerConfiguration.setName(TARGET_CONTAINER_NAME);
                        targetContainerConfiguration.setDomainName(TARGET_DOMAIN_NAME);
                        targetContainerConfiguration.setHost(TARGET_JMX_HOST);
                        targetContainerConfiguration.setJmxRMIConnectorPort(TARGET_JMX_PORT);

                        final TopologyServiceClient jmxTopologyServiceClientMock = EasyMock
                                .createMock(TopologyServiceClient.class);
                        EasyMock.expect(jmxTopologyServiceClientMock.retrieveTopology(TARGET_PASS_PHRASE, true))
                                .andReturn(targetTopoSvc.retrieveTopology(TARGET_PASS_PHRASE, true)).anyTimes();
                        EasyMock.expect(jmxTopologyServiceClientMock.retrieveSharedAreaCltImplementation())
                                .andReturn(DummySharedAreaImpl.class.getName()).anyTimes();
                        EasyMock.replay(jmxTopologyServiceClientMock);
                        this.jmxApiOnTargetContainer.setTopologyServiceClient(jmxTopologyServiceClientMock);

                        currentTopoSvc.attachContainerTo(targetContainerConfiguration, TARGET_PASS_PHRASE);
                    } finally {
                        this.stopTopologyService(currentTopoSvc);
                    }
                } finally {
                    ConfigurationTestUtils.stopConfigurationService(currentCfgSvc);
                }
            } finally {
                this.stopTopologyService(targetTopoSvc);
            }
        } finally {
            ConfigurationTestUtils.stopConfigurationService(targetCfgSvc);
        }
    }

    /**
     * <p>
     * Check {@link TopologyServiceImpl#attachContainerTo(ContainerConfiguration, String)} against:
     * <ul>
     * <li>the registry implementation of the current domain is different from the one of the target domain: no registry
     * implementation defined for the current domain, an explicit registry implementation is set to the target domain
     * </li>
     * </ul>
     * </p>
     * <p>
     * Expected results: An error occurs because the registry implementations are inconsistent
     * </p>
     */
    @Test(expected = InconsistentDomainRegistryImplException.class)
    public void testRegistryImplInconsistent_NoImplOnCur_AnImpOnTarget() throws Exception {

        // Create a basic target topology for the remote container
        final Topology basicTargetTopology = TopologyServiceImpl_AttachContainerToTest.createBasicTargetTopology();
        final Domain targetDomain = basicTargetTopology.getDomain();
        targetDomain.setMode(DomainMode.DYNAMIC);
        targetDomain.setRegistryImplementation("a.registry.implementation");
        final Properties targetServerProps = this.targetServerPropertiesBuilder
                .createMinimalServerProperties(TARGET_CONTAINER_NAME);
        targetServerProps.put(ConfigurationProperties.TOPOLOGY_PASSPHRASE, TARGET_PASS_PHRASE);
        final ConfigurationService targetCfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(targetCfgSvc);
        try {
            targetCfgSvc.loadConfiguration(targetServerProps, basicTargetTopology);
            final PetalsCompositeController targetPetalsCompositeCtrl = EasyMock
                    .createMock(PetalsCompositeController.class);
            EasyMock.expect(targetPetalsCompositeCtrl.getSharedAreaCltConfiguration(TARGET_PASS_PHRASE))
                    .andReturn(new Properties()).anyTimes();
            EasyMock.replay(targetPetalsCompositeCtrl);
            final TopologyService targetTopoSvc = new TopologyServiceImpl();
            ReflectionHelper.setPrivateField(targetTopoSvc, "configurationService", targetCfgSvc);
            ReflectionHelper.setPrivateField(targetTopoSvc, "petalsCompositeController", targetPetalsCompositeCtrl);
            this.startTopologyService(targetTopoSvc);
            try {
                // Create a basic sample topology for the current container
                final Topology basicCurrentTopology = ConfigurationTestUtils.createBasicSampleTopology();
                basicCurrentTopology.getDomain().setMode(DomainMode.DYNAMIC);

                // Create and initialize the configuration service
                final Properties currentServerProps = this.localServerPropertiesBuilder
                        .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);
                final ConfigurationService currentCfgSvc = new ConfigurationServiceImpl();
                ConfigurationTestUtils.startConfigurationService(currentCfgSvc);
                try {
                    currentCfgSvc.loadConfiguration(currentServerProps, basicCurrentTopology);

                    final TopologyService currentTopoSvc = createAndStartTopologySvc(currentCfgSvc, null, null);

                    try {
                        final ContainerConfiguration targetContainerConfiguration = new ContainerConfiguration();
                        targetContainerConfiguration.setName(TARGET_CONTAINER_NAME);
                        targetContainerConfiguration.setDomainName(TARGET_DOMAIN_NAME);
                        targetContainerConfiguration.setHost(TARGET_JMX_HOST);
                        targetContainerConfiguration.setJmxRMIConnectorPort(TARGET_JMX_PORT);

                        final TopologyServiceClient jmxTopologyServiceClientMock = EasyMock
                                .createMock(TopologyServiceClient.class);
                        EasyMock.expect(jmxTopologyServiceClientMock.retrieveTopology(TARGET_PASS_PHRASE, true))
                                .andReturn(targetTopoSvc.retrieveTopology(TARGET_PASS_PHRASE, true)).anyTimes();
                        EasyMock.expect(jmxTopologyServiceClientMock.retrieveSharedAreaCltImplementation())
                                .andReturn(DummySharedAreaImpl.class.getName()).anyTimes();
                        EasyMock.replay(jmxTopologyServiceClientMock);
                        this.jmxApiOnTargetContainer.setTopologyServiceClient(jmxTopologyServiceClientMock);

                        currentTopoSvc.attachContainerTo(targetContainerConfiguration, TARGET_PASS_PHRASE);
                    } finally {
                        this.stopTopologyService(currentTopoSvc);
                    }
                } finally {
                    ConfigurationTestUtils.stopConfigurationService(currentCfgSvc);
                }
            } finally {
                this.stopTopologyService(targetTopoSvc);
            }
        } finally {
            ConfigurationTestUtils.stopConfigurationService(targetCfgSvc);
        }
    }

    /**
     * <p>
     * Check {@link TopologyServiceImpl#attachContainerTo(ContainerConfiguration, String)} against:
     * <ul>
     * <li>the registry implementation of the current domain is different from the one of the target domain: an explicit
     * registry implementation is set to the current domain, no registry implementation defined for the target domain
     * </li>
     * </ul>
     * </p>
     * <p>
     * Expected results: An error occurs because the registry implementations are inconsistent
     * </p>
     */
    @Test(expected = InconsistentDomainRegistryImplException.class)
    public void testRegistryImplInconsistent_AnImplOnCur_NoImpOnTarget() throws Exception {

        // Create a basic target topology for the remote container
        final Topology basicTargetTopology = TopologyServiceImpl_AttachContainerToTest.createBasicTargetTopology();
        basicTargetTopology.getDomain().setMode(DomainMode.DYNAMIC);
        final Properties targetServerProps = this.targetServerPropertiesBuilder
                .createMinimalServerProperties(TARGET_CONTAINER_NAME);
        targetServerProps.put(ConfigurationProperties.TOPOLOGY_PASSPHRASE, TARGET_PASS_PHRASE);
        final ConfigurationService targetCfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(targetCfgSvc);
        try {
            targetCfgSvc.loadConfiguration(targetServerProps, basicTargetTopology);
            final PetalsCompositeController targetPetalsCompositeCtrl = EasyMock
                    .createMock(PetalsCompositeController.class);
            EasyMock.expect(targetPetalsCompositeCtrl.getSharedAreaCltConfiguration(TARGET_PASS_PHRASE))
                    .andReturn(new Properties()).anyTimes();
            EasyMock.replay(targetPetalsCompositeCtrl);
            final TopologyService targetTopoSvc = new TopologyServiceImpl();
            ReflectionHelper.setPrivateField(targetTopoSvc, "configurationService", targetCfgSvc);
            ReflectionHelper.setPrivateField(targetTopoSvc, "petalsCompositeController", targetPetalsCompositeCtrl);
            this.startTopologyService(targetTopoSvc);
            try {
                // Create a basic sample topology for the current container
                final Topology basicCurrentTopology = ConfigurationTestUtils.createBasicSampleTopology();
                final Domain currentDomain = basicCurrentTopology.getDomain();
                currentDomain.setMode(DomainMode.DYNAMIC);
                currentDomain.setRegistryImplementation("a.registry.implementation");

                // Create and initialize the configuration service
                final Properties currentServerProps = this.localServerPropertiesBuilder
                        .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);
                final ConfigurationService currentCfgSvc = new ConfigurationServiceImpl();
                ConfigurationTestUtils.startConfigurationService(currentCfgSvc);
                try {
                    currentCfgSvc.loadConfiguration(currentServerProps, basicCurrentTopology);

                    final TopologyService currentTopoSvc = createAndStartTopologySvc(currentCfgSvc, null, null);

                    try {
                        final ContainerConfiguration targetContainerConfiguration = new ContainerConfiguration();
                        targetContainerConfiguration.setName(TARGET_CONTAINER_NAME);
                        targetContainerConfiguration.setDomainName(TARGET_DOMAIN_NAME);
                        targetContainerConfiguration.setHost(TARGET_JMX_HOST);
                        targetContainerConfiguration.setJmxRMIConnectorPort(TARGET_JMX_PORT);

                        final TopologyServiceClient jmxTopologyServiceClientMock = EasyMock
                                .createMock(TopologyServiceClient.class);
                        EasyMock.expect(jmxTopologyServiceClientMock.retrieveTopology(TARGET_PASS_PHRASE, true))
                                .andReturn(targetTopoSvc.retrieveTopology(TARGET_PASS_PHRASE, true)).anyTimes();
                        EasyMock.expect(jmxTopologyServiceClientMock.retrieveSharedAreaCltImplementation())
                                .andReturn(DummySharedAreaImpl.class.getName()).anyTimes();

                        EasyMock.replay(jmxTopologyServiceClientMock);
                        this.jmxApiOnTargetContainer.setTopologyServiceClient(jmxTopologyServiceClientMock);

                        currentTopoSvc.attachContainerTo(targetContainerConfiguration, TARGET_PASS_PHRASE);
                    } finally {
                        this.stopTopologyService(currentTopoSvc);
                    }
                } finally {
                    ConfigurationTestUtils.stopConfigurationService(currentCfgSvc);
                }
            } finally {
                this.stopTopologyService(targetTopoSvc);
            }
        } finally {
            ConfigurationTestUtils.stopConfigurationService(targetCfgSvc);
        }
    }

    /**
     * <p>
     * Check {@link TopologyServiceImpl#attachContainerTo(ContainerConfiguration, String)} against:
     * <ul>
     * <li>the registry implementation of the current domain is different from the one of the target domain: an explicit
     * registry implementation is set to the current domain, another explicit registry implementation is set to the
     * target domain</li>
     * </ul>
     * </p>
     * <p>
     * Expected results: An error occurs because the registry implementations are inconsistent
     * </p>
     */
    @Test(expected = InconsistentDomainRegistryImplException.class)
    public void testRegistryImplInconsistent_DifferentImplSet() throws Exception {

        // Create a basic target topology for the remote container
        final Topology basicTargetTopology = TopologyServiceImpl_AttachContainerToTest.createBasicTargetTopology();
        final Domain targetDomain = basicTargetTopology.getDomain();
        targetDomain.setMode(DomainMode.DYNAMIC);
        targetDomain.setRegistryImplementation("another.registry.implementation");
        final Properties targetServerProps = this.targetServerPropertiesBuilder
                .createMinimalServerProperties(TARGET_CONTAINER_NAME);
        targetServerProps.put(ConfigurationProperties.TOPOLOGY_PASSPHRASE, TARGET_PASS_PHRASE);
        final ConfigurationService targetCfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(targetCfgSvc);
        try {
            targetCfgSvc.loadConfiguration(targetServerProps, basicTargetTopology);
            final PetalsCompositeController targetPetalsCompositeCtrl = EasyMock
                    .createMock(PetalsCompositeController.class);
            EasyMock.expect(targetPetalsCompositeCtrl.getSharedAreaCltConfiguration(TARGET_PASS_PHRASE))
                    .andReturn(new Properties()).anyTimes();
            EasyMock.replay(targetPetalsCompositeCtrl);
            final TopologyService targetTopoSvc = new TopologyServiceImpl();
            ReflectionHelper.setPrivateField(targetTopoSvc, "configurationService", targetCfgSvc);
            ReflectionHelper.setPrivateField(targetTopoSvc, "petalsCompositeController", targetPetalsCompositeCtrl);
            this.startTopologyService(targetTopoSvc);
            try {
                // Create a basic sample topology for the current container
                final Topology basicCurrentTopology = ConfigurationTestUtils.createBasicSampleTopology();
                final Domain currentDomain = basicCurrentTopology.getDomain();
                currentDomain.setMode(DomainMode.DYNAMIC);
                currentDomain.setRegistryImplementation("a.registry.implementation");

                // Create and initialize the configuration service
                final Properties currentServerProps = this.localServerPropertiesBuilder
                        .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);
                final ConfigurationService currentCfgSvc = new ConfigurationServiceImpl();
                ConfigurationTestUtils.startConfigurationService(currentCfgSvc);
                try {
                    currentCfgSvc.loadConfiguration(currentServerProps, basicCurrentTopology);

                    final TopologyService currentTopoSvc = createAndStartTopologySvc(currentCfgSvc, null, null);

                    try {
                        final ContainerConfiguration targetContainerConfiguration = new ContainerConfiguration();
                        targetContainerConfiguration.setName(TARGET_CONTAINER_NAME);
                        targetContainerConfiguration.setDomainName(TARGET_DOMAIN_NAME);
                        targetContainerConfiguration.setHost(TARGET_JMX_HOST);
                        targetContainerConfiguration.setJmxRMIConnectorPort(TARGET_JMX_PORT);

                        final TopologyServiceClient jmxTopologyServiceClientMock = EasyMock
                                .createMock(TopologyServiceClient.class);
                        EasyMock.expect(jmxTopologyServiceClientMock.retrieveTopology(TARGET_PASS_PHRASE, true))
                                .andReturn(targetTopoSvc.retrieveTopology(TARGET_PASS_PHRASE, true)).anyTimes();
                        EasyMock.expect(jmxTopologyServiceClientMock.retrieveSharedAreaCltImplementation())
                                .andReturn(DummySharedAreaImpl.class.getName()).anyTimes();
                        EasyMock.replay(jmxTopologyServiceClientMock);
                        this.jmxApiOnTargetContainer.setTopologyServiceClient(jmxTopologyServiceClientMock);

                        currentTopoSvc.attachContainerTo(targetContainerConfiguration, TARGET_PASS_PHRASE);
                    } finally {
                        this.stopTopologyService(currentTopoSvc);
                    }
                } finally {
                    ConfigurationTestUtils.stopConfigurationService(currentCfgSvc);
                }
            } finally {
                this.stopTopologyService(targetTopoSvc);
            }
        } finally {
            ConfigurationTestUtils.stopConfigurationService(targetCfgSvc);
        }
    }

    /**
     * <p>
     * Check {@link TopologyServiceImpl#attachContainerTo(ContainerConfiguration, String)} against:
     * <ul>
     * <li>all is ok</li>
     * </ul>
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>No error occurs</li>
     * <li>...
     * <li>
     * </ul>
     * </p>
     */
    @Test
    public void testAllIsOk() throws Exception {

        // Create a basic target topology for the remote container
        final Topology basicTargetTopology = TopologyServiceImpl_AttachContainerToTest.createBasicTargetTopology();
        final Domain targetDomain = basicTargetTopology.getDomain();
        targetDomain.setMode(DomainMode.DYNAMIC);
        final Properties targetServerProps = this.targetServerPropertiesBuilder
                .createMinimalServerProperties(TARGET_CONTAINER_NAME);
        targetServerProps.put(ConfigurationProperties.TOPOLOGY_PASSPHRASE, TARGET_PASS_PHRASE);
        final ConfigurationService targetCfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(targetCfgSvc);
        try {
            targetCfgSvc.loadConfiguration(targetServerProps, basicTargetTopology);
            final PetalsCompositeController targetPetalsCompositeCtrl = EasyMock
                    .createMock(PetalsCompositeController.class);
            EasyMock.expect(targetPetalsCompositeCtrl.getSharedAreaCltConfiguration(TARGET_PASS_PHRASE))
                    .andReturn(new Properties()).anyTimes();
            EasyMock.replay(targetPetalsCompositeCtrl);
            final TopologyService targetTopoSvc = new TopologyServiceImpl();
            ReflectionHelper.setPrivateField(targetTopoSvc, "configurationService", targetCfgSvc);
            ReflectionHelper.setPrivateField(targetTopoSvc, "petalsCompositeController", targetPetalsCompositeCtrl);
            this.startTopologyService(targetTopoSvc);
            try {
                // Create a basic sample topology for the current container
                final Topology basicCurrentTopology = ConfigurationTestUtils.createBasicSampleTopology();
                final Domain currentDomain = basicCurrentTopology.getDomain();
                currentDomain.setMode(DomainMode.DYNAMIC);

                // Create and initialize the configuration service
                final Properties currentServerProps = this.localServerPropertiesBuilder
                        .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);
                final ConfigurationService currentCfgSvc = new ConfigurationServiceImpl();
                ConfigurationTestUtils.startConfigurationService(currentCfgSvc);
                try {
                    currentCfgSvc.loadConfiguration(currentServerProps, basicCurrentTopology);

                    final PetalsCompositeController petalsCompositeCtrlMock = EasyMock
                            .createMock(PetalsCompositeController.class);
                    final AutoCloseable msPause = EasyMock.createMock(AutoCloseable.class);
                    msPause.close();
                    final AutoCloseable edPause = EasyMock.createMock(AutoCloseable.class);
                    edPause.close();
                    final AutoCloseable saPause = EasyMock.createMock(AutoCloseable.class);
                    saPause.close();
                    EasyMock.expect(petalsCompositeCtrlMock.pauseMessagingSubsystem()).andReturn(msPause);
                    EasyMock.expect(petalsCompositeCtrlMock.pauseEndpointDirectory()).andReturn(edPause);
                    EasyMock.expect(petalsCompositeCtrlMock.pauseSharedAreaComponent()).andReturn(saPause);
                    petalsCompositeCtrlMock.onConfigurationUpdated();
                    EasyMock.replay(petalsCompositeCtrlMock, msPause, edPause, saPause);

                    final TopologyService currentTopoSvc = createAndStartTopologySvc(currentCfgSvc,
                            basicTargetTopology, petalsCompositeCtrlMock);

                    try {
                        final ContainerConfiguration targetContainerConfiguration = new ContainerConfiguration();
                        targetContainerConfiguration.setName(TARGET_CONTAINER_NAME);
                        targetContainerConfiguration.setDomainName(TARGET_DOMAIN_NAME);
                        targetContainerConfiguration.setHost(TARGET_JMX_HOST);
                        targetContainerConfiguration.setJmxRMIConnectorPort(TARGET_JMX_PORT);

                        final TopologyServiceClient jmxTopologyServiceClientMock = EasyMock
                                .createMock(TopologyServiceClient.class);
                        EasyMock.expect(jmxTopologyServiceClientMock.retrieveTopology(TARGET_PASS_PHRASE, true))
                                .andReturn(targetTopoSvc.retrieveTopology(TARGET_PASS_PHRASE, true)).anyTimes();
                        EasyMock.expect(jmxTopologyServiceClientMock.retrieveSharedAreaCltImplementation())
                                .andReturn(DummySharedAreaImpl.class.getName()).anyTimes();
                        EasyMock.replay(jmxTopologyServiceClientMock);
                        this.jmxApiOnTargetContainer.setTopologyServiceClient(jmxTopologyServiceClientMock);

                        currentTopoSvc.attachContainerTo(targetContainerConfiguration, TARGET_PASS_PHRASE);
                    } finally {
                        this.stopTopologyService(currentTopoSvc);
                    }
                } finally {
                    ConfigurationTestUtils.stopConfigurationService(currentCfgSvc);
                }
            } finally {
                this.stopTopologyService(targetTopoSvc);
            }
        } finally {
            ConfigurationTestUtils.stopConfigurationService(targetCfgSvc);
        }
    }

    /**
     * Create a sample topology with only required parameters
     * 
     * @return A sample topology with only required parameters
     */
    private static Topology createBasicTargetTopology() {
        final Topology topology = new Topology();
        final Domain domain = new Domain();
        topology.setDomain(domain);

        domain.setName(TARGET_DOMAIN_NAME);
        domain.setMode(DomainMode.DYNAMIC);

        final Container container0 = new Container();
        domain.getContainer().add(container0);
        container0.setName(TARGET_CONTAINER_NAME);
        container0.setHost(TARGET_JMX_HOST);
        container0.setUser(TARGET_JMX_USER);
        container0.setPassword(TARGET_JMX_PWD);

        final JmxService jmxService0 = new JmxService();
        container0.setJmxService(jmxService0);

        final TransportService transportService0 = new TransportService();
        transportService0.setTcpPort(TARGET_CONTAINER_TCP_PORT);
        container0.setTransportService(transportService0);

        return topology;
    }

}
