/**
 * Copyright (c) 2014-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.communication.topology;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

import org.easymock.classextension.EasyMock;
import org.junit.Rule;
import org.junit.Test;
import org.ow2.petals.clientserverapi.topology.TopologyService.ContainerPropertyName;
import org.ow2.petals.clientserverapi.topology.TopologyService.DomainPropertyName;
import org.ow2.petals.clientserverapi.topology.TopologyService.DomainPropertyValues;
import org.ow2.petals.clientserverapi.topology.TopologyService.TopologyItem;
import org.ow2.petals.clientserverapi.topology.exception.InvalidPassphraseException;
import org.ow2.petals.launcher.api.server.conf.ConfigurationProperties;
import org.ow2.petals.microkernel.api.communication.topology.TopologyService;
import org.ow2.petals.microkernel.api.configuration.ConfigurationService;
import org.ow2.petals.microkernel.configuration.ConfigurationServiceImpl;
import org.ow2.petals.microkernel.configuration.ConfigurationTestUtils;
import org.ow2.petals.microkernel.configuration.LocalServerPropertiesBuilder;
import org.ow2.petals.microkernel.server.PetalsCompositeController;
import org.ow2.petals.topology.TopologyHelper;
import org.ow2.petals.topology.generated.Container;
import org.ow2.petals.topology.generated.Domain;
import org.ow2.petals.topology.generated.DomainMode;
import org.ow2.petals.topology.generated.JmxService;
import org.ow2.petals.topology.generated.Jndi;
import org.ow2.petals.topology.generated.Topology;
import org.ow2.petals.topology.generated.TransportService;

import com.ebmwebsourcing.easycommons.lang.reflect.ReflectionHelper;

/**
 * Unit tests of {@link TopologyServiceImpl#retrieveTopology(String)}
 * 
 * @author Christophe DENEUX - Linagora
 * 
 */
public class TopologyServiceImpl_RetrieveTopologyTest extends AbstractTopologyServiceImplTest {

    private final static String DOMAIN_JNDI_FACTORY = "domain.jndi.factory";

    private final static String DOMAIN_JNDI_PROVIDER_URL = "http://domain.jndi.provider.url";

    private final static String DOMAIN_JNDI_SECURITY_PRINCIPAL = "domain.jndi.security.principal";

    private final static String DOMAIN_JNDI_SECURITY_CREDENTIALS = "domain.jndi.security.credentials";

    private final static int DOMAIN_JNDI_POOL_SIZE = 10;

    private final static int DOMAIN_JNDI_BATCH_SIZE = 5;

    private final static String DOMAIN_REGISTRY_IMPLEMENTATION = "registry-implementation";

    private final static int CONTAINER_RMI_PORT_0_0 = 7700;

    private final static String CONTAINER_NAME_0_1 = "sample-0-1";

    private final static String CONTAINER_DESCRIPTION_0_1 = "description of the container 0-1";

    private final static int CONTAINER_RMI_PORT_0_1 = CONTAINER_RMI_PORT_0_0 + 1;

    private final static int CONTAINER_TCP_PORT_0_1 = ConfigurationTestUtils.CONTAINER_TCP_PORT_0 + 1;

    private final static String CONTAINER_NAME_1_0 = "sample-1-0";

    private final static String CONTAINER_DESCRIPTION_1_0 = "description of the container 1-0";

    private final static int CONTAINER_RMI_PORT_1_0 = CONTAINER_RMI_PORT_0_0 + 2;

    private final static int CONTAINER_TCP_PORT_1_0 = ConfigurationTestUtils.CONTAINER_TCP_PORT_0 + 2;

    private final static String CONTAINER_NAME_1_1 = "sample-1-1";

    private final static String CONTAINER_DESCRIPTION_1_1 = "description of the container 1-1";

    private final static int CONTAINER_RMI_PORT_1_1 = CONTAINER_RMI_PORT_0_0 + 3;

    private final static int CONTAINER_TCP_PORT_1_1 = ConfigurationTestUtils.CONTAINER_TCP_PORT_0 + 3;

    private final static String TOPOLOGY_PASSPHRASE = "a-pass-phrase";

    private final static String TOPOLOGY_INVALID_PASSPHRASE = "an-invalid-pass-phrase";

    @Rule
    public final LocalServerPropertiesBuilder localServerPropertiesBuilder = new LocalServerPropertiesBuilder();

    /**
     * <p>
     * Check the method {@link TopologyServiceImpl#retrieveTopology(String, boolean)} against a basic sample topology
     * definition where the sensible information are expected but the given passphrase is invalid
     * </p>
     * <p>
     * Expected result: the exception {@link InvalidPassphraseException} is thrown
     * </p>
     */
    @Test(expected = InvalidPassphraseException.class)
    public void testEXpectedSensibleInformationInvalidPassphrase() throws Exception {

        // Create a basic sample topology
        final Topology basicSampleTopology = ConfigurationTestUtils.createBasicSampleTopology();

        // Create and initialize the configuration service
        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);
        serverProps.put(ConfigurationProperties.TOPOLOGY_PASSPHRASE, TOPOLOGY_PASSPHRASE);
        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);
        try {
            cfgSvc.loadConfiguration(serverProps, basicSampleTopology);

            final TopologyService topoSvc = new TopologyServiceImpl();
            ReflectionHelper.setPrivateField(topoSvc, "configurationService", cfgSvc);
            this.startTopologyService(topoSvc);

            try {
                topoSvc.retrieveTopology(TOPOLOGY_INVALID_PASSPHRASE, true);
            } finally {
                this.stopTopologyService(topoSvc);
            }
        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check the method {@link TopologyServiceImpl#retrieveTopology(String)} against a basic sample topology definition.
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>no error,</li>
     * <li>all set parameters are returned with the right value, unset parameters are returned as empty or set to their
     * default value,</li>
     * <li>registry mutable implementation configuration is returned in the domain map</li>
     * </ul>
     * </p>
     */
    @Test
    public void testBasicSampleTopology() throws Exception {

        // Create a basic sample topology
        final Topology basicSampleTopology = ConfigurationTestUtils.createBasicSampleTopology();

        // Create and initialize the configuration service
        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);
        serverProps.put(ConfigurationProperties.TOPOLOGY_PASSPHRASE, TOPOLOGY_PASSPHRASE);
        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);
        try {
            cfgSvc.loadConfiguration(serverProps, basicSampleTopology);
            
            final Properties registryProperties = new Properties();
            final String registryPropertyName = "registry-property-name";
            final String registryPropertyValue = "registry-property-value";
            registryProperties.setProperty(registryPropertyName, registryPropertyValue);
            
            final PetalsCompositeController petalsCompositeCtrl = EasyMock.createMock(PetalsCompositeController.class);
            EasyMock.expect(petalsCompositeCtrl.getSharedAreaCltConfiguration(TOPOLOGY_PASSPHRASE))
                    .andReturn(registryProperties).anyTimes();
            EasyMock.replay(petalsCompositeCtrl);

            final TopologyService topoSvc = new TopologyServiceImpl();
            ReflectionHelper.setPrivateField(topoSvc, "configurationService", cfgSvc);
            ReflectionHelper.setPrivateField(topoSvc, "petalsCompositeController", petalsCompositeCtrl);
            this.startTopologyService(topoSvc);

            try {
                int domainMapFound = 0;
                int containerMapFound = 0;

                final Set<Map<String, String>> retrievedTopology = topoSvc.retrieveTopology(TOPOLOGY_PASSPHRASE, false);
                for (final Map<String, String> topoMap : retrievedTopology) {
                    if (topoMap.get(TopologyItem.TOPOLOGY_TYPE).equals(TopologyService.TopologyItem.DOMAIN)) {

                        // Remove the type property that is not a configuration property
                        topoMap.remove(TopologyItem.TOPOLOGY_TYPE);

                        assertEquals("Unexpected domain name", ConfigurationTestUtils.DOMAIN_NAME,
                                topoMap.remove(DomainPropertyName.CONF_DOMAIN_NAME));
                        assertEquals("Unexpected domain description", ConfigurationTestUtils.DOMAIN_DESCRIPTION,
                                topoMap.remove(DomainPropertyName.CONF_DOMAIN_DESCRIPTION));
                        assertTrue("Unexpected domain jndi factory",
                                topoMap.remove(DomainPropertyName.CONF_JNDI_FACTORY).isEmpty());
                        assertTrue("Unexpected domain jndi provider url",
                                topoMap.remove(DomainPropertyName.CONF_JNDI_PROVIDER_URL).isEmpty());
                        assertTrue("Unexpected domain jndi security principal",
                                topoMap.remove(DomainPropertyName.CONF_JNDI_SECURITY_PRINCIPAL).isEmpty());
                        assertTrue("Unexpected domain jndi security credentials",
                                topoMap.remove(DomainPropertyName.CONF_JNDI_SECURITY_CREDENTIALS).isEmpty());
                        assertTrue("Unexpected domain jndi pool size",
                                topoMap.remove(DomainPropertyName.CONF_JNDI_POOL_SIZE).isEmpty());
                        assertTrue("Unexpected domain batch pool size",
                                topoMap.remove(DomainPropertyName.CONF_JNDI_BATCH_SIZE).isEmpty());
                        assertEquals("Unexpected domain mode", DomainPropertyValues.MODE_STATIC,
                                topoMap.remove(DomainPropertyName.CONF_DOMAIN_MODE));
                        assertTrue("Unexpected domain registry implementation",
                                topoMap.remove(DomainPropertyName.CONF_SHARED_AREA_IMPLEMENTATION).isEmpty());
                        assertEquals("Unexpected domain registry mutable implementation configuration property",
                                registryPropertyValue, topoMap.remove(registryPropertyName));

                        assertEquals("Unexpected property remains in domain map: " + topoMap, 0, topoMap.size());

                        domainMapFound++;

                    } else if (topoMap.get(TopologyItem.TOPOLOGY_TYPE).equals(TopologyService.TopologyItem.CONTAINER)) {
                        // Remove the type property that is not a configuration property
                        topoMap.remove(TopologyItem.TOPOLOGY_TYPE);

                        // Check all container properties
                        assertEquals("Unexpected container name", ConfigurationTestUtils.CONTAINER_NAME_0,
                                topoMap.remove(ContainerPropertyName.CONF_CONTAINER_NAME));
                        assertEquals("Unexpected container description",
                                ConfigurationTestUtils.CONTAINER_DESCRIPTION_0,
                                topoMap.remove(ContainerPropertyName.CONF_CONTAINER_DESCRIPTION));
                        assertEquals("Unexpected container host", ConfigurationTestUtils.CONTAINER_HOST,
                                topoMap.remove(ContainerPropertyName.CONF_HOST));
                        assertEquals("Unexpected container jmx user", ConfigurationTestUtils.CONTAINER_JMX_USER,
                                topoMap.remove(ContainerPropertyName.CONF_USER));
                        assertEquals("Unexpected container jmx password", ConfigurationTestUtils.CONTAINER_JMX_PWD,
                                topoMap.remove(ContainerPropertyName.CONF_PWD));
                        assertTrue("Unexpected container jmx rmi port",
                                topoMap.remove(ContainerPropertyName.CONF_JMX_RMI_PORT).isEmpty());
                        assertEquals("Unexpected container tcp port",
                                Integer.toString(ConfigurationTestUtils.CONTAINER_TCP_PORT_0),
                                topoMap.remove(ContainerPropertyName.CONF_TRANSPORT_TCP_PORT));

                        assertNotNull("Container state is missing", topoMap.remove(ContainerPropertyName.CONF_STATE));
                        assertNotNull("Domain is missing", topoMap.remove(ContainerPropertyName.CONF_DOMAIN_NAME));

                        assertEquals("Unexpected property remains in container map: " + topoMap, 0, topoMap.size());

                        containerMapFound++;

                    } else {
                        fail("Unknown topology item type");
                    }
                }

                assertEquals("unexpected or missing domain map", domainMapFound, 1);
                assertEquals("unexpected or missing container found", containerMapFound, 1);
            } finally {
                this.stopTopologyService(topoSvc);
            }
        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check that the method {@link TopologyServiceImpl#retrieveTopology(String, boolean)} returns all possible
     * parameters of a topology definition when the given pass-phrase is the expected one.
     * </p>
     * <p>
     * Expected result: No error, all possible parameters are returned, including sensible information
     * </p>
     */
    @Test
    public void testAllTopologyParametersReturnedIncludingSensibleInformation() throws Exception {
        
        // Create a complete topology
        final Topology completeTopology = this.createSampleCompleteTopology();
        assertAllFieldsSet(completeTopology);
        
        // Create and initialize the configuration service
        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);
        serverProps.put(ConfigurationProperties.TOPOLOGY_PASSPHRASE, TOPOLOGY_PASSPHRASE);
        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);
        try {
            cfgSvc.loadConfiguration(serverProps, completeTopology);

            final PetalsCompositeController petalsCompositeCtrl = EasyMock.createMock(PetalsCompositeController.class);
            EasyMock.expect(petalsCompositeCtrl.getSharedAreaCltConfiguration(TOPOLOGY_PASSPHRASE))
                    .andReturn(new Properties()).anyTimes();
            EasyMock.replay(petalsCompositeCtrl);

            final TopologyService topoSvc = new TopologyServiceImpl();
            ReflectionHelper.setPrivateField(topoSvc, "configurationService", cfgSvc);
            ReflectionHelper.setPrivateField(topoSvc, "petalsCompositeController", petalsCompositeCtrl);
            this.startTopologyService(topoSvc);

            try {
                int domainMapFound = 0;
                int containerMapFound = 0;

                final Set<Map<String, String>> retrievedTopology = topoSvc.retrieveTopology(TOPOLOGY_PASSPHRASE, false);
                for (final Map<String, String> topoMap : retrievedTopology) {
                    if (topoMap.get(TopologyItem.TOPOLOGY_TYPE).equals(
                            TopologyService.TopologyItem.DOMAIN)) {

                        // Remove the type property that is not a configuration property
                        topoMap.remove(TopologyItem.TOPOLOGY_TYPE);

                        // Check all domain properties
                        final int expectPropertyNumber = this.getExpectedPropertyNumber(completeTopology.getDomain());
                        assertEquals("Unexpected property number in domain map", expectPropertyNumber, topoMap.size());
                        assertEquals("Unexpected domain name", ConfigurationTestUtils.DOMAIN_NAME,
                                topoMap.remove(DomainPropertyName.CONF_DOMAIN_NAME));
                        assertEquals("Unexpected domain description", ConfigurationTestUtils.DOMAIN_DESCRIPTION,
                                topoMap.remove(DomainPropertyName.CONF_DOMAIN_DESCRIPTION));
                        assertEquals("Unexpected domain jndi factory", DOMAIN_JNDI_FACTORY,
                                topoMap.remove(DomainPropertyName.CONF_JNDI_FACTORY));
                        assertEquals("Unexpected domain jndi provider url", DOMAIN_JNDI_PROVIDER_URL,
                                topoMap.remove(DomainPropertyName.CONF_JNDI_PROVIDER_URL));
                        assertEquals("Unexpected domain jndi security principal", DOMAIN_JNDI_SECURITY_PRINCIPAL,
                                topoMap.remove(DomainPropertyName.CONF_JNDI_SECURITY_PRINCIPAL));
                        assertEquals("Unexpected domain jndi security credentials", DOMAIN_JNDI_SECURITY_CREDENTIALS,
                                topoMap.remove(DomainPropertyName.CONF_JNDI_SECURITY_CREDENTIALS));
                        assertEquals("Unexpected domain jndi pool size", Integer.toString(DOMAIN_JNDI_POOL_SIZE),
                                topoMap.remove(DomainPropertyName.CONF_JNDI_POOL_SIZE));
                        assertEquals("Unexpected domain batch pool size", Integer.toString(DOMAIN_JNDI_BATCH_SIZE),
                                topoMap.remove(DomainPropertyName.CONF_JNDI_BATCH_SIZE));
                        assertEquals("Unexpected domain mode", DomainPropertyValues.MODE_STATIC,
                                topoMap.remove(DomainPropertyName.CONF_DOMAIN_MODE));
                        assertEquals("Unexpected domain registry implementation", DOMAIN_REGISTRY_IMPLEMENTATION,
                                topoMap.remove(DomainPropertyName.CONF_SHARED_AREA_IMPLEMENTATION));

                        assertEquals("Unexpected property remains in domain map: " + topoMap, 0, topoMap.size());

                        domainMapFound++;

                    } else if (topoMap.get(TopologyItem.TOPOLOGY_TYPE).equals(
                            TopologyService.TopologyItem.CONTAINER)) {
                        // Remove the type property that is not a configuration property
                        topoMap.remove(TopologyItem.TOPOLOGY_TYPE);

                        // We retrieve the container associated to the map in the topology
                        final String containerName = topoMap.get(ContainerPropertyName.CONF_CONTAINER_NAME);
                        assertNotNull("no container name in the map", containerName);
                        final Container container = TopologyHelper.findContainer(containerName, completeTopology);
                        assertNotNull("container not retrieve, ie. the map contains a container that does not exist",
                                container);

                        // Moreover container attributes contained in the topology container part, the container state
                        // and the domain name are added to the map
                        final int expectPropertyNumber = this.getExpectedPropertyNumber(container) + 2;
                        assertEquals("Unexpected property number in container map", expectPropertyNumber,
                                topoMap.size());
                        // Check all container properties
                        if (containerName.equals(ConfigurationTestUtils.CONTAINER_NAME_0)) {
                            assertEquals("Unexpected container name", ConfigurationTestUtils.CONTAINER_NAME_0,
                                    topoMap.remove(ContainerPropertyName.CONF_CONTAINER_NAME));
                            assertEquals("Unexpected container description",
                                    ConfigurationTestUtils.CONTAINER_DESCRIPTION_0,
                                    topoMap.remove(ContainerPropertyName.CONF_CONTAINER_DESCRIPTION));
                            assertEquals("Unexpected container host", ConfigurationTestUtils.CONTAINER_HOST,
                                    topoMap.remove(ContainerPropertyName.CONF_HOST));
                            assertEquals("Unexpected container jmx user", ConfigurationTestUtils.CONTAINER_JMX_USER,
                                    topoMap.remove(ContainerPropertyName.CONF_USER));
                            assertEquals("Unexpected container jmx password", ConfigurationTestUtils.CONTAINER_JMX_PWD,
                                    topoMap.remove(ContainerPropertyName.CONF_PWD));
                            assertEquals("Unexpected container jmx rmi port", Integer.toString(CONTAINER_RMI_PORT_0_0),
                                    topoMap.remove(ContainerPropertyName.CONF_JMX_RMI_PORT));
                            assertEquals("Unexpected container tcp port",
                                    Integer.toString(ConfigurationTestUtils.CONTAINER_TCP_PORT_0),
                                    topoMap.remove(ContainerPropertyName.CONF_TRANSPORT_TCP_PORT));
                        } else if (containerName.equals(CONTAINER_NAME_0_1)) {
                            assertEquals("Unexpected container name", CONTAINER_NAME_0_1,
                                    topoMap.remove(ContainerPropertyName.CONF_CONTAINER_NAME));
                            assertEquals("Unexpected container description", CONTAINER_DESCRIPTION_0_1,
                                    topoMap.remove(ContainerPropertyName.CONF_CONTAINER_DESCRIPTION));
                            assertEquals("Unexpected container host", ConfigurationTestUtils.CONTAINER_HOST,
                                    topoMap.remove(ContainerPropertyName.CONF_HOST));
                            assertEquals("Unexpected container jmx user", ConfigurationTestUtils.CONTAINER_JMX_USER,
                                    topoMap.remove(ContainerPropertyName.CONF_USER));
                            assertEquals("Unexpected container jmx password", ConfigurationTestUtils.CONTAINER_JMX_PWD,
                                    topoMap.remove(ContainerPropertyName.CONF_PWD));
                            assertEquals("Unexpected container jmx rmi port", Integer.toString(CONTAINER_RMI_PORT_0_1),
                                    topoMap.remove(ContainerPropertyName.CONF_JMX_RMI_PORT));
                            assertEquals("Unexpected container tcp port", Integer.toString(CONTAINER_TCP_PORT_0_1),
                                    topoMap.remove(ContainerPropertyName.CONF_TRANSPORT_TCP_PORT));
                        } else if (containerName.equals(CONTAINER_NAME_1_0)) {
                            assertEquals("Unexpected container name", CONTAINER_NAME_1_0,
                                    topoMap.remove(ContainerPropertyName.CONF_CONTAINER_NAME));
                            assertEquals("Unexpected container description", CONTAINER_DESCRIPTION_1_0,
                                    topoMap.remove(ContainerPropertyName.CONF_CONTAINER_DESCRIPTION));
                            assertEquals("Unexpected container host", ConfigurationTestUtils.CONTAINER_HOST,
                                    topoMap.remove(ContainerPropertyName.CONF_HOST));
                            assertEquals("Unexpected container jmx user", ConfigurationTestUtils.CONTAINER_JMX_USER,
                                    topoMap.remove(ContainerPropertyName.CONF_USER));
                            assertEquals("Unexpected container jmx password", ConfigurationTestUtils.CONTAINER_JMX_PWD,
                                    topoMap.remove(ContainerPropertyName.CONF_PWD));
                            assertEquals("Unexpected container jmx rmi port", Integer.toString(CONTAINER_RMI_PORT_1_0),
                                    topoMap.remove(ContainerPropertyName.CONF_JMX_RMI_PORT));
                            assertEquals("Unexpected container tcp port", Integer.toString(CONTAINER_TCP_PORT_1_0),
                                    topoMap.remove(ContainerPropertyName.CONF_TRANSPORT_TCP_PORT));
                        } else if (containerName.equals(CONTAINER_NAME_1_1)) {
                            assertEquals("Unexpected container name", CONTAINER_NAME_1_1,
                                    topoMap.remove(ContainerPropertyName.CONF_CONTAINER_NAME));
                            assertEquals("Unexpected container description", CONTAINER_DESCRIPTION_1_1,
                                    topoMap.remove(ContainerPropertyName.CONF_CONTAINER_DESCRIPTION));
                            assertEquals("Unexpected container host", ConfigurationTestUtils.CONTAINER_HOST,
                                    topoMap.remove(ContainerPropertyName.CONF_HOST));
                            assertEquals("Unexpected container jmx user", ConfigurationTestUtils.CONTAINER_JMX_USER,
                                    topoMap.remove(ContainerPropertyName.CONF_USER));
                            assertEquals("Unexpected container jmx password", ConfigurationTestUtils.CONTAINER_JMX_PWD,
                                    topoMap.remove(ContainerPropertyName.CONF_PWD));
                            assertEquals("Unexpected container jmx rmi port", Integer.toString(CONTAINER_RMI_PORT_1_1),
                                    topoMap.remove(ContainerPropertyName.CONF_JMX_RMI_PORT));
                            assertEquals("Unexpected container tcp port", Integer.toString(CONTAINER_TCP_PORT_1_1),
                                    topoMap.remove(ContainerPropertyName.CONF_TRANSPORT_TCP_PORT));
                        } else {
                            fail("Unexpected container");
                        }

                        assertNotNull("Container state is missing", topoMap.remove(ContainerPropertyName.CONF_STATE));
                        assertNotNull("Domain is missing", topoMap.remove(ContainerPropertyName.CONF_DOMAIN_NAME));

                        assertEquals("Unexpected property remains in container map: " + topoMap, 0, topoMap.size());

                        containerMapFound++;

                    } else {
                        fail("Unknown topology item type");
                    }
                }

                assertEquals("unexpected or missing domain map", 1, domainMapFound);
                assertEquals("unexpected or missing container found", 2, containerMapFound);
            } finally {
                this.stopTopologyService(topoSvc);
            }
        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check that the method {@link TopologyServiceImpl#retrieveTopology(String, boolean)} returns all possible
     * parameters of a topology definition when the given pass-phrase is not the expected one.
     * </p>
     * <p>
     * Expected result: No error, all possible parameters are returned, excluding sensible information
     * </p>
     */
    @Test
    public void testAllTopologyParametersReturnedExcludingSensibleInformation() throws Exception {

        // Create a complete topology
        final Topology completeTopology = this.createSampleCompleteTopology();
        assertAllFieldsSet(completeTopology);

        // Create and initialize the configuration service
        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);
        serverProps.put(ConfigurationProperties.TOPOLOGY_PASSPHRASE, TOPOLOGY_PASSPHRASE);
        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);
        try {
            cfgSvc.loadConfiguration(serverProps, completeTopology);

            final PetalsCompositeController petalsCompositeCtrl = EasyMock.createMock(PetalsCompositeController.class);
            EasyMock.expect(petalsCompositeCtrl.getSharedAreaCltConfiguration(TOPOLOGY_INVALID_PASSPHRASE))
                    .andReturn(new Properties()).anyTimes();
            EasyMock.replay(petalsCompositeCtrl);

            final TopologyService topoSvc = new TopologyServiceImpl();
            ReflectionHelper.setPrivateField(topoSvc, "configurationService", cfgSvc);
            ReflectionHelper.setPrivateField(topoSvc, "petalsCompositeController", petalsCompositeCtrl);
            this.startTopologyService(topoSvc);

            try {
                int domainMapFound = 0;
                int containerMapFound = 0;

                final Set<Map<String, String>> retrievedTopology = topoSvc.retrieveTopology(
                        TOPOLOGY_INVALID_PASSPHRASE, false);
                for (final Map<String, String> topoMap : retrievedTopology) {
                    if (topoMap.get(TopologyItem.TOPOLOGY_TYPE).equals(TopologyService.TopologyItem.DOMAIN)) {

                        // Remove the type property that is not a configuration property
                        topoMap.remove(TopologyItem.TOPOLOGY_TYPE);

                        // Check all domain properties
                        final int expectPropertyNumber = this.getExpectedPropertyNumber(completeTopology.getDomain());
                        assertEquals("Unexpected property number in domain map", expectPropertyNumber, topoMap.size());
                        assertEquals("Unexpected domain name", ConfigurationTestUtils.DOMAIN_NAME,
                                topoMap.remove(DomainPropertyName.CONF_DOMAIN_NAME));
                        assertEquals("Unexpected domain description", ConfigurationTestUtils.DOMAIN_DESCRIPTION,
                                topoMap.remove(DomainPropertyName.CONF_DOMAIN_DESCRIPTION));
                        assertEquals("Unexpected domain jndi factory", DOMAIN_JNDI_FACTORY,
                                topoMap.remove(DomainPropertyName.CONF_JNDI_FACTORY));
                        assertEquals("Unexpected domain jndi provider url", DOMAIN_JNDI_PROVIDER_URL,
                                topoMap.remove(DomainPropertyName.CONF_JNDI_PROVIDER_URL));
                        assertEquals("Unexpected domain jndi security principal", DOMAIN_JNDI_SECURITY_PRINCIPAL,
                                topoMap.remove(DomainPropertyName.CONF_JNDI_SECURITY_PRINCIPAL));
                        assertEquals("Unexpected domain jndi security credentials", DOMAIN_JNDI_SECURITY_CREDENTIALS,
                                topoMap.remove(DomainPropertyName.CONF_JNDI_SECURITY_CREDENTIALS));
                        assertEquals("Unexpected domain jndi pool size", Integer.toString(DOMAIN_JNDI_POOL_SIZE),
                                topoMap.remove(DomainPropertyName.CONF_JNDI_POOL_SIZE));
                        assertEquals("Unexpected domain batch pool size", Integer.toString(DOMAIN_JNDI_BATCH_SIZE),
                                topoMap.remove(DomainPropertyName.CONF_JNDI_BATCH_SIZE));
                        assertEquals("Unexpected domain mode", DomainPropertyValues.MODE_STATIC,
                                topoMap.remove(DomainPropertyName.CONF_DOMAIN_MODE));
                        assertEquals("Unexpected domain registry implementation", DOMAIN_REGISTRY_IMPLEMENTATION,
                                topoMap.remove(DomainPropertyName.CONF_SHARED_AREA_IMPLEMENTATION));

                        assertEquals("Unexpected property remains in domain map: " + topoMap, 0, topoMap.size());

                        domainMapFound++;

                    } else if (topoMap.get(TopologyItem.TOPOLOGY_TYPE).equals(TopologyService.TopologyItem.CONTAINER)) {
                        // Remove the type property that is not a configuration property
                        topoMap.remove(TopologyItem.TOPOLOGY_TYPE);

                        // We retrieve the container associated to the map in the topology
                        final String containerName = topoMap.get(ContainerPropertyName.CONF_CONTAINER_NAME);
                        assertNotNull("no container name in the map", containerName);
                        final Container container = TopologyHelper.findContainer(containerName, completeTopology);
                        assertNotNull("container not retrieve, ie. the map contains a container that does not exist",
                                container);

                        // Moreover container attributes contained in the topology container part, the container state
                        // and the domain name are added to the map.
                        // JMX User and JMX password are not returned because of invalid pass-phrase
                        final int expectPropertyNumber = this.getExpectedPropertyNumber(container) + 2 - 2;
                        assertEquals("Unexpected property number in container map", expectPropertyNumber,
                                topoMap.size());
                        // Check all container properties
                        if (containerName.equals(ConfigurationTestUtils.CONTAINER_NAME_0)) {
                            assertEquals("Unexpected container name", ConfigurationTestUtils.CONTAINER_NAME_0,
                                    topoMap.remove(ContainerPropertyName.CONF_CONTAINER_NAME));
                            assertEquals("Unexpected container description",
                                    ConfigurationTestUtils.CONTAINER_DESCRIPTION_0,
                                    topoMap.remove(ContainerPropertyName.CONF_CONTAINER_DESCRIPTION));
                            assertEquals("Unexpected container host", ConfigurationTestUtils.CONTAINER_HOST,
                                    topoMap.remove(ContainerPropertyName.CONF_HOST));
                            assertFalse("JMX user returned", topoMap.containsKey(ContainerPropertyName.CONF_USER));
                            assertFalse("JMX pwd returned", topoMap.containsKey(ContainerPropertyName.CONF_PWD));
                            assertEquals("Unexpected container jmx rmi port", Integer.toString(CONTAINER_RMI_PORT_0_0),
                                    topoMap.remove(ContainerPropertyName.CONF_JMX_RMI_PORT));
                            assertEquals("Unexpected container tcp port",
                                    Integer.toString(ConfigurationTestUtils.CONTAINER_TCP_PORT_0),
                                    topoMap.remove(ContainerPropertyName.CONF_TRANSPORT_TCP_PORT));
                        } else if (containerName.equals(CONTAINER_NAME_0_1)) {
                            assertEquals("Unexpected container name", CONTAINER_NAME_0_1,
                                    topoMap.remove(ContainerPropertyName.CONF_CONTAINER_NAME));
                            assertEquals("Unexpected container description", CONTAINER_DESCRIPTION_0_1,
                                    topoMap.remove(ContainerPropertyName.CONF_CONTAINER_DESCRIPTION));
                            assertEquals("Unexpected container host", ConfigurationTestUtils.CONTAINER_HOST,
                                    topoMap.remove(ContainerPropertyName.CONF_HOST));
                            assertFalse("JMX user returned", topoMap.containsKey(ContainerPropertyName.CONF_USER));
                            assertFalse("JMX pwd returned", topoMap.containsKey(ContainerPropertyName.CONF_PWD));
                            assertEquals("Unexpected container jmx rmi port", Integer.toString(CONTAINER_RMI_PORT_0_1),
                                    topoMap.remove(ContainerPropertyName.CONF_JMX_RMI_PORT));
                            assertEquals("Unexpected container tcp port", Integer.toString(CONTAINER_TCP_PORT_0_1),
                                    topoMap.remove(ContainerPropertyName.CONF_TRANSPORT_TCP_PORT));
                        } else if (containerName.equals(CONTAINER_NAME_1_0)) {
                            assertEquals("Unexpected container name", CONTAINER_NAME_1_0,
                                    topoMap.remove(ContainerPropertyName.CONF_CONTAINER_NAME));
                            assertEquals("Unexpected container description", CONTAINER_DESCRIPTION_1_0,
                                    topoMap.remove(ContainerPropertyName.CONF_CONTAINER_DESCRIPTION));
                            assertEquals("Unexpected container host", ConfigurationTestUtils.CONTAINER_HOST,
                                    topoMap.remove(ContainerPropertyName.CONF_HOST));
                            assertFalse("JMX user returned", topoMap.containsKey(ContainerPropertyName.CONF_USER));
                            assertFalse("JMX pwd returned", topoMap.containsKey(ContainerPropertyName.CONF_PWD));
                            assertEquals("Unexpected container jmx rmi port", Integer.toString(CONTAINER_RMI_PORT_1_0),
                                    topoMap.remove(ContainerPropertyName.CONF_JMX_RMI_PORT));
                            assertEquals("Unexpected container tcp port", Integer.toString(CONTAINER_TCP_PORT_1_0),
                                    topoMap.remove(ContainerPropertyName.CONF_TRANSPORT_TCP_PORT));
                        } else if (containerName.equals(CONTAINER_NAME_1_1)) {
                            assertEquals("Unexpected container name", CONTAINER_NAME_1_1,
                                    topoMap.remove(ContainerPropertyName.CONF_CONTAINER_NAME));
                            assertEquals("Unexpected container description", CONTAINER_DESCRIPTION_1_1,
                                    topoMap.remove(ContainerPropertyName.CONF_CONTAINER_DESCRIPTION));
                            assertEquals("Unexpected container host", ConfigurationTestUtils.CONTAINER_HOST,
                                    topoMap.remove(ContainerPropertyName.CONF_HOST));
                            assertFalse("JMX user returned", topoMap.containsKey(ContainerPropertyName.CONF_USER));
                            assertFalse("JMX pwd returned", topoMap.containsKey(ContainerPropertyName.CONF_PWD));
                            assertEquals("Unexpected container jmx rmi port", Integer.toString(CONTAINER_RMI_PORT_1_1),
                                    topoMap.remove(ContainerPropertyName.CONF_JMX_RMI_PORT));
                            assertEquals("Unexpected container tcp port", Integer.toString(CONTAINER_TCP_PORT_1_1),
                                    topoMap.remove(ContainerPropertyName.CONF_TRANSPORT_TCP_PORT));
                        } else {
                            fail("Unexpected container");
                        }

                        assertNotNull("Container state is missing", topoMap.remove(ContainerPropertyName.CONF_STATE));
                        assertNotNull("Domain is missing", topoMap.remove(ContainerPropertyName.CONF_DOMAIN_NAME));

                        assertEquals("Unexpected property remains in container map: " + topoMap, 0, topoMap.size());

                        containerMapFound++;

                    } else {
                        fail("Unknown topology item type");
                    }
                }

                assertEquals("unexpected or missing domain map", domainMapFound, 1);
                assertEquals("unexpected or missing container found", containerMapFound, 2);
            } finally {
                this.stopTopologyService(topoSvc);
            }
        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * Compute the number of property that a map returned by {@link TopologyService#retrieveTopology(String)} should
     * contain.
     * 
     * @param object
     *            A topology object
     * @return The number of property that a map associated to the topology object should contain.
     */
    private int getExpectedPropertyNumber(final Object object) throws IllegalArgumentException, IllegalAccessException,
            InvocationTargetException {
        int expectedPropertyNumber = 0;
        final String topologyPackageName = Topology.class.getPackage().getName();
        final Method[] methods = object.getClass().getMethods();
        for (final Method method : methods) {
            final Type methodReturnedGenericType = method.getGenericReturnType();

            if (method.getName().startsWith("get")) {
                if (methodReturnedGenericType instanceof ParameterizedType) {
                    final ParameterizedType methodReturnedParameterizedType = (ParameterizedType) methodReturnedGenericType;
                    if (((Class<?>) methodReturnedParameterizedType.getRawType()).getName()
                            .equals(List.class.getName())) {
                        // No list is converted as property by TopologyService.retrieveTopology
                    }
                } else if (methodReturnedGenericType instanceof Class) {

                    final Class<?> methodReturnedType = (Class<?>) methodReturnedGenericType;
                    if (!methodReturnedType.isPrimitive()
                            && methodReturnedType.getPackage().getName().equals(topologyPackageName)
                            && !methodReturnedType.getName().equals(DomainMode.class.getName())) {
                        final Object result = method.invoke(object, (Object[]) null);
                        expectedPropertyNumber += this.getExpectedPropertyNumber(result);
                    } else {
                        expectedPropertyNumber++;
                    }

                }
            } else if (method.getName().startsWith("is")) {
                final Class<?> methodReturnedType = (Class<?>) methodReturnedGenericType;
                if (methodReturnedType.getName().equals(Boolean.class.getName())) {
                    expectedPropertyNumber++;
                }

            }
        }

        return expectedPropertyNumber;
    }

    /**
     * Assert that all fields (mandatory and optional) of a topology are set
     * 
     * @param object
     *            A topology object
     */
    private void assertAllFieldsSet(final Object object) throws IllegalArgumentException, IllegalAccessException,
            InvocationTargetException {
        final String topologyPackageName = Topology.class.getPackage().getName();
        final Method[] methods = object.getClass().getMethods();
        for (final Method method : methods) {
            final Type methodReturnedGenericType = method.getGenericReturnType();

            if (method.getName().startsWith("get")) {
                if (methodReturnedGenericType instanceof ParameterizedType) {
                    final ParameterizedType methodReturnedParameterizedType = (ParameterizedType) methodReturnedGenericType;
                    if (((Class<?>) methodReturnedParameterizedType.getRawType()).getName()
                            .equals(List.class.getName())) {
                        final Type[] typeArguments = methodReturnedParameterizedType.getActualTypeArguments();
                        boolean topologyBasedArgFround = false;
                        for (final Type typeArgument : typeArguments) {
                            final Class<?> typeArgClass = (Class<?>) typeArgument;
                            if (typeArgClass.getPackage().getName().equals(topologyPackageName)) {
                                topologyBasedArgFround = true;
                            }
                        }
                        if (topologyBasedArgFround) {
                            final List<?> result = (List<?>) method.invoke(object, (Object[]) null);
                            for (final Object item : result) {
                                this.assertAllFieldsSet(item);
                            }
                        }
                    }
                } else if (methodReturnedGenericType instanceof Class) {

                    final Object result = method.invoke(object, (Object[]) null);

                    final Class<?> methodReturnedType = (Class<?>) methodReturnedGenericType;
                    if (methodReturnedType.isPrimitive()) {
                        if (methodReturnedType.equals(Short.TYPE)) {
                            assertFalse("A bean value is not set: " + method.getName(),
                                    ((Short) result).intValue() == 0);
                        } else if (methodReturnedType.equals(Integer.TYPE)) {
                            assertFalse("A bean value is not set: " + method.getName(),
                                    ((Integer) result).intValue() == 0);
                        } else if (methodReturnedType.equals(Long.TYPE)) {
                            assertFalse("A bean value is not set: " + method.getName(),
                                    ((Long) result).longValue() == 0l);
                        } else {
                            fail("Unsupported primitive");
                        }
                    } else {
                        assertNotNull("A bean value is not set: " + method.getName(), result);

                        if (methodReturnedType.getPackage().getName().equals(topologyPackageName)) {
                            this.assertAllFieldsSet(result);
                        }
                    }

                }
            } else if (method.getName().startsWith("is")) {
                final Class<?> methodReturnedType = (Class<?>) methodReturnedGenericType;
                if (methodReturnedType.getName().equals(Boolean.class.getName())) {
                    final Object result = method.invoke(object, (Object[]) null);
                    assertNotNull("A bean value is not set" + method.getName(), result);
                }

            }
        }
    }

    /**
     * Create a sample topology where all parameters are filled
     * 
     * @return A sample topology where all parameters are filled
     */
    private Topology createSampleCompleteTopology() {

        final Topology topology = new Topology();
        final Domain domain = new Domain();
        topology.setDomain(domain);

        domain.setName(ConfigurationTestUtils.DOMAIN_NAME);
        domain.setDescription(ConfigurationTestUtils.DOMAIN_DESCRIPTION);
        domain.setMode(ConfigurationTestUtils.DOMAIN_MODE);
        domain.setRegistryImplementation(DOMAIN_REGISTRY_IMPLEMENTATION);

        final Jndi jndi = new Jndi();
        domain.setJndi(jndi);
        jndi.setFactory(DOMAIN_JNDI_FACTORY);
        jndi.setProviderUrl(DOMAIN_JNDI_PROVIDER_URL);
        jndi.setSecurityPrincipal(DOMAIN_JNDI_SECURITY_PRINCIPAL);
        jndi.setSecurityCredentials(DOMAIN_JNDI_SECURITY_CREDENTIALS);
        jndi.setPoolSize(Integer.valueOf(DOMAIN_JNDI_POOL_SIZE));
        jndi.setBatchSize(Integer.valueOf(DOMAIN_JNDI_BATCH_SIZE));

        // Container #0
        final Container container0 = new Container();
        domain.getContainer().add(container0);
        container0.setName(ConfigurationTestUtils.CONTAINER_NAME_0);
        container0.setDescription(ConfigurationTestUtils.CONTAINER_DESCRIPTION_0);
        container0.setHost(ConfigurationTestUtils.CONTAINER_HOST);
        container0.setUser(ConfigurationTestUtils.CONTAINER_JMX_USER);
        container0.setPassword(ConfigurationTestUtils.CONTAINER_JMX_PWD);

        final JmxService jmxService0 = new JmxService();
        container0.setJmxService(jmxService0);
        jmxService0.setRmiPort(CONTAINER_RMI_PORT_0_0);

        final TransportService transportService0 = new TransportService();
        transportService0.setTcpPort(ConfigurationTestUtils.CONTAINER_TCP_PORT_0);
        container0.setTransportService(transportService0);

        // Container #1
        final Container container0_1 = new Container();
        domain.getContainer().add(container0_1);
        container0_1.setName(CONTAINER_NAME_0_1);
        container0_1.setDescription(CONTAINER_DESCRIPTION_0_1);
        container0_1.setHost(ConfigurationTestUtils.CONTAINER_HOST);
        container0_1.setUser(ConfigurationTestUtils.CONTAINER_JMX_USER);
        container0_1.setPassword(ConfigurationTestUtils.CONTAINER_JMX_PWD);

        final JmxService jmxService0_1 = new JmxService();
        container0_1.setJmxService(jmxService0_1);
        jmxService0_1.setRmiPort(CONTAINER_RMI_PORT_0_1);

        final TransportService transportService0_1 = new TransportService();
        transportService0_1.setTcpPort(CONTAINER_TCP_PORT_0_1);
        container0_1.setTransportService(transportService0_1);

        return topology;
    }

}
