/**
 * Copyright (c) 2014-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.configuration;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.fail;

import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.util.Properties;

import org.junit.Rule;
import org.junit.Test;
import org.ow2.petals.clientserverapi.configuration.ContainerConfiguration;
import org.ow2.petals.microkernel.api.configuration.ConfigurationService;
import org.ow2.petals.microkernel.api.configuration.exception.ConfigurationException;
import org.ow2.petals.microkernel.api.configuration.exception.ContainerAlreadyExistException;
import org.ow2.petals.microkernel.api.configuration.exception.DomainUnknownException;
import org.ow2.petals.microkernel.api.configuration.exception.InvalidContainerConfigurationProvidedException;

/**
 * Unit tests of {@link ConfigurationServiceImpl#addContainerConfiguration(ContainerConfiguration)}
 * 
 * @author Christophe DENEUX - Linagora
 * 
 */
public class ConfigurationServiceImpl_AddContainerConfigurationTest extends AbstractConfigurationServiceImplTest {

    private final static String CONTAINER_NAME_1 = "sample-1";

    @Rule
    public final LocalServerPropertiesBuilder localServerPropertiesBuilder = new LocalServerPropertiesBuilder();

    /**
     * <p>
     * Check the add of a container configuration against:
     * <ul>
     * <li>the container configuration provided has a name unset (ie. <code>null</code>)</li>
     * </ul>
     * </p>
     * <p>
     * Expected result: A error occurs because the container configuration is invalid.
     * </p>
     */
    @Test(expected = InvalidContainerConfigurationProvidedException.class)
    public void testContainerProvidedHasANullName() throws ConfigurationException, InvocationTargetException,
            IOException {

        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);

        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        try {
            cfgSvc.loadConfiguration(serverProps, this.topology);

            final ContainerConfiguration containerToAdd = new ContainerConfiguration();
            cfgSvc.addContainerConfiguration(containerToAdd);
        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check the add of a container configuration against:
     * <ul>
     * <li>the container configuration provided is the current container configuration</li>
     * </ul>
     * </p>
     * <p>
     * Expected result: A error occurs because the container configuration is invalid: the configuration of the current
     * container.
     * </p>
     */
    @Test(expected = InvalidContainerConfigurationProvidedException.class)
    public void testContainerProvidedIsTheCurrentOne() throws ConfigurationException, InvocationTargetException,
            IOException {

        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);

        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        try {
            cfgSvc.loadConfiguration(serverProps, this.topology);

            final ContainerConfiguration containerToAdd = new ContainerConfiguration();
            containerToAdd.setName(ConfigurationTestUtils.CONTAINER_NAME_0);
            cfgSvc.addContainerConfiguration(containerToAdd);
        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check the add of a container configuration against:
     * <ul>
     * <li>the container configuration provided has a name correctly set</li>
     * <li>the container configuration provided has a domain name unset (ie. <code>null</code>)</li>
     * </ul>
     * </p>
     * <p>
     * Expected result: A error occurs because the container configuration is invalid.
     * </p>
     */
    @Test(expected = InvalidContainerConfigurationProvidedException.class)
    public void testContainerProvidedHasANullDomainName()
            throws ConfigurationException, InvocationTargetException,
            IOException {

        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);

        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        try {
            cfgSvc.loadConfiguration(serverProps, this.topology);

            final ContainerConfiguration containerToAdd = new ContainerConfiguration();
            containerToAdd.setName(CONTAINER_NAME_1);
            cfgSvc.addContainerConfiguration(containerToAdd);
        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check the add of a container configuration against:
     * <ul>
     * <li>the container configuration provided has a name correctly set</li>
     * <li>the container configuration provided has a unknown domain name set</li>
     * </ul>
     * </p>
     * <p>
     * Expected result: A error occurs because the domain name is unknown
     * </p>
     */
    @Test(expected = DomainUnknownException.class)
    public void testContainerProvidedHasAUnknownDomainName()
            throws ConfigurationException, InvocationTargetException, IOException {

        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);

        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        final String unknownDomainName = "unknown-domain";
        try {
            cfgSvc.loadConfiguration(serverProps, this.topology);

            final ContainerConfiguration containerToAdd = new ContainerConfiguration();
            containerToAdd.setName(CONTAINER_NAME_1);
            containerToAdd.setDomainName(unknownDomainName);
            cfgSvc.addContainerConfiguration(containerToAdd);
        } catch (final DomainUnknownException e) {
            assertEquals("Unexpected domain name", unknownDomainName, e.getUnknownDomainName());
            throw e;
        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check the add of a container configuration against:
     * <ul>
     * <li>the topology contains 1 container</li>
     * <li>a container is added with a valid container configuration</li>
     * <li>the container is added again</li>
     * </ul>
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>On the first add:
     * <ul>
     * <li>No error occurs</li>
     * <li>The container is correctly added</li>
     * <li>The topology contains 2 containers</li>
     * </ul>
     * </li>
     * <li>An error occurs on the second add because the container already exist</li>
     * </ul>
     * </p>
     */
    @Test
    public void testContainerProvidedIsValid() throws ConfigurationException, InvocationTargetException, IOException {

        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);

        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        try {
            cfgSvc.loadConfiguration(serverProps, this.topology);

            final ContainerConfiguration containerToAdd = new ContainerConfiguration();
            containerToAdd.setName(CONTAINER_NAME_1);
            containerToAdd.setDomainName(ConfigurationTestUtils.DOMAIN_NAME);
            cfgSvc.addContainerConfiguration(containerToAdd);

            assertNotNull("container 0 no more exists",
                    cfgSvc.getContainerConfiguration(ConfigurationTestUtils.CONTAINER_NAME_0));
            assertNotNull("container 1 does not exist", cfgSvc.getContainerConfiguration(CONTAINER_NAME_1));
            assertEquals("unexpected container number", cfgSvc.getContainersConfiguration().size(), 2);

            try {
                cfgSvc.addContainerConfiguration(containerToAdd);
                fail("Exception " + ContainerAlreadyExistException.class.getName() + " not thrown");
            } catch (final ContainerAlreadyExistException e) {
                assertEquals("unexpected already existing container", CONTAINER_NAME_1,
                        e.getAlreadyExistingContainerName());
            }

        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

}
