/**
 * Copyright (c) 2014-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.configuration;

import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;

import java.util.Properties;

import org.junit.Rule;
import org.junit.Test;
import org.ow2.petals.launcher.api.server.conf.ConfigurationProperties;
import org.ow2.petals.microkernel.api.configuration.ConfigurationService;

/**
 * Unit tests of {@link ConfigurationServiceImpl#isSecurityTopologyPassphraseOk(String)}
 * 
 * @author Christophe DENEUX - Linagora
 * 
 */
public class ConfigurationServiceImpl_IsSecurityTopoPassPhraseTest extends AbstractConfigurationServiceImplTest {

    @Rule
    public final LocalServerPropertiesBuilder localServerPropertiesBuilder = new LocalServerPropertiesBuilder();

    /**
     * <p>
     * Check {@link ConfigurationServiceImpl#isSecurityTopologyPassphraseOk(String)} against:
     * <ul>
     * <li>the configured pass-phrase is null</li>
     * <li>the provided pass-phrase is null</li>
     * </ul>
     * </p>
     * <p>
     * Expected result: provided pass-phrase is invalid
     * </p>
     */
    @Test
    public void testConfNull_ProvidedNull() throws Exception {

        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);

        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        try {
            cfgSvc.loadConfiguration(serverProps, this.topology);

            assertFalse("Provided pass-phrase validated", cfgSvc.isSecurityTopologyPassphraseOk(null));
        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check {@link ConfigurationServiceImpl#isSecurityTopologyPassphraseOk(String)} against:
     * <ul>
     * <li>the configured pass-phrase is null</li>
     * <li>the provided pass-phrase is empty</li>
     * </ul>
     * </p>
     * <p>
     * Expected result: provided pass-phrase is invalid
     * </p>
     */
    @Test
    public void testConfNull_ProvidedEmpty() throws Exception {

        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);

        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        try {
            cfgSvc.loadConfiguration(serverProps, this.topology);

            assertFalse("Provided pass-phrase validated", cfgSvc.isSecurityTopologyPassphraseOk(""));
        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check {@link ConfigurationServiceImpl#isSecurityTopologyPassphraseOk(String)} against:
     * <ul>
     * <li>the configured pass-phrase is null</li>
     * <li>the provided pass-phrase is not null and non-empty</li>
     * </ul>
     * </p>
     * <p>
     * Expected result: provided pass-phrase is invalid
     * </p>
     */
    @Test
    public void testConfNull_Provided() throws Exception {

        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);

        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        try {
            cfgSvc.loadConfiguration(serverProps, this.topology);

            assertFalse("Provided pass-phrase validated",
                    cfgSvc.isSecurityTopologyPassphraseOk("a topology pass-phrase"));
        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check {@link ConfigurationServiceImpl#isSecurityTopologyPassphraseOk(String)} against:
     * <ul>
     * <li>the configured pass-phrase is empty</li>
     * <li>the provided pass-phrase is null</li>
     * </ul>
     * </p>
     * <p>
     * Expected result: provided pass-phrase is invalid
     * </p>
     */
    @Test
    public void testConfEmpty_ProvidedNull() throws Exception {

        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);
        serverProps.setProperty(ConfigurationProperties.TOPOLOGY_PASSPHRASE, "");

        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        try {
            cfgSvc.loadConfiguration(serverProps, this.topology);

            assertFalse("Provided pass-phrase validated", cfgSvc.isSecurityTopologyPassphraseOk(null));
        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check {@link ConfigurationServiceImpl#isSecurityTopologyPassphraseOk(String)} against:
     * <ul>
     * <li>the configured pass-phrase is empty</li>
     * <li>the provided pass-phrase is empty</li>
     * </ul>
     * </p>
     * <p>
     * Expected result: provided pass-phrase is invalid
     * </p>
     */
    @Test
    public void testConfEmpty_ProvidedEmpty() throws Exception {

        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);
        serverProps.setProperty(ConfigurationProperties.TOPOLOGY_PASSPHRASE, "");

        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        try {
            cfgSvc.loadConfiguration(serverProps, this.topology);

            assertFalse("Provided pass-phrase validated", cfgSvc.isSecurityTopologyPassphraseOk(""));
        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check {@link ConfigurationServiceImpl#isSecurityTopologyPassphraseOk(String)} against:
     * <ul>
     * <li>the configured pass-phrase is empty</li>
     * <li>the provided pass-phrase is not null and non-empty</li>
     * </ul>
     * </p>
     * <p>
     * Expected result: provided pass-phrase is invalid
     * </p>
     */
    @Test
    public void testConfEmpty_Provided() throws Exception {

        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);
        serverProps.setProperty(ConfigurationProperties.TOPOLOGY_PASSPHRASE, "");

        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        try {
            cfgSvc.loadConfiguration(serverProps, this.topology);

            assertFalse("Provided pass-phrase validated",
                    cfgSvc.isSecurityTopologyPassphraseOk("a topology pass-phrase"));
        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check {@link ConfigurationServiceImpl#isSecurityTopologyPassphraseOk(String)} against:
     * <ul>
     * <li>a configured pass-phrase is set</li>
     * <li>the provided pass-phrase is null</li>
     * </ul>
     * </p>
     * <p>
     * Expected result: provided pass-phrase is invalid
     * </p>
     */
    @Test
    public void testConfGiven_ProvidedNull() throws Exception {

        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);
        serverProps.setProperty(ConfigurationProperties.TOPOLOGY_PASSPHRASE, "a topology pass-phrase");

        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        try {
            cfgSvc.loadConfiguration(serverProps, this.topology);

            assertFalse("Provided pass-phrase validated", cfgSvc.isSecurityTopologyPassphraseOk(null));
        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check {@link ConfigurationServiceImpl#isSecurityTopologyPassphraseOk(String)} against:
     * <ul>
     * <li>a configured pass-phrase is set</li>
     * <li>the provided pass-phrase is empty</li>
     * </ul>
     * </p>
     * <p>
     * Expected result: provided pass-phrase is invalid
     * </p>
     */
    @Test
    public void testConfGiven_ProvidedEmpty() throws Exception {

        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);
        serverProps.setProperty(ConfigurationProperties.TOPOLOGY_PASSPHRASE, "a topology pass-phrase");

        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        try {
            cfgSvc.loadConfiguration(serverProps, this.topology);

            assertFalse("Provided pass-phrase validated", cfgSvc.isSecurityTopologyPassphraseOk(""));
        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check {@link ConfigurationServiceImpl#isSecurityTopologyPassphraseOk(String)} against:
     * <ul>
     * <li>a configured pass-phrase is set</li>
     * <li>the provided pass-phrase is not null, non-empty and not matches</li>
     * </ul>
     * </p>
     * <p>
     * Expected result: provided pass-phrase is invalid
     * </p>
     */
    @Test
    public void testConfGiven_ProvidedNotMatch() throws Exception {

        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);
        serverProps.setProperty(ConfigurationProperties.TOPOLOGY_PASSPHRASE, "a topology pass-phrase");

        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        try {
            cfgSvc.loadConfiguration(serverProps, this.topology);

            assertFalse("Provided pass-phrase validated",
                    cfgSvc.isSecurityTopologyPassphraseOk("a topology pass-phrase not matching"));
        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check {@link ConfigurationServiceImpl#isSecurityTopologyPassphraseOk(String)} against:
     * <ul>
     * <li>a configured pass-phrase is set</li>
     * <li>the provided pass-phrase is not null, non-empty and matches</li>
     * </ul>
     * </p>
     * <p>
     * Expected result: provided pass-phrase is invalid
     * </p>
     */
    @Test
    public void testConfGiven_ProvidedMatch() throws Exception {

        final String topologyPassphrase = "a topology pass-phrase";
        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);
        serverProps.setProperty(ConfigurationProperties.TOPOLOGY_PASSPHRASE, topologyPassphrase);

        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        try {
            cfgSvc.loadConfiguration(serverProps, this.topology);

            assertTrue("Provided pass-phrase not validated", cfgSvc.isSecurityTopologyPassphraseOk(topologyPassphrase));
        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }
}
