/**
 * Copyright (c) 2014-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.configuration;

import static org.junit.Assert.assertTrue;

import java.io.File;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.net.URISyntaxException;

import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.TemporaryFolder;
import org.ow2.petals.clientserverapi.configuration.ContainerConfiguration;
import org.ow2.petals.launcher.api.server.conf.ConfigurationProperties;
import org.ow2.petals.microkernel.api.configuration.ConfigurationService;
import org.ow2.petals.microkernel.api.configuration.exception.ConfigurationException;
import org.ow2.petals.topology.TopologyBuilder;
import org.ow2.petals.topology.TopologyException;
import org.ow2.petals.topology.generated.Topology;

import com.ebmwebsourcing.easycommons.lang.reflect.ReflectionHelper;
import com.ebmwebsourcing.easycommons.properties.PropertiesException;

/**
 * Unit tests of {@link ConfigurationServiceImpl#persistTopology(Topology)}
 * 
 * @author Christophe DENEUX - Linagora
 * 
 */
public class ConfigurationServiceImpl_PersistTopologyTest extends AbstractConfigurationServiceImplTest {

    private final static String CONTAINER_CONFIG_FIELD = "containerConfiguration";

    @Rule
    public final TemporaryFolder temporaryFolder = new TemporaryFolder();

    /**
     * <p>
     * Check the topology persistence against:
     * <ul>
     * <li>{@value ConfigurationProperties#WORKING_AREA_PATH_PROPERTY_NAME} that is set to an invalid value (it is a
     * file),</li>
     * <li>a sample topology is provided</li>
     * </ul>
     * </p>
     * <p>
     * Expected result: An error occurs about the invalid working area detected.
     * </p>
     */
    @Test(expected = ConfigurationException.class)
    public void testInvalidWorkingArea() throws ConfigurationException, InvocationTargetException, PropertiesException,
            URISyntaxException, IOException {

        final File invalidWorkingArea = this.temporaryFolder.newFile("invalid-working-area");
        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        final ContainerConfiguration containerConfig = new ContainerConfiguration();
        containerConfig.setWorkDirectoryPath(invalidWorkingArea.getAbsolutePath());
        ReflectionHelper.setPrivateField(cfgSvc, CONTAINER_CONFIG_FIELD, containerConfig);
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        try {
            cfgSvc.persistTopology(ConfigurationTestUtils.createBasicSampleTopology());

        } catch (final ConfigurationException e) {
            assertTrue("Unexpected configuration exception",
                    e.getMessage().contains(invalidWorkingArea.getAbsolutePath()));
            throw e;
        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check the topology persistence against:
     * <ul>
     * <li>{@value ConfigurationProperties#WORKING_AREA_PATH_PROPERTY_NAME} that is set to an unexisting directory,</li>
     * <li>a sample topology is provided</li>
     * </ul>
     * </p>
     * Expected result: An error occurs about the unexisting working area </p>
     */
    @Test(expected = ConfigurationException.class)
    public void testUnexistingWorkingArea() throws ConfigurationException, InvocationTargetException,
            PropertiesException, URISyntaxException {

        final String unexistingWorkingArea = "unexisting-working-area";
        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        final ContainerConfiguration containerConfig = new ContainerConfiguration();
        containerConfig.setWorkDirectoryPath(unexistingWorkingArea);
        ReflectionHelper.setPrivateField(cfgSvc, CONTAINER_CONFIG_FIELD, containerConfig);
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        try {
            cfgSvc.persistTopology(ConfigurationTestUtils.createBasicSampleTopology());

        } catch (final ConfigurationException e) {
            assertTrue("Unexpected configuration exception", e.getMessage().contains(unexistingWorkingArea));
            throw e;
        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check the topology persistence against:
     * <ul>
     * <li>the working area is valid,</li>
     * <li>a persisted topology already exists but is invalid (it is a directory),</li>
     * <li>a sample topology is provided</li>
     * </ul>
     * </p>
     * <p>
     * Expected result: An error occurs about the invalid existing persisted topology definition
     * </p>
     */
    @Test(expected = ConfigurationException.class)
    public void testInvalidPersistedTopologyFile() throws ConfigurationException, InvocationTargetException,
            PropertiesException, URISyntaxException, IOException {

        final File workingArea = this.temporaryFolder.newFolder("working-area");
        final File invalidPersistedTopologyFile = new File(workingArea, "topology.xml");
        assertTrue("Unable to create the invalid persisted topology file", invalidPersistedTopologyFile.mkdir());
        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        final ContainerConfiguration containerConfig = new ContainerConfiguration();
        containerConfig.setWorkDirectoryPath(workingArea.getAbsolutePath());
        ReflectionHelper.setPrivateField(cfgSvc, CONTAINER_CONFIG_FIELD, containerConfig);
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        try {
            cfgSvc.persistTopology(ConfigurationTestUtils.createBasicSampleTopology());

        } catch (final ConfigurationException e) {
            assertTrue("Unexpected configuration exception",
                    e.getMessage().contains(invalidPersistedTopologyFile.getAbsolutePath()));
            throw e;
        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check the topology persistence against:
     * <ul>
     * <li>the working area is valid,</li>
     * <li>a persisted topology already exists and is valid,</li>
     * <li>a sample topology is provided</li>
     * </ul>
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>No error occurs</li>
     * <li>the topology is correctly persisted</li>
     * </ul>
     * </p>
     */
    @Test
    public void testPersistedTopologyFileAlreadyExist() throws ConfigurationException, TopologyException, IOException,
            InvocationTargetException {

        final File workingArea = this.temporaryFolder.newFolder("working-area");
        final File persistedTopologyFile = new File(workingArea, "topology.xml");
        assertTrue("Unable to create the valid persisted topology file", persistedTopologyFile.createNewFile());
        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        final ContainerConfiguration containerConfig = new ContainerConfiguration();
        containerConfig.setWorkDirectoryPath(workingArea.getAbsolutePath());
        ReflectionHelper.setPrivateField(cfgSvc, CONTAINER_CONFIG_FIELD, containerConfig);
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        try {
            cfgSvc.persistTopology(ConfigurationTestUtils.createBasicSampleTopology());

            final Topology readTopology = TopologyBuilder.createTopology(persistedTopologyFile);
            ConfigurationTestUtils.assertBasicSampleTopology(readTopology);

        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check the topology persistence against:
     * <ul>
     * <li>the working area is valid,</li>
     * <li>no persisted topology exists in the working area,</li>
     * <li>a sample topology is provided</li>
     * </ul>
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>No error occurs</li>
     * <li>the topology is correctly persisted</li>
     * </ul>
     * </p>
     */
    @Test
    public void testPersistedTopologyFileDoesNotExist() throws ConfigurationException, TopologyException, IOException,
            InvocationTargetException {

        final File workingArea = this.temporaryFolder.newFolder("working-area");
        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        final ContainerConfiguration containerConfig = new ContainerConfiguration();
        containerConfig.setWorkDirectoryPath(workingArea.getAbsolutePath());
        ReflectionHelper.setPrivateField(cfgSvc, CONTAINER_CONFIG_FIELD, containerConfig);
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        try {
            cfgSvc.persistTopology(ConfigurationTestUtils.createBasicSampleTopology());

            final File persistedTopologyFile = new File(workingArea, "topology.xml");
            final Topology readTopology = TopologyBuilder.createTopology(persistedTopologyFile);
            ConfigurationTestUtils.assertBasicSampleTopology(readTopology);

        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

}
