/**
 * Copyright (c) 2014-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.configuration;

import static org.junit.Assert.assertEquals;

import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.List;
import java.util.Properties;

import org.junit.Rule;
import org.junit.Test;
import org.ow2.petals.clientserverapi.configuration.ContainerConfiguration;
import org.ow2.petals.microkernel.api.configuration.ConfigurationService;
import org.ow2.petals.microkernel.api.configuration.DomainConfiguration;
import org.ow2.petals.microkernel.api.configuration.DomainConfiguration.DomainMode;
import org.ow2.petals.microkernel.api.configuration.exception.ConfigurationException;
import org.ow2.petals.topology.generated.Topology;

/**
 * Unit tests of {@link ConfigurationServiceImpl#reconfigureDomain(DomainConfiguration)}
 * 
 * @author Christophe DENEUX - Linagora
 * 
 */
public class ConfigurationServiceImpl_ReconfigureDomainTest {

    private final static String DOMAIN_NAME_1 = "domain-1";

    private final static String DOMAIN_DESCRIPTION_1 = "A sample domain 1 configuration";

    private final static DomainMode DOMAIN_MODE_1 = DomainMode.STATIC;

    private final static List<Object> DOMAIN_EXTRA_PARAM_1 = new ArrayList<Object>();

    @Rule
    public final LocalServerPropertiesBuilder localServerPropertiesBuilder = new LocalServerPropertiesBuilder();

    /**
     * <p>
     * Check the reconfiguration of a domain against:
     * <ul>
     * <li>all is ok</li>
     * </ul>
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>No error occurs.</li>
     * <li>the new domain name is set in the current domain configuration</li>
     * <li>the new domain description is set in the current domain configuration</li>
     * </ul>
     * </p>
     */
    @Test
    public void testAllIsOk() throws ConfigurationException, InvocationTargetException, IOException {

        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);

        try {
            final Topology topology = ConfigurationTestUtils.createBasicSampleTopology();

            cfgSvc.loadConfiguration(serverProps, topology);

            final DomainConfiguration newDomainConfiguration = new DomainConfiguration();
            newDomainConfiguration.setName(DOMAIN_NAME_1);
            newDomainConfiguration.setDescription(DOMAIN_DESCRIPTION_1);
            newDomainConfiguration.setMode(DOMAIN_MODE_1);
            newDomainConfiguration.setExtraParameters(DOMAIN_EXTRA_PARAM_1);

            cfgSvc.reconfigureDomain(newDomainConfiguration);
            
            final DomainConfiguration reconfiguredDomain = cfgSvc.getDomainConfiguration();
            assertEquals("Unreconfigured domain name", DOMAIN_NAME_1, reconfiguredDomain.getName());
            assertEquals("Unreconfigured domain description", DOMAIN_DESCRIPTION_1, reconfiguredDomain.getDescription());
            assertEquals("Unreconfigured domain mode", DOMAIN_MODE_1, reconfiguredDomain.getMode());
            assertEquals("Unreconfigured domain extra param", DOMAIN_EXTRA_PARAM_1,
                    reconfiguredDomain.getExtraParameters());

            final ContainerConfiguration reconfiguredContainer = cfgSvc.getContainerConfiguration();
            assertEquals("Unreconfigured domain name", DOMAIN_NAME_1, reconfiguredContainer.getDomainName());

        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }
}
