/**
 * Copyright (c) 2014-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.configuration;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;

import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.util.Properties;

import org.junit.Rule;
import org.junit.Test;
import org.ow2.petals.clientserverapi.configuration.ContainerConfiguration;
import org.ow2.petals.microkernel.api.configuration.ConfigurationService;
import org.ow2.petals.microkernel.api.configuration.exception.ConfigurationException;
import org.ow2.petals.microkernel.api.configuration.exception.ContainerDoesNotExistException;
import org.ow2.petals.microkernel.api.configuration.exception.InvalidContainerConfigurationProvidedException;
import org.ow2.petals.topology.generated.Container;
import org.ow2.petals.topology.generated.Domain;
import org.ow2.petals.topology.generated.JmxService;
import org.ow2.petals.topology.generated.TransportService;

/**
 * Unit tests of {@link ConfigurationServiceImpl#removeContainerConfiguration(ContainerConfiguration)}
 * 
 * @author Christophe DENEUX - Linagora
 * 
 */
public class ConfigurationServiceImpl_RemoveContainerConfigurationTest extends AbstractConfigurationServiceImplTest {

    private final static String CONTAINER_NAME_1 = "sample-1";

    private final static String CONTAINER_DESCRIPTION_1 = "description of the container 1";

    @Rule
    public final LocalServerPropertiesBuilder localServerPropertiesBuilder = new LocalServerPropertiesBuilder();

    public ConfigurationServiceImpl_RemoveContainerConfigurationTest() {
        super();

        final Domain domain = this.topology.getDomain();

        final Container container1 = new Container();
        domain.getContainer().add(container1);
        container1.setName(CONTAINER_NAME_1);
        container1.setDescription(CONTAINER_DESCRIPTION_1);
        container1.setHost(ConfigurationTestUtils.CONTAINER_HOST);
        container1.setUser(ConfigurationTestUtils.CONTAINER_JMX_USER);
        container1.setPassword(ConfigurationTestUtils.CONTAINER_JMX_PWD);

        final JmxService jmxService1 = new JmxService();
        container1.setJmxService(jmxService1);

        final TransportService transportService1 = new TransportService();
        transportService1.setTcpPort(7781);
        container1.setTransportService(transportService1);
    }

    /**
     * <p>
     * Check the deletion of a container configuration against:
     * <ul>
     * <li>the container configuration provided has a name unset (ie. <code>null</code>)</li>
     * </ul>
     * </p>
     * <p>
     * Expected result: A error occurs because the container configuration is unknown.
     * </p>
     */
    @Test(expected = InvalidContainerConfigurationProvidedException.class)
    public void testContainerProvidedHasANullName() throws ConfigurationException, InvocationTargetException,
            IOException {

        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);

        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        try {
            cfgSvc.loadConfiguration(serverProps, this.topology);

            final ContainerConfiguration containerToRemove = new ContainerConfiguration();
            cfgSvc.removeContainerConfiguration(containerToRemove);
        } catch (final InvalidContainerConfigurationProvidedException e) {
            assertTrue("Unexpected container name", e.getMessage().contains("has no name"));
            throw e;
        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check the deletion of a container configuration against:
     * <ul>
     * <li>the container configuration provided is unknown</li>
     * </ul>
     * </p>
     * <p>
     * Expected result: A error occurs because the container configuration is unknown.
     * </p>
     */
    @Test(expected = ContainerDoesNotExistException.class)
    public void testContainerProvidedIsUnknown() throws ConfigurationException, InvocationTargetException, IOException {

        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);

        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        final String containerNameToRemove = "unknown-container";
        try {
            cfgSvc.loadConfiguration(serverProps, this.topology);

            final ContainerConfiguration containerToRemove = new ContainerConfiguration();
            containerToRemove.setName(containerNameToRemove);
            cfgSvc.removeContainerConfiguration(containerToRemove);
        } catch (final ContainerDoesNotExistException e) {
            assertEquals("Unexpected container name", e.getUnexistingContainerName(), containerNameToRemove);

            assertNotNull("container 0 no more exists",
                    cfgSvc.getContainerConfiguration(ConfigurationTestUtils.CONTAINER_NAME_0));
            assertNotNull("container 1 no more exists", cfgSvc.getContainerConfiguration(CONTAINER_NAME_1));
            assertEquals("unexpected container number", cfgSvc.getContainersConfiguration().size(), 2);
            throw e;
        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check the deletion of a container configuration against:
     * <ul>
     * <li>the container configuration provided is associated to the current container</li>
     * </ul>
     * </p>
     * <p>
     * Expected result: A error occurs because the current container configuration cannot be removed.
     * </p>
     */
    @Test(expected = InvalidContainerConfigurationProvidedException.class)
    public void testContainerProvidedIsCurrent() throws ConfigurationException, InvocationTargetException, IOException {

        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);

        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        try {
            cfgSvc.loadConfiguration(serverProps, this.topology);

            final ContainerConfiguration containerToRemove = new ContainerConfiguration();
            containerToRemove.setName(ConfigurationTestUtils.CONTAINER_NAME_0);
            cfgSvc.removeContainerConfiguration(containerToRemove);
        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

    /**
     * <p>
     * Check the deletion of a container configuration against:
     * <ul>
     * <li>the topology contains 2 containers</li>
     * <li>the container configuration provided is valid</li>
     * </ul>
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>No error occurs</li>
     * <li>The container is correctly removed</li>
     * <li>Only one container remains inside the topoogy</li>
     * </p>
     */
    @Test
    public void testContainerProvidedIsKnown() throws ConfigurationException, InvocationTargetException, IOException {

        final Properties serverProps = this.localServerPropertiesBuilder
                .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);

        final ConfigurationService cfgSvc = new ConfigurationServiceImpl();
        ConfigurationTestUtils.startConfigurationService(cfgSvc);

        try {
            cfgSvc.loadConfiguration(serverProps, this.topology);

            final ContainerConfiguration containerToRemove = new ContainerConfiguration();
            containerToRemove.setName(CONTAINER_NAME_1);
            cfgSvc.removeContainerConfiguration(containerToRemove);

            assertNotNull("container 0 no more exists",
                    cfgSvc.getContainerConfiguration(ConfigurationTestUtils.CONTAINER_NAME_0));
            assertNull("container 1 exists always", cfgSvc.getContainerConfiguration(CONTAINER_NAME_1));
            assertEquals("unexpected container number", cfgSvc.getContainersConfiguration().size(), 1);

        } finally {
            ConfigurationTestUtils.stopConfigurationService(cfgSvc);
        }
    }

}
