/**
 * Copyright (c) 2014-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.container;

import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;

import org.junit.Test;
import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.fractal.api.control.ContentController;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalContentException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.api.factory.InstantiationException;
import org.objectweb.fractal.util.Fractal;
import org.ow2.petals.jbi.descriptor.original.generated.ComponentType;
import org.ow2.petals.jbi.descriptor.original.generated.Jbi;
import org.ow2.petals.microkernel.api.container.ContainerService;
import org.ow2.petals.microkernel.api.container.Installer;
import org.ow2.petals.microkernel.api.server.FractalHelper;
import org.ow2.petals.microkernel.container.utest.DummyInstallerImpl;

import com.ebmwebsourcing.easycommons.lang.reflect.ReflectionHelper;

/**
 * Unit tests of {@link ContainerServiceImpl}.
 * 
 * @author Christophe DENEUX - Linagora
 */
public class ContainerServiceImplTest {

    /**
     * Name of the Fractal root composite
     */
    private final static String FRACTAL_ROOT_COMPOSITE_NAME = "org.ow2.petals.microkernel.container.ContainerServiceRoot";

    private Component getContainerCompositeStarted() throws NoSuchInterfaceException, IllegalLifeCycleException,
            InstantiationException {
        final Component containerComposite = FractalHelper.createCompositeComponent(FRACTAL_ROOT_COMPOSITE_NAME);
        FractalHelper.startComponent(containerComposite);

        return containerComposite;
    }

    private ContainerService getContainerService(final Component containerComposite) throws NoSuchInterfaceException {
        final ContentController containerCompositeCC = Fractal.getContentController(containerComposite);
        final Component containerServiceComp = FractalHelper.getRecursiveComponentByName(containerCompositeCC,
                FractalHelper.CONTAINER_SERVICE_COMPONENT);
        assertNotNull("Fractal component 'ContainerService' not found", containerServiceComp);
        final ContainerService containerService = (ContainerService) containerServiceComp
                .getFcInterface(ContainerService.FRACTAL_SRV_ITF_NAME);
        assertNotNull("Server interface of Fractal component 'ContainerService' not found", containerService);
        return containerService;
    }

    /**
     * @param containerComposite
     * @param componentName
     * @return The {@link Installer} implementation not wrapped by Fractal
     */
    private Component addAndStartInstaller(final Component containerComposite, final String componentName)
            throws NoSuchInterfaceException, IllegalContentException, IllegalLifeCycleException,
            IllegalBindingException, InstantiationException {

        // Creation of the installer
        final Component installerComponent = FractalHelper.createPrimitiveComponent(DummyInstallerImpl.class);

        // Add the new installer in the composite 'Container' ...
        FractalHelper.addComponent(installerComponent, containerComposite, null,
                ContainerController.PREFIX_INSTALLER_NAME + componentName);

        // ... and bind the installer to the container service
        final ContentController containerCompositeCC = Fractal.getContentController(containerComposite);
        final Component containerServiceComp = FractalHelper.getRecursiveComponentByName(containerCompositeCC,
                FractalHelper.CONTAINER_SERVICE_COMPONENT);
        final BindingController containerServiceComponentBC = Fractal.getBindingController(containerServiceComp);
        containerServiceComponentBC.bindFc(ContainerService.INSTALLERS_ITF_PREFIX + componentName,
                installerComponent.getFcInterface(Installer.FRACTAL_SRV_ITF_NAME));

        FractalHelper.startComponent(installerComponent);

        return installerComponent;
    }

    /**
     * <p>
     * Check {@link ContainerServiceImpl#isBindingComponent(String)} when no installer was loaded.
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>no error occurs,</li>
     * <li>the value <code>false</code> is returned.</li>
     * </ul>
     * </p>
     */
    @Test
    public void IsBinding_NoInstallerLoaded() throws NoSuchInterfaceException, IllegalLifeCycleException,
            InstantiationException {

        final ContainerService containerService = this.getContainerService(this.getContainerCompositeStarted());
        assertFalse(containerService.isBindingComponent("no-installer-loaded"));
    }


    /**
     * <p>
     * Check {@link ContainerServiceImpl#isBindingComponent(String)} when no installer was loaded for the requested
     * component, but other component installers are loaded.
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>no error occurs,</li>
     * <li>the value <code>false</code> is returned.</li>
     * </ul>
     * </p>
     */
    @Test
    public void IsBinding_NoInstallerLoadedForRequestedComponent() throws NoSuchInterfaceException,
            IllegalLifeCycleException, InstantiationException, IllegalContentException, IllegalBindingException {

        final Component containerComposite = this.getContainerCompositeStarted();
        this.addAndStartInstaller(containerComposite, "a-component-loaded");
        final ContainerService containerService = this.getContainerService(containerComposite);
        assertFalse(containerService.isBindingComponent("no-installer-loaded"));
    }

    /**
     * <p>
     * Check {@link ContainerServiceImpl#isBindingComponent(String)} when:
     * <ul>
     * <li>only the component installer is loaded,</li>
     * <li>the component is a binding component.</li>
     * </ul>
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>no error occurs,</li>
     * <li>the value <code>true</code> is returned.</li>
     * </ul>
     * </p>
     */
    @Test
    public void IsBinding_InstallerLoaded() throws NoSuchInterfaceException, IllegalLifeCycleException,
            InstantiationException, IllegalContentException, IllegalBindingException {

        final Component containerComposite = this.getContainerCompositeStarted();
        final String componentName = "a-component-loaded";
        final Component installer = this.addAndStartInstaller(containerComposite, componentName);
        final Jbi jbiDescriptor = new Jbi();
        final org.ow2.petals.jbi.descriptor.original.generated.Component jbiComponent = new org.ow2.petals.jbi.descriptor.original.generated.Component();
        jbiDescriptor.setComponent(jbiComponent);
        jbiComponent.setType(ComponentType.BINDING_COMPONENT);
        ReflectionHelper.setPrivateField(installer.getFcInterface("/content"), "jbiDescriptor", jbiDescriptor);

        final ContainerService containerService = this.getContainerService(containerComposite);
        assertTrue(containerService.isBindingComponent(componentName));
    }

    /**
     * <p>
     * Check {@link ContainerServiceImpl#isBindingComponent(String)} when:
     * <ul>
     * <li>only the component installer is loaded,</li>
     * <li>the component is a service engine.</li>
     * </ul>
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>no error occurs,</li>
     * <li>the value <code>false</code> is returned.</li>
     * </ul>
     * </p>
     */
    @Test
    public void IsBinding_InstallerLoadedButServiceEngine() throws NoSuchInterfaceException, IllegalLifeCycleException,
            InstantiationException, IllegalContentException, IllegalBindingException {

        final Component containerComposite = this.getContainerCompositeStarted();
        final String componentName = "a-component-loaded";
        final Component installer = this.addAndStartInstaller(containerComposite, componentName);
        final Jbi jbiDescriptor = new Jbi();
        final org.ow2.petals.jbi.descriptor.original.generated.Component jbiComponent = new org.ow2.petals.jbi.descriptor.original.generated.Component();
        jbiDescriptor.setComponent(jbiComponent);
        jbiComponent.setType(ComponentType.SERVICE_ENGINE);
        ReflectionHelper.setPrivateField(installer.getFcInterface("/content"), "jbiDescriptor", jbiDescriptor);

        final ContainerService containerService = this.getContainerService(containerComposite);
        assertFalse(containerService.isBindingComponent(componentName));
    }

    /**
     * <p>
     * Check {@link ContainerServiceImpl#isServiceEngine(String)} when no installer was loaded.
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>no error occurs,</li>
     * <li>the value <code>false</code> is returned.</li>
     * </ul>
     * </p>
     */
    @Test
    public void IsEngine_NoInstallerLoaded() throws NoSuchInterfaceException, IllegalLifeCycleException,
            InstantiationException {

        final ContainerService containerService = this.getContainerService(this.getContainerCompositeStarted());
        assertFalse(containerService.isServiceEngine("no-installer-loaded"));
    }

    /**
     * <p>
     * Check {@link ContainerServiceImpl#isServiceEngine(String)} when no installer was loaded for the requested
     * component, but other component installers are loaded.
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>no error occurs,</li>
     * <li>the value <code>false</code> is returned.</li>
     * </ul>
     * </p>
     */
    @Test
    public void IsEngine_NoInstallerLoadedForRequestedComponent() throws NoSuchInterfaceException,
            IllegalLifeCycleException, InstantiationException, IllegalContentException, IllegalBindingException {

        final Component containerComposite = this.getContainerCompositeStarted();
        this.addAndStartInstaller(containerComposite, "a-component-loaded");
        final ContainerService containerService = this.getContainerService(containerComposite);
        assertFalse(containerService.isServiceEngine("no-installer-loaded"));
    }

    /**
     * <p>
     * Check {@link ContainerServiceImpl#isServiceEngine(String)} when:
     * <ul>
     * <li>only the component installer is loaded,</li>
     * <li>the component is a service engine.</li>
     * </ul>
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>no error occurs,</li>
     * <li>the value <code>true</code> is returned.</li>
     * </ul>
     * </p>
     */
    @Test
    public void IsEngine_InstallerLoaded() throws NoSuchInterfaceException, IllegalLifeCycleException,
            InstantiationException, IllegalContentException, IllegalBindingException {

        final Component containerComposite = this.getContainerCompositeStarted();
        final String componentName = "a-component-loaded";
        final Component installer = this.addAndStartInstaller(containerComposite, componentName);
        final Jbi jbiDescriptor = new Jbi();
        final org.ow2.petals.jbi.descriptor.original.generated.Component jbiComponent = new org.ow2.petals.jbi.descriptor.original.generated.Component();
        jbiDescriptor.setComponent(jbiComponent);
        jbiComponent.setType(ComponentType.SERVICE_ENGINE);
        ReflectionHelper.setPrivateField(installer.getFcInterface("/content"), "jbiDescriptor", jbiDescriptor);

        final ContainerService containerService = this.getContainerService(containerComposite);
        assertTrue(containerService.isServiceEngine(componentName));
    }

    /**
     * <p>
     * Check {@link ContainerServiceImpl#isServiceEngine(String)} when:
     * <ul>
     * <li>only the component installer is loaded,</li>
     * <li>the component is a service engine.</li>
     * </ul>
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>no error occurs,</li>
     * <li>the value <code>false</code> is returned.</li>
     * </ul>
     * </p>
     */
    @Test
    public void IsEngine_InstallerLoadedButBindingComponent() throws NoSuchInterfaceException,
            IllegalLifeCycleException, InstantiationException, IllegalContentException, IllegalBindingException {

        final Component containerComposite = this.getContainerCompositeStarted();
        final String componentName = "a-component-loaded";
        final Component installer = this.addAndStartInstaller(containerComposite, componentName);
        final Jbi jbiDescriptor = new Jbi();
        final org.ow2.petals.jbi.descriptor.original.generated.Component jbiComponent = new org.ow2.petals.jbi.descriptor.original.generated.Component();
        jbiDescriptor.setComponent(jbiComponent);
        jbiComponent.setType(ComponentType.BINDING_COMPONENT);
        ReflectionHelper.setPrivateField(installer.getFcInterface("/content"), "jbiDescriptor", jbiDescriptor);

        final ContainerService containerService = this.getContainerService(containerComposite);
        assertFalse(containerService.isServiceEngine(componentName));
    }

}
