/**
 * Copyright (c) 2015-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.container.lifecycle.util;

import java.net.URL;
import java.net.URLClassLoader;
import java.util.List;
import java.util.concurrent.Callable;
import java.util.concurrent.CompletionService;
import java.util.concurrent.ExecutorCompletionService;
import java.util.concurrent.Executors;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.logging.Logger;

import javax.jbi.JBIException;

import org.junit.Assert;
import org.junit.Before;
import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.ExpectedException;

import com.ebmwebsourcing.easycommons.lang.reflect.ReflectionHelper;
import com.ebmwebsourcing.easycommons.log.LoggingUtil;

/**
 * 
 * @author vnoel
 *
 */
public class SeparateIsolatingThreadTest extends Assert {

    @Rule
    public ExpectedException exception = ExpectedException.none();

    private static final String THREAD_NAME = "Test Thread";

    private SeparateIsolatingThread separateThread;

    private URLClassLoader urlClassLoader;

    @Before
    public void setup() {
        urlClassLoader = new URLClassLoader(new URL[] {});
        separateThread = new SeparateIsolatingThread(THREAD_NAME, 1000, new LoggingUtil(Logger.getAnonymousLogger()));
    }

    public void teardown() {
        separateThread.shutdown();
    }

    private ThreadPoolExecutor getPool() {
        return (ThreadPoolExecutor) ReflectionHelper.getFieldValue(SeparateIsolatingThread.class, separateThread,
                "executor", false);
    }

    @Test(timeout = 3000L)
    public void testConfiguration() throws Exception {
        separateThread.execute(new Callable<Object>() {
            @Override
            public Object call() throws Exception {
                assertEquals(THREAD_NAME, Thread.currentThread().getName());
                assertEquals(urlClassLoader, Thread.currentThread().getContextClassLoader());
                return null;
            }
        }, urlClassLoader);

        final ThreadPoolExecutor executor = getPool();
        final List<Runnable> remainingTasks = executor.shutdownNow();
        assertEquals(0, remainingTasks.size());
    }

    @Test(timeout = 3000L)
    public void testTimeout() throws Exception {
        exception.expect(JBIException.class);
        exception.expectMessage("Timeout exceeded");
        separateThread.execute(new Callable<Object>() {
            @Override
            public Object call() throws Exception {
                Thread.sleep(2000);
                return null;
            }
        }, urlClassLoader);
        // we should never arrive here because execute blocks
        fail();
    }

    @Test(timeout = 10000L)
    public void testManyTasks() throws Exception {

        final int loopMax = 10;

        // let's use 10 thread to simulate multi-threaded call to the executor
        final CompletionService<Object> ecs = new ExecutorCompletionService<Object>(
                Executors.newFixedThreadPool(10));
        
        // this is used to verify there is only one thread executed at a time
        final AtomicBoolean taken = new AtomicBoolean(false);
        
        for (int i = 1; i <= loopMax; i++) {
            ecs.submit(new Callable<Object>() {
                @Override
                public Object call() throws Exception {
                    separateThread.execute(new Callable<Object>() {
                        @Override
                        public Object call() throws Exception {
                            // this test that there is always only one task executed at once
                            assertFalse(taken.getAndSet(true));
                            // let's wait to clog up the execution and be sure some tasks wait more than the timeout in
                            // the queue (which is ok)
                            Thread.sleep(500);
                            assertTrue(taken.getAndSet(false));
                            return null;
                        }
                    }, urlClassLoader);
                    return null;
                }
            });
        }

        // wait for completion of the tasks
        for (int i = 1; i <= loopMax; i++) {
            ecs.take().get();
        }
    }

}
