/**
 * Copyright (c) 2014-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.extension;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

import java.io.IOException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Properties;

import org.junit.Rule;
import org.junit.Test;
import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.control.ContentController;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.util.Fractal;
import org.ow2.petals.microkernel.api.configuration.ConfigurationService;
import org.ow2.petals.microkernel.api.extension.PetalsExtensionController;
import org.ow2.petals.microkernel.api.extension.PetalsPreExtensionController;
import org.ow2.petals.microkernel.api.server.FractalHelper;
import org.ow2.petals.microkernel.configuration.ConfigurationTestUtils;
import org.ow2.petals.microkernel.configuration.LocalServerPropertiesBuilder;
import org.ow2.petals.microkernel.extension.utest.PreExtensionController;
import org.ow2.petals.microkernel.extension.utest.PreExtensionWithDependencyController;
import org.ow2.petals.topology.generated.Topology;

/**
 * Unit tests of {@link PreExtensionsManagerImpl}
 * 
 * @author Christophe DENEUX - Linagora
 * 
 */
public class PreExtensionsManagerImplTest {

    /**
     * Name of the Fractal root composite
     */
    private final static String FRACTAL_ROOT_COMPOSITE_NAME = "org.ow2.petals.microkernel.extension.PreExtensionManagerRoot";

    @Rule
    public final LocalServerPropertiesBuilder localServerPropertiesBuilder = new LocalServerPropertiesBuilder();

    /**
     * <p>
     * Check the loading of a pre-extension:
     * <ul>
     * <li>only-one pre-extension is available on the classpath,</li>
     * <li>the extension has no dependency,</li>
     * <li>the extension is activated.</li>
     * </ul>
     * </p>
     * <p>
     * Expected result: The implementation available on the classpath is correctly loaded.
     * </p>
     */
    @Test
    public void testStart_Nominal() throws Exception {

        final ClassLoader oldClassloader = Thread.currentThread().getContextClassLoader();
        Thread.currentThread().setContextClassLoader(new ClassLoader(oldClassloader) {

            @Override
            public Enumeration<URL> getResources(final String name) throws IOException {
                if (name.equals("META-INF/services/" + PetalsPreExtensionController.class.getName())) {
                    return super.getResources("services/utest/" + PetalsPreExtensionController.class.getName());
                } else {
                    return super.getResources(name);
                }
            }
        });

        try {

            final PreExtensionsManager preExtensionMgr = this
                    .getPreExtensionManagerStarted(this.localServerPropertiesBuilder
                            .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0));

            final Map<PetalsExtensionController, Component> extensionControllers = preExtensionMgr
                    .getExtensionControllers();
            assertNotNull("Extension controller list is null", extensionControllers);
            assertEquals("Unexpected number of pre-extensions loaded", 1, extensionControllers.size());
            final PetalsExtensionController extensionController = extensionControllers.keySet().toArray(
                    new PetalsExtensionController[1])[0];
            assertTrue("Unexpected extension controller", extensionController instanceof PreExtensionController);

        } finally {
            Thread.currentThread().setContextClassLoader(oldClassloader);
        }
    }

    /**
     * <p>
     * Check the loading of a pre-extension:
     * <ul>
     * <li>only-one pre-extension is available on the classpath,</li>
     * <li>the extension has no dependency,</li>
     * <li>the extension is not activated.</li>
     * </ul>
     * </p>
     * <p>
     * Expected result: No implementation is loaded.
     * </p>
     */
    @Test
    public void testStart_ExtensionNotActivated() throws Exception {

        final ClassLoader oldClassloader = Thread.currentThread().getContextClassLoader();
        Thread.currentThread().setContextClassLoader(new ClassLoader(oldClassloader) {

            @Override
            public Enumeration<URL> getResources(final String name) throws IOException {
                if (name.equals("META-INF/services/" + PetalsPreExtensionController.class.getName())) {
                    return super.getResources("services/utest/" + PetalsPreExtensionController.class.getName());
                } else {
                    return super.getResources(name);
                }
            }
        });

        try {

            final Properties serverProps = this.localServerPropertiesBuilder
                    .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0);
            serverProps.setProperty(PreExtensionController.ACTIVATION_PROPERTY_NAME, Boolean.FALSE.toString());

            final PreExtensionsManager preExtensionMgr = this.getPreExtensionManagerStarted(serverProps);

            final Map<PetalsExtensionController, Component> extensionControllers = preExtensionMgr
                    .getExtensionControllers();
            assertNotNull("Extension controller list is null", extensionControllers);
            assertEquals("Unexpected number of pre-extensions loaded", 0, extensionControllers.size());

        } finally {
            Thread.currentThread().setContextClassLoader(oldClassloader);
        }
    }
    
    /**
     * An {@link Enumeration} base on a {@link Iterator}
     * 
     * @author Christophe DENEUX - Linagora
     */
    private class IteratorEnum<E> implements Enumeration<E> {
        
        private final Iterator<E> iterator;
        
        public IteratorEnum(final Iterator<E> iterator) {
            this.iterator = iterator;
        }

        @Override
        public boolean hasMoreElements() {
            return this.iterator.hasNext();
        }

        @Override
        public E nextElement() {
            return this.iterator.next();
        }
        
    }

    /**
     * <p>
     * Check the loading of a pre-extension:
     * <ul>
     * <li>two pre-extensions are available on the classpath,</li>
     * <li>one extension has a dependency loadable,</li>
     * <li>the extensions are activated.</li>
     * </ul>
     * </p>
     * <p>
     * Expected result: The implementations available on the classpath are correctly loaded.
     * </p>
     */
    @Test
    public void testStart_DependencySatisfied() throws Exception {

        final ClassLoader oldClassloader = Thread.currentThread().getContextClassLoader();
        Thread.currentThread().setContextClassLoader(new ClassLoader(oldClassloader) {

            @Override
            public Enumeration<URL> getResources(final String name) throws IOException {
                if (name.equals("META-INF/services/" + PetalsPreExtensionController.class.getName())) {
                    final Enumeration<URL> enum1 = super.getResources("services/utest/dependency-satisfied/"
                            + PetalsPreExtensionController.class.getName());
                    final Enumeration<URL> enum2 = super.getResources("services/utest/"
                            + PetalsPreExtensionController.class.getName());
                    final List<URL> resources = new ArrayList<URL>();
                    while (enum1.hasMoreElements()) {
                        resources.add(enum1.nextElement());
                    }
                    while (enum2.hasMoreElements()) {
                        resources.add(enum2.nextElement());
                    }
                    return new IteratorEnum<URL>(resources.iterator());
                } else {
                    return super.getResources(name);
                }
            }
        });

        try {

            final PreExtensionsManager preExtensionMgr = this
                    .getPreExtensionManagerStarted(this.localServerPropertiesBuilder
                            .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0));

            final Map<PetalsExtensionController, Component> extensionControllers = preExtensionMgr
                    .getExtensionControllers();
            assertNotNull("Extension controller list is null", extensionControllers);
            assertEquals("Unexpected number of pre-extensions loaded", 2, extensionControllers.size());
            boolean pecFound = false;
            boolean pewdcFound = false;
            for (final PetalsExtensionController pec : extensionControllers.keySet()) {
                if (pec instanceof PreExtensionController) {
                    assertTrue("Unexpected extension name",
                            PreExtensionController.EXTENSION_NAME.equals(pec.getExtensionName()));
                    pecFound = true;
                } else if (pec instanceof PreExtensionWithDependencyController) {
                    assertTrue("Unexpected extension name",
                            PreExtensionWithDependencyController.EXTENSION_NAME.equals(pec.getExtensionName()));
                    pewdcFound = true;
                } else {
                    fail("Unexpected extension controller");
                }
            }
            assertTrue("PreExtensionController not loaded", pecFound);
            assertTrue("PreExtensionWithDependencyController not loaded", pewdcFound);

        } finally {
            Thread.currentThread().setContextClassLoader(oldClassloader);
        }
    }

    /**
     * <p>
     * Check the loading of a pre-extension:
     * <ul>
     * <li>only-one pre-extension is available on the classpath,</li>
     * <li>the extension has a dependency not loaded,</li>
     * <li>the extension is activated.</li>
     * </ul>
     * </p>
     * <p>
     * Expected result: An error occurs during the startup of the pre-extension manager because the dependency is not
     * satisfied
     * </p>
     */
    // TODO: Fractal does not permit to catch the real error 'NotSatisfiedDependenciesException' occurring in Fractal
    // component startup method. Investigate how this can be done.
    @Test(expected = IllegalLifeCycleException.class)
    public void testStart_DependencyNotSatisfied() throws Exception {

        final ClassLoader oldClassloader = Thread.currentThread().getContextClassLoader();
        Thread.currentThread().setContextClassLoader(new ClassLoader(oldClassloader) {

            @Override
            public Enumeration<URL> getResources(final String name) throws IOException {
                if (name.equals("META-INF/services/" + PetalsPreExtensionController.class.getName())) {
                    return super.getResources("services/utest/dependency-not-satisfied/"
                            + PetalsPreExtensionController.class.getName());
                } else {
                    return super.getResources(name);
                }
            }
        });

        try {

            this.getPreExtensionManagerStarted(this.localServerPropertiesBuilder
                    .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0));

            fail("No error occurs");

        } finally {
            Thread.currentThread().setContextClassLoader(oldClassloader);
        }
    }

    /**
     * <p>
     * Check the loading of a pre-extension:
     * <ul>
     * <li>only-one pre-extension is available on the classpath,</li>
     * <li>the extension has no dependency,</li>
     * <li>the extension is activated,</li>
     * <li>the extension throws an error during its creation,</li>
     * <li>the extension is designed to fail the Petals container startup.</li>
     * </ul>
     * </p>
     * <p>
     * Expected result: An error occurs during the startup of the pre-extension manager because the dependency is not
     * satisfied
     * </p>
     */
    // TODO: Fractal does not permit to catch the real error 'PetalsExtensionException' occurring in Fractal
    // component startup method. Investigate how this can be done.
    @Test(expected = IllegalLifeCycleException.class)
    public void testStart_ErrorOnCreationFaillingContainer() throws Exception {

        final ClassLoader oldClassloader = Thread.currentThread().getContextClassLoader();
        Thread.currentThread().setContextClassLoader(new ClassLoader(oldClassloader) {

            @Override
            public Enumeration<URL> getResources(final String name) throws IOException {
                if (name.equals("META-INF/services/" + PetalsPreExtensionController.class.getName())) {
                    return super.getResources("services/utest/error-creating-extension-failling-container/"
                            + PetalsPreExtensionController.class.getName());
                } else {
                    return super.getResources(name);
                }
            }
        });

        try {
            this.getPreExtensionManagerStarted(this.localServerPropertiesBuilder
                    .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0));
        } finally {
            Thread.currentThread().setContextClassLoader(oldClassloader);
        }
    }

    /**
     * <p>
     * Check the loading of a pre-extension:
     * <ul>
     * <li>only-one pre-extension is available on the classpath,</li>
     * <li>the extension has no dependency,</li>
     * <li>the extension is activated,</li>
     * <li>the extension throws an error during its start,</li>
     * <li>the extension is designed to fail the Petals container startup.</li>
     * </ul>
     * </p>
     * <p>
     * Expected result: An error occurs during the startup of the pre-extension manager because the dependency is not
     * satisfied
     * </p>
     */
    // TODO: Fractal does not permit to catch the real error 'PetalsExtensionException' occurring in Fractal
    // component startup method. Investigate how this can be done.
    @Test(expected = IllegalLifeCycleException.class)
    public void testStart_ErrorOnStartupFaillingContainer() throws Exception {

        final ClassLoader oldClassloader = Thread.currentThread().getContextClassLoader();
        Thread.currentThread().setContextClassLoader(new ClassLoader(oldClassloader) {

            @Override
            public Enumeration<URL> getResources(final String name) throws IOException {
                if (name.equals("META-INF/services/" + PetalsPreExtensionController.class.getName())) {
                    return super.getResources("services/utest/error-starting-extension-failling-container/"
                            + PetalsPreExtensionController.class.getName());
                } else {
                    return super.getResources(name);
                }
            }
        });

        try {
            this.getPreExtensionManagerStarted(this.localServerPropertiesBuilder
                    .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0));

        } finally {
            Thread.currentThread().setContextClassLoader(oldClassloader);
        }
    }

    /**
     * <p>
     * Check the loading of a pre-extension:
     * <ul>
     * <li>only-one pre-extension is available on the classpath,</li>
     * <li>the extension has no dependency,</li>
     * <li>the extension is activated,</li>
     * <li>the extension throws an error during its creation,</li>
     * <li>the extension is designed to not fail the Petals container startup.</li>
     * </ul>
     * </p>
     * <p>
     * Expected result: An error occurs during the startup of the pre-extension manager because the dependency is not
     * satisfied
     * </p>
     */
    // TODO: Fractal does not permit to catch the real error 'PetalsExtensionException' occurring in Fractal
    // component startup method. Investigate how this can be done.
    @Test
    public void testStart_ErrorOnCreationNotFaillingContainer() throws Exception {

        final ClassLoader oldClassloader = Thread.currentThread().getContextClassLoader();
        Thread.currentThread().setContextClassLoader(new ClassLoader(oldClassloader) {

            @Override
            public Enumeration<URL> getResources(final String name) throws IOException {
                if (name.equals("META-INF/services/" + PetalsPreExtensionController.class.getName())) {
                    return super.getResources("services/utest/error-creating-extension-not-failing-container/"
                            + PetalsPreExtensionController.class.getName());
                } else {
                    return super.getResources(name);
                }
            }
        });

        try {
            final PreExtensionsManager preExtensionMgr = this
                    .getPreExtensionManagerStarted(this.localServerPropertiesBuilder
                            .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0));

            final Map<PetalsExtensionController, Component> extensionControllers = preExtensionMgr
                    .getExtensionControllers();
            assertNotNull("Extension controller list is null", extensionControllers);
            assertEquals("Unexpected number of pre-extensions loaded", 0, extensionControllers.size());

        } finally {
            Thread.currentThread().setContextClassLoader(oldClassloader);
        }
    }

    /**
     * <p>
     * Check the loading of a pre-extension:
     * <ul>
     * <li>only-one pre-extension is available on the classpath,</li>
     * <li>the extension has no dependency,</li>
     * <li>the extension is activated,</li>
     * <li>the extension throws an error during its start,</li>
     * <li>the extension is designed to not fail the Petals container startup.</li>
     * </ul>
     * </p>
     * <p>
     * Expected result: An error occurs during the startup of the pre-extension manager because the dependency is not
     * satisfied
     * </p>
     */
    // TODO: Fractal does not permit to catch the real error 'PetalsExtensionException' occurring in Fractal
    // component startup method. Investigate how this can be done.
    @Test
    public void testStart_ErrorOnStartupNotFaillingContainer() throws Exception {

        final ClassLoader oldClassloader = Thread.currentThread().getContextClassLoader();
        Thread.currentThread().setContextClassLoader(new ClassLoader(oldClassloader) {

            @Override
            public Enumeration<URL> getResources(final String name) throws IOException {
                if (name.equals("META-INF/services/" + PetalsPreExtensionController.class.getName())) {
                    return super.getResources("services/utest/error-starting-extension-not-failling-container/"
                            + PetalsPreExtensionController.class.getName());
                } else {
                    return super.getResources(name);
                }
            }
        });

        try {
            final PreExtensionsManager preExtensionMgr = this.getPreExtensionManagerStarted(this.localServerPropertiesBuilder
                    .createMinimalServerProperties(ConfigurationTestUtils.CONTAINER_NAME_0));

            final Map<PetalsExtensionController, Component> extensionControllers = preExtensionMgr
                    .getExtensionControllers();
            assertNotNull("Extension controller list is null", extensionControllers);
            assertEquals("Unexpected number of pre-extensions loaded", 0, extensionControllers.size());

        } finally {
            Thread.currentThread().setContextClassLoader(oldClassloader);
        }
    }

    /**
     * <p>
     * Instantiates and starts the Fractal environment, including:
     * <ul>
     * <li>{@link ConfigurationService}</li>
     * <li>and {@link PreExtensionsManagerImpl}</li>
     * </ul>
     * </p>
     * 
     * @param serverProps
     *            The server properties (ie. server.properties)
     * @return The pre-extension manager
     */
    private final PreExtensionsManager getPreExtensionManagerStarted(final Properties serverProps) throws Exception {
        final Component rootComposite = FractalHelper.createCompositeComponent(FRACTAL_ROOT_COMPOSITE_NAME);
        final ContentController rootCompositeCC = Fractal.getContentController(rootComposite);
        final Component configurationComp = FractalHelper.getComponentByName(rootCompositeCC,
                FractalHelper.CONFIGURATION_COMPONENT);
        assertNotNull("Fractal component 'Configuration' not found", configurationComp);
        final ConfigurationService configurationService = (ConfigurationService) configurationComp
                .getFcInterface(ConfigurationService.FRACTAL_SRV_ITF_NAME);

        // We start manually the components to be able to inject the right configuration
        // First, the configuration service is started ...
        FractalHelper.startComponent(configurationComp);
        final Topology basicSampleTopology = ConfigurationTestUtils.createBasicSampleTopology();
        configurationService.loadConfiguration(serverProps, basicSampleTopology);

        // ... Second, the pre-extension manager is started
        final Component preExtensionMgrComp = FractalHelper.getRecursiveComponentByName(rootCompositeCC,
                FractalHelper.PRE_EXTENSIONSMANAGER_COMPONENT);
        assertNotNull("Fractal component 'pre-extension manager ' not found", preExtensionMgrComp);
        FractalHelper.startComponent(preExtensionMgrComp);

        final PreExtensionsManager preExtensionMgr = (PreExtensionsManager) preExtensionMgrComp
                .getFcInterface(PetalsPreExtensionController.FRACTAL_SRV_ITF_NAME);

        assertNotNull("Implementation of Fractal component 'pre-extension manager ' not found", preExtensionMgr);
        return preExtensionMgr;
    }

}
