/**
 * Copyright (c) 2005-2012 Fossil E-Commerce, 2005-2012 EBM WebSourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.jbi.management.installation;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;

import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.easymock.EasyMock;
import org.easymock.IAnswer;
import org.junit.Before;
import org.junit.Ignore;
import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.TemporaryFolder;
import org.ow2.petals.clientserverapi.configuration.ContainerConfiguration;
import org.ow2.petals.microkernel.api.configuration.ConfigurationService;
import org.ow2.petals.microkernel.api.container.ContainerService;
import org.ow2.petals.microkernel.api.container.Installer;
import org.ow2.petals.microkernel.api.container.SharedLibraryBean;
import org.ow2.petals.microkernel.api.container.SharedLibraryLifeCycle;
import org.ow2.petals.microkernel.api.extension.InstallationExtension;
import org.ow2.petals.microkernel.api.jbi.management.AdminService;
import org.ow2.petals.microkernel.api.jbi.management.InstallationService;
import org.ow2.petals.microkernel.extension.InstallationExtensionsManager;

import com.ebmwebsourcing.easycommons.lang.reflect.ReflectionHelper;

/**
 * Unit test of class {@link InstallationServiceImpl}.
 * 
 * @author Christophe DENEUX - Linagora
 */
public class InstallationServiceImplTest {
    
    /**
     * The installation service to test, implemented with {@link InstallationServiceImpl}. It's re-created before each test.
     */
    InstallationService instalationService;

    @Rule
    public TemporaryFolder testFolder = new TemporaryFolder();

    @Before
    public void createInstallationService() throws Exception {

        final ConfigurationService configurationServiceMock = EasyMock
                .createMock(ConfigurationService.class);
        EasyMock.expect(configurationServiceMock.getContainerConfiguration()).andAnswer(
                new IAnswer<ContainerConfiguration>() {

                    @Override
                    public ContainerConfiguration answer() throws Throwable {
                        final ContainerConfiguration contCfg = new ContainerConfiguration();
                        contCfg.setWorkDirectoryPath(InstallationServiceImplTest.this.testFolder
                                .newFolder("work").getAbsolutePath());
                        return contCfg;
                    }
                });
        EasyMock.replay(configurationServiceMock);

        final InstallationExtensionsManager installationExtensionsMgrMock = EasyMock
                .createMock(InstallationExtensionsManager.class);
        EasyMock.expect(installationExtensionsMgrMock.getExtensions()).andAnswer(
                new IAnswer<List<InstallationExtension>>() {

                    @Override
                    public List<InstallationExtension> answer() throws Throwable {
                        return new ArrayList<InstallationExtension>();
                    }
                });
        EasyMock.replay(installationExtensionsMgrMock);

        this.instalationService = new InstallationServiceImpl();

        // Inject dependencies
        ReflectionHelper.setPrivateField(instalationService, "configurationService",
                configurationServiceMock);
        ReflectionHelper.setPrivateField(instalationService, "installationExtensionManager",
                installationExtensionsMgrMock);

        // Start the system state implementation
        final Method methodStart = instalationService.getClass().getDeclaredMethod("start");
        methodStart.setAccessible(true);
        methodStart.invoke(instalationService);

        EasyMock.verify(installationExtensionsMgrMock);

    }

    /**
     * Test the method
     * {@link InstallationServiceImpl#uninstallAllNoMoreUsedSharedLibraries()}
     * where:
     * <ul>
     * <li>no SL was previously installed (empty map returned by
     * {@link AdminService#getSharedLibraries()}).</li>
     * </ul>
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>no error,</li>
     * <li>no SL uninstalled.</li>
     * </ul>
     * </p>
     */
    @Test
    public void testUninstallAllNoMoreUsedSharedLibraries_WithoutPreviouslyInstalledSL_EmptyMap() {
        
        final ContainerService containerServiceMock = EasyMock.createMock(ContainerService.class);
        EasyMock.expect(containerServiceMock.getSharedLibraries()).andReturn(
                new HashMap<SharedLibraryBean, SharedLibraryLifeCycle>());
        EasyMock.replay(containerServiceMock);

        // Inject dependencies
        ReflectionHelper.setPrivateField(this.instalationService, "containerService", containerServiceMock);

        final String[] uninstalledSls = this.instalationService
                .uninstallAllNoMoreUsedSharedLibraries();
        assertEquals("SL has been uninstalled", 0, uninstalledSls.length);

        EasyMock.verify(containerServiceMock);
    }

    /**
     * Test the method
     * {@link InstallationServiceImpl#uninstallAllNoMoreUsedSharedLibraries()}
     * where:
     * <ul>
     * <li>SLs was previously installed,</li>
     * <li>no component reference these libraries.</li>
     * </ul>
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>no error,</li>
     * <li>SL uninstalled are returned.</li>
     * </ul>
     * </p>
     */
    @Test
    @Ignore("Enable this test when mocking sl uninstallation will be easier.")
    public void testUninstallAllNoMoreUsedSharedLibraries_WithSLsNotReferenced() {

        final String slName1 = "sl-1";
        final String slVersion1 = "version-1";
        final SharedLibraryLifeCycle sllc1 = EasyMock.createMock(SharedLibraryLifeCycle.class);
        EasyMock.expect(sllc1.getRegisteredComponents()).andReturn(new ArrayList<String>());
        EasyMock.replay(sllc1);

        final String slName2 = "sl-2";
        final String slVersion2 = "version-2";
        final SharedLibraryLifeCycle sllc2 = EasyMock.createMock(SharedLibraryLifeCycle.class);
        EasyMock.expect(sllc2.getRegisteredComponents()).andReturn(null);
        EasyMock.replay(sllc2);

        final SharedLibraryLifeCycle sllc3 = EasyMock.createMock(SharedLibraryLifeCycle.class);
        EasyMock.expect(sllc3.getRegisteredComponents()).andReturn(null);
        EasyMock.replay(sllc3);

        final ContainerService containerServiceMock = EasyMock.createMock(ContainerService.class);
        EasyMock.expect(containerServiceMock.getSharedLibraries()).andAnswer(
                new IAnswer<Map<SharedLibraryBean, SharedLibraryLifeCycle>>() {

                    @Override
                    public Map<SharedLibraryBean, SharedLibraryLifeCycle> answer() throws Throwable {
                        final Map<SharedLibraryBean, SharedLibraryLifeCycle> sharedLibraries = new HashMap<SharedLibraryBean, SharedLibraryLifeCycle>();
                        sharedLibraries.put(new SharedLibraryBean(slName1, slVersion1), sllc1);
                        sharedLibraries.put(new SharedLibraryBean(slName2, slVersion2), sllc2);
                        sharedLibraries.put(new SharedLibraryBean(slName1, slVersion2), sllc3);
                        return sharedLibraries;
                    }
                });
        EasyMock.replay(containerServiceMock);

        // Inject dependencies
        ReflectionHelper.setPrivateField(this.instalationService, "adminService", containerServiceMock);

        final String[] uninstalledSls = this.instalationService
                .uninstallAllNoMoreUsedSharedLibraries();
        assertEquals("Uninstalled shared library number", 3, uninstalledSls.length);

        final List<String> lstUninstalledSls = Arrays.asList(uninstalledSls);
        assertTrue(slName1 + " not uninstalled", lstUninstalledSls.contains(slName1));
        assertTrue(slName2 + " not uninstalled", lstUninstalledSls.contains(slName2));

        EasyMock.verify(containerServiceMock);
    }

    /**
     * Test the method {@link InstallationServiceImpl#uninstallAllNoMoreUsedSharedLibraries()} where:
     * <ul>
     * <li>SLs was previously installed,</li>
     * <li>all SLs are referenced by components.</li>
     * </ul>
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>no error,</li>
     * <li>No SL are uninstalled.</li>
     * </ul>
     * </p>
     */
    @Test
    public void testUninstallAllNoMoreUsedSharedLibraries_WithSLsReferenced() {

        final String slName1 = "sl-1";
        final String slVersion1 = "version-1";
        final SharedLibraryLifeCycle sllc1 = EasyMock.createMock(SharedLibraryLifeCycle.class);
        EasyMock.expect(sllc1.getRegisteredComponents()).andAnswer(new IAnswer<List<String>>() {

            @Override
            public List<String> answer() throws Throwable {
                final List<String> components = new ArrayList<String>();
                components.add("comp-1");
                return components;
            }
        });
        EasyMock.replay(sllc1);

        final String slVersion2 = "version-2";
        final SharedLibraryLifeCycle sllc2 = EasyMock.createMock(SharedLibraryLifeCycle.class);
        EasyMock.expect(sllc2.getRegisteredComponents()).andAnswer(new IAnswer<List<String>>() {

            @Override
            public List<String> answer() throws Throwable {
                final List<String> components = new ArrayList<String>();
                components.add("comp-2");
                return components;
            }
        });
        EasyMock.replay(sllc2);

        final ContainerService containerServiceMock = EasyMock.createMock(ContainerService.class);
        EasyMock.expect(containerServiceMock.getSharedLibraries()).andAnswer(
                new IAnswer<Map<SharedLibraryBean, SharedLibraryLifeCycle>>() {

                    @Override
                    public Map<SharedLibraryBean, SharedLibraryLifeCycle> answer() throws Throwable {
                        final Map<SharedLibraryBean, SharedLibraryLifeCycle> sharedLibraries = new HashMap<SharedLibraryBean, SharedLibraryLifeCycle>();
                        sharedLibraries.put(new SharedLibraryBean(slName1, slVersion1), sllc1);
                        sharedLibraries.put(new SharedLibraryBean(slName1, slVersion2), sllc2);
                        return sharedLibraries;
                    }
                });
        EasyMock.replay(containerServiceMock);

        // Inject dependencies
        ReflectionHelper.setPrivateField(this.instalationService, "containerService", containerServiceMock);

        final String[] uninstalledSls = this.instalationService
                .uninstallAllNoMoreUsedSharedLibraries();
        assertEquals("Uninstalled shared library number", 0, uninstalledSls.length);

        EasyMock.verify(sllc1);
        EasyMock.verify(sllc2);
        EasyMock.verify(containerServiceMock);
    }

    /**
     * Test the method
     * {@link InstallationServiceImpl#shutdownAllNoMoreUsedComponents()()}
     * where:
     * <ul>
     * <li>no component was previously installed (an empty value is returned by
     * {@link AdminService#getInstallerList()}).</li>
     * </ul>
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>no error,</li>
     * <li>no SL uninstalled.</li>
     * </ul>
     * </p>
     */
    @Test
    public void testShutdownAllNoMoreUsedComponents_WithoutPreviouslyInstalledSL_Empty() {

        final ContainerService containerServiceMock = EasyMock.createMock(ContainerService.class);
        EasyMock.expect(containerServiceMock.getInstallers()).andReturn(new HashMap<String, Installer>());
        EasyMock.replay(containerServiceMock);

        // Inject dependencies
        ReflectionHelper.setPrivateField(this.instalationService, "containerService", containerServiceMock);

        final String[] shutdownComponents = this.instalationService
                .shutdownAllNoMoreUsedComponents();
        assertEquals("component has been shutdown", 0, shutdownComponents.length);

        EasyMock.verify(containerServiceMock);

    }

}
