/**
 * Copyright (c) 2011-2012 EBM WebSourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.jbi.management.task.download.jetty.server;

import java.util.logging.Logger;

import org.mortbay.jetty.Connector;
import org.mortbay.jetty.Handler;
import org.mortbay.jetty.Server;
import org.mortbay.jetty.handler.ContextHandlerCollection;
import org.mortbay.jetty.nio.SelectChannelConnector;
import org.mortbay.jetty.servlet.Context;
import org.mortbay.jetty.servlet.ServletHolder;
import org.mortbay.thread.BoundedThreadPool;
import org.ow2.petals.microkernel.jbi.management.task.download.jetty.servlets.CorrectContentTypeServlet;
import org.ow2.petals.microkernel.jbi.management.task.download.jetty.servlets.CorrectContentTypeWithCharsetServlet;
import org.ow2.petals.microkernel.jbi.management.task.download.jetty.servlets.WithoutContentTypeServlet;
import org.ow2.petals.microkernel.jbi.management.task.download.jetty.servlets.WrongContentTypeServlet;

/**
 * @author Nicolas Oddoux - EBM WebSourcing
 */
public class JettyServer {

    /**
     * The logger
     */
    private Logger log = Logger.getLogger(JettyServer.class.getName());

    public static final int HEADER_BUFFER_SIZE = 4096 * 4;

    private static JettyServer instance = null;

    private CorrectContentTypeServlet correctContentTypeServlet;

    private CorrectContentTypeWithCharsetServlet correctContentTypeWithCharsetServlet;

    private WrongContentTypeServlet wrongContentTypeSevlet;
    
    private WithoutContentTypeServlet withoutContentTypeSevlet;

    private ServerConfig config;

    private Server server;

    private BoundedThreadPool threadPool;

    /**
     * Creates a new instance of {@link JettyServer}
     * 
     * @param configContext
     * @param config
     * @param componentContext
     * @param logger
     * @throws AxisFault
     */
    private JettyServer(final ServerConfig config) {
        this.config = config;

        // jetty threapool configuration
        this.threadPool = new BoundedThreadPool();
        this.threadPool.setName("SoapJettyThreadPool");
        this.threadPool.setMaxThreads(this.config.getJettyThreadMaxPoolSize());
        this.threadPool.setMinThreads(this.config.getJettyThreadMinPoolSize());

        // jetty http connector configuration
        final SelectChannelConnector nioConnector = new SelectChannelConnector();
        nioConnector.setPort(this.config.getPort());

        // If we assign the host, we will only be able to contact server
        // on it. No value or a null one is a wildcard so connection is possible
        // on network interface
        // @see java.net.InetSocketAddress
        if (this.config.isRestrict()) {
            nioConnector.setHost(this.config.getHost());
        }

        nioConnector.setHeaderBufferSize(HEADER_BUFFER_SIZE);
        nioConnector.setStatsOn(false);
        nioConnector.setAcceptors(this.config.getJettyAcceptors());

        this.server = new Server();
        this.server.setConnectors(new Connector[] { nioConnector });
        this.server.setThreadPool(this.threadPool);

    }

    public void createAllServlets() {
        // add welcome servlet
        ContextHandlerCollection contexts = new ContextHandlerCollection();
        // create context handlers
        this.server.setHandler(contexts);

        createCorrectContentTypeServlet(contexts);

        createCorrectContentTypeWithCharsetServlet(contexts);

        createWrongContentTypeServlet(contexts);
        
        createWithoutContentTypeServlet(contexts);
    }

    private void createWithoutContentTypeServlet(ContextHandlerCollection contexts) {
        ServletHolder repositoryServletHolder;
        final Context withoutContentTypeContext = new Context(contexts, "/WithoutContentType",
                Context.SESSIONS);
        this.withoutContentTypeSevlet = new WithoutContentTypeServlet();
        repositoryServletHolder = new ServletHolder(withoutContentTypeSevlet);
        repositoryServletHolder.setName("WithoutContentTypeSevlet");
        repositoryServletHolder.setInitOrder(3);
        withoutContentTypeContext.addServlet(repositoryServletHolder, "/");
    }
    
    private void createWrongContentTypeServlet(ContextHandlerCollection contexts) {
        ServletHolder repositoryServletHolder;
        final Context wrongContentTypeContext = new Context(contexts, "/WrongContentType",
                Context.SESSIONS);
        this.wrongContentTypeSevlet = new WrongContentTypeServlet();
        repositoryServletHolder = new ServletHolder(wrongContentTypeSevlet);
        repositoryServletHolder.setName("WrongContentTypeSevlet");
        repositoryServletHolder.setInitOrder(1);
        wrongContentTypeContext.addServlet(repositoryServletHolder, "/");
    }
    
    private void createCorrectContentTypeServlet(ContextHandlerCollection contexts) {
        final Context repositoryContext = new Context(contexts, "/CorrectContentType", Context.SESSIONS);
        this.correctContentTypeServlet = new CorrectContentTypeServlet();
        ServletHolder correctContentTypeServletHolder = new ServletHolder(correctContentTypeServlet);
        correctContentTypeServletHolder.setName("CorrectContentTypeServlet");
        correctContentTypeServletHolder.setInitOrder(2);
        repositoryContext.addServlet(correctContentTypeServletHolder, "/");
    }

    private void createCorrectContentTypeWithCharsetServlet(ContextHandlerCollection contexts) {
        final Context repositoryContext = new Context(contexts, "/CorrectContentTypeWithCharset",
                Context.SESSIONS);
        this.correctContentTypeWithCharsetServlet = new CorrectContentTypeWithCharsetServlet();
        ServletHolder correctContentTypeWithCharsetServletHolder = new ServletHolder(
                correctContentTypeWithCharsetServlet);
        correctContentTypeWithCharsetServletHolder.setName("CorrectContentTypeWithCharsetServlet");
        correctContentTypeWithCharsetServletHolder.setInitOrder(2);
        repositoryContext.addServlet(correctContentTypeWithCharsetServletHolder, "/");
    }

    public static JettyServer getInstance() throws Exception {
        if (instance == null) {
            instance = new JettyServer(new ServerConfig());
            instance.createAllServlets();
            instance.start();
        }
        if (!instance.getServer().isStarted()) {
            instance.createAllServlets();
            instance.start();
        }

        return instance;
    }


    /*
     * (non-Javadoc)
     * 
     * @see org.apache.axis2.transport.TransportListener#start()
     */
    public void start() throws Exception {
        this.log.info("Starting Jetty server...");
        this.log.info("Host : "
                + (!this.config.isRestrict() ? "*" : this.config.getHost() + " (restricted)")
                + " / Port : " + this.config.getPort() + " / Jetty Max poolsize : "
                + this.config.getJettyThreadMaxPoolSize() + " / Jetty Min poolsize : "
                + this.config.getJettyThreadMinPoolSize() + " / Jetty Acceptors size : "
                + this.config.getJettyAcceptors());

        this.server.start();

    }

    /*
     * (non-Javadoc)
     * 
     * @see org.apache.axis2.transport.TransportListener#stop()
     */
    public void stop() throws Exception {
        this.log.info("Stop and kill Jetty server...");

        // stop jetty server
        for (Handler h : this.server.getHandlers()) {
            this.server.removeHandler(h);
        }
        for (Handler h : this.server.getChildHandlers()) {
            this.server.removeHandler(h);
        }
        // for(Connector c : this.server.getConnectors()) {
        // this.server.removeConnector(c);
        // }
        this.server.stop();
        // this.server.destroy();

    }

    public Server getServer() {
        return server;
    }

    public ServerConfig getConfig() {
        return config;
    }

}
