/**
 * Copyright (c) 2011-2012 EBM WebSourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.jbi.management.task.installation.install.sl;

import java.util.logging.Logger;

import org.easymock.EasyMock;
import org.easymock.IAnswer;
import org.junit.Test;
import org.ow2.petals.microkernel.api.jbi.management.Context;
import org.ow2.petals.microkernel.api.jbi.management.Task;
import org.ow2.petals.microkernel.api.system.SystemStateService;
import org.ow2.petals.microkernel.jbi.management.task.installation.uninstall.sl.RemoveSLStateHolderTask;
import org.ow2.petals.microkernel.util.AbstractMavenTest;
import org.ow2.petals.systemstate.generated.SharedLibrary;

import com.ebmwebsourcing.easycommons.log.LoggingUtil;

import junit.framework.TestCase;

/**
 * Test class for the task {@link RemoveSLStateHolderTask}.
 * @author Christophe DENEUX - EBM WebSourcing
 */
public class RemoveSLStateHolderTaskTest extends AbstractMavenTest {

    final static private String SL_NAME = "mySharedLibraryName";

    final static private String SL_VERSION = "1.0.0";

    final LoggingUtil log = new LoggingUtil(Logger.getLogger(
            RemoveSLStateHolderTaskTest.class.getName()));

    /**
     * Check interaction between the tested task and the system state service,
     * when processing the task.
     * 
     * @throws Exception
     */
    @Test    
    public void testRemoveSLStateHolderTask_000() throws Exception {
        final SystemStateService systemStateSvc = EasyMock.createMock(SystemStateService.class);
        EasyMock.expect(systemStateSvc.deleteSharedLibraryStateHolder(SL_NAME, SL_VERSION))
                .andAnswer(new IAnswer<SharedLibrary>() {

                    @Override
                    public SharedLibrary answer() throws Throwable {
                        final SharedLibrary sharedLibrary = new SharedLibrary();
                        sharedLibrary.setName(SL_NAME);
                        sharedLibrary.setVersion(SL_VERSION);
                        return sharedLibrary;
                    }
                });
        EasyMock.replay(systemStateSvc);

        final Context ctx = new Context();
        ctx.setEntityName(SL_NAME);
        ctx.setEntityVersion(SL_VERSION);
        final Task task = new RemoveSLStateHolderTask(log, systemStateSvc, null);

        // Process the task.
        task.execute(ctx);
        // Check returned value of the task
        TestCase.assertEquals(SL_NAME, ctx.getSharedLibraryStateHolder().getName());
        TestCase.assertEquals(SL_VERSION, ctx.getSharedLibraryStateHolder().getVersion());

        // Verify interaction between the current test and the system state.
        EasyMock.verify(systemStateSvc);
    }

    /**
     * Check interaction between the tested task and the system state service,
     * when processing the task twice for the same shared library.
     * 
     * @throws Exception
     */
    @Test    
    public void testRemoveSLStateHolderTask_001() throws Exception {
        final SystemStateService systemStateSvc = EasyMock.createMock(SystemStateService.class);
        EasyMock.expect(systemStateSvc.deleteSharedLibraryStateHolder(SL_NAME, SL_VERSION))
                .andAnswer(new IAnswer<SharedLibrary>() {

                    @Override
                    public SharedLibrary answer() throws Throwable {
                        final SharedLibrary sharedLibrary = new SharedLibrary();
                        sharedLibrary.setName(SL_NAME);
                        sharedLibrary.setVersion(SL_VERSION);
                        return sharedLibrary;
                    }
                });
        EasyMock.expect(systemStateSvc.deleteSharedLibraryStateHolder(SL_NAME, SL_VERSION))
                .andReturn(null);
        EasyMock.replay(systemStateSvc);

        // First, process the task for the 1st installation of the shared
        // library.
        {
            final Context ctx1 = new Context();
            ctx1.setEntityName(SL_NAME);
            ctx1.setEntityVersion(SL_VERSION);
            final Task task1 = new RemoveSLStateHolderTask(log, systemStateSvc, null);

            task1.execute(ctx1);

            // Check returned value of the task
            TestCase.assertEquals(SL_NAME, ctx1.getSharedLibraryStateHolder().getName());
            TestCase.assertEquals(SL_VERSION, ctx1.getSharedLibraryStateHolder().getVersion());
        }

        // Second, process the task for the 2nd installation of the shared
        // library.
        {
            final Context ctx2 = new Context();
            ctx2.setEntityName(SL_NAME);
            ctx2.setEntityVersion(SL_VERSION);
            final Task task2 = new RemoveSLStateHolderTask(log, systemStateSvc, null);

            task2.execute(ctx2);

            // Check returned value of the task
            TestCase.assertNull("Returned shared library state holder not null",
                    ctx2.getSharedLibraryStateHolder());
        }

        // Verify interaction between the current test and the system state
        // service.
        EasyMock.verify(systemStateSvc);
    }

}
