/**
 * Copyright (c) 2005-2012 EBM WebSourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.system.classloader;

import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;

import java.io.DataInputStream;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.util.Arrays;
import java.util.Collections;
import java.util.Enumeration;

import org.junit.Test;

/**
 * @author ddesjardins - EBM WebSourcing
 */
public class ComponentClassLoaderTest {

    private ComponentClassLoader getClassLoader(final boolean parentFirst) throws Exception {

        final URL oneResourceURL = Thread.currentThread().getContextClassLoader().getResource("resourceBoth.txt");
        final File oneResourceFile = new File(oneResourceURL.toURI());
        final File testClassesDir = oneResourceFile.getParentFile();
        final File targetDir = testClassesDir.getParentFile();

        SharedLibraryClassLoader sharedLibraryClassLoader = new SharedLibraryClassLoader(
                new URL[] { testClassesDir.toURI().toURL(), targetDir.toURI().toURL() },
                Arrays.asList("test-classloader-project.jar"), ClassLoader.getSystemClassLoader(), false);

        SharedLibrariesClassLoader sharedLibrariesClassLoader = new SharedLibrariesClassLoader(
                this.getClass().getClassLoader(), Collections.singletonList("1"),
                Collections.singletonMap("1", sharedLibraryClassLoader));

        return new ComponentClassLoader(new URL[] { testClassesDir.toURI().toURL() },
                Arrays.asList("test-classloader-project-2.jar"), sharedLibrariesClassLoader, parentFirst);
    }

    /**
     * Test to get a resource from itself The resource is in the component classpath and in the shared library classpath
     * 
     * @throws Exception
     */
    @Test
    public void testGetResourceInSelfAndParentSelfFirst() throws Exception {
        ComponentClassLoader cl = getClassLoader(false);
        URL url = cl.getResource("resourceBoth.txt");
        assertNotNull("testGetResourceInSelfAndParentSelfFirst - resource not found", url);
    }

    /**
     * Test to get a resource from its parent The resource is in the component classpath and in the shared library
     * classpath
     * 
     * @throws Exception
     */
    @Test
    public void testGetResourceInSelfAndParentParentFirst() throws Exception {
        PetalsClassLoader componentClassLoader = getClassLoader(true);
        URL url = componentClassLoader.getResource("resourceBoth.txt");
        assertNotNull("testGetResourceInSelfAndParentParentFirst - resource not found", url);
    }

    /**
     * Test to get a resource from itself The resource is in the component classpath and in the shared library classpath
     * 
     * @throws IOException
     */
    @Test
    public void testGetResourcesInSelfAndParentSelfFirst() throws Exception {
        ComponentClassLoader cl = getClassLoader(false);
        Enumeration<URL> urls = cl.getResources("resourceBoth.txt");
        assertTrue("testGetResourceInSelfAndParentSelfFirst - resource not found", urls.hasMoreElements());
    }

    /**
     * Test to get a resource from its parent The resource is in the component classpath and in the shared library
     * classpath
     * 
     * @throws IOException
     */
    @Test
    public void testGetResourcesInSelfAndParentParentFirst() throws Exception {
        PetalsClassLoader componentClassLoader = getClassLoader(true);
        Enumeration<URL> urls = componentClassLoader.getResources("resourceBoth.txt");
        assertTrue("testGetResourceInSelfAndParentParentFirst - resource not found", urls.hasMoreElements());
    }

    /**
     * Test to get a resource as stream from itself The resource is in the
     * component classpath and in the shared library classpath
     * 
     * @throws IOException
     */
    @Test
    public void testGetResourceAsStreamInSelfAndParentSelfFirst() throws Exception {
        ComponentClassLoader cl = getClassLoader(false);
        InputStream inputStream = cl.getResourceAsStream("resourceBoth.txt");
        assertNotNull("testGetResourceAsStreamInSelfAndParentSelfFirst - resource not found",
                inputStream);

        DataInputStream dis = new DataInputStream(inputStream);
        byte[] content = new byte[dis.available()];
        dis.readFully(content);

    }

    /**
     * Test to get a resource as stream from its parent The resource is in the
     * component classpath and in the shared library classpath
     * 
     * @throws IOException
     * 
     */
    @Test
    public void testGetResourceAsStreamInSelfAndParentParentFirst() throws Exception {
        PetalsClassLoader componentClassLoader = getClassLoader(true);
        InputStream inputStream = componentClassLoader.getResourceAsStream("resourceBoth.txt");
        assertNotNull("testGetResourceAsStreamInSelfAndParentParentFirst - resource not found",
                inputStream);

        DataInputStream dis = new DataInputStream(inputStream);
        byte[] content = new byte[dis.available()];
        dis.readFully(content);
    }

    /**
     * Test to get a resource as stream from its self The resource is in the
     * shared library classpath
     * 
     * @throws IOException
     */
    @Test
    public void testGetResourceAsStreamInParentSelfFirst() throws Exception {
        ComponentClassLoader cl = getClassLoader(false);
        InputStream inputStream = cl.getResourceAsStream("resourceCore.txt");
        assertNotNull("testGetResourceAsStreamInParentSelfFirst - resource not found", inputStream);

        DataInputStream dis = new DataInputStream(inputStream);
        byte[] content = new byte[dis.available()];
        dis.readFully(content);
    }

    /**
     * Test to get a resource as stream from its parent The resource is in the
     * component classpath
     * 
     * @throws IOException
     */
    @Test
    public void testGetResourceAsStreamInSelfParentFirst() throws Exception {
        PetalsClassLoader componentClassLoader = getClassLoader(true);
        InputStream inputStream = componentClassLoader.getResourceAsStream("resourceContainer.txt");
        assertNotNull("testGetResourceAsStreamInSelfParentFirst - resource not found", inputStream);

        DataInputStream dis = new DataInputStream(inputStream);
        byte[] content = new byte[dis.available()];
        dis.readFully(content);
    }

    /**
     * Test to load a class from its parent
     */
    @Test
    public void testLoadClassParentFirst() throws Exception {
        PetalsClassLoader componentClassLoader = getClassLoader(true);
        final Class<?> clazz = componentClassLoader.loadClass(
                "pkg.TestClass", false);
        assertNotNull("testLoadClassParentFirst - can't be null, an exception must happen if the class is not found",
                clazz);
    }

    /**
     * Test to load a class from its self
     */
    @Test
    public void testLoadClassSelfFirst() throws Exception {

        ComponentClassLoader cl = getClassLoader(false);
        final Class<?> clazz = cl.loadClass(
                "pkg.TestClass2", false);
        assertNotNull("testLoadClassSelfFirst - can't be null, an exception must happen if the class is not found",
                clazz);
    }

    /**
     * Test to load a class from its parent The class is in both classpath
     * (parent and self)
     */
    @Test
    public void testLoadClassInSelfAndParentParentFirst() throws Exception {
        PetalsClassLoader componentClassLoader = getClassLoader(true);
        final Class<?> clazz = componentClassLoader.loadClass("org.objectweb.petals.classloader.TestClassloader",
                false);
        assertNotNull(
                "testLoadClassInSelfAndParentParentFirst - can't be null, an exception must happen if the class is not found",
                clazz);
    }

    /**
     * Test to load a class from its self The class is in both classpath (parent
     * and self)
     */
    @Test
    public void testLoadClassInSelfAndParentSelfFirst() throws Exception {

        ComponentClassLoader cl = getClassLoader(false);
        final Class<?> clazz = cl.loadClass("org.objectweb.petals.classloader.TestClassloader", false);
        assertNotNull(
                "testLoadClassInSelfAndParentSelfFirst - can't be null, an exception must happen if the class is not found",
                clazz);
    }

}
