/**
 * Copyright (c) 2005-2012 EBM WebSourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.system.classloader.factory;

import java.io.File;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;

import org.junit.Assert;
import org.junit.Test;

/**
 * Tests of the URLFactory.
 * 
 * @author cdeneux
 */
public class URLFactoryTest extends Assert {

    @Test
    public void test0() throws MalformedURLException {
        URL tmp = new File("/tmp").toURI().toURL();
        System.out.println("/tmp : " + tmp.toString());
        URL toto = new File("/toto").toURI().toURL();
        System.out.println("/toto : " + toto.toString());
    }

    /**
     * Check {@link URLFactory#getFactory(URL)} against a URL locating a JAR file on a Unix platform.
     */
    @Test
    public void testGetFactory_JarFile_Unix() throws MalformedURLException, IOException {
        final URLFactory urlFactory = URLFactory.getFactory(new URL("file:///home/user/file.jar"));
        assertNotNull(urlFactory);
        assertEquals(true, urlFactory instanceof JarURLFactory);
    }

    /**
     * Check {@link URLFactory#getFactory(URL)} against a URL locating a JAR file on a Windows platform.
     */
    @Test
    public void testGetFactory_JarFile_Windows() throws MalformedURLException, IOException {
        final URLFactory urlFactory = URLFactory.getFactory(new URL("file://C:/temp/file.jar"));
        assertNotNull(urlFactory);
        assertEquals(true, urlFactory instanceof JarURLFactory);
    }

    /**
     * Check {@link URLFactory#getFactory(URL)} against a URL locating a JAR file as a remote HTTP resource.
     */
    @Test
    public void testGetFactory_JarHttp() throws MalformedURLException, IOException {
        final URLFactory urlFactory = URLFactory.getFactory(new URL("http://www.ow2.org/file.jar"));
        assertNotNull(urlFactory);
        assertEquals(true, urlFactory instanceof JarURLFactory);
    }

    /**
     * Check {@link URLFactory#getFactory(URL)} against a URL locating a WAR file on a Unix platform.
     */
    @Test
    public void testGetFactory_WarFile_Unix() throws MalformedURLException, IOException {
        final URLFactory urlFactory = URLFactory.getFactory(new URL("file:///home/user/file.war"));
        assertNotNull(urlFactory);
        assertEquals(true, urlFactory instanceof JarURLFactory);
    }

    /**
     * Check {@link URLFactory#getFactory(URL)} against a URL locating a EAR file on a Unix platform.
     */
    @Test
    public void testGetFactory_EarFile_Unix() throws MalformedURLException, IOException {
        final URLFactory urlFactory = URLFactory.getFactory(new URL("file:///home/user/file.ear"));
        assertNotNull(urlFactory);
        assertEquals(true, urlFactory instanceof JarURLFactory);
    }

    /**
     * Check {@link URLFactory#getFactory(URL)} against a URL locating a directory ended with a '/'.
     */
    @Test
    public void testGetFactory_DirWithEndingSlash() throws MalformedURLException, IOException {
        final URL dirURL = Thread.currentThread().getContextClassLoader().getResource("fakearchive");
        assertNotNull("Directory not found in resources of the classloader.", dirURL);
        final URL url = new URL(dirURL.toString() + "/");
        assertTrue(url.getPath().endsWith("/"));
        final URLFactory urlFactory = URLFactory.getFactory(url);
        assertNotNull(urlFactory);
        assertEquals(true, urlFactory instanceof DirURLFactory);
    }

    /**
     * Check {@link URLFactory#getFactory(URL)} against a URL locating a directory not ended with a '/'.
     */
    @Test
    public void testGetFactory_DirWithoutEndingSlash() throws MalformedURLException, IOException {
        final URL dirURL = Thread.currentThread().getContextClassLoader().getResource("fakearchive");
        assertNotNull("Directory not found in resources of the classloader.", dirURL);
        final URL url = new URL(dirURL.toString());
        assertFalse(url.getPath().endsWith("/"));
        final URLFactory urlFactory = URLFactory.getFactory(url);
        assertNotNull(urlFactory);
        assertEquals(true, urlFactory instanceof DirURLFactory);
    }

    /**
     * Check {@link URLFactory#getFactory(URL)} against a URL locating any file.
     */
    @Test(expected = IOException.class)
    public void testGetFactory_File() throws IOException {
        final URL dirURL = Thread.currentThread().getContextClassLoader().getResource("fakearchive/archive");
        assertNotNull("File not found in resources of the classloader.", dirURL);
        URLFactory.getFactory(new URL(dirURL.toString()));
    }

}
